@@ -96,12 +96,20 @@ class Options {
       return value;
   }
 
+  applyLocked(node, key) {
+    const locked = this.configs.$isLocked(key);
+    node.disabled = locked;
+    const label = node.closest('label') || (node.id && node.ownerDocument.querySelector(`label[for="${node.id}"]`)) || node;
+    if (label)
+      label.classList.toggle('locked', locked);
+  }
+
   bindToCheckbox(key, node) {
     node.checked = this.configValueToUIValue(this.configs[key]);
     node.addEventListener('change', () => {
       this.throttledUpdate(key, node, node.checked);
     });
-    node.disabled = this.configs.$isLocked(key);
+    this.applyLocked(node, key);
     this.addResetMethod(key, node);
     const nodes = this.uiNodes.get(key) || [];
     nodes.push(node);
@@ -116,7 +124,7 @@ class Options {
       const nodes = this.uiNodes.get(id) || [];
       if (nodes.includes(radio))
         continue;
-      radio.disabled = this.configs.$isLocked(key);
+      this.applyLocked(radio, key);
       nodes.push(radio);
       this.uiNodes.set(id, nodes);
       radio.addEventListener('change', () => {
@@ -139,7 +147,7 @@ class Options {
     node.addEventListener('input', () => {
       this.throttledUpdate(key, node, node.value);
     });
-    node.disabled = this.configs.$isLocked(key);
+    this.applyLocked(node, key);
     this.addResetMethod(key, node);
     const nodes = this.uiNodes.get(key) || [];
     nodes.push(node);
@@ -150,7 +158,7 @@ class Options {
     node.addEventListener('change', () => {
       this.throttledUpdate(key, node, node.value);
     });
-    node.disabled = this.configs.$isLocked(key);
+    this.applyLocked(node, key);
     this.addResetMethod(key, node);
     const nodes = this.uiNodes.get(key) || [];
     nodes.push(node);
@@ -223,7 +231,7 @@ class Options {
       else {
         node.value = this.configValueToUIValue(this.configs[key]);
       }
-      node.disabled = this.configs.$isLocked(key);
+      this.applyLocked(node, key);
     }
   }
 
@@ -260,6 +268,7 @@ class Options {
         <tbody>${rows.join('')}</tbody>
       </table>
       <div>
+        <button id="allconfigs-reset-all">Reset All</button>
         <button id="allconfigs-export">Export</button>
         <a id="allconfigs-export-file"
            type="application/json"
@@ -295,6 +304,15 @@ class Options {
           input.$reset();
       });
     }
+    const resetAllButton = document.getElementById('allconfigs-reset-all');
+    resetAllButton.addEventListener('keydown', event => {
+      if (event.key == 'Enter' || event.key == ' ')
+        this.resetAll();
+    });
+    resetAllButton.addEventListener('click', event => {
+      if (event.button == 0)
+        this.resetAll();
+    });
     const exportButton = document.getElementById('allconfigs-export');
     exportButton.addEventListener('keydown', event => {
       if (event.key == 'Enter' || event.key == ' ')
@@ -319,13 +337,19 @@ class Options {
       reader.onload = event => {
         const values = JSON.parse(event.target.result);
         for (const key of Object.keys(this.configs.$default)) {
-          this.configs[key] = values[key] !== undefined ? values[key] : this.configs.$default[key]
+          this.configs[key] = values[key] !== undefined ? values[key] : this.configs.$default[key];
         }
       };
       reader.readAsText(fileField.files.item(0), 'utf-8');
     });
   }
 
+  resetAll() {
+    for (const key of Object.keys(this.configs.$default)) {
+      this.configs[key] = this.configs.$default[key];
+    }
+  }
+
   importFromFile() {
     document.getElementById('allconfigs-import-file').click();
   }
