@@ -8,6 +8,9 @@
 import {
   log as internalLogger,
   dumpTab,
+  mapAndFilter,
+  mapAndFilterUniq,
+  toLines,
   configs
 } from './common.js';
 
@@ -16,6 +19,7 @@ import * as ApiTabs from '/common/api-tabs.js';
 import * as TabsStore from './tabs-store.js';
 import * as UniqueId from './unique-id.js';
 import * as SidebarConnection from './sidebar-connection.js';
+import * as ContextualIdentities from './contextual-identities.js';
 
 import Window from './Window.js';
 
@@ -104,6 +108,9 @@ export default class Tab {
       incompletelyTrackedTabsPerWindow.delete(tab);
       Tab.onTracked.dispatch(tab);
     });
+
+    // We should initialize private properties with blank value for better performance with a fixed shape.
+    this.delayedInheritSoundStateFromChildren = null;
   }
 
   destroy() {
@@ -128,15 +135,16 @@ export default class Tab {
       if (this.element.parentNode) {
         this.element.parentNode.removeChild(this.element);
       }
-      delete this.element.$TST;
-      delete this.element.apiTab;
-      delete this.element;
-      delete this.classList;
+      this.element.$TST = null;
+      this.element.apiTab = null;
+      this.element = null;
+      this.classList = null;
     }
-    delete this;
-    delete this.tab;
-    delete this.promisedUniqueId;
-    delete this.uniqueId;
+    // this.tab.$TST = null; // tab.$TST is used by destruction processes.
+    this.tab = null;
+    this.promisedUniqueId = null;
+    this.uniqueId = null;
+    this.destroyed = true;
   }
 
   clear() {
@@ -150,6 +158,8 @@ export default class Tab {
   }
 
   bindElement(element) {
+    element.$TST   = this;
+    element.apiTab = this.tab;
     this.element = element;
     this.classList = element.classList;
     setTimeout(() => { // wait until initialization processes are completed
@@ -257,6 +267,11 @@ export default class Tab {
     return following && following.pinned;
   }
 
+  get followsUnpinnedTab() {
+    const preceding = this.nearestVisiblePrecedingTab;
+    return preceding && !preceding.pinned;
+  }
+
   get duplicating() {
     return this.states.has(Constants.kTAB_STATE_DUPLICATING);
   }
@@ -264,7 +279,25 @@ export default class Tab {
   get isNewTabCommandTab() {
     if (!configs.guessNewOrphanTabAsOpenedByNewTabCommand)
       return false;
-    return this.tab.url == configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl;
+
+    const newTabUrl = configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl;
+    if (this.tab.url != newTabUrl)
+      return false;
+
+    // Firefox always opens a blank tab as the placeholder, when trying to
+    // open a bookmark in a new tab. So, we cannot determine is the tab
+    // really opened as a new blank tab or just as a placeholder for an
+    // "Open in New Tab" operation, when the user choose the "Blank Page"
+    // as the new tab page.
+    // But, when "Blank Page" is chosen as the new tab page, Firefox loads
+    // "about:blank" into a newly opened blank tab. As the result both current
+    // URL and the previous URL become "about:blank". This is an important
+    // difference between "a new blank tab" and "a blank tab opened for an
+    // Open in New Tab command".
+    if (newTabUrl == 'about:blank')
+      return this.tab.previousUrl == 'about:blank';
+
+    return true;
   }
 
   get isGroupTab() {
@@ -293,18 +326,43 @@ export default class Tab {
   // Temporary Containers
   // https://addons.mozilla.org/firefox/addon/temporary-containers/
   get mayBeReplacedWithContainer() {
-    const prevTab = this.unsafePreviousTab;
-    return (
-      !!prevTab &&
-      this.tab.cookieStoreId != prevTab.cookieStoreId &&
-      this.tab.url == prevTab.url
+    return !!(
+      this.$possiblePredecessorPreviousTab ||
+      this.$possiblePredecessorNextTab
     );
   }
-  get mayBeOriginalOfReplacedWithContainer() {
-    const firstChild = this.firstChild;
+  get $possiblePredecessorPreviousTab() {
+    const prevTab = this.unsafePreviousTab;
     return (
-      !!firstChild &&
-      firstChild.$TST.mayBeReplacedWithContainer
+      prevTab &&
+      this.tab.cookieStoreId != prevTab.cookieStoreId &&
+      this.tab.url == prevTab.url
+    ) ? prevTab : null;
+  }
+  get $possiblePredecessorNextTab() {
+    const nextTab = this.unsafeNextTab;
+    return (
+      nextTab &&
+      this.tab.cookieStoreId != nextTab.cookieStoreId &&
+      this.tab.url == nextTab.url
+    ) ? nextTab : null;
+  }
+  get possibleSuccessorWithDifferentContainer() {
+    const firstChild = this.firstChild;
+    const nextTab = this.nextTab;
+    const prevTab = this.previousTab;
+    return (
+      (firstChild &&
+       firstChild.$TST.$possiblePredecessorPreviousTab == this.tab &&
+       firstChild) ||
+      (nextTab &&
+       !nextTab.$TST.openedCompletely &&
+       nextTab.$TST.$possiblePredecessorPreviousTab == this.tab &&
+       nextTab) ||
+      (prevTab &&
+       !prevTab.$TST.openedCompletely &&
+       prevTab.$TST.$possiblePredecessorNextTab == this.tab &&
+       prevTab)
     );
   }
 
@@ -324,6 +382,36 @@ export default class Tab {
     return !!(highlightedTabs && highlightedTabs.size > 1);
   }
 
+  get promisedPossibleOpenerBookmarks() {
+    if ('possibleOpenerBookmarks' in this)
+      return Promise.resolve(this.possibleOpenerBookmarks);
+    return new Promise(async (resolve, _reject) => {
+      if (!browser.bookmarks)
+        return resolve(this.possibleOpenerBookmarks = []);
+      // A new tab from bookmark is opened with a title: its URL without the scheme part.
+      const url = this.tab.$possibleInitialUrl;
+      try {
+        const possibleBookmarks = await Promise.all([
+          browser.bookmarks.search({ url: `http://${url}` }).catch(_error => []),
+          browser.bookmarks.search({ url: `https://${url}` }).catch(_error => []),
+          browser.bookmarks.search({ url: `ftp://${url}` }).catch(_error => []),
+          browser.bookmarks.search({ url: `moz-extension://${url}` }).catch(_error => [])
+        ]);
+        resolve(this.possibleOpenerBookmarks = possibleBookmarks.flat());
+      }
+      catch(_error) {
+        // If it is detected as "not a valid URL", then
+        // it cannot be a tab opened from a bookmark.
+        resolve(this.possibleOpenerBookmarks = []);
+      }
+    });
+  }
+
+  get cookieStoreName() {
+    const identity = this.tab.cookieStoreId && ContextualIdentities.get(this.tab.cookieStoreId);
+    return identity ? identity.name : null;
+  }
+
   //===================================================================
   // neighbor tabs
   //===================================================================
@@ -429,6 +517,109 @@ export default class Tab {
     });
   }
 
+  get nearestLoadedTab() {
+    const tabs = TabsStore.visibleTabsInWindow.get(this.tab.windowId);
+    return (
+      // nearest following tab
+      TabsStore.query({
+        windowId:  this.tab.windowId,
+        tabs,
+        discarded: false,
+        fromId:    this.tab.id,
+        visible:   true,
+        index:     (index => index > this.tab.index)
+      }) ||
+      // nearest preceding tab
+      TabsStore.query({
+        windowId:  this.tab.windowId,
+        tabs,
+        discarded: false,
+        fromId:    this.tab.id,
+        visible:   true,
+        index:     (index => index < this.tab.index),
+        last:      true
+      })
+    );
+  }
+
+  get nearestLoadedTabInTree() {
+    let tab = this.tab;
+    const tabs = TabsStore.visibleTabsInWindow.get(tab.windowId);
+    let lastLastDescendant;
+    while (tab) {
+      const parent = tab.$TST.parent;
+      if (!parent)
+        return null;
+      const lastDescendant = parent.$TST.lastDescendant;
+      const loadedTab = (
+        // nearest following tab
+        TabsStore.query({
+          windowId:     tab.windowId,
+          tabs,
+          descendantOf: parent.id,
+          discarded:    false,
+          '!id':        this.tab.id,
+          fromId:       (lastLastDescendant || this.tab).id,
+          toId:         lastDescendant.id,
+          visible:      true,
+          index:        (index => index > this.tab.index)
+        }) ||
+        // nearest preceding tab
+        TabsStore.query({
+          windowId:     tab.windowId,
+          tabs,
+          descendantOf: parent.id,
+          discarded:    false,
+          '!id':        this.tab.id,
+          fromId:       tab.id,
+          toId:         parent.$TST.firstChild.id,
+          visible:      true,
+          index:        (index => index < tab.index),
+          last:         true
+        })
+      );
+      if (loadedTab)
+        return loadedTab;
+      if (!parent.discarded)
+        return parent;
+      lastLastDescendant = lastDescendant;
+      tab = tab.$TST.parent;
+    }
+    return null;
+  }
+
+  get nearestLoadedSiblingTab() {
+    const parent = this.parent;
+    if (!parent)
+      return null;
+    const tabs = TabsStore.visibleTabsInWindow.get(this.tab.windowId);
+    return (
+      // nearest following tab
+      TabsStore.query({
+        windowId:  this.tab.windowId,
+        tabs,
+        childOf:   parent.id,
+        discarded: false,
+        fromId:    this.tab.id,
+        toId:      parent.$TST.lastChild.id,
+        visible:   true,
+        index:     (index => index > this.tab.index)
+      }) ||
+      // nearest preceding tab
+      TabsStore.query({
+        windowId:  this.tab.windowId,
+        tabs,
+        childOf:   parent.id,
+        discarded: false,
+        fromId:    this.tab.id,
+        toId:      parent.$TST.firstChild.id,
+        visible:   true,
+        index:     (index => index < this.tab.index),
+        last:      true
+      })
+    );
+  }
+
   //===================================================================
   // tree relations
   //===================================================================
@@ -478,10 +669,15 @@ export default class Tab {
     return !!this.parentId;
   }
 
-  get ancestors() {
+  get ancestorIds() {
     if (!this.cachedAncestorIds)
-      return this.updateAncestors();
-    return this.cachedAncestorIds.map(id => Tab.get(id)).filter(TabsStore.ensureLivingTab);
+      this.updateAncestors();
+    return this.cachedAncestorIds;
+  }
+
+  get ancestors() {
+    return mapAndFilter(this.ancestorIds,
+                        id => TabsStore.ensureLivingTab(Tab.get(id)) || undefined);
   }
 
   updateAncestors() {
@@ -537,20 +733,19 @@ export default class Tab {
   }
 
   set children(tabs) {
-    const ancestorsOfSelf = this.ancestors;
-    tabs = tabs.filter(tab => {
-      if (!ancestorsOfSelf.includes(tab))
-        return true;
+    const ancestorIds = this.ancestorIds;
+    const newChildIds = mapAndFilter(tabs, tab => {
+      const id = typeof tab == 'number' ? tab : tab && tab.id;
+      if (!ancestorIds.includes(id))
+        return TabsStore.ensureLivingTab(Tab.get(id)) ? id : undefined;
       console.log('FATAL ERROR: Cyclic tree structure has detected and prevented. ', {
-        ancestorsOfSelf,
+        ancestorsOfSelf: this.ancestors,
         tabs,
         tab,
         stack: new Error().stack
       });
-      return false;
+      return undefined;
     });
-
-    const newChildIds = tabs.map(tab => typeof tab == 'number' ? tab : tab && tab.id).filter(id => id);
     if (newChildIds.join('|') == this.childIds.join('|'))
       return tabs;
 
@@ -575,7 +770,8 @@ export default class Tab {
     return tabs;
   }
   get children() {
-    return this.childIds.map(id => Tab.get(id)).filter(TabsStore.ensureLivingTab);
+    return mapAndFilter(this.childIds,
+                        id => TabsStore.ensureLivingTab(Tab.get(id)) || undefined);
   }
 
   get firstChild() {
@@ -583,13 +779,21 @@ export default class Tab {
     return children.length > 0 ? children[0] : null ;
   }
 
+  get firstVisibleChild() {
+    const firstChild = this.firstChild;
+    return firstChild && !firstChild.$TST.collapsed && !firstChild.hidden && firstChild;
+  }
+
   get lastChild() {
     const children = this.children;
     return children.length > 0 ? children[children.length - 1] : null ;
   }
 
   sortChildren() {
-    this.childIds = Tab.sort(this.childIds.map(id => Tab.get(id))).map(tab => tab && tab.id);
+    // Tab.get(tabId) calls into TabsStore.tabs.get(tabId), which is just a
+    // Map. This is acceptable to repeat in order to avoid two array copies,
+    // especially on larger tab sets.
+    this.childIds.sort((a, b) => Tab.compare(Tab.get(a), Tab.get(b)));
     this.invalidateCachedDescendants();
   }
 
@@ -600,7 +804,8 @@ export default class Tab {
   get descendants() {
     if (!this.cachedDescendantIds)
       return this.updateDescendants();
-    return this.cachedDescendantIds.map(id => Tab.get(id)).filter(TabsStore.ensureLivingTab);
+    return mapAndFilter(this.cachedDescendantIds,
+                        id => TabsStore.ensureLivingTab(Tab.get(id)) || undefined);
   }
 
   updateDescendants() {
@@ -650,6 +855,11 @@ export default class Tab {
     }
   }
 
+  get nextVisibleSiblingTab() {
+    const nextSibling = this.nextSiblingTab;
+    return nextSibling && !nextSibling.$TST.collapsed && !nextSibling.hidden && nextSibling;
+  }
+
   get previousSiblingTab() {
     const parent = this.parent;
     if (parent) {
@@ -708,6 +918,21 @@ export default class Tab {
     return opener && opener.pinned;
   }
 
+  get bundledTab() {
+    if (!this.tab)
+      return null;
+    if (this.tab.pinned)
+      return Tab.getGroupTabForOpener(this.tab);
+    if (this.isGroupTab)
+      return Tab.getOpenerFromGroupTab(this.tab);
+    return null;
+  }
+
+  get bundledTabId() {
+    const tab = this.bundledTab;
+    return tab ? tab.id : -1;
+  }
+
   findSuccessor(options = {}) {
     if (typeof options != 'object')
       options = {};
@@ -914,7 +1139,7 @@ export default class Tab {
       clearTimeout(this.delayedInheritSoundStateFromChildren);
 
     this.delayedInheritSoundStateFromChildren = setTimeout(() => {
-      delete this.delayedInheritSoundStateFromChildren;
+      this.delayedInheritSoundStateFromChildren = null;
       if (!TabsStore.ensureLivingTab(this.tab))
         return;
 
@@ -1001,6 +1226,9 @@ export default class Tab {
   }
 
   memorizeNeighbors() {
+    if (!this.tab) // already closed tab
+      return;
+
     const previousTab = this.unsafePreviousTab;
     this.lastPreviousTabId = previousTab && previousTab.id;
 
@@ -1010,20 +1238,24 @@ export default class Tab {
 
   get isSubstantiallyMoved() {
     const previousTab = this.unsafePreviousTab;
-    if (this.lastPreviousTabId != (previousTab && previousTab.id))
+    if (this.lastPreviousTabId &&
+        this.lastPreviousTabId != (previousTab && previousTab.id))
       return true;
 
     const nextTab = this.unsafeNextTab;
-    if (this.lastNextTabId != (nextTab && nextTab.id))
+    if (this.lastNextTabId &&
+        this.lastNextTabId != (nextTab && nextTab.id))
       return true;
 
     return false;
   }
 
   get sanitized() {
-    const sanitized = Object.assign({}, this.tab, {
+    const sanitized = {
+      ...this.tab,
+      '$possibleInitialUrl': null,
       '$TST': null
-    });
+    };
     delete sanitized.$TST;
     return sanitized;
   }
@@ -1038,7 +1270,10 @@ export default class Tab {
       childIds:   this.childIds
     };
     if (full)
-      return Object.assign({}, this.sanitized, { $TST: exported });
+      return {
+        ...this.sanitized,
+        $TST: exported
+      };
     return exported;
   }
 
@@ -1063,6 +1298,24 @@ export default class Tab {
 
     TabsStore.updateIndexesForTab(this.tab);
   }
+
+
+  /* element utilities */
+
+  invalidateElement(targets) {
+    if (this.element && this.element.invalidate)
+      this.element.invalidate(targets);
+  }
+
+  updateElement(targets) {
+    if (this.element && this.element.update)
+      this.element.update(targets);
+  }
+
+  set favIconUrl(url) {
+    if (this.element && 'favIconUrl' in this.element)
+      this.element.favIconUrl = url;
+  }
 }
 
 // The list of properties which should be ignored when synchronization from the
@@ -1138,15 +1391,9 @@ Tab.needToWaitTracked = (windowId) => {
 };
 
 Tab.waitUntilTrackedAll = async (windowId, options = {}) => {
-  const tabSets = [];
-  if (windowId) {
-    tabSets.push(mIncompletelyTrackedTabs.get(windowId));
-  }
-  else {
-    for (const tabs of mIncompletelyTrackedTabs.values()) {
-      tabSets.push(tabs);
-    }
-  }
+  const tabSets = windowId ?
+    [mIncompletelyTrackedTabs.get(windowId)] :
+    [...mIncompletelyTrackedTabs.values()];
   return Promise.all(tabSets.map(tabs => {
     if (!tabs)
       return;
@@ -1173,7 +1420,7 @@ async function waitUntilTracked(tabId, options = {}) {
         Tab.onTracked.removeListener(onTracked);
       // eslint-disable-next-line no-use-before-define
       Tab.onDestroyed.removeListener(onDestroyed);
-      reject(new Error(`Tab.waitUntilTracked for ${tabId} is timed out (in ${TabsStore.getWindow() || 'bg'})\b${stack}`));
+      reject(new Error(`Tab.waitUntilTracked for ${tabId} is timed out (in ${TabsStore.getCurrentWindowId() || 'bg'})\b${stack}`));
     }, configs.maximumDelayUntilTabIsTracked); // Tabs.moveTabs() between windows may take much time
     const onDestroyed = (tab) => {
       if (tab.id != tabId)
@@ -1183,7 +1430,7 @@ async function waitUntilTracked(tabId, options = {}) {
       else // eslint-disable-next-line no-use-before-define
         Tab.onTracked.removeListener(onTracked);
       Tab.onDestroyed.removeListener(onDestroyed);
-      reject(new Error(`Tab.waitUntilTracked: ${tabId} is removed while waiting (in ${TabsStore.getWindow() || 'bg'})\n${stack}`));
+      reject(new Error(`Tab.waitUntilTracked: ${tabId} is removed while waiting (in ${TabsStore.getCurrentWindowId() || 'bg'})\n${stack}`));
     };
     const onTracked = (tab) => {
       if (tab.id != tabId)
@@ -1219,7 +1466,7 @@ Tab.waitUntilTracked = async (tabId, options = {}) => {
   if (Array.isArray(tabId))
     return Promise.all(tabId.map(id => Tab.waitUntilTracked(id, options)));
 
-  const windowId = TabsStore.getWindow();
+  const windowId = TabsStore.getCurrentWindowId();
   if (windowId) {
     const tabs = TabsStore.removedTabsInWindow.get(windowId);
     if (tabs && tabs.has(tabId))
@@ -1437,12 +1684,13 @@ Tab.getActiveTabs = () => {
 };
 
 Tab.getAllTabs = (windowId = null, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:     windowId && TabsStore.livingTabsInWindow.get(windowId),
     living:   true,
-    ordered:  true
-  }, options));
+    ordered:  true,
+    ...options
+  });
 };
 
 Tab.getTabAt = (windowId, index) => {
@@ -1459,12 +1707,13 @@ Tab.getTabAt = (windowId, index) => {
 };
 
 Tab.getTabs = (windowId, options = {}) => { // only visible, including collapsed and pinned
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:         TabsStore.controllableTabsInWindow.get(windowId),
     controllable: true,
-    ordered:      true
-  }, options));
+    ordered:      true,
+    ...options
+  });
 };
 
 Tab.getTabsBetween = (begin, end) => {
@@ -1488,49 +1737,59 @@ Tab.getTabsBetween = (begin, end) => {
 };
 
 Tab.getNormalTabs = (windowId, options = {}) => { // only visible, including collapsed, not pinned
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.unpinnedTabsInWindow.get(windowId),
     normal:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getVisibleTabs = (windowId, options = {}) => { // visible, not-collapsed, not-hidden
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.visibleTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getPinnedTabs = (windowId, options = {}) => { // visible, pinned
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.pinnedTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 
 Tab.getUnpinnedTabs = (windowId, options = {}) => { // visible, not pinned
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.unpinnedTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getRootTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:         TabsStore.rootTabsInWindow.get(windowId),
     controllable: true,
-    ordered:      true
-  }, options));
+    ordered:      true,
+    ...options
+  });
+};
+
+Tab.getLastRootTab = (windowId, options = {}) => {
+  const tabs = this.getRootTabs(windowId, options);
+  return tabs[tabs.length - 1];
 };
 
 Tab.collectRootTabs = tabs => {
@@ -1543,89 +1802,99 @@ Tab.collectRootTabs = tabs => {
 };
 
 Tab.getSubtreeCollapsedTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:     TabsStore.subtreeCollapsableTabsInWindow.get(windowId),
     living:   true,
     hidden:   false,
-    ordered:  true
-  }, options));
+    ordered:  true,
+    ...options
+  });
 };
 
 Tab.getCollapsingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
-    tabs: TabsStore.collapsingTabsInWindow.get(windowId)
-  }, options));
+    tabs: TabsStore.collapsingTabsInWindow.get(windowId),
+    ...options
+  });
 };
 
 Tab.getExpandingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs: TabsStore.expandingTabsInWindow.get(windowId),
-  }, options));
+    ...options
+  });
 };
 
 Tab.getGroupTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.groupTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getLoadingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.loadingTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getDraggingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.draggingTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getRemovingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.removingTabsInWindow.get(windowId),
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getDuplicatingTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.duplicatingTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getHighlightedTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.highlightedTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
 };
 
 Tab.getSelectedTabs = (windowId, options = {}) => {
-  const selectedTabs = TabsStore.queryAll(Object.assign({
+  const selectedTabs = TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.selectedTabsInWindow.get(windowId),
     living:  true,
-    ordered: true
-  }, options));
+    ordered: true,
+    ...options
+  });
   const highlightedTabs = TabsStore.highlightedTabsInWindow.get(windowId);
   if (!highlightedTabs ||
       highlightedTabs.size < 2)
@@ -1648,11 +1917,12 @@ Tab.getSelectedTabs = (windowId, options = {}) => {
 };
 
 Tab.getNeedToBeSynchronizedTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:    TabsStore.unsynchronizedTabsInWindow.get(windowId),
-    visible: true
-  }, options));
+    visible: true,
+    ...options
+  });
 };
 
 Tab.hasLoadingTab = windowId => {
@@ -1664,11 +1934,13 @@ Tab.hasLoadingTab = windowId => {
 };
 
 Tab.hasMultipleTabs = (windowId, options = {}) => {
-  const tabs = TabsStore.queryAll(Object.assign({
+  const tabs = TabsStore.queryAll({
     windowId,
     tabs:   TabsStore.livingTabsInWindow.get(windowId),
-    living: true
-  }, options, { iterator: true }));
+    living: true,
+    ...options,
+    iterator: true
+  });
   let count = 0;
   // eslint-disable-next-line no-unused-vars
   for (const tab of tabs) {
@@ -1681,13 +1953,14 @@ Tab.hasMultipleTabs = (windowId, options = {}) => {
 
 // "Recycled tab" is an existing but reused tab for session restoration.
 Tab.getRecycledTabs = (windowId, options = {}) => {
-  return TabsStore.queryAll(Object.assign({
+  return TabsStore.queryAll({
     windowId,
     tabs:       TabsStore.livingTabsInWindow.get(windowId),
     living:     true,
     states:     [Constants.kTAB_STATE_RESTORED, false],
-    attributes: [Constants.kCURRENT_URI, new RegExp(`^(|${configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl}|about:blank|about:privatebrowsing)$`)]
-  }, options));
+    attributes: [Constants.kCURRENT_URI, new RegExp(`^(|${configs.guessNewOrphanTabAsOpenedByNewTabCommandUrl}|about:blank|about:privatebrowsing)$`)],
+    ...options
+  });
 };
 
 
@@ -1704,6 +1977,7 @@ Tab.onHidden           = new EventListenerManager();
 Tab.onShown            = new EventListenerManager();
 Tab.onTabInternallyMoved     = new EventListenerManager();
 Tab.onCollapsedStateChanged  = new EventListenerManager();
+Tab.onMutedStateChanged      = new EventListenerManager();
 
 Tab.onBeforeCreate     = new EventListenerManager();
 Tab.onCreating         = new EventListenerManager();
@@ -1720,6 +1994,10 @@ Tab.onWindowRestoring  = new EventListenerManager();
 Tab.onAttached         = new EventListenerManager();
 Tab.onDetached         = new EventListenerManager();
 
+Tab.onMultipleTabsRemoving = new EventListenerManager();
+Tab.onMultipleTabsRemoved  = new EventListenerManager();
+Tab.onChangeMultipleTabsRestorability = new EventListenerManager();
+
 
 //===================================================================
 // utilities
@@ -1748,7 +2026,7 @@ Tab.getOtherTabs = (windowId, ignoreTabs, options = {}) => {
   if (Array.isArray(ignoreTabs) &&
       ignoreTabs.length > 0)
     query['!id'] = ignoreTabs.map(tab => tab.id);
-  return TabsStore.queryAll(Object.assign(query, options));
+  return TabsStore.queryAll({ ...query, ...options });
 };
 
 function getTabIndex(tab, options = {}) {
@@ -1779,28 +2057,27 @@ Tab.doAndGetNewTabs = async (asyncTask, windowId) => {
     tabsQueryOptions.windowId = windowId;
   }
   const beforeTabs = await browser.tabs.query(tabsQueryOptions).catch(ApiTabs.createErrorHandler());
-  const beforeIds  = beforeTabs.map(tab => tab.id);
+  const beforeIds  = mapAndFilterUniq(beforeTabs, tab => tab.id, { set: true });
   await asyncTask();
   const afterTabs = await browser.tabs.query(tabsQueryOptions).catch(ApiTabs.createErrorHandler());
-  const addedTabs = afterTabs.filter(afterTab => !beforeIds.includes(afterTab.id));
-  return addedTabs.map(tab => Tab.get(tab.id));
+  const addedTabs = mapAndFilter(afterTabs,
+                                 tab => !beforeIds.has(tab.id) && Tab.get(tab.id) || undefined);
+  return addedTabs;
 };
 
-Tab.sort = tabs => {
-  if (tabs.length == 0)
-    return tabs;
-  return tabs.sort((a, b) => a.index - b.index);
-};
+Tab.compare = (a, b) => a.index - b.index;
+
+Tab.sort = tabs => tabs.length == 0 ? tabs : tabs.sort(Tab.compare);
 
 Tab.dumpAll = windowId => {
   if (!configs.debug)
     return;
-  const output = ['dumpAllTabs'];
+  let output = 'dumpAllTabs';
   for (const tab of Tab.getAllTabs(windowId, {iterator: true })) {
-    output.push(tab.$TST.ancestors.reverse().concat([tab])
-      .map(tab => tab.id + (tab.pinned ? ' [pinned]' : ''))
-      .join(' => '));
+    output += '\n' + toLines([...tab.$TST.ancestors.reverse(), tab],
+                             tab => `${tab.id}${tab.pinned ? ' [pinned]' : ''}`,
+                             ' => ');
   }
-  log(output.join('\n'));
+  log(output);
 };
 
