@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -26,11 +26,10 @@
  * ***** END LICENSE BLOCK ******/
 'use strict';
 
-import TabIdFixer from '/extlib/TabIdFixer.js';
-
 import {
   log as internalLogger,
   dumpTab,
+  toLines,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -53,6 +52,8 @@ function logUpdated(...args) {
   internalLogger('common/tabs-update', ...args);
 }
 
+let mAppIsActive = false;
+
 export function init() {
   browser.tabs.onActivated.addListener(onActivated);
   browser.tabs.onUpdated.addListener(onUpdated);
@@ -64,6 +65,10 @@ export function init() {
   browser.tabs.onDetached.addListener(onDetached);
   browser.windows.onCreated.addListener(onWindowCreated);
   browser.windows.onRemoved.addListener(onWindowRemoved);
+
+  browser.windows.getAll({}).then(windows => {
+    mAppIsActive = windows.some(window => window.focused);
+  });
 }
 
 let mPromisedStartedResolver;
@@ -138,7 +143,7 @@ async function onActivated(activeInfo) {
     const byTabDuplication = parseInt(window.duplicatingTabsCount) > 0;
 
     if (!Tab.isTracked(activeInfo.tabId))
-      await Tab.waitUntilTracked(activeInfo.tabId, { element: !!TabsStore.getWindow() });
+      await Tab.waitUntilTracked(activeInfo.tabId, { element: !!TabsStore.getCurrentWindowId() });
 
     const newActiveTab = Tab.get(activeInfo.tabId);
     if (!newActiveTab ||
@@ -158,12 +163,13 @@ async function onActivated(activeInfo) {
       return;
     }
 
-    let focusOverridden = Tab.onActivating.dispatch(newActiveTab, Object.assign({}, activeInfo, {
+    let focusOverridden = Tab.onActivating.dispatch(newActiveTab, {
+      ...activeInfo,
       byActiveTabRemove,
       byTabDuplication,
       byInternalOperation,
       silently
-    }));
+    });
     // don't do await if not needed, to process things synchronously
     if (focusOverridden instanceof Promise)
       focusOverridden = await focusOverridden;
@@ -179,13 +185,14 @@ async function onActivated(activeInfo) {
       return;
     }
 
-    const onActivatedReuslt = Tab.onActivated.dispatch(newActiveTab, Object.assign({}, activeInfo, {
+    const onActivatedReuslt = Tab.onActivated.dispatch(newActiveTab, {
+      ...activeInfo,
       oldActiveTabs,
       byActiveTabRemove,
       byTabDuplication,
       byInternalOperation,
       silently
-    }));
+    });
     // don't do await if not needed, to process things synchronously
     if (onActivatedReuslt instanceof Promise)
       await onActivatedReuslt;
@@ -207,11 +214,8 @@ async function onUpdated(tabId, changeInfo, tab) {
   if (mPromisedStarted)
     await mPromisedStarted;
 
-  TabIdFixer.fixTab(tab);
-  tabId = tab.id;
-
   if (!Tab.isTracked(tabId))
-    await Tab.waitUntilTracked(tabId, { element: !!TabsStore.getWindow() });
+    await Tab.waitUntilTracked(tabId, { element: !!TabsStore.getCurrentWindowId() });
 
   const [onCompleted, previous] = addTabOperationQueue();
   if (!configs.acceleratedTabOperations && previous)
@@ -228,8 +232,11 @@ async function onUpdated(tabId, changeInfo, tab) {
 
     logUpdated('tabs.onUpdated ', tabId, changeInfo, tab, updatedTab);
 
-    if ('url' in changeInfo)
+    const oldState = {};
+
+    if ('url' in changeInfo) {
       changeInfo.previousUrl = updatedTab.url;
+    }
     /*
       Updated openerTabId is not notified via tabs.onUpdated due to
       https://bugzilla.mozilla.org/show_bug.cgi?id=1409262 , so it can be
@@ -239,8 +246,11 @@ async function onUpdated(tabId, changeInfo, tab) {
       outside of TST (in other words, by any other addon.)
     */
     for (const key of Object.keys(changeInfo)) {
-      if (key != 'index')
-        updatedTab[key] = changeInfo[key];
+      if (key == 'index')
+        continue;
+      if ('key' in updatedTab)
+        oldState[key] = updatedTab[key];
+      updatedTab[key] = changeInfo[key];
     }
     if (changeInfo.url ||
         changeInfo.status == 'complete') {
@@ -261,7 +271,7 @@ async function onUpdated(tabId, changeInfo, tab) {
       updatedTab.$TST.updatedOpenerTabId = updatedTab.openerTabId = changeInfo.openerTabId = tab.openerTabId;
     }
 
-    TabsUpdate.updateTab(updatedTab, changeInfo, { tab });
+    TabsUpdate.updateTab(updatedTab, changeInfo, { tab, old: oldState });
 
     const onUpdatedResult = Tab.onUpdated.dispatch(updatedTab, changeInfo);
     // don't do await if not needed, to process things synchronously
@@ -323,6 +333,8 @@ async function onNewTabTracked(tab, info) {
   const duplicatedInternally = window.duplicatingTabsCount > 0;
   const maybeOrphan          = window.toBeOpenedOrphanTabs > 0;
   const activeTab            = Tab.getActiveTab(window.id);
+  const fromExternal         = !mAppIsActive;
+  const initialOpenerTabId   = tab.openerTabId;
 
   // New tab's index can become invalid because the value of "index" is same to
   // the one given to browser.tabs.create() (new tab) or the original index
@@ -338,6 +350,10 @@ async function onNewTabTracked(tab, info) {
   tab.index = Math.max(0, Math.min(tab.index, window.tabs.size));
   tab.reindexedBy = `onNewTabTracked (${tab.index})`;
 
+  // New tab from a bookmark always have its URL as the title
+  // (but the scheme part is missing.)
+  tab.$possibleInitialUrl = tab.title;
+
   // We need to track new tab after getting old active tab. Otherwise, this
   // operation updates the latest active tab in the window amd it becomes
   // impossible to know which tab was previously active.
@@ -353,7 +369,8 @@ async function onNewTabTracked(tab, info) {
     positionedBySelf,
     mayBeReplacedWithContainer,
     maybeOrphan,
-    activeTab
+    activeTab,
+    fromExternal
   });
 
   if (Tab.needToWaitTracked(tab.windowId, { exceptionTabId: tab.id }))
@@ -365,6 +382,23 @@ async function onNewTabTracked(tab, info) {
 
   log(`onNewTabTracked(${dumpTab(tab)}): start to create tab element`);
 
+  // Cached tree information may be expired when there are multiple new tabs
+  // opened at just same time and some of others are attached on listeners of
+  // "onCreating" and other points. Thus we need to refresh cached information
+  // dynamically.
+  // See also: https://github.com/piroor/treestyletab/issues/2419
+  let treeForActionDetection;
+  const onTreeModified = (_child, _info) => {
+    if (!treeForActionDetection)
+      return;
+    treeForActionDetection = Tree.snapshotForActionDetection(tab);
+    log('Tree modification is detected while waiting. Cached tree for action detection is updated: ', treeForActionDetection);
+  };
+  // We should refresh ceched information only when tabs are creaetd and
+  // attached, because the cacheed information was originally introduced for
+  // failsafe around problems from tabs closed while waiting.
+  Tree.onAttached.addListener(onTreeModified);
+
   try {
     tab = Tab.init(tab, { inBackground: false });
 
@@ -393,8 +427,21 @@ async function onNewTabTracked(tab, info) {
       forceApply: true
     });
 
-    // tabs can be removed and detached while waiting, so cache them here for `detectTabActionFromNewPosition()`.
-    const treeForActionDetection = Tree.snapshotForActionDetection(tab);
+    const duplicated = duplicatedInternally || uniqueId.duplicated;
+    const restored   = uniqueId.restored;
+    const skipFixupTree = !nextTab;
+
+    const maybeNeedToFixupTree = (
+      (info.mayBeReplacedWithContainer ||
+       (!duplicated &&
+        !restored &&
+        !skipFixupTree)) &&
+      !info.positionedBySelf
+    );
+    // Tabs can be removed and detached while waiting, so cache them here for `detectTabActionFromNewPosition()`.
+    // This operation takes too much time so it should be skipped if unnecessary.
+    // See also: https://github.com/piroor/treestyletab/issues/2278#issuecomment-521534290
+    treeForActionDetection = maybeNeedToFixupTree ? Tree.snapshotForActionDetection(tab) : null;
 
     if (positionedBySelf)
       window.toBeOpenedTabsWithPositions--;
@@ -403,8 +450,6 @@ async function onNewTabTracked(tab, info) {
     if (duplicatedInternally)
       window.duplicatingTabsCount--;
 
-    const duplicated = duplicatedInternally || uniqueId.duplicated;
-    const restored   = uniqueId.restored;
     if (restored) {
       window.restoredCount = window.restoredCount || 0;
       window.restoredCount++;
@@ -440,6 +485,7 @@ async function onNewTabTracked(tab, info) {
       onCompleted(uniqueId);
       Tab.untrack(tab.id);
       warnTabDestroyedWhileWaiting(tab.id, tab);
+      Tree.onAttached.removeListener(onTreeModified);
       return;
     }
 
@@ -450,7 +496,8 @@ async function onNewTabTracked(tab, info) {
       restored,
       duplicated,
       duplicatedInternally,
-      activeTab
+      activeTab,
+      fromExternal
     });
     // don't do await if not needed, to process things synchronously
     if (moved instanceof Promise)
@@ -480,6 +527,7 @@ async function onNewTabTracked(tab, info) {
       onCompleted(uniqueId);
       Tab.untrack(tab.id);
       warnTabDestroyedWhileWaiting(tab.id, tab);
+      Tree.onAttached.removeListener(onTreeModified);
       return;
     }
 
@@ -489,12 +537,13 @@ async function onNewTabTracked(tab, info) {
       positionedBySelf,
       mayBeReplacedWithContainer,
       movedBySelfWhileCreation: moved,
-      skipFixupTree: !nextTab,
+      skipFixupTree,
       restored,
       duplicated,
       duplicatedInternally,
       originalTab: duplicated && Tab.get(uniqueId.originalTabId),
-      treeForActionDetection
+      treeForActionDetection,
+      fromExternal
     });
     tab.$TST.resolveOpened();
 
@@ -502,6 +551,7 @@ async function onNewTabTracked(tab, info) {
       type:     Constants.kCOMMAND_NOTIFY_TAB_CREATED,
       windowId: tab.windowId,
       tabId:    tab.id,
+      active:   tab.active,
       maybeMoved: moved
     });
 
@@ -519,12 +569,33 @@ async function onNewTabTracked(tab, info) {
     const renewedTab = await browser.tabs.get(tab.id).catch(ApiTabs.createErrorHandler());
     if (!renewedTab)
       throw new Error(`tab ${tab.id} is closed while tracking`);
+    const updatedOpenerTabId = tab.openerTabId;
     const changedProps = {};
     for (const key of Object.keys(renewedTab)) {
       if (tab[key] != renewedTab[key])
         changedProps[key] = renewedTab[key];
     }
 
+    // When the active tab is duplicated, Firefox creates a duplicated tab
+    // with its `openerTabId` filled with the ID of the source tab.
+    // It is the `initialOpenerTabId`.
+    // On the other hand, TST may attach the duplicated tab to any other
+    // parent while it is initializing, based on a configuration
+    // `configs.autoAttachOnDuplicated`. It is the `updatedOpenerTabId`.
+    // At this scenario `renewedTab.openerTabId` becomes `initialOpenerTabId`
+    // and `updatedOpenerTabId` is lost.
+    // Thus we need to re-apply `updatedOpenerTabId` as the `openerTabId` of
+    // the tab again, to keep the tree structure managed by TST.
+    // See also: https://github.com/piroor/treestyletab/issues/2388
+    if (duplicated &&
+        tab.active &&
+        'openerTabId' in changedProps &&
+        changedProps.openerTabId == initialOpenerTabId &&
+        changedProps.openerTabId != updatedOpenerTabId) {
+      delete changedProps.openerTabId;
+      browser.tabs.update(tab.id, { openerTabId: updatedOpenerTabId });
+    }
+
     if (Object.keys(renewedTab).length > 0)
       onUpdated(tab.id, changedProps, renewedTab);
 
@@ -543,12 +614,15 @@ async function onNewTabTracked(tab, info) {
     if (tab.$TST.unsafeNextTab)
       tab.$TST.unsafeNextTab.$TST.memorizeNeighbors();
 
+    Tree.onAttached.removeListener(onTreeModified);
+
     return tab;
   }
   catch(e) {
     console.log(e, e.stack);
     onCompleted();
     tab.$TST.removeState(Constants.kTAB_STATE_CREATING);
+    Tree.onAttached.removeListener(onTreeModified);
   }
 }
 
@@ -607,11 +681,12 @@ async function onRemoved(tabId, removeInfo) {
 
     TabsStore.addRemovedTab(oldTab);
 
-    removeInfo = Object.assign({}, removeInfo, {
+    removeInfo = {
+      ...removeInfo,
       byInternalOperation,
       oldChildren: oldTab.$TST.children,
       oldParent:   oldTab.$TST.parent
-    });
+    };
 
     if (!removeInfo.isWindowClosing) {
       const closeParentBehavior = TreeBehavior.getCloseParentBehaviorForTabWithSidebarOpenState(oldTab, removeInfo);
@@ -629,9 +704,10 @@ async function onRemoved(tabId, removeInfo) {
       });
     }
 
-    const onRemovingResult = Tab.onRemoving.dispatch(oldTab, Object.assign({}, removeInfo, {
+    const onRemovingResult = Tab.onRemoving.dispatch(oldTab, {
+      ...removeInfo,
       byInternalOperation
-    }));
+    });
     // don't do await if not needed, to process things synchronously
     if (onRemovingResult instanceof Promise)
       await onRemovingResult;
@@ -643,7 +719,9 @@ async function onRemoved(tabId, removeInfo) {
     oldTab.$TST.addState(Constants.kTAB_STATE_REMOVING);
     TabsStore.addRemovingTab(oldTab);
 
-    TabsStore.windows.get(removeInfo.windowId).detachTab(oldTab.id);
+    TabsStore.windows.get(removeInfo.windowId).detachTab(oldTab.id, {
+      toBeRemoved: true
+    });
 
     const onRemovedReuslt = Tab.onRemoved.dispatch(oldTab, removeInfo);
     // don't do await if not needed, to process things synchronously
@@ -688,7 +766,7 @@ async function onMoved(tabId, moveInfo) {
   const maybeInternalOperation = window.internalMovingTabs.has(tabId);
 
   if (!Tab.isTracked(tabId))
-    await Tab.waitUntilTracked(tabId, { element: !!TabsStore.getWindow() });
+    await Tab.waitUntilTracked(tabId, { element: !!TabsStore.getCurrentWindowId() });
   if (Tab.needToWaitMoved(moveInfo.windowId))
     await Tab.waitUntilMovedAll(moveInfo.windowId);
 
@@ -715,8 +793,8 @@ async function onMoved(tabId, moveInfo) {
       return;
     }
 
-    let oldPreviousTab = movedTab.$TST.previousTab;
-    let oldNextTab     = movedTab.$TST.nextTab;
+    let oldPreviousTab = movedTab.hidden ? movedTab.$TST.unsafePreviousTab : movedTab.$TST.previousTab;
+    let oldNextTab     = movedTab.hidden ? movedTab.$TST.unsafeNextTab : movedTab.$TST.nextTab;
     if (movedTab.index != moveInfo.toIndex) { // already moved
       oldPreviousTab = Tab.getTabAt(moveInfo.windowId, moveInfo.toIndex < moveInfo.fromIndex ? moveInfo.fromIndex : moveInfo.fromIndex - 1);
       oldNextTab     = Tab.getTabAt(moveInfo.windowId, moveInfo.toIndex < moveInfo.fromIndex ? moveInfo.fromIndex + 1 : moveInfo.fromIndex);
@@ -728,13 +806,14 @@ async function onMoved(tabId, moveInfo) {
       alreadyMoved = true;
     }
 
-    const extendedMoveInfo = Object.assign({}, moveInfo, {
+    const extendedMoveInfo = {
+      ...moveInfo,
       byInternalOperation: maybeInternalOperation,
       alreadyMoved,
       oldPreviousTab,
       oldNextTab,
       isSubstantiallyMoved: movedTab.$TST.isSubstantiallyMoved
-    });
+    };
     log('tabs.onMoved: ', movedTab, extendedMoveInfo);
 
     let canceled = Tab.onMoving.dispatch(movedTab, extendedMoveInfo);
@@ -763,9 +842,8 @@ async function onMoved(tabId, moveInfo) {
         movedTab.reindexedBy = `tabs.onMoved (${movedTab.index})`;
         window.trackTab(movedTab);
         log('Tab nodes rearranged by tabs.onMoved listener:\n'+(!configs.debug ? '' :
-          Array.from(window.getOrderedTabs())
-            .map(tab => ' - '+tab.index+': '+tab.id+(tab.id == movedTab.id ? '[MOVED]' : ''))
-            .join('\n')),
+          toLines(Array.from(window.getOrderedTabs()),
+                  tab => ` - ${tab.index}: ${tab.id}${tab.id == movedTab.id ? '[MOVED]' : ''}`)),
             { moveInfo });
       }
       const onMovedResult = Tab.onMoved.dispatch(movedTab, extendedMoveInfo);
@@ -814,7 +892,7 @@ async function onAttached(tabId, attachInfo) {
 
   try {
     log('tabs.onAttached, id: ', tabId, attachInfo);
-    let tab = Tab.get(tabId);
+    const tab = Tab.get(tabId);
     const attachedTab = await browser.tabs.get(tabId).catch(ApiTabs.createErrorHandler());
     if (!attachedTab) {
       onCompleted();
@@ -826,13 +904,12 @@ async function onAttached(tabId, attachInfo) {
       tab.index    = attachedTab.index;
       tab.reindexedBy = `tabs.onAttached (${tab.index})`;
     }
-    else {
-      tab = attachedTab;
-      TabIdFixer.fixTab(tab);
-    }
 
     TabsInternalOperation.clearOldActiveStateInWindow(attachInfo.newWindowId);
-    const info = Object.assign({}, attachInfo, mTreeInfoForTabsMovingAcrossWindows.get(tabId));
+    const info = {
+      ...attachInfo,
+      ...mTreeInfoForTabsMovingAcrossWindows.get(tabId)
+    };
     mTreeInfoForTabsMovingAcrossWindows.delete(tabId);
 
     const window = TabsStore.windows.get(attachInfo.newWindowId);
@@ -884,11 +961,12 @@ async function onDetached(tabId, detachInfo) {
     if (byInternalOperation)
       oldWindow.toBeDetachedTabs.delete(tabId);
 
-    const info = Object.assign({}, detachInfo, {
+    const info = {
+      ...detachInfo,
       byInternalOperation,
       windowId:    detachInfo.oldWindowId,
       descendants: oldTab.$TST.descendants
-    });
+    };
     mTreeInfoForTabsMovingAcrossWindows.set(tabId, info);
 
     if (!byInternalOperation) // we should process only tabs detached by others.
@@ -902,8 +980,10 @@ async function onDetached(tabId, detachInfo) {
     });
 
     TabsStore.addRemovedTab(oldTab);
-    oldWindow.detachTab(oldTab.id);
-    if (!TabsStore.getWindow() && // only in the background page - the sidebar has no need to destroy itself manually.
+    oldWindow.detachTab(oldTab.id, {
+      toBeDetached: true
+    });
+    if (!TabsStore.getCurrentWindowId() && // only in the background page - the sidebar has no need to destroy itself manually.
         oldWindow.tabs &&
         oldWindow.tabs.size == 0) { // not destroyed yet case
       if (oldWindow.delayedDestroy)
@@ -945,7 +1025,7 @@ async function onWindowRemoved(windowId) {
     log('onWindowRemoved ', windowId);
     const window = TabsStore.windows.get(windowId);
     if (window &&
-        !TabsStore.getWindow()) // skip destructor on sidebar
+        !TabsStore.getCurrentWindowId()) // skip destructor on sidebar
       window.destroy();
 
     onCompleted();
@@ -956,3 +1036,7 @@ async function onWindowRemoved(windowId) {
   }
 }
 
+
+browser.windows.onFocusChanged.addListener(windowId => {
+  mAppIsActive = windowId > 0;
+});
