@@ -9,6 +9,7 @@ import {
   log as internalLogger,
   dumpTab,
   wait,
+  mapAndFilter,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -26,6 +27,8 @@ function log(...args) {
   internalLogger('background/background-cache', ...args);
 }
 
+const kCONTENTS_VERSION = 5;
+
 let mActivated = false;
 
 export function activate() {
@@ -49,7 +52,7 @@ export async function restoreWindowFromEffectiveWindowCache(windowId, options =
   cancelReservedCacheTree(windowId); // prevent to break cache before loading
   const tabs = options.tabs || await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
   if (configs.debug)
-    log(`restoreWindowFromEffectiveWindowCache for ${windowId} tabs: `, tabs.map(dumpTab));
+    log(`restoreWindowFromEffectiveWindowCache for ${windowId} tabs: `, () => tabs.map(dumpTab));
   const actualSignature = getWindowSignature(tabs);
   let cache = options.caches && options.caches.get(owner.id) || await MetricsData.addAsync('restoreWindowFromEffectiveWindowCache: window cache', getWindowCache(owner, Constants.kWINDOW_STATE_CACHED_TABS));
   if (!cache) {
@@ -94,7 +97,7 @@ export async function restoreWindowFromEffectiveWindowCache(windowId, options =
     cache, actualSignature, cachedSignature, signatureMatched
   });
   if (!cache ||
-      cache.version != Constants.kBACKGROUND_CONTENTS_VERSION ||
+      cache.version != kCONTENTS_VERSION ||
       !signatureMatched) {
     log(`restoreWindowFromEffectiveWindowCache for ${windowId}: no effective cache`);
     TabsInternalOperation.clearCache(owner);
@@ -150,7 +153,7 @@ function signatureFromTabsCache(cachedTabs) {
 
 async function restoreTabsFromCache(windowId, params = {}) {
   if (!params.cache ||
-      params.cache.version != Constants.kBACKGROUND_CONTENTS_VERSION)
+      params.cache.version != kCONTENTS_VERSION)
     return false;
 
   return (await restoreTabsFromCacheInternal({
@@ -195,7 +198,7 @@ async function fixupTabsRestoredFromCache(tabs, permanentStates, cachedTabs) {
   MetricsData.add('fixupTabsRestoredFromCache: start');
   if (tabs.length != cachedTabs.length)
     throw new Error(`fixupTabsRestoredFromCache: Mismatched number of tabs restored from cache, tabs=${tabs.length}, cachedTabs=${cachedTabs.length}`);
-  log('fixupTabsRestoredFromCache start ', { tabs: tabs.map(dumpTab), cachedTabs });
+  log('fixupTabsRestoredFromCache start ', () => ({ tabs: tabs.map(dumpTab), cachedTabs }));
   const idMap = new Map();
   // step 1: build a map from old id to new id
   tabs = tabs.map((tab, index) => {
@@ -216,24 +219,33 @@ async function fixupTabsRestoredFromCache(tabs, permanentStates, cachedTabs) {
   for (let i = tabs.length - 1; i > -1; i--) {
     fixupTabRestoredFromCache(tabs[i], permanentStates[i], cachedTabs[i], idMap);
   }
+  // step 3: restore collapsed/expanded state of tabs and finalize the
+  // restoration process
+  // Do this from top to bottom, because a tab can be placed under an
+  // expanded parent but the parent can be placed under a collapsed parent.
+  for (const tab of tabs) {
+    fixupTabRestoredFromCachePostProcess(tab);
+  }
   MetricsData.add('fixupTabsRestoredFromCache: step 2 done.');
 }
 
 function fixupTabRestoredFromCache(tab, permanentStates, cachedTab, idMap) {
   tab.$TST.clear();
-  tab.$TST.states = new Set([...cachedTab.$TST.states, ...permanentStates]);
+  const tabStates = new Set([...cachedTab.$TST.states, ...permanentStates]);
   for (const state of Constants.kTAB_TEMPORARY_STATES) {
-    tab.$TST.states.delete(state);
+    tabStates.delete(state);
   }
+  tab.$TST.states = tabStates;
   tab.$TST.attributes = cachedTab.$TST.attributes;
 
   log('fixupTabRestoredFromCache children: ', cachedTab.$TST.childIds);
-  const childTabs = cachedTab.$TST.childIds
-    .map(oldId => idMap.get(oldId))
-    .filter(tab => !!tab);
-  tab.$TST.children = childTabs;
-  if (childTabs.length > 0)
-    tab.$TST.setAttribute(Constants.kCHILDREN, `|${childTabs.map(tab => tab.id).join('|')}|`);
+  const childIds = mapAndFilter(cachedTab.$TST.childIds, oldId => {
+    const tab = idMap.get(oldId);
+    return tab && tab.id || undefined;
+  });
+  tab.$TST.children = childIds;
+  if (childIds.length > 0)
+    tab.$TST.setAttribute(Constants.kCHILDREN, `|${childIds.join('|')}|`);
   else
     tab.$TST.removeAttribute(Constants.kCHILDREN);
   log('fixupTabRestoredFromCache children: => ', tab.$TST.childIds);
@@ -246,7 +258,10 @@ function fixupTabRestoredFromCache(tab, permanentStates, cachedTab, idMap) {
   else
     tab.$TST.removeAttribute(Constants.kPARENT);
   log('fixupTabRestoredFromCache parent: => ', tab.$TST.parentId);
+}
 
+function fixupTabRestoredFromCachePostProcess(tab) {
+  const parentTab = tab.$TST.parent;
   if (parentTab &&
       (parentTab.$TST.collapsed ||
        parentTab.$TST.subtreeCollapsed)) {
@@ -373,7 +388,7 @@ async function cacheTree(windowId) {
     return;
   log('cacheTree for window ', windowId, { stack: configs.debug && new Error().stack });
   updateWindowCache(window.lastWindowCacheOwner, Constants.kWINDOW_STATE_CACHED_TABS, {
-    version:         Constants.kBACKGROUND_CONTENTS_VERSION,
+    version:         kCONTENTS_VERSION,
     tabs:            TabsStore.windows.get(windowId).export(true),
     pinnedTabsCount: Tab.getPinnedTabs(windowId).length,
     signature
