@@ -6,12 +6,12 @@
 'use strict';
 
 import RichConfirm from '/extlib/RichConfirm.js';
-import TabIdFixer from '/extlib/TabIdFixer.js';
 
 import {
   log as internalLogger,
   wait,
-  configs
+  configs,
+  sanitizeForHTMLText
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
@@ -23,6 +23,8 @@ import * as ContextualIdentities from '/common/contextual-identities.js';
 import * as Permissions from '/common/permissions.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as UserOperationBlocker from '/common/user-operation-blocker.js';
+import '/common/bookmark.js'; // we need to load this once in the background page to register the global listener
 
 import Tab from '/common/Tab.js';
 import Window from '/common/Window.js';
@@ -33,6 +35,7 @@ import * as Tree from './tree.js';
 import * as TreeStructure from './tree-structure.js';
 import * as BackgroundCache from './background-cache.js';
 import * as TabContextMenu from './tab-context-menu.js';
+import * as ContextMenu from './context-menu.js';
 import * as Migration from './migration.js';
 import './browser-action-menu.js';
 import './successor-tab.js';
@@ -97,6 +100,7 @@ export async function init() {
   EventListenerManager.debug = configs.debug;
 
   Migration.migrateConfigs();
+  Migration.migrateBookmarkUrls();
   configs.grantedRemovingTabIds = []; // clear!
   MetricsData.add('init: Migration.migrateConfigs');
 
@@ -110,18 +114,14 @@ export async function init() {
   log('init: Start to process messages including queued ones');
   ApiTabsListener.start();
 
-  // Open new tab now (after listening is started, before the end of initialization),
-  // because the sidebar may fail to track tabs.onCreated for the tab while its
-  // initializing process.
-  const promisedNotificationTab = Migration.notifyNewFeatures();
-  if (promisedNotificationTab)
-    await promisedNotificationTab;
+  Migration.tryNotifyNewFeatures();
 
   ContextualIdentities.startObserve();
   onBuilt.dispatch(); // after this line, this master process may receive "kCOMMAND_PING_TO_BACKGROUND" requests from sidebars.
   MetricsData.add('init: started listening');
 
   TabContextMenu.init();
+  ContextMenu.init();
   MetricsData.add('init: started initializing of context menu');
 
   Permissions.clearRequest();
@@ -177,7 +177,7 @@ async function notifyReadyToSidebars() {
 }
 
 function updatePanelUrl() {
-  const panel = browser.extension.getURL(`/sidebar/sidebar.html?style=${encodeURIComponent(configs.style)}`);
+  const panel = `${Constants.kSHORTHAND_URIS.tabbar}?style=${encodeURIComponent(configs.style)}`;
   browser.sidebarAction.setPanel({ panel });
 }
 
@@ -235,7 +235,6 @@ async function rebuildAll(windows) {
         trackedWindow = Window.init(window.id);
 
       for (const tab of window.tabs) {
-        TabIdFixer.fixTab(tab);
         Tab.track(tab);
         Tab.init(tab, { existing: true });
         tryStartHandleAccelKeyOnTab(tab);
@@ -341,7 +340,7 @@ async function updateInsertionPosition(tab) {
   if (!TabsStore.ensureLivingTab(tab))
     return;
 
-  const prev = tab.$TST.previousTab;
+  const prev = tab.hidden ? tab.$TST.unsafePreviousTab : tab.$TST.previousTab;
   if (prev)
     browser.sessions.setTabValue(
       tab.id,
@@ -354,7 +353,7 @@ async function updateInsertionPosition(tab) {
       Constants.kPERSISTENT_INSERT_AFTER
     ).catch(ApiTabs.createErrorSuppressor());
 
-  const next = tab.$TST.nextTab;
+  const next = tab.hidden ? tab.$TST.unsafeNextTab : tab.$TST.nextTab;
   if (next)
     browser.sessions.setTabValue(
       tab.id,
@@ -446,42 +445,88 @@ async function updateSubtreeCollapsed(tab) {
     tab.$TST.removeState(Constants.kTAB_STATE_SUBTREE_COLLAPSED, { permanently: true });
 }
 
-export async function confirmToCloseTabs(tabs, options = {}) {
-  tabs = tabs.filter(tab => !configs.grantedRemovingTabIds.includes(tab.id));
-  const tabIds = tabs.map(tab => tab.id);
-  const count = tabIds.length;
-  log('confirmToCloseTabs ', { tabIds, count, options });
+export async function confirmToCloseTabs(tabs, { windowId, configKey, messageKey, titleKey } = {}) {
+  const grantedIds = new Set(configs.grantedRemovingTabIds);
+  let count = 0;
+  const tabIds = [];
+  tabs = tabs.map(tab => tab && Tab.get(tab.id)).filter(tab => {
+    if (tab && !grantedIds.has(tab.id)) {
+      count++;
+      tabIds.push(tab.id);
+      return true;
+    }
+    return false;
+  });
+  const shouldConfirm = configs[configKey || 'warnOnCloseTabs'];
+  log('confirmToCloseTabs ', { tabIds, count, windowId, configKey, grantedIds, shouldConfirm });
   if (count <= 1 ||
-      !configs.warnOnCloseTabs ||
-      Date.now() - configs.lastConfirmedToCloseTabs < 500)
+      !shouldConfirm ||
+      Date.now() - configs.lastConfirmedToCloseTabs < 500) {
+    log('confirmToCloseTabs: skip confirmation and treated as granted');
     return true;
+  }
 
-  const activeTabs = await browser.tabs.query({
-    active:   true,
-    windowId: options.windowId
-  }).catch(ApiTabs.createErrorHandler());
-
-  const granted = await Permissions.isGranted(Permissions.ALL_URLS);
-  if (!granted ||
-      /^(about|chrome|resource):/.test(activeTabs[0].url) ||
-      (!options.showInTab &&
-       SidebarConnection.isOpen(options.windowId) &&
-       SidebarConnection.hasFocus(options.windowId)))
-    return browser.runtime.sendMessage({
-      type: Constants.kCOMMAND_CONFIRM_TO_CLOSE_TABS,
-      tabs,
-      windowId: options.windowId
+  if (!windowId) {
+    const activeTabs = await browser.tabs.query({
+      active:   true,
+      windowId
     }).catch(ApiTabs.createErrorHandler());
+    windowId = activeTabs[0].windowId;
+  }
 
-  const result = await RichConfirm.showInTab(activeTabs[0].id, {
-    message: browser.i18n.getMessage('warnOnCloseTabs_message', [count]),
+  const listing = configs.warnOnCloseTabsWithListing ?
+    tabsToHTMLList(tabs, { maxRows: configs.warnOnCloseTabsWithListingMaxRows }) :
+    '';
+  const dialogParams = {
+    content: `
+      <div>${sanitizeForHTMLText(browser.i18n.getMessage(messageKey || 'warnOnCloseTabs_message', [count]))}</div>${listing}
+    `.trim(),
     buttons: [
       browser.i18n.getMessage('warnOnCloseTabs_close'),
       browser.i18n.getMessage('warnOnCloseTabs_cancel')
     ],
     checkMessage: browser.i18n.getMessage('warnOnCloseTabs_warnAgain'),
     checked: true
-  });
+  };
+  let result;
+  UserOperationBlocker.blockIn(windowId, { throbber: false });
+  try {
+    if (configs.showDialogInSidebar &&
+        SidebarConnection.isOpen(windowId)/* &&
+        SidebarConnection.hasFocus(windowId)*/) {
+      result = await browser.runtime.sendMessage({
+        type:   Constants.kCOMMAND_SHOW_DIALOG,
+        params: dialogParams,
+        windowId
+      }).catch(ApiTabs.createErrorHandler());
+    }
+    else {
+      log('confirmToCloseTabs: show confirmation in a popup window on ', windowId);
+      result = await RichConfirm.showInPopup(windowId, {
+        ...dialogParams,
+        onShown(container) {
+          setTimeout(() => { // this need to be done on the next tick, to use the height of the box for     calculation of dialog size
+            const style = container.querySelector('ul').style;
+            style.height = '0px'; // this makes the box shrinkable
+            style.maxHeight = 'none';
+            style.minHeight = '0px';
+          }, 0);
+        },
+        modal: true,
+        type:  'common-dialog',
+        url:   '/resources/blank.html', // required on Firefox ESR68
+        title: browser.i18n.getMessage(titleKey || 'warnOnCloseTabs_title')
+      });
+    }
+  }
+  catch(_error) {
+    result = { buttonIndex: -1 };
+  }
+  finally {
+    UserOperationBlocker.unblockIn(windowId, { throbber: false });
+  }
+
+  log('confirmToCloseTabs: result = ', result);
   switch (result.buttonIndex) {
     case 0:
       if (!result.checked)
@@ -495,9 +540,41 @@ export async function confirmToCloseTabs(tabs, options = {}) {
   }
 }
 Commands.onTabsClosing.addListener((tabIds, options = {}) => {
-  return confirmToCloseTabs(tabIds, options);
+  return confirmToCloseTabs(tabIds.map(Tab.get), options);
 });
 
+export function tabsToHTMLList(tabs, { maxRows }) {
+  const rootLevelOffset = tabs.map(tab => parseInt(tab.$TST.getAttribute(Constants.kLEVEL) || 0)).sort()[0];
+  return (
+    `<ul style="border: 1px inset;
+                display: flex;
+                flex-direction: column;
+                flex-grow: 1;
+                flex-shrink: 1;
+                margin: 0.5em 0;
+                min-height: ${(Math.max(1, maxRows || 0)) + 1}em;
+                max-height: ${(Math.max(1, maxRows || 0)) + 1}em;
+                overflow: auto;
+                padding: 0.5em;">` +
+      tabs.map(tab => `<li style="align-items: center;
+                                  display: flex;
+                                  flex-direction: row;
+                                  padding-left: calc((${tab.$TST.getAttribute(Constants.kLEVEL)} - ${rootLevelOffset}) * 0.25em);"
+                           title="${sanitizeForHTMLText(tab.title)}"
+                          ><img style="display: flex;
+                                       max-height: 1em;
+                                       max-width: 1em;"
+                                alt=""
+                                src="${sanitizeForHTMLText(tab.favIconUrl || browser.extension.getURL('resources/icons/globe-16.svg'))}"
+                               ><span style="display: flex;
+                                             margin-left: 0.25em;
+                                             overflow: hidden;
+                                             white-space: nowrap;"
+                                     >${sanitizeForHTMLText(tab.title)}</span></li>`).join('') +
+      `</ul>`
+  );
+}
+
 function reserveToClearGrantedRemovingTabs() {
   const lastGranted = configs.grantedRemovingTabIds.join(',');
   setTimeout(() => {
@@ -515,8 +592,8 @@ Tab.onCreated.addListener((tab, info = {}) => {
   reserveToUpdateChildren(tab);
   reserveToUpdateInsertionPosition([
     tab,
-    tab.$TST.nextTab,
-    tab.$TST.previousTab
+    tab.hidden ? tab.$TST.unsafePreviousTab : tab.$TST.previousTab,
+    tab.hidden ? tab.$TST.unsafeNextTab : tab.$TST.nextTab
   ]);
 });
 
@@ -545,11 +622,77 @@ Tab.onUpdated.addListener((tab, changeInfo) => {
     tryStartHandleAccelKeyOnTab(tab);
 });
 
+Tab.onShown.addListener(tab => {
+  if (configs.fixupTreeOnTabVisibilityChanged) {
+    reserveToUpdateAncestors(tab);
+    reserveToUpdateChildren(tab);
+  }
+  reserveToUpdateInsertionPosition([
+    tab,
+    tab.hidden ? tab.$TST.unsafePreviousTab : tab.$TST.previousTab,
+    tab.hidden ? tab.$TST.unsafeNextTab : tab.$TST.nextTab
+  ]);
+});
+
+Tab.onMutedStateChanged.addListener((root, toBeMuted) => {
+  // Spread muted state of a parent tab to its collapsed descendants
+  if (!root.$TST.subtreeCollapsed ||
+      // We don't need to spread muted state to descendants of multiselected
+      // tabs here, because tabs.update() was called with all multiselected tabs.
+      root.$TST.multiselected ||
+      // We should not spread muted state to descendants of collapsed tab
+      // recursively, because they were already controlled from a visible
+      // ancestor.
+      root.$TST.collapsed)
+    return;
+
+  const tabs = root.$TST.descendants;
+  for (const tab of tabs) {
+    const playing = tab.$TST.soundPlaying;
+    const muted   = tab.$TST.muted;
+    log(`tab ${tab.id}: playing=${playing}, muted=${muted}`);
+    if (configs.spreadMutedStateOnlyToSoundPlayingTabs &&
+        !playing &&
+        playing != toBeMuted)
+      continue;
+
+    log(` => set muted=${toBeMuted}`);
+    browser.tabs.update(tab.id, {
+      muted: toBeMuted
+    }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+
+    const add = [];
+    const remove = [];
+    if (toBeMuted) {
+      add.push(Constants.kTAB_STATE_MUTED);
+      tab.$TST.addState(Constants.kTAB_STATE_MUTED);
+    }
+    else {
+      remove.push(Constants.kTAB_STATE_MUTED);
+      tab.$TST.removeState(Constants.kTAB_STATE_MUTED);
+    }
+
+    if (tab.audible && !toBeMuted) {
+      add.push(Constants.kTAB_STATE_SOUND_PLAYING);
+      tab.$TST.addState(Constants.kTAB_STATE_SOUND_PLAYING);
+    }
+    else {
+      remove.push(Constants.kTAB_STATE_SOUND_PLAYING);
+      tab.$TST.removeState(Constants.kTAB_STATE_SOUND_PLAYING);
+    }
+
+    // tabs.onUpdated is too slow, so users will be confused
+    // from still-not-updated tabs (in other words, they tabs
+    // are unresponsive for quick-clicks).
+    Tab.broadcastState(tab, { add, remove });
+  }
+});
+
 Tab.onTabInternallyMoved.addListener((tab, info = {}) => {
   reserveToUpdateInsertionPosition([
     tab,
-    tab.$TST.previousTab,
-    tab.$TST.nextTab,
+    tab.hidden ? tab.$TST.unsafePreviousTab : tab.$TST.previousTab,
+    tab.hidden ? tab.$TST.unsafeNextTab : tab.$TST.nextTab,
     info.oldPreviousTab,
     info.oldNextTab
   ]);
@@ -562,8 +705,8 @@ Tab.onMoved.addListener((tab, moveInfo) => {
     tab,
     moveInfo.oldPreviousTab,
     moveInfo.oldNextTab,
-    tab.$TST.previousTab,
-    tab.$TST.nextTab
+    tab.hidden ? tab.$TST.unsafePreviousTab : tab.$TST.previousTab,
+    tab.hidden ? tab.$TST.unsafeNextTab : tab.$TST.nextTab
   ]);
 });
 
@@ -587,7 +730,7 @@ Tree.onSubtreeCollapsedStateChanging.addListener((tab, _info) => { reserveToUpda
 //   https://bugzilla.mozilla.org/show_bug.cgi?id=1367042
 function applyThemeColorToIcon() {
   if (configs.applyThemeColorToIcon) {
-    const icons = { path: browser.runtime.getManifest().variable_color_icons };
+    const icons = { path: browser.runtime.getManifest().icons };
     browser.browserAction.setIcon(icons);
     browser.sidebarAction.setIcon(icons);
   }
