@@ -74,12 +74,6 @@ const mItems = [
             value: 'plain',
             type:  'radio'
           },
-          {
-            title: browser.i18n.getMessage('config_style_plain_dark'),
-            key:   'style',
-            value: 'plain-dark',
-            type:  'radio'
-          },
           {
             title: browser.i18n.getMessage('config_style_vertigo'),
             key:   'style',
@@ -92,12 +86,6 @@ const mItems = [
             value: 'mixed',
             type:  'radio'
           },
-          {
-            title: browser.i18n.getMessage('config_style_metal'),
-            key:   'style',
-            value: 'metal',
-            type:  'radio'
-          },
           {
             title: browser.i18n.getMessage('config_style_sidebar'),
             key:   'style',
@@ -118,6 +106,11 @@ const mItems = [
           }
         ]
       },
+      {
+        title: browser.i18n.getMessage('config_animation_label'),
+        key:   'animation',
+        type:  'checkbox'
+      },
       {
         title:    browser.i18n.getMessage('config_labelOverflowStyle_caption'),
         children: [
@@ -133,26 +126,30 @@ const mItems = [
             value: 'crop',
             type:  'radio'
           }
-        ]
+        ],
+        expert: true
       },
       {
         title: browser.i18n.getMessage('config_faviconizePinnedTabs_label'),
         key:   'faviconizePinnedTabs',
-        type:  'checkbox'
-      },
-      {
-        title: browser.i18n.getMessage('config_animation_label'),
-        key:   'animation',
-        type:  'checkbox'
+        type:  'checkbox',
+        expert: true
       },
       {
         title: browser.i18n.getMessage('config_showCollapsedDescendantsByTooltip_label'),
         key:   'showCollapsedDescendantsByTooltip',
-        type:  'checkbox'
+        type:  'checkbox',
+        expert: true
       },
       {
         title: browser.i18n.getMessage('config_applyThemeColorToIcon_label'),
         key:   'applyThemeColorToIcon',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: browser.i18n.getMessage('config_showDialogInSidebar_label'),
+        key:   'showDialogInSidebar',
         type:  'checkbox'
       }
     ]
@@ -161,7 +158,7 @@ const mItems = [
     title:    browser.i18n.getMessage('config_context_caption'),
     children: [
       {
-        title: browser.i18n.getMessage('config_extraItems_topLevel'),
+        title: browser.i18n.getMessage('config_extraItems_tabs_topLevel'),
         enabled: false
       },
       {
@@ -194,6 +191,11 @@ const mItems = [
         key:   'context_topLevel_collapseTree',
         type:  'checkbox'
       },
+      {
+        title: indent() + browser.i18n.getMessage('context_collapseTreeRecursively_label'),
+        key:   'context_topLevel_collapseTreeRecursively',
+        type:  'checkbox'
+      },
       {
         title: indent() + browser.i18n.getMessage('context_collapseAll_label'),
         key:   'context_topLevel_collapseAll',
@@ -204,6 +206,11 @@ const mItems = [
         key:   'context_topLevel_expandTree',
         type:  'checkbox'
       },
+      {
+        title: indent() + browser.i18n.getMessage('context_expandTreeRecursively_label'),
+        key:   'context_topLevel_expandTreeRecursively',
+        type:  'checkbox'
+      },
       {
         title: indent() + browser.i18n.getMessage('context_expandAll_label'),
         key:   'context_topLevel_expandAll',
@@ -221,7 +228,7 @@ const mItems = [
       },
       { type: 'separator' },
       {
-        title: browser.i18n.getMessage('config_extraItems_subMenu'),
+        title: browser.i18n.getMessage('config_extraItems_tabs_subMenu'),
         enabled: false
       },
       {
@@ -254,6 +261,11 @@ const mItems = [
         key:   'context_collapseTree',
         type:  'checkbox'
       },
+      {
+        title: indent() + browser.i18n.getMessage('context_collapseTreeRecursively_label'),
+        key:   'context_collapseTreeRecursively',
+        type:  'checkbox'
+      },
       {
         title: indent() + browser.i18n.getMessage('context_collapseAll_label'),
         key:   'context_collapseAll',
@@ -264,6 +276,11 @@ const mItems = [
         key:   'context_expandTree',
         type:  'checkbox'
       },
+      {
+        title: indent() + browser.i18n.getMessage('context_expandTreeRecursively_label'),
+        key:   'context_expandTreeRecursively',
+        type:  'checkbox'
+      },
       {
         title: indent() + browser.i18n.getMessage('context_expandAll_label'),
         key:   'context_expandAll',
@@ -278,6 +295,27 @@ const mItems = [
         title: indent() + browser.i18n.getMessage('context_groupTabs_label'),
         key:   'context_groupTabs',
         type:  'checkbox'
+      },
+      { type: 'separator' },
+      {
+        title: browser.i18n.getMessage('config_extraItems_bookmarks_caption'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_openAllBookmarksWithStructure_label'),
+        key:   'context_openAllBookmarksWithStructure',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('context_openAllBookmarksWithStructureRecursively_label'),
+        key:   'context_openAllBookmarksWithStructureRecursively',
+        type:  'checkbox'
+      },
+      { type: 'separator' },
+      {
+        title: indent() + browser.i18n.getMessage('config_openAllBookmarksWithStructureDiscarded_label'),
+        key:   'openAllBookmarksWithStructureDiscarded',
+        type:  'checkbox'
       }
     ]
   },
@@ -285,7 +323,53 @@ const mItems = [
     title:    browser.i18n.getMessage('config_newTab_caption'),
     children: [
       {
-        title: browser.i18n.getMessage('config_newTabButton_caption')
+        title: browser.i18n.getMessage('config_newTabAction_caption'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_before'),
+        children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_OPEN_AS_ORPHAN,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_child') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_OPEN_AS_CHILD,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_sibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_OPEN_AS_SIBLING,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
+            key:   'autoAttachOnNewTabCommand',
+            value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
+            type:  'radio'
+          }
+        ]
+      },
+      {
+        title: browser.i18n.getMessage('config_inheritContextualIdentityToNewChildTab_label'),
+        key:   'inheritContextualIdentityToNewChildTab',
+        type:  'checkbox'
+      },
+      { type: 'separator' },
+      {
+        title: browser.i18n.getMessage('config_newTabButton_caption'),
+        enabled: false
       },
       {
         title: indent() + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_before'),
@@ -319,6 +403,12 @@ const mItems = [
             key:   'autoAttachOnNewTabButtonMiddleClick',
             value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
             type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_nextSiblingWithInheritedContainer') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabButtonMiddleClick_after'),
+            key:   'autoAttachOnNewTabButtonMiddleClick',
+            value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING_WITH_INHERITED_CONTAINER,
+            type:  'radio'
           }
         ]
       },
@@ -357,44 +447,9 @@ const mItems = [
       },
       { type: 'separator' },
       {
-        title: browser.i18n.getMessage('config_newTabAction_caption'),
+        title: browser.i18n.getMessage('config_autoAttachWithURL_caption'),
         enabled: false
       },
-      {
-        title:    indent() + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_before'),
-        children: [
-          {
-            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
-            key:   'autoAttachOnNewTabCommand',
-            value: Constants.kNEWTAB_DO_NOTHING,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
-            key:   'autoAttachOnNewTabCommand',
-            value: Constants.kNEWTAB_OPEN_AS_ORPHAN,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_child') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
-            key:   'autoAttachOnNewTabCommand',
-            value: Constants.kNEWTAB_OPEN_AS_CHILD,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_sibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
-            key:   'autoAttachOnNewTabCommand',
-            value: Constants.kNEWTAB_OPEN_AS_SIBLING,
-            type:  'radio'
-          },
-          {
-            title: browser.i18n.getMessage('config_autoAttachOnNewTabCommand_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnNewTabCommand_after'),
-            key:   'autoAttachOnNewTabCommand',
-            value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
-            type:  'radio'
-          }
-        ]
-      },
       {
         title:    indent() + browser.i18n.getMessage('config_autoAttachOnDuplicated_before'),
         children: [
@@ -431,13 +486,7 @@ const mItems = [
         ]
       },
       {
-        title: browser.i18n.getMessage('config_inheritContextualIdentityToNewChildTab_label'),
-        key:   'inheritContextualIdentityToNewChildTab',
-        type:  'checkbox'
-      },
-      { type: 'separator' },
-      {
-        title:    browser.i18n.getMessage('config_sameSiteOrphan_caption'),
+        title:    indent() + browser.i18n.getMessage('config_sameSiteOrphan_caption'),
         children: [
           {
             title: browser.i18n.getMessage('config_autoAttachSameSiteOrphan_before') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachSameSiteOrphan_after'),
@@ -477,6 +526,48 @@ const mItems = [
           }
         ]
       },
+      {
+        title:    indent() + browser.i18n.getMessage('config_fromExternal_caption'),
+        children: [
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_before') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_noControl') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_after'),
+            key:   'autoAttachOnOpenedFromExternal',
+            value: Constants.kNEWTAB_DO_NOTHING,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_before') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_independent') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_after'),
+            key:   'autoAttachOnOpenedFromExternal',
+            value: Constants.kNEWTAB_OPEN_AS_ORPHAN,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_before') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_child') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_after'),
+            key:   'autoAttachOnOpenedFromExternal',
+            value: Constants.kNEWTAB_OPEN_AS_CHILD,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_before') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_sibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_after'),
+            key:   'autoAttachOnOpenedFromExternal',
+            value: Constants.kNEWTAB_OPEN_AS_SIBLING,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_before') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_nextSibling') + delimiter + browser.i18n.getMessage('config_autoAttachOnOpenedFromExternal_after'),
+            key:   'autoAttachOnOpenedFromExternal',
+            value: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
+            type:  'radio'
+          },
+          { type: 'separator' },
+          {
+            title: browser.i18n.getMessage('config_inheritContextualIdentityToTabsFromExternal_label'),
+            key:   'inheritContextualIdentityToTabsFromExternal',
+            type:  'checkbox'
+          }
+        ]
+      },
+      { type: 'separator' },
       {
         title:    browser.i18n.getMessage('config_insertNewChildAt_caption'),
         children: [
@@ -529,22 +620,43 @@ const mItems = [
         enabled: false
       },
       {
-        dynamicTitle: true,
-        get title() {
-          return indent() + browser.i18n.getMessage('config_autoGroupNewTabs_before') + delimiter + configs.autoGroupNewTabsTimeout + delimiter + browser.i18n.getMessage('config_autoGroupNewTabs_after');
-        },
-        key:   'autoGroupNewTabs',
-        type:  'checkbox'
-      },
-      {
-        title: indent() + browser.i18n.getMessage('config_insertNewTabFromPinnedTabAt_caption'),
+        title: indent() + browser.i18n.getMessage('config_autoGroupNewTabsFromPinned_label'),
         key:   'autoGroupNewTabsFromPinned',
         type:  'checkbox'
       },
+      {
+        dynamicTitle: true,
+        get title() {
+          return indent() + browser.i18n.getMessage('config_autoGroupNewTabsFromBookmarks_before') + delimiter + configs.autoGroupNewTabsTimeout + delimiter + browser.i18n.getMessage('config_autoGroupNewTabsFromBookmarks_after');
+        },
+        key:   'autoGroupNewTabsFromBookmarks',
+        type:  'checkbox'
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_autoGroupNewTabsFromOthers_label'),
+        key:   'autoGroupNewTabsFromOthers',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: indent(2) + browser.i18n.getMessage('config_warnOnAutoGroupNewTabs_label'),
+        key:   'warnOnAutoGroupNewTabs',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: indent(4) + browser.i18n.getMessage('config_warnOnAutoGroupNewTabsWithListing_label'),
+        key:   'warnOnAutoGroupNewTabsWithListing',
+        type:  'checkbox',
+        expert: true
+      },
+      { type: 'separator',
+        expert: true },
       {
         title: indent() + browser.i18n.getMessage('config_renderTreeInGroupTabs_label'),
         key:   'renderTreeInGroupTabs',
-        type:  'checkbox'
+        type:  'checkbox',
+        expert: true
       }
     ]
   },
@@ -564,12 +676,48 @@ const mItems = [
       {
         title: indent() + browser.i18n.getMessage('config_autoCollapseExpandSubtreeOnSelectExceptActiveTabRemove_label'),
         key:   'autoCollapseExpandSubtreeOnSelectExceptActiveTabRemove',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: browser.i18n.getMessage('config_unfocusableCollapsedTab_label'),
+        key:   'unfocusableCollapsedTab',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: browser.i18n.getMessage('config_autoDiscardTabForUnexpectedFocus_label'),
+        key:   'autoDiscardTabForUnexpectedFocus',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: browser.i18n.getMessage('config_avoidDiscardedTabToBeActivatedIfPossible_label'),
+        key:   'avoidDiscardedTabToBeActivatedIfPossible',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('config_collapseExpandSubtreeByDblClick_label'),
-        key:   'collapseExpandSubtreeByDblClick',
-        type:  'checkbox'
+        title:    browser.i18n.getMessage('config_treeDoubleClickBehavior_caption'),
+        children: [
+          {
+            title: browser.i18n.getMessage('config_treeDoubleClickBehavior_toggleCollapsed'),
+            key:   'treeDoubleClickBehavior',
+            value: Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_TOGGLE_COLLAPSED,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_treeDoubleClickBehavior_close'),
+            key:   'treeDoubleClickBehavior',
+            value: Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_CLOSE,
+            type:  'radio'
+          },
+          {
+            title: browser.i18n.getMessage('config_treeDoubleClickBehavior_none'),
+            key:   'treeDoubleClickBehavior',
+            value: Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_NONE,
+            type:  'radio'
+          }
+        ]
       },
       {
         title:    browser.i18n.getMessage('config_successorTabControlLevel_caption'),
@@ -811,6 +959,12 @@ const mItems = [
           }
         ]
       },
+      {
+        title: indent() + browser.i18n.getMessage('config_treatTreeAsExpandedOnClosedWithNoSidebar_label'),
+        key:   'treatTreeAsExpandedOnClosedWithNoSidebar',
+        type:  'checkbox',
+        expert: true
+      },
       {
         title:    browser.i18n.getMessage('config_fixupTreeOnTabVisibilityChanged_caption'),
         children: [
@@ -833,6 +987,13 @@ const mItems = [
   {
     title:    browser.i18n.getMessage('config_drag_caption'),
     children: [
+      {
+        title:  browser.i18n.getMessage('config_moveDroppedTabToNewWindowForUnhandledDragEvent_label'),
+        key:    'moveDroppedTabToNewWindowForUnhandledDragEvent',
+        type:   'checkbox',
+        expert: true
+      },
+      { type: 'separator', expert: true },
       {
         title: browser.i18n.getMessage('config_tabDragBehavior_caption'),
         enabled: false
@@ -907,11 +1068,6 @@ const mItems = [
           }
         ]
       },
-      {
-        title: browser.i18n.getMessage('config_showTabDragHandle_label'),
-        key:   'showTabDragHandle',
-        type:  'checkbox'
-      },
       {
         title: browser.i18n.getMessage('config_showTabDragBehaviorNotification_label'),
         key:   'showTabDragBehaviorNotification',
@@ -940,42 +1096,112 @@ const mItems = [
             type:  'radio'
           }
         ]
+      },
+      { type: 'separator', expert: true },
+      {
+        dynamicTitle: true,
+        get title() {
+          return browser.i18n.getMessage('config_autoExpandOnLongHoverDelay_before') + delimiter + configs.autoExpandOnLongHoverDelay + delimiter + browser.i18n.getMessage('config_autoExpandOnLongHoverDelay_after');
+        },
+        key:    'autoExpandOnLongHover',
+        type:   'checkbox',
+        expert: true
+      },
+      {
+        title:   indent() + browser.i18n.getMessage('config_autoExpandOnLongHoverRestoreIniitalState_label'),
+        key:    'autoExpandOnLongHoverRestoreIniitalState',
+        type:   'checkbox',
+        expert: true
+      },
+      {
+        title:   indent() + browser.i18n.getMessage('config_autoExpandIntelligently_label'),
+        key:    'autoExpandIntelligently',
+        type:   'checkbox',
+        expert: true
       }
     ]
   },
   {
-    title:    browser.i18n.getMessage('config_advanced_caption') + ' / ' + browser.i18n.getMessage('config_debug_caption'),
+    title:    browser.i18n.getMessage('config_more_caption'),
     children: [
       {
-        title: browser.i18n.getMessage('config_warnOnCloseTabs_label'),
+        title:   browser.i18n.getMessage('config_shortcuts_caption'),
+        enabled: false,
+        expert:  true
+      },
+      {
+        dynamicTitle: true,
+        get title() {
+          return indent() + browser.i18n.getMessage('config_autoExpandOnTabSwitchingShortcutsDelay_before') + delimiter + configs.autoExpandOnTabSwitchingShortcutsDelay + delimiter + browser.i18n.getMessage('config_autoExpandOnTabSwitchingShortcutsDelay_after');
+        },
+        key:   'autoExpandOnTabSwitchingShortcuts',
+        type:  'checkbox',
+        expert: true
+      },
+      { type: 'separator', expert: true },
+      {
+        title:   browser.i18n.getMessage('config_advanced_caption'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_warnOnCloseTabs_label'),
         key:   'warnOnCloseTabs',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('config_useCachedTree_label'),
+        title: indent(2) + browser.i18n.getMessage('config_warnOnCloseTabsByClosebox_label'),
+        key:   'warnOnCloseTabsByClosebox',
+        type:  'checkbox',
+      },
+      {
+        title: indent(2) + browser.i18n.getMessage('config_warnOnCloseTabsWithListing_label'),
+        key:   'warnOnCloseTabsWithListing',
+        type:  'checkbox',
+        expert: true
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_useCachedTree_label'),
         key:   'useCachedTree',
         type:  'checkbox'
       },
       {
-        title: browser.i18n.getMessage('config_simulateCloseTabByDblclick_label'),
-        key:   'simulateCloseTabByDblclick',
-        type:  'checkbox'
+        title: indent(2) + browser.i18n.getMessage('config_storeCacheAsWindowValue_label'),
+        key:   'storeCacheAsWindowValue',
+        type:  'checkbox',
+        expert: true
       },
       {
-        title: browser.i18n.getMessage('config_supportTabsMultiselect_label'),
+        title: indent() + browser.i18n.getMessage('config_supportTabsMultiselect_label'),
         key:   'supportTabsMultiselect',
-        type:  'checkbox'
+        type:  'checkbox',
+        expert: true
       },
       {
-        title: browser.i18n.getMessage('config_loggingQueries_label'),
+        title: indent() + browser.i18n.getMessage('config_undoMultipleTabsClose_label'),
+        key:   'undoMultipleTabsClose',
+        type:  'checkbox'
+      },
+      { type: 'separator' },
+      {
+        title:   browser.i18n.getMessage('config_debug_caption'),
+        enabled: false
+      },
+      {
+        title: indent() + browser.i18n.getMessage('config_loggingQueries_label'),
         key:   'loggingQueries',
         type:  'checkbox'
       },
       { type: 'separator' },
       {
-        title: browser.i18n.getMessage('config_debug_label'),
+        title: indent() + browser.i18n.getMessage('config_debug_label'),
         key:   'debug',
         type:  'checkbox'
+      },
+      { type: 'separator' },
+      {
+        title: browser.i18n.getMessage('config_showExpertOptions_label'),
+        key:   'showExpertOptions',
+        type:  'checkbox'
       }
     ]
   }
@@ -983,6 +1209,7 @@ const mItems = [
 
 const mItemsById = new Map();
 const mUpdatableItemsById = new Map();
+const mExpertItems = new Set();
 
 function createItem(id, item, parent) {
   if (item.visible === false)
@@ -1010,7 +1237,9 @@ function createItem(id, item, parent) {
   if ('enabled' in item)
     params.enabled = item.enabled;
   log('create: ', params);
-  browser.menus.create(params);
+  id = browser.menus.create(params);
+  if (item.expert)
+    mExpertItems.add(id);
   if (item.children) {
     for (let i = 0, maxi = item.children.length; i < maxi; i++) {
       const child = item.children[i];
@@ -1058,3 +1287,16 @@ browser.menus.onClicked.addListener((info, _tab) => {
 
   configs[item.key] = 'value' in item ? item.value : !configs[item.key];
 });
+
+
+function updateExpertOptionsVisibility() {
+  for (const id of mExpertItems) {
+    browser.menus.update(id, { visible: configs.showExpertOptions });
+  }
+  browser.menus.refresh();
+}
+configs.$addObserver(key => {
+  if (key == 'showExpertOptions')
+    updateExpertOptionsVisibility();
+});
+configs.$loaded.then(updateExpertOptionsVisibility);
