@@ -40,6 +40,8 @@ import * as BackgroundConnection from './background-connection.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
+import { TabInvalidationTarget } from './components/TabElement.js';
+
 function log(...args) {
   internalLogger('sidebar/collapse-expand', ...args);
 }
@@ -193,30 +195,40 @@ export function setCollapsed(tab, info = {}) {
   });
 }
 
+const BUFFER_KEY_PREFIX = 'collapse-expand-';
+
 BackgroundConnection.onMessage.addListener(async message => {
   switch (message.type) {
     case Constants.kCOMMAND_NOTIFY_SUBTREE_COLLAPSED_STATE_CHANGED: {
+      if (BackgroundConnection.handleBufferedMessage(message, `${BUFFER_KEY_PREFIX}${message.tabId}`))
+        return;
       await Tab.waitUntilTracked(message.tabId, { element: true });
       const tab = Tab.get(message.tabId);
-      if (!tab)
+      const lastMessage = BackgroundConnection.fetchBufferedMessage(message.type, `${BUFFER_KEY_PREFIX}${message.tabId}`);
+      if (!tab ||
+          !lastMessage)
         return;
-      if (message.collapsed)
+      if (lastMessage.collapsed)
         tab.$TST.addState(Constants.kTAB_STATE_SUBTREE_COLLAPSED);
       else
         tab.$TST.removeState(Constants.kTAB_STATE_SUBTREE_COLLAPSED);
-      tab.$TST.tooltipIsDirty = true;
+      tab.$TST.invalidateElement(TabInvalidationTarget.Twisty | TabInvalidationTarget.Tooltip);
     }; break;
 
     case Constants.kCOMMAND_NOTIFY_TAB_COLLAPSED_STATE_CHANGED: {
+      if (BackgroundConnection.handleBufferedMessage(message, `${BUFFER_KEY_PREFIX}${message.tabId}`))
+        return;
       await Tab.waitUntilTracked(message.tabId, { element: true });
       const tab = Tab.get(message.tabId);
-      if (!tab)
+      const lastMessage = BackgroundConnection.fetchBufferedMessage(message.type, `${BUFFER_KEY_PREFIX}${message.tabId}`);
+      if (!tab ||
+          !lastMessage)
         return;
       setCollapsed(tab, {
-        collapsed: message.collapsed,
-        justNow:   message.justNow,
-        anchor:    Tab.get(message.anchorId),
-        last:      message.last
+        collapsed: lastMessage.collapsed,
+        justNow:   lastMessage.justNow,
+        anchor:    Tab.get(lastMessage.anchorId),
+        last:      lastMessage.last
       });
     }; break;
   }
