@@ -8,8 +8,8 @@
 import {
   log as internalLogger,
   dumpTab,
-  notify,
   wait,
+  countMatched,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -19,11 +19,13 @@ import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
 import * as Bookmark from '/common/bookmark.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as ContextualIdentities from '/common/contextual-identities.js';
 
 import Tab from '/common/Tab.js';
 
 import * as TabsOpen from './tabs-open.js';
 import * as TabsMove from './tabs-move.js';
+import * as TabsGroup from './tabs-group.js';
 import * as Tree from './tree.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
@@ -36,62 +38,77 @@ export const onTabsClosing = new EventListenerManager();
 export const onMoveUp      = new EventListenerManager();
 export const onMoveDown    = new EventListenerManager();
 
-export function reloadTree(rootTab) {
-  const tabs = [rootTab].concat(rootTab.$TST.descendants);
-  for (const tab of tabs) {
+function uniqTabsAndDescendantsSet(tabs) {
+  if (!Array.isArray(tabs))
+    tabs = [tabs];
+  return Array.from(new Set(tabs.map(tab => [tab].concat(tab.$TST.descendants)).flat())).sort(Tab.compare);
+}
+
+export function reloadTree(tabs) {
+  for (const tab of uniqTabsAndDescendantsSet(tabs)) {
     browser.tabs.reload(tab.id)
       .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
   }
 }
 
-export function reloadDescendants(rootTab) {
-  const tabs = rootTab.$TST.descendants;
-  for (const tab of tabs) {
+export function reloadDescendants(rootTabs) {
+  const rootTabsSet = new Set(rootTabs);
+  for (const tab of uniqTabsAndDescendantsSet(rootTabs)) {
+    if (rootTabsSet.has(tab))
+      continue;
     browser.tabs.reload(tab.id)
       .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
   }
 }
 
-export async function closeTree(rootTab) {
-  const tabs = [rootTab].concat(rootTab.$TST.descendants);
-  const canceled = (await onTabsClosing.dispatch(tabs.map(tab => tab.id), { windowId: rootTab.windowId })) === false;
+export async function closeTree(tabs) {
+  tabs = uniqTabsAndDescendantsSet(tabs);
+  const windowId = tabs[0].windowId;
+  const canceled = (await onTabsClosing.dispatch(tabs.map(tab => tab.id), { windowId })) === false;
   if (canceled)
     return;
   tabs.reverse(); // close bottom to top!
   TabsInternalOperation.removeTabs(tabs);
 }
 
-export async function closeDescendants(rootTab) {
-  const tabs = rootTab.$TST.descendants;
-  const canceled = (await onTabsClosing.dispatch(tabs.map(tab => tab.id), { windowId: rootTab.windowId })) === false;
+export async function closeDescendants(rootTabs) {
+  const rootTabsSet = new Set(rootTabs);
+  const tabs     = uniqTabsAndDescendantsSet(rootTabs).filter(tab => !rootTabsSet.has(tab));
+  const windowId = rootTabs[0].windowId;
+  const canceled = (await onTabsClosing.dispatch(tabs.map(tab => tab.id), { windowId })) === false;
   if (canceled)
     return;
   tabs.reverse(); // close bottom to top!
   TabsInternalOperation.removeTabs(tabs);
 }
 
-export async function closeOthers(rootTab) {
-  const exceptionTabs = [rootTab].concat(rootTab.$TST.descendants);
-  const tabs          = Tab.getNormalTabs(rootTab.windowId, { iterator: true, reversed: true }); // except pinned or hidden tabs, close bottom to top!
+export async function closeOthers(exceptionRoots) {
+  const exceptionTabs = uniqTabsAndDescendantsSet(exceptionRoots);
+  const windowId      = exceptionTabs[0].windowId;
+  const tabs          = Tab.getNormalTabs(windowId, { iterator: true, reversed: true }); // except pinned or hidden tabs, close bottom to top!
   const closeTabs     = [];
   for (const tab of tabs) {
     if (!exceptionTabs.includes(tab))
       closeTabs.push(tab);
   }
-  const canceled = (await onTabsClosing.dispatch(closeTabs.map(tab => tab.id), { windowId: rootTab.windowId })) === false;
+  const canceled = (await onTabsClosing.dispatch(closeTabs.map(tab => tab.id), { windowId })) === false;
   if (canceled)
     return;
   TabsInternalOperation.removeTabs(closeTabs);
 }
 
-export function collapseTree(rootTab) {
-  if (!rootTab.$TST.hasChild ||
-      rootTab.$TST.subtreeCollapsed)
-    return;
-  Tree.collapseExpandSubtree(rootTab, {
-    collapsed: true,
-    broadcast: true
-  });
+export function collapseTree(rootTabs, options = {}) {
+  const tabs = (
+    options.recursively ?
+      uniqTabsAndDescendantsSet(rootTabs) :
+      Array.isArray(rootTabs) && rootTabs || [rootTabs]
+  ).filter(tab => tab.$TST.hasChild && !tab.$TST.subtreeCollapsed);
+  for (const tab of tabs) {
+    Tree.collapseExpandSubtree(tab, {
+      collapsed: true,
+      broadcast: true
+    });
+  }
 }
 
 export function collapseAll(windowId) {
@@ -100,14 +117,18 @@ export function collapseAll(windowId) {
   }
 }
 
-export function expandTree(rootTab) {
-  if (!rootTab.$TST.hasChild ||
-      !rootTab.$TST.subtreeCollapsed)
-    return;
-  Tree.collapseExpandSubtree(rootTab, {
-    collapsed: false,
-    broadcast: true
-  });
+export function expandTree(rootTabs, options = {}) {
+  const tabs = (
+    options.recursively ?
+      uniqTabsAndDescendantsSet(rootTabs) :
+      Array.isArray(rootTabs) && rootTabs || [rootTabs]
+  ).filter(tab => tab.$TST.hasChild && tab.$TST.subtreeCollapsed);
+  for (const tab of tabs) {
+    Tree.collapseExpandSubtree(tab, {
+      collapsed: false,
+      broadcast: true
+    });
+  }
 }
 
 export function expandAll(windowId) {
@@ -116,33 +137,28 @@ export function expandAll(windowId) {
   }
 }
 
-export async function bookmarkTree(root, options = {}) {
-  const tabs = [root].concat(root.$TST.descendants);
+export async function bookmarkTree(rootTabs, options = {}) {
+  const tabs = uniqTabsAndDescendantsSet(rootTabs);
   if (tabs.length > 1 &&
       tabs[0].$TST.isGroupTab)
     tabs.shift();
 
   const tab = tabs[0];
-  if (SidebarConnection.isOpen(tab.windowId)) {
+  if (configs.showDialogInSidebar &&
+      SidebarConnection.isOpen(tab.windowId)) {
     return SidebarConnection.sendMessage({
       type:     Constants.kCOMMAND_BOOKMARK_TABS_WITH_DIALOG,
       windowId: tab.windowId,
-      tabIds:   tabs.map(tab => tab.id)
+      tabIds:   tabs.map(tab => tab.id),
+      options
+    });
+  }
+  else {
+    return Bookmark.bookmarkTabs(tabs, {
+      ...options,
+      showDialog: true
     });
   }
-
-  const folder = await Bookmark.bookmarkTabs(tabs, options);
-  if (!folder)
-    return null;
-  notify({
-    title:   browser.i18n.getMessage('bookmarkTree_notification_success_title'),
-    message: browser.i18n.getMessage('bookmarkTree_notification_success_message', [
-      root.title,
-      tabs.length,
-      folder.title
-    ])
-  });
-  return folder;
 }
 
 
@@ -253,11 +269,11 @@ export async function outdent(tab, options = {}) {
 }
 
 // drag and drop helper
-export async function performTabsDragDrop(params = {}) {
-  const windowId = params.windowId || TabsStore.getWindow();
+async function performTabsDragDrop(params = {}) {
+  const windowId = params.windowId || TabsStore.getCurrentWindowId();
   const destinationWindowId = params.destinationWindowId || windowId;
 
-  log('performTabsDragDrop ', {
+  log('performTabsDragDrop ', () => ({
     tabs:                params.tabs.map(dumpTab),
     attachTo:            dumpTab(params.attachTo),
     insertBefore:        dumpTab(params.insertBefore),
@@ -265,12 +281,13 @@ export async function performTabsDragDrop(params = {}) {
     windowId:            params.windowId,
     destinationWindowId: params.destinationWindowId,
     action:              params.action
-  });
+  }));
 
-  const movedTabs = await moveTabsWithStructure(params.tabs, Object.assign({}, params, {
+  const movedTabs = await moveTabsWithStructure(params.tabs, {
+    ...params,
     windowId, destinationWindowId,
     broadcast: true
-  }));
+  });
   if (movedTabs.length == 0)
     return;
   if (windowId != destinationWindowId) {
@@ -283,7 +300,7 @@ export async function performTabsDragDrop(params = {}) {
 
 // useful utility for general purpose
 export async function moveTabsWithStructure(tabs, params = {}) {
-  log('moveTabsWithStructure ', tabs.map(dumpTab));
+  log('moveTabsWithStructure ', () => tabs.map(dumpTab));
 
   let movedTabs = tabs.filter(tab => !!tab);
   if (!movedTabs.length)
@@ -299,7 +316,7 @@ export async function moveTabsWithStructure(tabs, params = {}) {
         movedWholeTree.push(descendant);
     }
   }
-  log('=> movedTabs: ', movedTabs.map(dumpTab).join(' / '));
+  log('=> movedTabs: ', () => movedTabs.map(dumpTab).join(' / '));
 
   while (params.insertBefore &&
          movedWholeTree.includes(params.insertBefore)) {
@@ -352,9 +369,10 @@ export async function moveTabsWithStructure(tabs, params = {}) {
         console.log(error);
       }
       if (!importedTab)
-        importedTab = await browser.tabs.create(Object.assign({}, createParams, {
+        importedTab = await browser.tabs.create({
+          ...createParams,
           url: `about:blank?${tab.url}`
-        }));
+        });
       movedTabs.push(importedTab);
       index++;
     }
@@ -378,11 +396,13 @@ export async function moveTabsWithStructure(tabs, params = {}) {
   }
 
   log('try attach/detach');
+  let shouldExpand = false;
   if (!params.attachTo) {
     log('=> detach');
     detachTabsWithStructure(movedRoots, {
       broadcast: params.broadcast
     });
+    shouldExpand = true;
   }
   else {
     log('=> attach');
@@ -392,9 +412,10 @@ export async function moveTabsWithStructure(tabs, params = {}) {
       draggedTabs:  movedTabs,
       broadcast:    params.broadcast
     });
+    shouldExpand = !params.attachTo.$TST.subtreeCollapsed;
   }
 
-  log('=> moving tabs ', movedTabs.map(dumpTab));
+  log('=> moving tabs ', () => movedTabs.map(dumpTab));
   if (params.insertBefore)
     await TabsMove.moveTabsBefore(
       movedTabs,
@@ -435,13 +456,27 @@ export async function moveTabsWithStructure(tabs, params = {}) {
   }, 0);
   */
 
+  if (shouldExpand) {
+    log('=> expand dropped tabs');
+    // Collapsed tabs may be moved to the root level,
+    // then we need to expand them.
+    await Promise.all(movedRoots.map(tab => {
+      if (!tab.$TST.collapsed)
+        return;
+      return Tree.collapseExpandTabAndSubtree(tab, {
+        collapsed: false,
+        broadcast: params.broadcast
+      });
+    }));
+  }
+
   log('=> finished');
 
   return movedTabs;
 }
 
 async function attachTabsWithStructure(tabs, parent, options = {}) {
-  log('attachTabsWithStructure: start ', tabs.map(dumpTab), dumpTab(parent));
+  log('attachTabsWithStructure: start ', () => [tabs.map(dumpTab), dumpTab(parent)]);
   if (parent &&
       !options.insertBefore &&
       !options.insertAfter) {
@@ -467,30 +502,33 @@ async function attachTabsWithStructure(tabs, parent, options = {}) {
       { broadcast: options.broadcast }
     );
 
-  const memberOptions = Object.assign({}, options, {
+  const memberOptions = {
+    ...options,
     insertBefore: null,
     insertAfter:  null,
     dontMove:     true,
     forceExpand:  options.draggedTabs.some(tab => tab.active)
-  });
-  for (const tab of tabs) {
+  };
+  return Promise.all(tabs.map(async tab => {
     if (parent)
-      Tree.attachTabTo(tab, parent, memberOptions);
+      await Tree.attachTabTo(tab, parent, memberOptions);
     else
-      Tree.detachTab(tab, memberOptions);
-    Tree.collapseExpandTabAndSubtree(tab, Object.assign({}, memberOptions, {
+      await Tree.detachTab(tab, memberOptions);
+    return Tree.collapseExpandTabAndSubtree(tab, {
+      ...memberOptions,
       collapsed: false
-    }));
-  }
+    });
+  }));
 }
 
 function detachTabsWithStructure(tabs, options = {}) {
-  log('detachTabsWithStructure: start ', tabs.map(dumpTab));
+  log('detachTabsWithStructure: start ', () => tabs.map(dumpTab));
   for (const tab of tabs) {
     Tree.detachTab(tab, options);
-    Tree.collapseExpandTabAndSubtree(tab, Object.assign({}, options, {
+    Tree.collapseExpandTabAndSubtree(tab, {
+      ...options,
       collapsed: false
-    }));
+    });
   }
 }
 
@@ -498,9 +536,10 @@ export async function moveUp(tab, options = {}) {
   const previousTab = tab.$TST.nearestVisiblePrecedingTab;
   if (!previousTab)
     return false;
-  const moved = await moveBefore(tab, Object.assign({}, options, {
+  const moved = await moveBefore(tab, {
+    ...options,
     referenceTabId: previousTab.id
-  }));
+  });
   if (moved && !options.followChildren)
     await onMoveUp.dispatch(tab);
   return moved;
@@ -510,9 +549,10 @@ export async function moveDown(tab, options = {}) {
   const nextTab = options.followChildren ? tab.$TST.nearestFollowingForeignerTab : tab.$TST.nearestVisibleFollowingTab;
   if (!nextTab)
     return false;
-  const moved = await moveAfter(tab, Object.assign({}, options, {
+  const moved = await moveAfter(tab, {
+    ...options,
     referenceTabId: nextTab.id
-  }));
+  });
   if (moved && !options.followChildren)
     await onMoveDown.dispatch(tab);
   return moved;
@@ -536,6 +576,7 @@ export async function moveBefore(tab, options = {}) {
   }
   else {
     const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(tab, {
+      context: Constants.kINSERTION_CONTEXT_MOVED,
       insertBefore
     });
     if (!referenceTabs.insertBefore &&
@@ -569,6 +610,7 @@ export async function moveAfter(tab, options = {}) {
   }
   else {
     const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(tab, {
+      context: Constants.kINSERTION_CONTEXT_MOVED,
       insertAfter
     });
     if (!referenceTabs.insertBefore && !referenceTabs.insertAfter)
@@ -620,10 +662,17 @@ export async function moveTabsToStart(movedTabs) {
     return;
   const tab       = movedTabs[0];
   const allTabs   = tab.pinned ? Tab.getPinnedTabs(tab.windowId) : Tab.getUnpinnedTabs(tab.windowId);
-  const otherTabs = allTabs.filter(tab => !movedTabs.includes(tab));
-  if (otherTabs.length > 0)
+  const movedTabsSet = new Set(movedTabs);
+  let firstOtherTab;
+  for (const tab of allTabs) {
+    if (movedTabsSet.has(tab))
+      continue;
+    firstOtherTab = tab;
+    break;
+  }
+  if (firstOtherTab)
     await moveTabsWithStructure(movedTabs, {
-      insertBefore: otherTabs[0],
+      insertBefore: firstOtherTab,
       broadcast:    true
     });
 }
@@ -638,10 +687,18 @@ export async function moveTabsToEnd(movedTabs) {
     return;
   const tab       = movedTabs[0];
   const allTabs   = tab.pinned ? Tab.getPinnedTabs(tab.windowId) : Tab.getUnpinnedTabs(tab.windowId);
-  const otherTabs = allTabs.filter(tab => !movedTabs.includes(tab));
-  if (otherTabs.length > 0)
+  const movedTabsSet = new Set(movedTabs);
+  let lastOtherTabs;
+  for (let i = allTabs.length - 1; i > -1; i--) {
+    const tab = allTabs[i];
+    if (movedTabsSet.has(tab))
+      continue;
+    lastOtherTabs = tab;
+    break;
+  }
+  if (lastOtherTabs)
     await moveTabsWithStructure(movedTabs, {
-      insertAfter: otherTabs[otherTabs.length-1],
+      insertAfter: lastOtherTabs,
       broadcast:   true
     });
 }
@@ -664,11 +721,41 @@ export async function openTabsInWindow(tabs) {
   return movedTabs.length > 0 ? movedTabs[0].windowId : null;
 }
 
+
+export async function restoreTabs(count) {
+  const toBeRestoredTabSessions = (await browser.sessions.getRecentlyClosed({
+    maxResults: browser.sessions.MAX_SESSION_RESULTS
+  }).catch(ApiTabs.createErrorHandler())).filter(session => session.tab).slice(0, count);
+  log('restoreTabs: toBeRestoredTabSessions = ', toBeRestoredTabSessions);
+  const promisedRestoredTabs = [];
+  for (const session of toBeRestoredTabSessions.reverse()) {
+    log('restoreTabs: Tabrestoring session = ', session);
+    promisedRestoredTabs.push(Tab.doAndGetNewTabs(async () => {
+      browser.sessions.restore(session.tab.sessionId).catch(ApiTabs.createErrorSuppressor());
+      await Tab.waitUntilTrackedAll();
+    }));
+  }
+  const restoredTabs = Array.from(new Set((await Promise.all(promisedRestoredTabs)).flat()));
+  log('restoreTabs: restoredTabs = ', restoredTabs);
+  await Promise.all(restoredTabs.map(tab => tab && Tab.get(tab.id).$TST.opened));
+
+  if (restoredTabs.length > 0) {
+    // Parallelly restored tabs can have ghost "active" state, so we need to clear them
+    const activeTab = Tab.getActiveTab(restoredTabs[0].windowId);
+    if (restoredTabs.some(tab => tab.id == activeTab.id))
+      await TabsInternalOperation.setTabActive(activeTab);
+  }
+
+  return Tab.sort(restoredTabs);
+}
+
+
 export async function bookmarkTab(tab, options = {}) {
   if (options.multiselected !== false && tab.$TST.multiselected)
     return bookmarkTabs(Tab.getSelectedTabs(tab.windowId));
 
-  if (SidebarConnection.isOpen(tab.windowId)) {
+  if (configs.showDialogInSidebar &&
+      SidebarConnection.isOpen(tab.windowId)) {
     SidebarConnection.sendMessage({
       type:     Constants.kCOMMAND_BOOKMARK_TAB_WITH_DIALOG,
       windowId: tab.windowId,
@@ -676,13 +763,8 @@ export async function bookmarkTab(tab, options = {}) {
     });
   }
   else {
-    await Bookmark.bookmarkTab(tab);
-    notify({
-      title:   browser.i18n.getMessage('bookmarkTab_notification_success_title'),
-      message: browser.i18n.getMessage('bookmarkTab_notification_success_message', [
-        tab.title
-      ]),
-      icon:    Constants.kNOTIFICATION_DEFAULT_ICON
+    Bookmark.bookmarkTab(tab, {
+      showDialog: true
     });
   }
 }
@@ -690,7 +772,8 @@ export async function bookmarkTab(tab, options = {}) {
 export async function bookmarkTabs(tabs) {
   if (tabs.length == 0)
     return;
-  if (SidebarConnection.isOpen(tabs[0].windowId)) {
+  if (configs.showDialogInSidebar &&
+      SidebarConnection.isOpen(tabs[0].windowId)) {
     SidebarConnection.sendMessage({
       type:     Constants.kCOMMAND_BOOKMARK_TABS_WITH_DIALOG,
       windowId: tabs[0].windowId,
@@ -698,17 +781,9 @@ export async function bookmarkTabs(tabs) {
     });
   }
   else {
-    const folder = await Bookmark.bookmarkTabs(tabs);
-    if (folder)
-      notify({
-        title:   browser.i18n.getMessage('bookmarkTabs_notification_success_title'),
-        message: browser.i18n.getMessage('bookmarkTabs_notification_success_message', [
-          tabs[0].title,
-          tabs.length,
-          folder.title
-        ]),
-        icon:    Constants.kNOTIFICATION_DEFAULT_ICON
-      });
+    Bookmark.bookmarkTabs(tabs, {
+      showDialog: true
+    });
   }
 }
 
@@ -758,12 +833,196 @@ SidebarConnection.onMessage.addListener(async (windowId, message) => {
         message.insertAfterId
       ]));
       log('perform tabs dragdrop requested: ', message);
-      performTabsDragDrop(Object.assign({}, message, {
+      performTabsDragDrop({
+        ...message,
         tabs:         message.import ? message.tabs : draggedTabIds.map(id => Tab.get(id)),
         attachTo:     message.attachToId && Tab.get(message.attachToId),
         insertBefore: message.insertBeforeId && Tab.get(message.insertBeforeId),
         insertAfter:  message.insertAfterId && Tab.get(message.insertAfterId)
-      }));
+      });
+    }; break;
+
+    case Constants.kCOMMAND_TOGGLE_MUTED: {
+      await Tab.waitUntilTracked(message.tabId);
+      const root = Tab.get(message.tabId);
+      log('toggle muted state: ', message, root);
+      if (!root)
+        break;
+
+      const multiselected = root.$TST.multiselected;
+      const tabs = multiselected ?
+        Tab.getSelectedTabs(root.windowId, { iterator: true }) :
+        [root] ;
+      const toBeMuted = (!multiselected && root.$TST.subtreeCollapsed) ?
+        root.$TST.maybeSoundPlaying :
+        root.$TST.soundPlaying ;
+
+      log('  toBeMuted: ', toBeMuted);
+      if (!multiselected &&
+          root.$TST.subtreeCollapsed &&
+          !root.$TST.soundPlaying) {
+        const toBeUpdatedTabs = root.$TST.descendants.filter(tab =>
+          // The "audible" possibly become "false" when the tab is
+          // really audible but muted.
+          // However, we can think more simply and robustly.
+          //  - We need to mute "audible" tabs.
+          //  - We need to unmute "muted" tabs.
+          // So, tabs which any of "audible" or "muted" is "true"
+          // have enough reason to be updated.
+          (tab.audible || tab.mutedInfo.muted) &&
+            // And we really need to update only tabs not been the
+            // expected state.
+            (tab.$TST.soundPlaying == toBeMuted)
+        );
+        log('  toBeUpdatedTabs: ', toBeUpdatedTabs);
+        for (const tab of toBeUpdatedTabs) {
+          browser.tabs.update(tab.id, {
+            muted: toBeMuted
+          }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+        }
+      }
+      else {
+        log('  tabs: ', tabs);
+        for (const tab of tabs) {
+          browser.tabs.update(tab.id, {
+            muted: toBeMuted
+          }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+        }
+      }
     }; break;
   }
 });
+
+
+const DESCENDANT_MATCHER = /^(>+) /;
+
+async function collectBookmarkItems(root, recursively) {
+  let items = await browser.bookmarks.getChildren(root.id);
+  if (recursively) {
+    let expandedItems = [];
+    for (const item of items) {
+      switch (item.type) {
+        case 'bookmark':
+          expandedItems.push(item);
+          break;
+        case 'folder':
+          expandedItems = expandedItems.concat(await collectBookmarkItems(item, recursively));
+          break;
+      }
+    }
+    items = expandedItems;
+  }
+  else {
+    items = items.filter(item => item.type == 'bookmark');
+  }
+  if (countMatched(items, item => !DESCENDANT_MATCHER.test(item.title)) > 1) {
+    for (const item of items) {
+      item.title = DESCENDANT_MATCHER.test(item.title) ?
+        item.title.replace(DESCENDANT_MATCHER, '>$1 ') :
+        `> ${item.title}`;
+    }
+    items.unshift({
+      title: '',
+      url:   TabsGroup.makeGroupTabURI({
+        title:               root.title,
+        temporaryAggressive: true
+      }),
+      group: true,
+      discarded: false
+    });
+  }
+  return items;
+}
+
+export async function openBookmarksWithStructure(items, { activeIndex = 0, discarded } = {}) {
+  if (typeof discarded == 'undefined')
+    discarded = configs.openAllBookmarksWithStructureDiscarded;
+
+  const lastItemIndicesWithLevel = new Map();
+  let lastMaxLevel = 0;
+  const containerMatcher = new RegExp(`#${configs.containerRedirectKey}-(.+)$`);
+  const structure = items.reduce((result, item, index) => {
+    // Respect container type stored by Container Bookmarks
+    // https://addons.mozilla.org/firefox/addon/container-bookmarks/
+    const matchedContainer = item.url.match(containerMatcher);
+    if (matchedContainer) {
+      const cookieStoreId = ContextualIdentities.getIdFromName(decodeURIComponent(matchedContainer[matchedContainer.length-1]));
+      if (cookieStoreId) {
+        item.cookieStoreId = cookieStoreId;
+        item.url = item.url.replace(containerMatcher, '');
+      }
+    }
+
+    let level = 0;
+    if (lastItemIndicesWithLevel.size > 0 &&
+        item.title.match(DESCENDANT_MATCHER)) {
+      level = RegExp.$1.length;
+      if (level - lastMaxLevel > 1) {
+        level = lastMaxLevel + 1;
+      }
+      else {
+        while (lastMaxLevel > level) {
+          lastItemIndicesWithLevel.delete(lastMaxLevel--);
+        }
+      }
+      lastItemIndicesWithLevel.set(level, index);
+      lastMaxLevel = level;
+      result.push(lastItemIndicesWithLevel.get(level - 1) - lastItemIndicesWithLevel.get(0));
+      item.title = item.title.replace(DESCENDANT_MATCHER, '')
+    }
+    else {
+      result.push(-1);
+      lastItemIndicesWithLevel.clear();
+      lastItemIndicesWithLevel.set(0, index);
+    }
+    return result;
+  }, []);
+
+  const windowId = TabsStore.getCurrentWindowId() || (await browser.windows.getCurrent()).id;
+  const tabs = await TabsOpen.openURIsInTabs(items, {
+    windowId,
+    isOrphan:     true,
+    inBackground: true,
+    fixPositions: true,
+    discarded
+  });
+
+  if (tabs.length > activeIndex)
+    TabsInternalOperation.activateTab(tabs[activeIndex]);
+  if (tabs.length == structure.length)
+    await Tree.applyTreeStructureToTabs(tabs, structure);
+
+  // tabs can be opened at middle of an existing tree due to browser.tabs.insertAfterCurrent=true
+  const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(tabs, {
+    context:      Constants.kINSERTION_CONTEXT_CREATED,
+    insertAfter:  tabs[0].$TST.previousTab,
+    insertBefore: tabs[tabs.length - 1].$TST.nextTab
+  });
+  if (referenceTabs.parent)
+    await Tree.attachTabTo(tabs[0], referenceTabs.parent, {
+      insertAfter:  referenceTabs.insertAfter,
+      insertBefore: referenceTabs.insertBefore
+    });
+}
+
+export async function openAllBookmarksWithStructure(id, { discarded, recursively } = {}) {
+  if (typeof discarded == 'undefined')
+    discarded = configs.openAllBookmarksWithStructureDiscarded;
+
+  let item = await browser.bookmarks.get(id);
+  if (Array.isArray(item))
+    item = item[0];
+  if (!item)
+    return;
+
+  if (item.type != 'folder') {
+    item = await browser.bookmarks.get(item.parentId);
+    if (Array.isArray(item))
+      item = item[0];
+  }
+
+  const items = await collectBookmarkItems(item, recursively);
+  const activeIndex = items.findIndex(item => !item.group);
+
+  openBookmarksWithStructure(items, { activeIndex, discarded });
+}
