@@ -7,12 +7,41 @@
 
 import Configs from '/extlib/Configs.js';
 
-/* global
-  uneval: false,
- */
-
 import * as Constants from './constants.js';
 
+const localKeys = mapAndFilter(`
+  baseIndent
+  colorScheme
+  debug
+  faviconizedTabScale
+  grantedRemovingTabIds
+  lastConfirmedToCloseTabs
+  lastDraggedTabs
+  logFor
+  loggingConnectionMessages
+  loggingQueries
+  logTimestamp
+  maximumDelayForBug1561879
+  migratedBookmarkUrls
+  minimumIntervalToProcessDragoverEvent
+  minIndent
+  notifiedFeaturesVersion
+  optionsExpandedGroups
+  optionsExpandedSections
+  requestingPermissions
+  requestingPermissionsNatively
+  sidebarDirection
+  sidebarPosition
+  startDragTimeout
+  style
+  subMenuCloseDelay
+  subMenuOpenDelay
+  testKey
+`.trim().split('\n'), key => {
+  key = key.trim();
+  return key && key.indexOf('//') != 0 && key;
+});
+
 export const configs = new Configs({
   optionsExpandedSections: ['section-appearance'],
   optionsExpandedGroups: [],
@@ -21,14 +50,15 @@ export const configs = new Configs({
   sidebarPosition: Constants.kTABBAR_POSITION_LEFT,
   sidebarDirection: Constants.kTABBAR_DIRECTION_LTR,
 
-  sidebarScrollbarPosition: 0, // obsolete, migrated to user stylesheet
-  scrollbarMode: -1, // obsolete, migrated to user stylesheet
+  sidebarScrollbarPosition: null, // obsolete, migrated to user stylesheet
+  scrollbarMode: null, // obsolete, migrated to user stylesheet
 
   style:
     /^Mac/i.test(navigator.platform) ? 'sidebar' :
       'mixed',
   colorScheme: /^Linux/i.test(navigator.platform) ? 'system-color' : 'photon' ,
-  applyBrowserTheme: true,
+
+  unrepeatableBGImageAspectRatio: 4,
 
   faviconizePinnedTabs: true,
   maxFaviconizedPinnedTabsInOneRow: 0, // auto
@@ -49,9 +79,11 @@ export const configs = new Configs({
   zoomable: false,
   showCollapsedDescendantsByTooltip: true,
 
+  showDialogInSidebar: false,
+
 
   // context menu
-  fakeContextMenu: true, // obsolete, migrated to emulateDefaultContextMenu
+  fakeContextMenu: null, // obsolete, migrated to emulateDefaultContextMenu
   emulateDefaultContextMenu: true,
 
   context_reloadTree: true,
@@ -60,8 +92,10 @@ export const configs = new Configs({
   context_closeDescendants: false,
   context_closeOthers: false,
   context_collapseTree: false,
+  context_collapseTreeRecursively: true,
   context_collapseAll: true,
   context_expandTree: false,
+  context_expandTreeRecursively: true,
   context_expandAll: true,
   context_bookmarkTree: true,
   context_groupTabs: true,
@@ -72,20 +106,27 @@ export const configs = new Configs({
   context_topLevel_closeDescendants: false,
   context_topLevel_closeOthers: false,
   context_topLevel_collapseTree: false,
+  context_topLevel_collapseTreeRecursively: false,
   context_topLevel_collapseAll: false,
   context_topLevel_expandTree: false,
+  context_topLevel_expandTreeRecursively: false,
   context_topLevel_expandAll: false,
   context_topLevel_bookmarkTree: false,
   context_topLevel_groupTabs: false,
 
-  context_closeTabOptions_closeTree: false, // obsolete, migrated to context_topLevel_closeTree
-  context_closeTabOptions_closeDescendants: false, // obsolete, migrated to context_topLevel_closeDescendants
-  context_closeTabOptions_closeOthers: false, // obsolete, migrated to context_topLevel_closeOthers
+  context_closeTabOptions_closeTree: null, // obsolete, migrated to context_topLevel_closeTree
+  context_closeTabOptions_closeDescendants: null, // obsolete, migrated to context_topLevel_closeDescendants
+  context_closeTabOptions_closeOthers: null, // obsolete, migrated to context_topLevel_closeOthers
 
   context_collapsed: false,
   context_pinnedTab: false,
   context_unpinnedTab: false,
 
+  context_openAllBookmarksWithStructure: true,
+  context_openAllBookmarksWithStructureRecursively: false,
+
+  openAllBookmarksWithStructureDiscarded: true,
+
 
   // tree behavior
   shouldDetectClickOnIndentSpaces: true,
@@ -94,10 +135,12 @@ export const configs = new Configs({
   autoCollapseExpandSubtreeOnSelect: true,
   autoCollapseExpandSubtreeOnSelectExceptActiveTabRemove: true,
 
-  collapseExpandSubtreeByDblClick: false,
+  collapseExpandSubtreeByDblClick: null, // obsolete, migrated to treeDoubleClickBehavior
+  treeDoubleClickBehavior: Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_NONE,
 
   autoExpandIntelligently: true,
-  autoExpandOnCollapsedChildActive: true,
+  unfocusableCollapsedTab: true,
+  autoExpandOnCollapsedChildActive: null, // obsolete, migrate to unfocusableCollapsedTab
   autoExpandOnTabSwitchingShortcuts: true,
   autoExpandOnTabSwitchingShortcutsDelay: 800,
   autoExpandOnLongHover: true,
@@ -110,23 +153,29 @@ export const configs = new Configs({
 
   dropLinksOnTabBehavior: Constants.kDROPLINK_ASK,
 
-  showTabDragHandle:    false,
   tabDragBehavior:      Constants.kDRAG_BEHAVIOR_TEAR_OFF | Constants.kDRAG_BEHAVIOR_WHOLE_TREE,
   tabDragBehaviorShift: Constants.kDRAG_BEHAVIOR_WHOLE_TREE | Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK,
   showTabDragBehaviorNotification: true,
+  guessDraggedNativeTabs: true,
 
   fixupTreeOnTabVisibilityChanged: false,
 
   scrollToExpandedTree: true,
 
+  spreadMutedStateOnlyToSoundPlayingTabs: true,
+
 
   // grouping
-  autoGroupNewTabs: true,
+  autoGroupNewTabsFromBookmarks: true,
+  tabsFromSameFolderMinThresholdPercentage: 50,
+  autoGroupNewTabsFromOthers: false,
   autoGroupNewTabsTimeout: 100,
   autoGroupNewTabsDelayOnNewWindow: 500,
   autoGroupNewTabsFromPinned: true,
   renderTreeInGroupTabs: true,
   warnOnAutoGroupNewTabs: true,
+  warnOnAutoGroupNewTabsWithListing: true,
+  warnOnAutoGroupNewTabsWithListingMaxRows: 5,
   showAutoGroupOptionHint: true,
 
 
@@ -141,12 +190,15 @@ export const configs = new Configs({
   autoAttachOnOpenedWithOwner: Constants.kNEWTAB_OPEN_AS_CHILD,
   autoAttachOnNewTabCommand: Constants.kNEWTAB_DO_NOTHING,
   autoAttachOnNewTabButtonMiddleClick: Constants.kNEWTAB_OPEN_AS_CHILD,
+  autoAttachOnNewTabButtonAccelClick: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING_WITH_INHERITED_CONTAINER,
   autoAttachOnDuplicated: Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING,
   autoAttachSameSiteOrphan: Constants.kNEWTAB_OPEN_AS_CHILD,
+  autoAttachOnOpenedFromExternal: Constants.kNEWTAB_DO_NOTHING,
   guessNewOrphanTabAsOpenedByNewTabCommand: true,
   guessNewOrphanTabAsOpenedByNewTabCommandUrl: 'about:newtab',
   inheritContextualIdentityToNewChildTab: false,
   inheritContextualIdentityToSameSiteOrphan: true,
+  inheritContextualIdentityToTabsFromExternal: false,
 
 
   // behavior around closed tab
@@ -154,15 +206,20 @@ export const configs = new Configs({
   closeParentBehavior:                Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
   closeParentBehavior_outsideSidebar: Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
   closeParentBehavior_noSidebar:      Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD,
-  promoteFirstChildForClosedRoot:     true, // obsolete, migrated to Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY of closeParentBehavior
-  parentTabBehaviorForChanges:        Constants.kPARENT_TAB_BEHAVIOR_ALWAYS, // obsolete, migrated to closeParentBehaviorMode
+  promoteFirstChildForClosedRoot:     null, // obsolete, migrated to Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_INTELLIGENTLY of closeParentBehavior
+  parentTabBehaviorForChanges:        null, // obsolete, migrated to closeParentBehaviorMode
   moveTabsToBottomWhenDetachedFromClosedParent: false,
   promoteAllChildrenWhenClosedParentIsLastChild: true,
-  moveFocusInTreeForClosedActiveTab: true, // obsolete, migrated to "successorTabControlLevel"
+  treatTreeAsExpandedOnClosedWithNoSidebar: true,
+  moveFocusInTreeForClosedActiveTab: null, // obsolete, migrated to "successorTabControlLevel"
   successorTabControlLevel: Constants.kSUCCESSOR_TAB_CONTROL_IN_TREE,
   simulateSelectOwnerOnClose: true,
   supportTabsMultiselect: typeof browser.menus.overrideContext == 'function',
   warnOnCloseTabs: true,
+  warnOnCloseTabsNotificationTimeout: 20 * 1000,
+  warnOnCloseTabsByClosebox: true,
+  warnOnCloseTabsWithListing: true,
+  warnOnCloseTabsWithListingMaxRows: 5,
   lastConfirmedToCloseTabs: 0,
   grantedRemovingTabIds: [],
 
@@ -177,18 +234,24 @@ export const configs = new Configs({
   outOfViewTabNotifyDuration: 750,
   subMenuOpenDelay: 300,
   subMenuCloseDelay: 300,
-  tabDragHandleDelay: 750,
-  tabDragHandleFeedbackDuration: 1000,
+
+
+  // subpanel
+  lastSelectedSubPanelProviderId: null,
+  lastSubPanelHeight: 0,
 
 
   // misc.
+  showExpertOptions: false,
   bookmarkTreeFolderName: browser.i18n.getMessage('bookmarkFolder_label_default', ['%TITLE%', '%YEAR%', '%MONTH%', '%DATE%']),
   defaultBookmarkParentId: 'unfiled_____',
   defaultSearchEngine: 'https://www.google.com/search?q=%s',
   acceleratedTabOperations: true,
   acceleratedTabCreation: false,
   enableWorkaroundForBug1409262: false,
-  simulateCloseTabByDblclick: false,
+  enableWorkaroundForBug1548949: true,
+  maximumDelayForBug1561879: 100,
+  workaroundForBug1548949DroppedTabs: null,
   maximumAcceptableDelayForTabDuplication: 10 * 1000,
   maximumDelayUntilTabIsTracked: 10 * 60 * 1000,
   delayToBlockUserOperationForTabsRestoration: 1000,
@@ -197,19 +260,18 @@ export const configs = new Configs({
   acceptableDelayForInternalFocusMoving: 150,
   delayToRetrySyncTabsOrder: 100,
   notificationTimeout: 10 * 1000,
-  startDragTimeout: 400, // obsolete, migrated to longPressDuration
+  startDragTimeout: null, // obsolete, migrated to longPressDuration
+  simulateCloseTabByDblclick: null, // obsolete, migrated to "treeDoubleClickBehavior=kTREE_DOUBLE_CLICK_BEHAVIOR_CLOSE"
   longPressDuration: 400,
   minimumIntervalToProcessDragoverEvent: 50,
   delayToApplyHighlightedState: 50,
+  acceptableFlickerToIgnoreClickOnTabAndTabbar: 10,
   moveDroppedTabToNewWindowForUnhandledDragEvent: true, // see also: https://github.com/piroor/treestyletab/issues/1646
   autoDiscardTabForUnexpectedFocus: true,
   autoDiscardTabForUnexpectedFocusDelay: 500,
-  knownExternalAddons: [
-    'multipletab@piro.sakura.ne.jp'
-  ],
-  cachedExternalAddons: [],
-  grantedExternalAddonPermissions: {},
-  incognitoAllowedExternalAddons: [],
+  avoidDiscardedTabToBeActivatedIfPossible: false,
+  undoMultipleTabsClose: true,
+  migratedBookmarkUrls: [],
   notifiedFeaturesVersion: 0,
 
   useCachedTree: true,
@@ -223,14 +285,16 @@ export const configs = new Configs({
 
   requestingPermissions: null,
   requestingPermissionsNatively: null,
+  lastDraggedTabs: null,
 
   // https://dxr.mozilla.org/mozilla-central/rev/2535bad09d720e71a982f3f70dd6925f66ab8ec7/browser/base/content/browser.css#137
   newTabAnimationDuration: 100,
 
+  // obsolete, migrated to userStyleRules0-5
   userStyleRules: `
 /* Show title of unread tabs with red and italic font */
 /*
-.tab.unread .label {
+:root.sidebar tab-item.unread .label-content {
   color: red !important;
   font-style: italic !important;
 }
@@ -238,11 +302,31 @@ export const configs = new Configs({
 
 /* Add private browsing indicator per tab */
 /*
-.tab.private-browsing .label:before {
+:root.sidebar tab-item.private-browsing tab-label:before {
   content: "🕶";
 }
 */
 `.trim(),
+  userStyleRules0: '',
+  userStyleRules1: '',
+  userStyleRules2: '',
+  userStyleRules3: '',
+  userStyleRules4: '',
+  userStyleRules5: '',
+
+
+  // Compatibility with other addons
+  knownExternalAddons: [
+    'multipletab@piro.sakura.ne.jp'
+  ],
+  cachedExternalAddons: [],
+  grantedExternalAddonPermissions: {},
+  incognitoAllowedExternalAddons: [],
+
+  // This must be same to the redirect key of Container Bookmarks.
+  // https://addons.mozilla.org/firefox/addon/container-bookmarks/
+  containerRedirectKey: 'container',
+
 
   debug:     false,
   logTimestamp: true,
@@ -273,7 +357,7 @@ export const configs = new Configs({
     'common/Tab': false,
     'common/Window': false,
     'common/api-tabs': false,
-    'common/bookmarks': false,
+    'common/bookmark': false,
     'common/contextual-identities': false,
     'common/permissions': false,
     'common/sidebar-connection': false,
@@ -285,7 +369,6 @@ export const configs = new Configs({
     'common/user-operation-blocker': false,
     'sidebar/background-connection': false,
     'sidebar/collapse-expand': false,
-    'sidebar/color': false,
     'sidebar/drag-and-drop': false,
     'sidebar/event-utils': false,
     'sidebar/indent': false,
@@ -296,8 +379,8 @@ export const configs = new Configs({
     'sidebar/sidebar-tabs': false,
     'sidebar/sidebar': false,
     'sidebar/size': false,
-    'sidebar/tab-context-menu': false,
-    'sidebar/tab-drag-handle': false
+    'sidebar/subpanel': false,
+    'sidebar/tab-context-menu': false
   },
   loggingConnectionMessages: false,
 
@@ -305,28 +388,12 @@ export const configs = new Configs({
 
   testKey: 0 // for tests/utils.js
 }, {
-  localKeys: `
-    optionsExpandedSections
-    sidebarPosition
-    sidebarDirection
-    style
-    colorScheme
-    faviconizedTabScale
-    baseIndent
-    minIndent
-    lastConfirmedToCloseTabs
-    grantedRemovingTabIds
-    subMenuOpenDelay
-    subMenuCloseDelay
-    minimumIntervalToProcessDragoverEvent
-    cachedExternalAddons
-    notifiedFeaturesVersion
-    requestingPermissions
-    requestingPermissionsNatively
-    testKey
-  `.trim().split('\n').map(key => key.trim()).filter(key => key && key.indexOf('//') != 0)
+  localKeys
 });
 
+// cleanup old data
+browser.storage.sync.remove(localKeys);
+
 configs.$loaded.then(() => {
   log.forceStore = false;
   if (!configs.debug)
@@ -334,6 +401,46 @@ configs.$loaded.then(() => {
 });
 
 
+export function loadUserStyleRules() {
+  let style = '';
+  let count = 0;
+  while (true) {
+    const key = `userStyleRules${count}`;
+    if (!(key in configs))
+      break;
+    style += configs[key];
+    count++;
+  }
+  return style;
+}
+
+export function saveUserStyleRules(style) {
+  chunkString(style, Constants.kSYNC_STORAGE_SAFE_QUOTA).forEach((chunk, index) => {
+    const key = `userStyleRules${index}`;
+    if (key in configs)
+      configs[key] = chunk;
+  });
+}
+
+// https://stackoverflow.com/questions/57068850/how-to-split-a-string-into-chunks-of-a-particular-byte-size
+function chunkString(input, maxBytes) {
+  const decoder = new TextDecoder('utf-8');
+  let buffer = new TextEncoder('utf-8').encode(input);
+  const chunks = [];
+  while (buffer.length) {
+    let index = buffer.lastIndexOf(32, maxBytes + 1);
+    if (index < 0)
+      index = buffer.indexOf(32, maxBytes);
+    if (index < 0)
+      index = buffer.length;
+    chunks.push(decoder.decode(buffer.slice(0, index)));
+    buffer = buffer.slice(index + 1);
+  }
+  return chunks;
+}
+
+
+
 export function log(module, ...args)
 {
   const isModuleLog = module in configs.$default.logFor;
@@ -343,6 +450,8 @@ export function log(module, ...args)
   if (!logging)
     return;
 
+  args = args.map(arg => typeof arg == 'function' ? arg() : arg);
+
   const nest = (new Error()).stack.split('\n').length;
   let indent = '';
   for (let i = 0; i < nest; i++) {
@@ -358,7 +467,10 @@ export function log(module, ...args)
   if (useConsole)
     console.log(line, ...args);
 
-  log.logs.push(`${line} ${args.map(arg => uneval(arg)).join(', ')}`);
+  log.logs.push(`${line} ${args.reduce((output, arg, index) => {
+    output += `${index == 0 ? '' : ', '}${uneval(arg)}`;
+    return output;
+  }, '')}`);
   log.logs = log.logs.slice(-log.max);
 }
 log.context = '?';
@@ -366,6 +478,28 @@ log.max  = 2000;
 log.logs = [];
 log.forceStore = true;
 
+// uneval() is no more available after https://bugzilla.mozilla.org/show_bug.cgi?id=1565170
+function uneval(value) {
+  switch (typeof value) {
+    case 'undefined':
+      return 'undefined';
+
+    case 'function':
+      return value.toString();
+
+    case 'object':
+      if (!value)
+        return 'null';
+    default:
+      try {
+        return JSON.stringify(value);
+      }
+      catch(e) {
+        return `${String(value)} (couldn't be stringified due to an error: ${String(e)})`;
+      }
+  }
+}
+
 function getTimeStamp() {
   const time = new Date();
   const hours = `0${time.getHours()}`.substr(-2);
@@ -390,7 +524,7 @@ export async function wait(task = 0, timeout = 0) {
     timeout = task;
     task    = null;
   }
-  return new Promise((resolve, _aReject) => {
+  return new Promise((resolve, _reject) => {
     setTimeout(async () => {
       if (task)
         await task();
@@ -400,42 +534,150 @@ export async function wait(task = 0, timeout = 0) {
 }
 
 export function nextFrame() {
-  return new Promise((resolve, _aReject) => {
+  return new Promise((resolve, _reject) => {
     window.requestAnimationFrame(resolve);
   });
 }
 
-export async function notify(params = {}) {
+export async function notify({ icon, title, message, timeout, url } = {}) {
   const id = await browser.notifications.create({
     type:    'basic',
-    iconUrl: params.icon || Constants.kNOTIFICATION_DEFAULT_ICON,
-    title:   params.title,
-    message: params.message
+    iconUrl: icon || Constants.kNOTIFICATION_DEFAULT_ICON,
+    title,
+    message
   });
 
   let onClicked;
-  if (params.url) {
+  let onClosed;
+  return new Promise(async (resolve, _reject) => {
+    let resolved = false;
+
     onClicked = notificationId => {
       if (notificationId != id)
         return;
-      browser.tabs.create({
-        url: params.url
-      });
-      browser.notifications.onClicked.removeListener(onClicked);
-      onClicked = null;
+      if (url) {
+        browser.tabs.create({
+          url
+        });
+      }
+      resolved = true;
+      resolve(true);
     };
     browser.notifications.onClicked.addListener(onClicked);
-  }
 
-  let timeout = params.timeout;
-  if (typeof timeout != 'number')
-    timeout = configs.notificationTimeout;
-  if (timeout >= 0)
-    await wait(timeout);
+    onClosed = notificationId => {
+      if (notificationId != id)
+        return;
+      if (!resolved) {
+        resolved = true;
+        resolve(false);
+      }
+    };
+    browser.notifications.onClosed.addListener(onClosed);
 
-  if (onClicked) {
+    if (typeof timeout != 'number')
+      timeout = configs.notificationTimeout;
+    if (timeout >= 0) {
+      await wait(timeout);
+    }
+    await browser.notifications.clear(id);
+    if (!resolved)
+      resolve(false);
+  }).then(clicked => {
     browser.notifications.onClicked.removeListener(onClicked);
     onClicked = null;
+    browser.notifications.onClosed.removeListener(onClosed);
+    onClosed = null;
+    return clicked;
+  });
+}
+
+
+// Helper functions for optimization
+// Originally implemented by @bb010g at
+// https://github.com/piroor/treestyletab/pull/2368/commits/9d184c4ac6c9977d2557cd17cec8c2a0f21dd527
+
+// For better performance the callback function must return "undefined"
+// when the item should not be included. "null", "false", and other false
+// values will be included to the mapped result.
+export function mapAndFilter(values, mapper) {
+  /* This function logically equals to:
+  return values.reduce((mappedValues, value) => {
+    value = mapper(value);
+    if (value !== undefined)
+      mappedValues.push(value);
+    return mappedValues;
+  }, []);
+  */
+  const maxi = ('length' in values ? values.length : values.size) >>> 0; // define as unsigned int
+  const mappedValues = new Array(maxi); // prepare with enough size at first, to avoid needless re-allocation
+  let count = 0,
+      value, // this must be defined outside of the loop, to avoid needless re-allocation
+      mappedValue; // this must be defined outside of the loop, to avoid needless re-allocation
+  for (value of values) {
+    mappedValue = mapper(value);
+    if (mappedValue !== undefined)
+      mappedValues[count++] = mappedValue;
   }
-  await browser.notifications.clear(id);
+  mappedValues.length = count; // shrink the array at last
+  return mappedValues;
+}
+
+export function mapAndFilterUniq(values, mapper, options = {}) {
+  const mappedValues = new Set();
+  let value, // this must be defined outside of the loop, to avoid needless re-allocation
+      mappedValue; // this must be defined outside of the loop, to avoid needless re-allocation
+  for (value of values) {
+    mappedValue = mapper(value);
+    if (mappedValue !== undefined)
+      mappedValues.add(mappedValue);
+  }
+  return options.set ? mappedValues : Array.from(mappedValues);
+}
+
+export function countMatched(values, matcher) {
+  /* This function logically equals to:
+  return values.reduce((count, value) => {
+    if (matcher(value))
+      count++;
+    return count;
+  }, 0);
+  */
+  let count = 0,
+      value; // this must be defined outside of the loop, to avoid needless re-allocation
+  for (value of values) {
+    if (matcher(value))
+      count++;
+  }
+  return count;
+}
+
+export function toLines(values, mapper, separator = '\n') {
+  /* This function logically equals to:
+  return values.reduce((output, value, index) => {
+    output += `${index == 0 ? '' : '\n'}${mapper(value)}`;
+    return output;
+  }, '');
+  */
+  const maxi = values.length >>> 0; // define as unsigned int
+  let i = 0,
+      lines = '';
+  while (i < maxi) { // use "while" loop instead "for" loop, for better performance
+    lines += `${i == 0 ? '' : separator}${mapper(values[i])}`;
+    i++;
+  }
+  return lines;
+}
+
+export async function sha1sum(string) {
+  const encoder = new TextEncoder();
+  const data = encoder.encode(string);
+  const hashBuffer = await crypto.subtle.digest('SHA-1', data);
+  const hashArray = Array.from(new Uint8Array(hashBuffer));
+  const hashHex = hashArray.map(byte => byte.toString(16).padStart(2, '0')).join('');
+  return hashHex;
+}
+
+export function sanitizeForHTMLText(text) {
+  return text.replace(/&/g, '&amp;').replace(/</g, '&lt;').replace(/>/g, '&gt;');
 }
