@@ -5,15 +5,14 @@
 */
 'use strict';
 
-export const kBACKGROUND_CONTENTS_VERSION = 5;
-export const kSIDEBAR_CONTENTS_VERSION    = 9;
-
 export const kCOMMAND_GET_INSTANCE_ID                = 'treestyletab:get-instance-id';
 export const kCOMMAND_RELOAD                         = 'treestyletab:reload';
 export const kCOMMAND_PING_TO_BACKGROUND             = 'treestyletab:ping-to-background';
 export const kCOMMAND_PING_TO_SIDEBAR                = 'treestyletab:ping-to-sidebar';
 export const kCOMMAND_REQUEST_CONNECT_PREFIX         = 'treestyletab:request-connect-from:';
 export const kCOMMAND_REQUEST_UNIQUE_ID              = 'treestyletab:request-unique-id';
+export const kCOMMAND_GET_THEME_DECLARATIONS         = 'treestyletab:get-theme-declarations';
+export const kCOMMAND_GET_CONTEXTUAL_IDENTITIES_COLOR_INFO = 'treestyletab:get-contextual-identities-color-info';
 export const kCOMMAND_GET_CONFIG_VALUE               = 'treestyletab:get-config-value';
 export const kCOMMAND_SET_CONFIG_VALUE               = 'treestyletab:set-config-value';
 export const kCOMMAND_PULL_TABS                      = 'treestyletab:pull-tabs';
@@ -28,7 +27,7 @@ export const kCOMMAND_NEW_TAB_AS                     = 'treestyletab:open-new-ta
 export const kCOMMAND_REMOVE_TABS_INTERNALLY         = 'treestyletab:remove-tabs-internally';
 export const kCOMMAND_UPDATE_LOADING_STATE           = 'treestyletab:update-loading-state';
 export const kCOMMAND_CONFIRM_TO_CLOSE_TABS          = 'treestyletab:confirm-to-close-tabs';
-export const kCOMMAND_CONFIRM_TO_AUTO_GROUP_NEW_TABS = 'treestyletab:confirm-to-auto-group-new-tabs';
+export const kCOMMAND_SHOW_DIALOG                    = 'treestyletab:show-dialog';
 export const kCOMMAND_NOTIFY_TAB_CREATING            = 'treestyletab:notify-tab-creating';
 export const kCOMMAND_NOTIFY_TAB_CREATED             = 'treestyletab:notify-tab-created';
 export const kCOMMAND_NOTIFY_TAB_UPDATED             = 'treestyletab:notify-tab-updated';
@@ -74,9 +73,13 @@ export const kCOMMAND_REQUEST_CONNECTION_MESSAGE_LOGS  = 'treestyletab:request-c
 export const kCOMMAND_RESPONSE_CONNECTION_MESSAGE_LOGS = 'treestyletab:response-connection-message-logs';
 export const kCOMMAND_NOTIFY_TEST_KEY_CHANGED        = 'treestyletab:notify-test-key-changed';
 export const kCOMMAND_SIMULATE_SIDEBAR_MESSAGE       = 'treestyletab:simulate-sidebar-message';
+export const kCOMMAND_GET_CONTEXT_MENU_ITEMS         = 'treestyletab:contextMenu-get-items';
+export const kCOMMAND_NOTIFY_CONTEXT_MENU_UPDATED    = 'treestyletab:contextMenu-updated';
+export const kCOMMAND_NOTIFY_CONTEXT_ITEM_CHECKED_STATUS_CHANGED = 'treestyletab:contextMenu-item-checked-status-changed';
+export const kCOMMAND_NOTIFY_CONTEXT_OVERRIDDEN      = 'treestyletab:notify-context-overridden';
 
 export const kCOMMAND_SELECT_TAB              = 'treestyletab:select-tab';
-export const kCOMMAND_SET_SUBTREE_MUTED       = 'treestyletab:set-subtree-muted';
+export const kCOMMAND_TOGGLE_MUTED            = 'treestyletab:toggle-muted';
 export const kCOMMAND_PERFORM_TABS_DRAG_DROP  = 'treestyletab:perform-tabs-drag-drop';
 export const kCOMMAND_BLOCK_USER_OPERATIONS   = 'treestyletab:block-user-operations';
 export const kCOMMAND_UNBLOCK_USER_OPERATIONS = 'treestyletab:unblock-user-operations';
@@ -100,6 +103,7 @@ export const kCHILDREN      = 'data-child-ids';
 export const kLEVEL         = 'data-level';
 export const kCLOSED_SET_ID = 'data-closed-set-id';
 export const kCURRENT_URI   = 'data-current-uri';
+export const kCONTEXTUAL_IDENTITY_NAME = 'data-contextual-identity-name';
 export const kMAX_TREE_LEVEL = 'data-max-tree-level';
 export const kLABEL_OVERFLOW = 'data-label-overflow';
 
@@ -113,18 +117,12 @@ export const kPERSISTENT_SUBTREE_COLLAPSED = 'subtree-collapsed'; // obsolete
 export const kPERSISTENT_ORIGINAL_OPENER_TAB_ID            = 'data-original-opener-tab-id';
 export const kPERSISTENT_ALREADY_GROUPED_FOR_PINNED_OPENER = 'data-already-grouped-for-pinned-opener';
 
-export const kFAVICON         = 'favicon';
 export const kFAVICON_IMAGE   = 'favicon-image';
 export const kFAVICON_BUILTIN = 'favicon-builtin';
 export const kFAVICON_DEFAULT = 'favicon-default'; // just for backward compatibility, and this should be removed from future versions
 export const kTHROBBER        = 'throbber';
 export const kHIGHLIGHTER     = 'highlighter';
 export const kBURSTER         = 'burster';
-export const kSOUND_BUTTON    = 'sound-button';
-export const kTWISTY          = 'twisty';
-export const kLABEL           = 'label';
-export const kCOUNTER         = 'counter';
-export const kCLOSEBOX        = 'closebox';
 export const kNEWTAB_BUTTON   = 'newtab-button';
 export const kEXTRA_ITEMS_CONTAINER        = 'extra-items-container';
 export const kCONTEXTUAL_IDENTITY_MARKER   = 'contextual-identity-marker';
@@ -201,6 +199,7 @@ export const kTAB_SAFE_STATES = new Set([
   kTAB_STATE_SUBTREE_COLLAPSED,
   kTAB_STATE_GROUP_TAB
 ]);
+export const kTAB_SAFE_STATES_ARRAY = Array.from(kTAB_SAFE_STATES);
 
 export const kTABBAR_STATE_OVERFLOW               = 'overflow';
 export const kTABBAR_STATE_BLOCKING               = 'blocking';
@@ -280,6 +279,11 @@ export const kSUCCESSOR_TAB_CONTROL_NEVER            = 0;
 export const kSUCCESSOR_TAB_CONTROL_SIMULATE_DEFAULT = 1;
 export const kSUCCESSOR_TAB_CONTROL_IN_TREE          = 2;
 
+export const kTREE_DOUBLE_CLICK_BEHAVIOR_NONE                  = 0;
+export const kTREE_DOUBLE_CLICK_BEHAVIOR_TOGGLE_COLLAPSED      = 1;
+// 2 is a retired number for a removed feature
+export const kTREE_DOUBLE_CLICK_BEHAVIOR_CLOSE                 = 3;
+
 export const kDRAG_BEHAVIOR_NONE           = 0;
 export const kDRAG_BEHAVIOR_WHOLE_TREE     = 1 << 0;
 export const kDRAG_BEHAVIOR_ALLOW_BOOKMARK = 1 << 1;
@@ -290,6 +294,7 @@ export const kNEWTAB_OPEN_AS_ORPHAN       = 0;
 export const kNEWTAB_OPEN_AS_CHILD        = 1;
 export const kNEWTAB_OPEN_AS_SIBLING      = 2;
 export const kNEWTAB_OPEN_AS_NEXT_SIBLING = 3;
+export const kNEWTAB_OPEN_AS_NEXT_SIBLING_WITH_INHERITED_CONTAINER = 4;
 
 export const kSCROLL_TO_NEW_TAB_IGNORE      = 0;
 export const kSCROLL_TO_NEW_TAB_IF_POSSIBLE = 1;
@@ -307,9 +312,10 @@ export const kDEFAULT_MIN_INDENT = 3;
 export const kGROUP_TAB_URI = browser.extension.getURL('resources/group-tab.html');
 export const kGROUP_TAB_DEFAULT_TITLE_MATCHER = new RegExp(`^${browser.i18n.getMessage('groupTab_label', '.+')}$`);
 export const kGROUP_TAB_FROM_PINNED_DEFAULT_TITLE_MATCHER = new RegExp(`^${browser.i18n.getMessage('groupTab_fromPinnedTab_label', '.+')}$`);
-export const kSHORTHAND_CUSTOM_URI = /^ext\+treestyletab:([^:]+)(?::(.*))?$/;
-export const kSHORTHAND_ABOUT_URI = /^about:treestyletab-([^\?]+)/;
+export const kSHORTHAND_CUSTOM_URI = /^ext\+treestyletab:([^:?]+)(?:[:?](.*))?$/;
+export const kSHORTHAND_ABOUT_URI = /^about:treestyletab-([^?]+)/;
 export const kSHORTHAND_URIS = {
+  tabbar:  browser.extension.getURL('sidebar/sidebar.html'),
   group:   kGROUP_TAB_URI,
   options: browser.extension.getURL('options/options.html?independent=true'),
   startup: browser.extension.getURL('resources/startup.html'),
@@ -317,4 +323,16 @@ export const kSHORTHAND_URIS = {
   'test-runner': browser.extension.getURL('tests/runner.html')
 };
 
-export const kNOTIFICATION_DEFAULT_ICON = '/resources/icon64.png';
+export const kINSERTION_CONTEXT_MOVED = 1;
+export const kINSERTION_CONTEXT_SHOWN = 2;
+export const kINSERTION_CONTEXT_CREATED = 3;
+
+export const kAGGRESSIVE_OPENER_TAB_DETECTION_RULES_WITH_URL = [
+  { opener: /^about:addons/,
+    child:  /^https:\/\/addons.mozilla.org\/([^\/]+\/)?[^\/]+\/search\// }
+];
+
+export const kNOTIFICATION_DEFAULT_ICON = '/resources/64x64.svg';
+
+// https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/storage/sync
+export const kSYNC_STORAGE_SAFE_QUOTA = (2 ^ 10) * 1024;
