@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2010-2019
+ * Portions created by the Initial Developer are Copyright (C) 2010-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -30,7 +30,9 @@ import RichConfirm from '/extlib/RichConfirm.js';
 import {
   log as internalLogger,
   wait,
-  configs
+  mapAndFilter,
+  configs,
+  sha1sum
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
@@ -52,7 +54,9 @@ function log(...args) {
 const kTREE_DROP_TYPE   = 'application/x-treestyletab-tree';
 const kTYPE_X_MOZ_URL   = 'text/x-moz-url';
 const kTYPE_URI_LIST    = 'text/uri-list';
+const kTYPE_MOZ_TEXT_INTERNAL = 'text/x-moz-text-internal';
 const kBOOKMARK_FOLDER  = 'x-moz-place:';
+const kTYPE_ADDON_DRAG_DATA = `application/x-moz-addon-drag-data;provider=${browser.runtime.id}&id=`;
 
 const kDROP_BEFORE  = 'before';
 const kDROP_ON_SELF = 'self';
@@ -140,7 +144,7 @@ function setDragData(dragData) {
 
 /* helpers */
 
-function getDragDataFromOneTab(tab) {
+function getDragDataFromOneTab(tab, options = {}) {
   if (!tab)
     return {
       tab:      null,
@@ -149,7 +153,7 @@ function getDragDataFromOneTab(tab) {
       instanceId: mInstanceId,
       structure:  []
     };
-  const tabs = getDraggedTabsFromOneTab(tab);
+  const tabs = getDraggedTabsFromOneTab(tab, options);
   return {
     tab,
     tabs,
@@ -159,10 +163,12 @@ function getDragDataFromOneTab(tab) {
   };
 }
 
-function getDraggedTabsFromOneTab(tab) {
+function getDraggedTabsFromOneTab(tab, { asTree } = {}) {
   if (tab.$TST.multiselected)
     return Tab.getSelectedTabs(tab.windowId);
-  return [tab].concat(tab.$TST.descendants);
+  if (asTree)
+    return [tab].concat(tab.$TST.descendants);
+  return [tab];
 }
 
 function sanitizeDragData(dragData) {
@@ -215,41 +221,47 @@ function getDropAction(event) {
     const dragData = info.dragData;
     if (dragData && dragData.instanceId != mInstanceId)
       return [];
-    const tabs     = dragData && dragData.tabs;
-    return tabs && tabs.map(tab => tab && Tab.get(tab.id) || tab).filter(tab => !!tab) || [];
+    const tabIds = dragData && dragData.tabs;
+    return !tabIds ? [] : mapAndFilter(tabIds, id => Tab.get(id) || undefined);
   });
   info.defineGetter('draggedTabIds', () => {
     return info.draggedTabs.map(tab => tab.id);
   });
   info.defineGetter('firstTargetTab', () => {
-    return Tab.getFirstNormalTab(TabsStore.getWindow()) || Tab.getFirstTab(TabsStore.getWindow());
+    return Tab.getFirstNormalTab(TabsStore.getCurrentWindowId()) || Tab.getFirstTab(TabsStore.getCurrentWindowId());
   });
   info.defineGetter('lastTargetTab', () => {
-    return Tab.getLastTab(TabsStore.getWindow());
+    return Tab.getLastTab(TabsStore.getCurrentWindowId());
   });
   info.defineGetter('canDrop', () => {
-    if (info.dropPosition == kDROP_IMPOSSIBLE)
+    if (info.dropPosition == kDROP_IMPOSSIBLE) {
+      log('canDrop:undroppable: dropPosition == kDROP_IMPOSSIBLE');
       return false;
+    }
 
     const draggedTab = info.dragData && info.dragData.tab;
     const isPrivateBrowsingTabDragged = draggedTab && draggedTab.incognito;
-    const isPrivateBrowsingDropTarget = (info.dragOverTab || Tab.getFirstTab(TabsStore.getWindow())).incognito;
+    const isPrivateBrowsingDropTarget = (info.dragOverTab || Tab.getFirstTab(TabsStore.getCurrentWindowId())).incognito;
     if (draggedTab &&
         isPrivateBrowsingTabDragged != isPrivateBrowsingDropTarget) {
+      log('canDrop:undroppable: mispatched incognito status');
       return false;
     }
     else if (info.draggedTab) {
       if (info.action & Constants.kACTION_ATTACH) {
-        if (info.draggedTab.windowId != TabsStore.getWindow()) {
+        if (info.draggedTab.windowId != TabsStore.getCurrentWindowId()) {
           return true;
         }
         if (info.parent &&
             info.parent.id == info.draggedTab.id) {
+          log('canDrop:undroppable: drop on child');
           return false;
         }
         if (info.dragOverTab) {
-          if (info.draggedTabIds.includes(info.dragOverTab.id))
+          if (info.draggedTabIds.includes(info.dragOverTab.id)) {
+            log('canDrop:undroppable: on self');
             return false;
+          }
           const ancestors = info.dragOverTab.$TST.ancestors;
           /* too many function call in this way, so I use alternative way for better performance.
           return !info.draggedTabIds.includes(info.dragOverTab.id) &&
@@ -259,8 +271,10 @@ function getDropAction(event) {
           */
           for (const tab of info.draggedTabs.slice(0).reverse()) {
             const parent = tab.$TST.parent;
-            if (!parent && ancestors.includes(parent))
+            if (!parent && ancestors.includes(parent)) {
+              log('canDrop:undroppable: on descendant');
               return false;
+            }
           }
           return true;
         }
@@ -270,8 +284,10 @@ function getDropAction(event) {
     if (info.dragOverTab &&
         (info.dragOverTab.hidden ||
          (info.dragOverTab.$TST.collapsed &&
-          info.dropPosition != kDROP_AFTER)))
+          info.dropPosition != kDROP_AFTER))) {
+      log('canDrop:undroppable: on hidden tab');
       return false;
+    }
 
     return true;
   });
@@ -288,8 +304,10 @@ function getDropAction(event) {
       info.action       = action;
       if (info.draggedTab &&
           !info.draggedTab.pinned &&
-          info.targetTab.pinned)
+          info.targetTab.pinned) {
+        log('undroppable: above the first tab');
         info.dropPosition = kDROP_IMPOSSIBLE;
+      }
     }
     else if (event.clientY > info.lastTargetTab.$TST.element.getBoundingClientRect().bottom) {
       //log('dragging below the last tab');
@@ -298,8 +316,10 @@ function getDropAction(event) {
       info.action       = action;
       if (info.draggedTab &&
           info.draggedTab.pinned &&
-          !info.targetTab.pinned)
+          !info.targetTab.pinned) {
+        log('undroppable: below the last tab');
         info.dropPosition = kDROP_IMPOSSIBLE;
+      }
     }
     return info;
   }
@@ -310,11 +330,11 @@ function getDropAction(event) {
    * So, if a tab is dragged and the target tab is pinned, then, we
    * have to ignore the [center] area.
    */
-  const onPinnedTab         = targetTab.pinned;
-  const dropAreasCount      = (info.draggedTab && onPinnedTab) ? 2 : 3 ;
+  const onFaviconizedTab    = targetTab.pinned && configs.faviconizePinnedTabs;
+  const dropAreasCount      = (info.draggedTab && onFaviconizedTab) ? 2 : 3 ;
   const targetTabRect       = targetTab.$TST.element.getBoundingClientRect();
-  const targetTabCoordinate = onPinnedTab ? targetTabRect.left : targetTabRect.top ;
-  const targetTabSize       = onPinnedTab ? targetTabRect.width : targetTabRect.height ;
+  const targetTabCoordinate = onFaviconizedTab ? targetTabRect.left : targetTabRect.top ;
+  const targetTabSize       = onFaviconizedTab ? targetTabRect.width : targetTabRect.height ;
   let beforeOrAfterDropAreaSize;
   if (dropAreasCount == 2) {
     beforeOrAfterDropAreaSize = Math.round(targetTabSize / dropAreasCount);
@@ -322,7 +342,7 @@ function getDropAction(event) {
   else { // enlarge the area to dop something on the tab itself
     beforeOrAfterDropAreaSize = Math.round(targetTabSize / 4);
   }
-  const eventCoordinate = onPinnedTab ? event.clientX : event.clientY;
+  const eventCoordinate = onFaviconizedTab ? event.clientX : event.clientY;
   //log('coordinates: ', {
   //  event: eventCoordinate,
   //  targetTab: targetTabCoordinate,
@@ -355,6 +375,12 @@ function getDropAction(event) {
       if (info.draggedTab &&
           info.draggedTab.pinned != targetTab.pinned)
         info.dropPosition = kDROP_IMPOSSIBLE;
+      if (info.draggedTab &&
+          info.insertBefore == info.draggedTab) // failsafe
+        info.insertBefore = configs.insertNewChildAt == Constants.kINSERT_FIRST ?
+          info.draggedTab.$TST.unsafeNextTab :
+          (info.draggedTab.$TST.nextSiblingTab ||
+           info.draggedTab.$TST.unsafeNearestFollowingForeignerTab);
       if (configs.debug)
         log(' calculated info: ', info);
     }; break;
@@ -362,7 +388,9 @@ function getDropAction(event) {
     case kDROP_BEFORE: {
       log('drop position = before ', info.targetTab.id);
       const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(info.draggedTab, {
-        insertBefore: targetTab
+        context:      Constants.kINSERTION_CONTEXT_MOVED,
+        insertBefore: targetTab,
+        windowId:     TabsStore.getCurrentWindowId()
       });
       if (referenceTabs.parent)
         info.parent = referenceTabs.parent;
@@ -374,7 +402,10 @@ function getDropAction(event) {
       //if (info.insertBefore)
       //  log('insertBefore = ', dumpTab(info.insertBefore));
       if (info.draggedTab &&
-          info.draggedTab.pinned != targetTab.pinned)
+          ((info.draggedTab.pinned &&
+            targetTab.$TST.followsUnpinnedTab) ||
+           (!info.draggedTab.pinned &&
+            targetTab.pinned)))
         info.dropPosition = kDROP_IMPOSSIBLE;
       if (configs.debug)
         log(' calculated info: ', info);
@@ -401,7 +432,10 @@ function getDropAction(event) {
              +-----------------------------------------------------
         */
         if (info.draggedTab &&
-            info.draggedTab.id == info.insertBefore.id) {
+            info.draggedTab.$TST &&
+            info.draggedTab.$TST.nearestVisibleFollowingTab &&
+            info.draggedTab.$TST.nearestVisibleFollowingTab.id == info.insertBefore.id) {
+          log('special case: promote tab');
           info.action      = Constants.kACTION_MOVE | Constants.kACTION_ATTACH;
           info.parent      = targetTab.$TST.parent;
           let insertBefore = targetTab.$TST.nextSiblingTab;
@@ -415,7 +449,10 @@ function getDropAction(event) {
         }
       }
       if (info.draggedTab &&
-          info.draggedTab.pinned != !!targetTab.$TST.precedesPinnedTab)
+          ((info.draggedTab.pinned &&
+            !targetTab.pinned) ||
+           (!info.draggedTab.pinned &&
+            targetTab.$TST.precedesPinnedTab)))
         info.dropPosition = kDROP_IMPOSSIBLE;
       if (configs.debug)
         log(' calculated info: ', info);
@@ -444,14 +481,14 @@ export function clearDropPosition() {
 }
 
 export function clearDraggingTabsState() {
-  for (const tab of Tab.getDraggingTabs(TabsStore.getWindow(), { iterator: true })) {
+  for (const tab of Tab.getDraggingTabs(TabsStore.getCurrentWindowId(), { iterator: true })) {
     tab.$TST.removeState(Constants.kTAB_STATE_DRAGGING);
     TabsStore.removeDraggingTab(tab);
   }
 }
 
 export function clearDraggingState() {
-  const window = TabsStore.windows.get(TabsStore.getWindow());
+  const window = TabsStore.windows.get(TabsStore.getCurrentWindowId());
   window.classList.remove(kTABBAR_STATE_TAB_DRAGGING);
   document.documentElement.classList.remove(kTABBAR_STATE_TAB_DRAGGING);
   document.documentElement.classList.remove(kTABBAR_STATE_LINK_DRAGGING);
@@ -482,7 +519,7 @@ function collapseAutoExpandedTabsWhileDragging() {
 async function handleDroppedNonTabItems(event, dropActionInfo) {
   event.stopPropagation();
 
-  const uris = retrieveURIsFromDragEvent(event);
+  const uris = await retrieveURIsFromDragEvent(event);
   // uris.forEach(uRI => {
   //   if (uRI.indexOf(Constants.kURI_BOOKMARK_FOLDER) != 0)
   //     securityCheck(uRI, event);
@@ -511,35 +548,61 @@ async function handleDroppedNonTabItems(event, dropActionInfo) {
   BackgroundConnection.sendMessage({
     type:           Constants.kCOMMAND_NEW_TABS,
     uris,
-    windowId:       TabsStore.getWindow(),
+    windowId:       TabsStore.getCurrentWindowId(),
     parentId:       dropActionInfo.parent && dropActionInfo.parent.id,
     insertBeforeId: dropActionInfo.insertBefore && dropActionInfo.insertBefore.id,
     insertAfterId:  dropActionInfo.insertAfter && dropActionInfo.insertAfter.id
   });
 }
 
-function retrieveURIsFromDragEvent(event) {
+const ACCEPTABLE_DRAG_DATA_TYPES = [
+  kTYPE_URI_LIST,
+  kTYPE_X_MOZ_URL,
+  kTYPE_MOZ_TEXT_INTERNAL,
+  'text/plain'
+];
+
+async function retrieveURIsFromDragEvent(event) {
   log('retrieveURIsFromDragEvent');
-  const dt    = event.dataTransfer;
-  const types = [
-    kTYPE_URI_LIST,
-    kTYPE_X_MOZ_URL,
-    'text/plain'
-  ];
+  const dt = event.dataTransfer;
   let urls = [];
-  for (const type of types) {
-    const urlData  = dt.getData(type);
+  for (const type of ACCEPTABLE_DRAG_DATA_TYPES) {
+    const urlData = dt.getData(type);
     if (urlData)
       urls = urls.concat(retrieveURIsFromData(urlData, type));
     if (urls.length)
       break;
   }
+  for (const type of dt.types) {
+    if (!/^application\/x-moz-addon-drag-data;(.+)$/.test(type))
+      continue;
+    const params     = RegExp.$1;
+    const providerId = /provider=([^;&]+)/.test(params) && RegExp.$1;
+    const dataId     = /id=([^;&]+)/.test(params) && RegExp.$1;
+    try {
+      const dragData = await browser.runtime.sendMessage(providerId, {
+        type: 'get-drag-data',
+        id:   dataId
+      });
+      if (!dragData || typeof dragData != 'object')
+        break;
+      for (const type of ACCEPTABLE_DRAG_DATA_TYPES) {
+        const urlData = dragData[type];
+        if (urlData)
+          urls = urls.concat(retrieveURIsFromData(urlData, type));
+        if (urls.length)
+          break;
+      }
+    }
+    catch(_error) {
+    }
+  }
   log(' => retrieved: ', urls);
-  urls = urls.filter(uRI =>
-    uRI &&
-      uRI.length &&
-      uRI.indexOf(kBOOKMARK_FOLDER) == 0 ||
-      !/^\s*(javascript|data):/.test(uRI)
+  urls = urls.filter(uri =>
+    uri &&
+      uri.length &&
+      uri.indexOf(kBOOKMARK_FOLDER) == 0 ||
+      !/^\s*(javascript|data):/.test(uri)
   );
   log('  => filtered: ', urls);
 
@@ -571,6 +634,13 @@ function retrieveURIsFromData(data, type) {
           return index % 2 == 0;
         });
 
+    case kTYPE_MOZ_TEXT_INTERNAL:
+      return data
+        .replace(/\r/g, '\n')
+        .replace(/\n\n+/g, '\n')
+        .trim()
+        .split('\n');
+
     case 'text/plain':
       return data
         .replace(/\r/g, '\n')
@@ -628,18 +698,88 @@ async function getDroppedLinksOnTabBehavior() {
 
 let mFinishCanceledDragOperation;
 let mLastDragStartTime = 0;
+let mCurrentDragDataForExternalsId = null;
+let mCurrentDragDataForExternals = null;
 
 export const onDragStart = EventUtils.wrapWithErrorHandler(function onDragStart(event, options = {}) {
   log('onDragStart: start ', event, options);
   clearDraggingTabsState(); // clear previous state anyway
+  if (configs.enableWorkaroundForBug1548949)
+    configs.workaroundForBug1548949DroppedTabs = '';
 
-  const behavior = 'behavior' in options ? options.behavior :
+  let draggedTab = options.tab || EventUtils.getTabFromEvent(event);
+  let behavior = 'behavior' in options ? options.behavior :
     event.shiftKey ? configs.tabDragBehaviorShift :
       configs.tabDragBehavior;
-  const allowBookmark = !!(behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK);
 
-  const dragData = getDragDataFromOneTab(options.tab || EventUtils.getTabFromEvent(event));
-  dragData.individualOnOutside = !dragData.tab.$TST.multiselected && !(behavior & Constants.kDRAG_BEHAVIOR_WHOLE_TREE);
+  mCurrentDragDataForExternalsId = `${parseInt(Math.random() * 65000)}-${Date.now()}`;
+  mCurrentDragDataForExternals = {};
+
+  const originalTarget = EventUtils.getElementOriginalTarget(event);
+  const extraTabContentsDragData = originalTarget && JSON.parse(originalTarget.dataset && originalTarget.dataset.dragData || 'null');
+  log('onDragStart: extraTabContentsDragData = ', extraTabContentsDragData);
+  let dataOverridden = false;
+  if (extraTabContentsDragData) {
+    const dataSet = detectOverrideDragDataSet(extraTabContentsDragData, event);
+    log('onDragStart: detected override data set = ', dataSet);
+    /*
+      expected drag data format:
+        Tab:
+          { type: 'tab',
+            data: { asTree:      (boolean),
+                    allowDetach: (boolean, will detach the tab to new window),
+                    allowLink:   (boolean, will create link/bookmark from the tab) }}
+        other arbitrary types:
+          { type:          'text/plain',
+            data:          'something text',
+            effectAllowed: 'copy' }
+          { type:          'text/x-moz-url',
+            data:          'http://example.com/\nExample Link',
+            effectAllowed: 'copyMove' }
+          ...
+    */
+    let tabIsGiven = false;
+    for (const data of dataSet) {
+      if (!data)
+        continue;
+      switch (data.type) {
+        case 'tab':
+          if (data.data.id) {
+            const tab = Tab.get(data.data.id);
+            if (tab) {
+              tabIsGiven = true;
+              draggedTab = tab;
+              behavior   = Constants.kDRAG_BEHAVIOR_NONE;
+              if (data.data.allowDetach)
+                behavior |= Constants.kDRAG_BEHAVIOR_TEAR_OFF;
+              if (data.data.allowLink)
+                behavior |= Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK;
+              if (data.data.asTree)
+                behavior |= Constants.kDRAG_BEHAVIOR_WHOLE_TREE;
+            }
+          }
+          break;
+        default: {
+          const dt = event.dataTransfer;
+          dt.effectAllowed = data.effectAllowed || 'copy';
+          const type       = String(data.type);
+          const stringData = String(data.data);
+          dt.setData(type, stringData);
+          //*** We need to sanitize drag data from helper addons, because
+          //they can have sensitive data...
+          //mCurrentDragDataForExternals[type] = stringData;
+          dataOverridden = true;
+        }; break;
+      }
+    }
+    if (!tabIsGiven && dataOverridden)
+      return;
+  }
+
+  const allowBookmark = !!(behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK);
+  const asTree = !!(behavior & Constants.kDRAG_BEHAVIOR_WHOLE_TREE);
+  const dragData = getDragDataFromOneTab(draggedTab, { asTree });
+  dragData.individualOnOutside = !dragData.tab.$TST.multiselected && !asTree
   dragData.behavior = behavior;
   if (!dragData.tab) {
     log('onDragStart: canceled / no dragged tab from drag data');
@@ -656,8 +796,8 @@ export const onDragStart = EventUtils.wrapWithErrorHandler(function onDragStart(
     mLastDragEnteredTarget = tab.$TST.element;
     const startOnClosebox = mDragTargetIsClosebox = mousedown.detail.closebox;
     if (startOnClosebox)
-      mLastDragEnteredTarget = SidebarTabs.getClosebox(tab);
-    const windowId = TabsStore.getWindow();
+      mLastDragEnteredTarget = tab.$TST.element.closeBoxElement;
+    const windowId = TabsStore.getCurrentWindowId();
     TSTAPI.sendMessage({
       type:   TSTAPI.kNOTIFY_TAB_DRAGSTART,
       tab:    new TSTAPI.TreeItem(tab),
@@ -698,34 +838,39 @@ export const onDragStart = EventUtils.wrapWithErrorHandler(function onDragStart(
   mCurrentDragData = sanitizedDragData;
   browser.runtime.sendMessage({
     type:     Constants.kCOMMAND_BROADCAST_CURRENT_DRAG_DATA,
-    windowId: TabsStore.getWindow(),
+    windowId: TabsStore.getCurrentWindowId(),
     dragData: sanitizedDragData
   }).catch(ApiTabs.createErrorSuppressor());
 
-  const mozUrl  = [];
-  const urlList = [];
-  for (const draggedTab of dragData.tabs) {
-    draggedTab.$TST.addState(Constants.kTAB_STATE_DRAGGING);
-    TabsStore.addDraggingTab(draggedTab);
-    if (!dragData.individualOnOutside ||
-        mozUrl.length == 0) {
-      mozUrl.push(`${draggedTab.url}\n${draggedTab.title}`);
-      urlList.push(`#${draggedTab.title}\n${draggedTab.url}`);
+  if (!dataOverridden) {
+    const mozUrl  = [];
+    const urlList = [];
+    for (const draggedTab of dragData.tabs) {
+      draggedTab.$TST.addState(Constants.kTAB_STATE_DRAGGING);
+      TabsStore.addDraggingTab(draggedTab);
+      if (!dragData.individualOnOutside ||
+          mozUrl.length == 0) {
+        mozUrl.push(`${draggedTab.url}\n${draggedTab.title}`);
+        urlList.push(`#${draggedTab.title}\n${draggedTab.url}`);
+      }
+    }
+    mCurrentDragDataForExternals[kTYPE_X_MOZ_URL] = mozUrl.join('\n');
+    mCurrentDragDataForExternals[kTYPE_URI_LIST] = urlList.join('\n');
+    if (allowBookmark) {
+      log('set kTYPE_X_MOZ_URL');
+      dt.setData(kTYPE_X_MOZ_URL, mCurrentDragDataForExternals[kTYPE_X_MOZ_URL]);
+      log('set kTYPE_URI_LIST');
+      dt.setData(kTYPE_URI_LIST, mCurrentDragDataForExternals[kTYPE_URI_LIST]);
     }
   }
-  if (allowBookmark) {
-    log('set kTYPE_X_MOZ_URL');
-    dt.setData(kTYPE_X_MOZ_URL, mozUrl.join('\n'));
-    log('set kTYPE_URI_LIST');
-    dt.setData(kTYPE_URI_LIST, urlList.join('\n'));
-  }
+  dt.setData(`${kTYPE_ADDON_DRAG_DATA}${mCurrentDragDataForExternalsId}`, JSON.stringify(mCurrentDragDataForExternals));
 
   if (options.tab) {
     const tabRect = options.tab.$TST.element.getBoundingClientRect();
     dt.setDragImage(options.tab.$TST.element, event.clientX - tabRect.left, event.clientY - tabRect.top);
   }
 
-  TabsStore.windows.get(TabsStore.getWindow()).classList.add(kTABBAR_STATE_TAB_DRAGGING);
+  TabsStore.windows.get(TabsStore.getCurrentWindowId()).classList.add(kTABBAR_STATE_TAB_DRAGGING);
   document.documentElement.classList.add(kTABBAR_STATE_TAB_DRAGGING);
 
   // The drag operation can be canceled by something, then
@@ -759,14 +904,70 @@ export const onDragStart = EventUtils.wrapWithErrorHandler(function onDragStart(
   TSTAPI.sendMessage({
     type:     TSTAPI.kNOTIFY_NATIVE_TAB_DRAGSTART,
     tab:      new TSTAPI.TreeItem(tab),
-    windowId: TabsStore.getWindow()
+    windowId: TabsStore.getCurrentWindowId()
   }, { tabProperties: ['tab'] }).catch(_error => {});
 
   mLastDragStartTime = Date.now();
+  // Don't store raw URLs to save privacy!
+  sha1sum(dragData.tabs.map(tab => tab.url).join('\n')).then(digest => {
+    configs.lastDraggedTabs = {
+      tabIds:     dragData.tabs.map(tab => tab.id),
+      urlsDigest: digest
+    };
+  });
 
   log('onDragStart: started');
 });
 
+/* acceptable input:
+  {
+    "default":    { type: ..., data: ... },
+    "Ctrl":       { type: ..., data: ... },
+    "MacCtrl":    { type: ..., data: ... },
+    "Ctrl+Shift": { type: ..., data: ... },
+    "Alt-Shift":  { type: ..., data: ... },
+    ...
+  }
+*/
+const isMac = /^Mac/i.test(navigator.platform);
+function detectOverrideDragDataSet(dataSet, event) {
+  if (Array.isArray(dataSet))
+    return dataSet.map(oneDataSet => detectOverrideDragDataSet(oneDataSet, event)).flat();
+
+  if ('type' in dataSet)
+    return [dataSet];
+
+  const keys = [];
+  if (event.altKey)
+    keys.push('alt');
+  if (event.ctrlKey) {
+    if (isMac)
+      keys.push('macctrl');
+    else
+      keys.push('ctrl');
+  }
+  if (event.metaKey) {
+    if (isMac)
+      keys.push('command');
+    else
+      keys.push('meta');
+  }
+  if (event.shiftKey)
+    keys.push('shift');
+  const findKey = keys.sort().join('+') || 'default';
+
+  for (const key of Object.keys(dataSet)) {
+    const normalizedKey = key.split(/[-\+]/).filter(part => !!part).sort().join('+').toLowerCase();
+    if (normalizedKey != findKey)
+      continue;
+    if (Array.isArray(dataSet[key]))
+      return dataSet[key];
+    else
+      return [dataSet[key]];
+  }
+  return [];
+}
+
 function getTabDragBehaviorNotificationMessageType(behavior, count) {
   if (behavior & Constants.kDRAG_BEHAVIOR_WHOLE_TREE && count > 1) {
     if (behavior & Constants.kDRAG_BEHAVIOR_ALLOW_BOOKMARK)
@@ -872,7 +1073,7 @@ function onDragEnter(event) {
         info.dragData.tabs.some(tab => tab.id == enteredTab.id)
       );
     }
-    TabsStore.windows.get(TabsStore.getWindow()).classList.add(kTABBAR_STATE_TAB_DRAGGING);
+    TabsStore.windows.get(TabsStore.getCurrentWindowId()).classList.add(kTABBAR_STATE_TAB_DRAGGING);
     document.documentElement.classList.add(kTABBAR_STATE_TAB_DRAGGING);
   }
   catch(_e) {
@@ -968,7 +1169,7 @@ function onDragLeave(event) {
         }, 10);
       }
       else {
-        leftFromTabBar = !enteredTab || enteredTab.windowId != TabsStore.getWindow();
+        leftFromTabBar = !enteredTab || enteredTab.windowId != TabsStore.getCurrentWindowId();
         if (onDragLeave.delayedLeftFromDraggedTabs) {
           clearTimeout(onDragLeave.delayedLeftFromDraggedTabs);
           delete onDragLeave.delayedLeftFromDraggedTabs;
@@ -1008,6 +1209,7 @@ function onDrop(event) {
   }
 
   const dropActionInfo = getDropAction(event);
+  log('onDrop ', dropActionInfo, event.dataTransfer);
 
   if (!dropActionInfo.canDrop) {
     log('undroppable');
@@ -1026,6 +1228,10 @@ function onDrop(event) {
   if (dropActionInfo.dragData &&
       dropActionInfo.dragData.tab) {
     log('there are dragged tabs');
+    if (configs.enableWorkaroundForBug1548949) {
+      configs.workaroundForBug1548949DroppedTabs = dropActionInfo.dragData.tabs.map(tab => `${mInstanceId}/${tab.id}`).join('\n');
+      log('workaround for bug 1548949: setting last dropped tabs: ', configs.workaroundForBug1548949DroppedTabs);
+    }
     const fromOtherProfile = dropActionInfo.dragData.instanceId != mInstanceId;
     BackgroundConnection.sendMessage({
       type:                Constants.kCOMMAND_PERFORM_TABS_DRAG_DROP,
@@ -1036,20 +1242,54 @@ function onDrop(event) {
       attachToId:          dropActionInfo.parent && dropActionInfo.parent.id,
       insertBeforeId:      dropActionInfo.insertBefore && dropActionInfo.insertBefore.id,
       insertAfterId:       dropActionInfo.insertAfter && dropActionInfo.insertAfter.id,
-      destinationWindowId: TabsStore.getWindow(),
+      destinationWindowId: TabsStore.getCurrentWindowId(),
       duplicate:           !fromOtherProfile && dt.dropEffect == 'copy',
       import:              fromOtherProfile
     });
     return;
   }
 
+  if (dt.types.includes(kTYPE_MOZ_TEXT_INTERNAL) &&
+      configs.guessDraggedNativeTabs) {
+    log('there are dragged native tabs');
+    const url = dt.getData(kTYPE_MOZ_TEXT_INTERNAL);
+    browser.tabs.query({ url }).then(tabs => {
+      if (!tabs.length) {
+        log('=> from other profile');
+        handleDroppedNonTabItems(event, dropActionInfo);
+        return;
+      }
+      log('=> possible dragged tabs: ', tabs);
+      tabs = tabs.sort((a, b) => b.lastAccessed - a.lastAccessed);
+      if (configs.enableWorkaroundForBug1548949) {
+        configs.workaroundForBug1548949DroppedTabs = tabs.map(tab => `${mInstanceId}/${tab.id}`).join('\n');
+        log('workaround for bug 1548949: setting last dropped tabs: ', configs.workaroundForBug1548949DroppedTabs);
+      }
+      const recentTab = tabs[0];
+      BackgroundConnection.sendMessage({
+        type:                Constants.kCOMMAND_PERFORM_TABS_DRAG_DROP,
+        windowId:            recentTab.windowId,
+        tabs:                [recentTab],
+        structure:           [-1],
+        action:              dropActionInfo.action,
+        attachToId:          dropActionInfo.parent && dropActionInfo.parent.id,
+        insertBeforeId:      dropActionInfo.insertBefore && dropActionInfo.insertBefore.id,
+        insertAfterId:       dropActionInfo.insertAfter && dropActionInfo.insertAfter.id,
+        destinationWindowId: TabsStore.getCurrentWindowId(),
+        duplicate:           dt.dropEffect == 'copy',
+        import:              false
+      });
+    });
+    return;
+  }
+
   log('link or bookmark item is dropped');
   handleDroppedNonTabItems(event, dropActionInfo);
 }
 onDrop = EventUtils.wrapWithErrorHandler(onDrop);
 
 async function onDragEnd(event) {
-  log('onDragEnd, ', { mDraggingOnSelfWindow, mDraggingOnDraggedTabs });
+  log('onDragEnd, ', { mDraggingOnSelfWindow, mDraggingOnDraggedTabs, dropEffect: event.dataTransfer.dropEffect });
 
   let dragData = event.dataTransfer.getData(kTREE_DROP_TYPE);
   dragData = (dragData && JSON.parse(dragData)) || mCurrentDragData;
@@ -1058,19 +1298,46 @@ async function onDragEnd(event) {
     dragData.tabs = dragData.tabs && dragData.tabs.map(tab => tab && Tab.get(tab.id) || tab);
   }
 
+  TSTAPI.sendMessage({
+    type:     TSTAPI.kNOTIFY_NATIVE_TAB_DRAGEND,
+    windowId: TabsStore.getCurrentWindowId()
+  }).catch(_error => {});
+
   // Don't clear flags immediately, because they are referred by following operations in this function.
   setTimeout(finishDrag, 0);
 
   if (!(dragData.behavior & Constants.kDRAG_BEHAVIOR_TEAR_OFF))
     return;
 
+  let handledBySomeone = event.dataTransfer.dropEffect != 'none';
+
   if (event.dataTransfer.getData(kTYPE_URI_LIST)) {
     log('do nothing by TST for dropping just for bookmarking or linking');
     return;
   }
+  else if (configs.enableWorkaroundForBug1548949) {
+    // Due to the bug 1548949, "dropEffect" can become "move" even if no one
+    // actually handles the drop. Basically kTREE_DROP_TYPE is not processible
+    // by anyone except TST, so, we can treat the dropend as "dropped outside
+    // the sidebar" when all dragged tabs are exact same to last tabs dropped
+    // to a sidebar on this Firefox instance.
+    // The only one exception is the case: tabs have been dropped to a TST
+    // sidebar on any other Firefox instance. In this case tabs dropped to the
+    // foreign Firefox will become duplicated: imported to the foreign Firefox
+    // and teared off from the source window. This is clearly undesider
+    // behavior from misdetection, but I decide to ignore it because it looks
+    // quite rare case.
+    await wait(250); // wait until "workaroundForBug1548949DroppedTabs" is synchronized
+    const draggedTabs = dragData.tabs.map(tab => `${mInstanceId}/${tab.id}`).join('\n');
+    const lastDroppedTabs = configs.workaroundForBug1548949DroppedTabs;
+    handledBySomeone = draggedTabs == lastDroppedTabs;
+    log('workaround for bug 1548949: detect dragged tabs are handled by me or not.',
+        { handledBySomeone, draggedTabs, lastDroppedTabs });
+    configs.workaroundForBug1548949DroppedTabs = null;
+  }
 
   if (event.dataTransfer.mozUserCancelled ||
-      event.dataTransfer.dropEffect != 'none' ||
+      handledBySomeone ||
       //event.shiftKey || // don't ignore shift-drop, because it can be used to drag a parent tab as an individual tab.
       !configs.moveDroppedTabToNewWindowForUnhandledDragEvent) {
     log('dragged items are processed by someone: ', event.dataTransfer.dropEffect);
@@ -1081,7 +1348,7 @@ async function onDragEnd(event) {
   const windowY = window.mozInnerScreenY;
   const windowW = window.innerWidth;
   const windowH = window.innerHeight;
-  const offset  = dragData.tab.$TST.element.getBoundingClientRect().height / 2;
+  const offset  = dragData.tab.$TST.element && dragData.tab.$TST.element.getBoundingClientRect().height / 2;
   // Workaround for https://bugzilla.mozilla.org/show_bug.cgi?id=1561522
   const fixedEventScreenX = event.screenX / window.devicePixelRatio;
   const fixedEventScreenY = event.screenY / window.devicePixelRatio;
@@ -1108,8 +1375,9 @@ async function onDragEnd(event) {
        fixedEventScreenY >= windowY - offset &&
        fixedEventScreenX <= windowX + windowW + offset &&
        fixedEventScreenY <= windowY + windowH + offset) ||
+      // Workaround for https://bugzilla.mozilla.org/show_bug.cgi?id=1561879
       // On macOS sometimes drag gesture is canceled immediately with (0,0) coordinates.
-      (Date.now() - mLastDragStartTime < 100 &&
+      (Date.now() - mLastDragStartTime < configs.maximumDelayForBug1561879 &&
        event.screenX == 0 &&
        event.screenY == 0)) {
     log('dropped near the tab bar (from coordinates): detaching is canceled');
@@ -1149,9 +1417,11 @@ function finishDrag() {
 
   wait(100).then(() => {
     mCurrentDragData = null;
+    mCurrentDragDataForExternalsId = null;
+    mCurrentDragDataForExternals = null;
     browser.runtime.sendMessage({
       type:     Constants.kCOMMAND_BROADCAST_CURRENT_DRAG_DATA,
-      windowId: TabsStore.getWindow(),
+      windowId: TabsStore.getCurrentWindowId(),
       dragData: null
     }).catch(ApiTabs.createErrorSuppressor());
   });
@@ -1180,7 +1450,7 @@ function onTSTAPIDragEnter(event) {
     return;
   let target = tab.$TST.element;
   if (mDragTargetIsClosebox && EventUtils.isEventFiredOnClosebox(event))
-    target = SidebarTabs.getClosebox(tab);
+    target = tab.$TST.element.closeBoxElement;
   cancelDelayedTSTAPIDragExitOn(target);
   if (tab &&
       (!mDragTargetIsClosebox ||
@@ -1206,7 +1476,7 @@ function onTSTAPIDragExit(event) {
     return;
   let target = tab.$TST.element;
   if (mDragTargetIsClosebox && EventUtils.isEventFiredOnClosebox(event))
-    target = SidebarTabs.getClosebox(tab);
+    target = tab.$TST.element.closeBoxElement;
   cancelDelayedTSTAPIDragExitOn(target);
   target.onTSTAPIDragExitTimeout = setTimeout(() => {
     delete target.onTSTAPIDragExitTimeout;
@@ -1241,3 +1511,13 @@ function onMessage(message, _sender, _respond) {
   }
 }
 
+
+TSTAPI.onMessageExternal.addListener((message, _sender) => {
+  switch (message.type) {
+    case TSTAPI.kGET_DRAG_DATA:
+      if (mCurrentDragDataForExternals &&
+          message.id == mCurrentDragDataForExternalsId)
+        return Promise.resolve(mCurrentDragDataForExternals);
+      break;
+  }
+});
