@@ -11,7 +11,8 @@ import {
   log as internalLogger,
   dumpTab,
   wait,
-  configs
+  configs,
+  sanitizeForHTMLText
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
@@ -19,14 +20,15 @@ import * as ApiTabs from '/common/api-tabs.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TSTAPI from '/common/tst-api.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
-import * as Permissions from '/common/permissions.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
+import * as UserOperationBlocker from '/common/user-operation-blocker.js';
 
 import Tab from '/common/Tab.js';
 
 import * as TabsGroup from './tabs-group.js';
 import * as TabsOpen from './tabs-open.js';
 import * as Tree from './tree.js';
+import * as Background from './background.js';
 
 function log(...args) {
   internalLogger('background/handle-group-tabs', ...args);
@@ -49,30 +51,34 @@ export async function tryInitGroupTab(tab) {
     matchAboutBlank: true
   };
   try {
-    const initialized = await browser.tabs.executeScript(tab.id, Object.assign({}, scriptOptions, {
-      code:  'window.initialized',
-    })).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-    if (initialized[0])
+    const ready = await browser.tabs.executeScript(tab.id, {
+      ...scriptOptions,
+      code:  'window.ready',
+    }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+    if (ready[0])
       return;
   }
   catch(_e) {
   }
   try {
-    const titleElementExists = await browser.tabs.executeScript(tab.id, Object.assign({}, scriptOptions, {
+    const titleElementExists = await browser.tabs.executeScript(tab.id, {
+      ...scriptOptions,
       code:  '!!document.querySelector("#title")',
-    })).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+    }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
     if (!titleElementExists[0] && tab.status == 'complete') // we need to load resources/group-tab.html at first.
       return browser.tabs.update(tab.id, { url: tab.url }).catch(ApiTabs.createErrorSuppressor());
   }
   catch(_e) {
   }
-  browser.tabs.executeScript(tab.id, Object.assign({}, scriptOptions, {
+  browser.tabs.executeScript(tab.id, {
+    ...scriptOptions,
     //file:  '/common/l10n.js'
     file:  '/extlib/l10n-classic.js' // ES module does not supported as a content script...
-  })).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-  browser.tabs.executeScript(tab.id, Object.assign({}, scriptOptions, {
+  }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+  browser.tabs.executeScript(tab.id, {
+    ...scriptOptions,
     file:  '/resources/group-tab.js'
-  })).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
+  }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
 
   if (tab.$TST.states.has(Constants.kTAB_STATE_UNREAD)) {
     tab.$TST.removeState(Constants.kTAB_STATE_UNREAD, { permanently: true });
@@ -303,7 +309,8 @@ Tab.onBeforeCreate.addListener(async (tab, info) => {
       window.openedNewTabs.set(tab.id, {
         id:       tab.id,
         openerId: openerTab && openerTab.id,
-        openerIsPinned: openerTab && openerTab.pinned
+        openerIsPinned: openerTab && openerTab.pinned,
+        maybeFromBookmark: tab.$TST.maybeFromBookmark
       });
     }
   }
@@ -325,11 +332,12 @@ async function onNewTabsTimeout(window) {
     await Tab.waitUntilMovedAll(window.id);
 
   let tabReferences = Array.from(window.openedNewTabs.values());
-  log('onNewTabsTimeout ', tabReferences);
+  log('onNewTabsTimeout for ', tabReferences);
 
   window.openedNewTabs.clear();
 
   const blocked = TSTAPI.isGroupingBlocked();
+  log('  blocked?: ', blocked);
   tabReferences = tabReferences.filter(tabReference => {
     if (blocked || !tabReference.id)
       return false;
@@ -339,8 +347,10 @@ async function onNewTabsTimeout(window) {
     const uniqueId = tab && tab.$TST && tab.$TST.uniqueId;
     return !uniqueId || (!uniqueId.duplicated && !uniqueId.restored);
   });
-  if (tabReferences.length == 0)
+  if (tabReferences.length == 0) {
+    log(' => no tab to be grouped.');
     return;
+  }
 
   if (tabReferences.length > 1) {
     for (const tabReference of tabReferences) {
@@ -360,45 +370,69 @@ async function tryGroupNewTabs() {
   if (!tabReferences)
     return;
 
-  log('tryGroupNewTabs ', tabReferences);
+  log('tryGroupNewTabs for ', tabReferences);
   tryGroupNewTabs.running = true;
   try {
+    const fromPinned    = [];
+    const fromOthers    = [];
     // extract only pure new tabs
-    let tabs = tabReferences.map(tabReference => {
+    for (const tabReference of tabReferences) {
       const tab = Tab.get(tabReference.id);
       if (!tab)
-        return null;
-      // We should check the config here, because to-be-grouped tabs should be
-      // ignored by the handler for "autoAttachSameSiteOrphan" behavior.
-      const shouldBeGrouped = tabReference.openerIsPinned ? configs.autoGroupNewTabsFromPinned : configs.autoGroupNewTabs;
-      if (!shouldBeGrouped)
-        return null;
+        continue;
       if (tabReference.openerTabId)
         tab.openerTabId = parseInt(tabReference.openerTabId); // restore the opener information
-      return tab;
-    }).filter(tab => !!tab);
-    const uniqueIds = tabs.map(tab => tab.$TST.uniqueId);
-    tabs = tabs.filter((id, index) => {
-      const uniqueId = uniqueIds[index];
-      return !uniqueId.duplicated && !uniqueId.restored;
-    });
-    Tab.sort(tabs);
-
-    let newRootTabs = Tab.collectRootTabs(tabs)
-      .filter(tab => !tab.$TST.isGroupTab);
-    if (newRootTabs.length <= 0)
-      return;
-
-    const newRootTabsFromPinned = newRootTabs.filter(tab => tab.$TST.hasPinnedOpener);
-    if (newRootTabsFromPinned.length > 0) {
-      newRootTabs = newRootTabs.filter(tab => !newRootTabsFromPinned.includes(tab));
-      await tryGroupNewTabsFromPinnedOpener(newRootTabsFromPinned);
+      const uniqueId = tab.$TST.uniqueId;
+      if (tab.$TST.isGroupTab || uniqueId.duplicated || uniqueId.restored)
+        continue;
+      if (tabReference.openerIsPinned) {
+        // We should check the "autoGroupNewTabsFromPinned" config here,
+        // because to-be-grouped tabs should be ignored by the handler for
+        // "autoAttachSameSiteOrphan" behavior.
+        if (tab.$TST.hasPinnedOpener &&
+            configs.autoGroupNewTabsFromPinned)
+          fromPinned.push(tab);
+      }
+      else {
+        fromOthers.push(tab);
+      }
     }
-    if (newRootTabs.length > 1 &&
-        configs.autoGroupNewTabs) {
-      const granted = await confirmToAutoGroupNewTabs(tabs);
-      if (granted)
-        await TabsGroup.groupTabs(newRootTabs, { broadcast: true });
+
+    if (fromPinned.length > 0) {
+      const newRootTabs = Tab.collectRootTabs(Tab.sort(fromPinned));
+      if (newRootTabs.length > 0) {
+        await tryGroupNewTabsFromPinnedOpener(newRootTabs);
+      }
+    }
+
+    // We can assume that new tabs from a bookmark folder and from other
+    // sources won't be mixed.
+    const openedFromBookmarkFolder = fromOthers.length > 0 && await new Promise((resolve, _reject) => {
+      const maybeFromBookmarks = [];
+      let restCount = fromOthers.length;
+      for (const tab of fromOthers) {
+        tab.$TST.promisedPossibleOpenerBookmarks.then(bookmarks => {
+          restCount--;
+          if (bookmarks.length > 0)
+            maybeFromBookmarks.push(tab);
+          if (areMostTabsFromSameBookmarkFolder(maybeFromBookmarks, tabReferences.length))
+            resolve(true);
+          else if (restCount == 0)
+            resolve(false);
+        });
+      }
+    });
+    const newRootTabs = Tab.collectRootTabs(Tab.sort(fromOthers));
+    if (newRootTabs.length > 1) {
+      if (openedFromBookmarkFolder) {
+        if (configs.autoGroupNewTabsFromBookmarks)
+          await TabsGroup.groupTabs(newRootTabs, { broadcast: true });
+      }
+      else if (configs.autoGroupNewTabsFromOthers) {
+        const granted = await confirmToAutoGroupNewTabsFromOthers(fromOthers);
+        if (granted)
+          await TabsGroup.groupTabs(newRootTabs, { broadcast: true });
+      }
     }
   }
   catch(e) {
@@ -411,32 +445,81 @@ async function tryGroupNewTabs() {
   }
 }
 
-async function confirmToAutoGroupNewTabs(tabs) {
+function areMostTabsFromSameBookmarkFolder(bookmarkedTabs, allNewTabsCount) {
+  const parentIds = bookmarkedTabs.map(tab => tab.$TST.possibleOpenerBookmarks.map(bookmark => bookmark.parentId)).flat();
+  const counts    = [];
+  const countById = {};
+  for (const id of parentIds) {
+    if (!(id in countById))
+      counts.push(countById[id] = { id, count: 0 });
+    countById[id].count++;
+  }
+  if (counts.length == 0)
+    return false;
+
+  const greatestCount = counts.sort((a, b) => b.count - a.count)[0];
+  const minCount = allNewTabsCount * configs.tabsFromSameFolderMinThresholdPercentage / 100;
+  return greatestCount.count > minCount;
+}
+
+async function confirmToAutoGroupNewTabsFromOthers(tabs) {
   if (tabs.length <= 1 ||
       !configs.warnOnAutoGroupNewTabs)
     return true;
 
   const windowId = tabs[0].windowId;
-  const granted = await Permissions.isGranted(Permissions.ALL_URLS);
-  if (!granted ||
-      /^(about|chrome|resource):/.test(tabs[0].url) ||
-      (SidebarConnection.isOpen(windowId) &&
-       SidebarConnection.hasFocus(windowId)))
-    return browser.runtime.sendMessage({
-      type:     Constants.kCOMMAND_CONFIRM_TO_AUTO_GROUP_NEW_TABS,
-      tabIds:   tabs.map(tab => tab.id),
-      windowId: windowId
-    }).catch(ApiTabs.createErrorHandler());
 
-  const result = await RichConfirm.showInTab(tabs[0].id, {
-    message: browser.i18n.getMessage('warnOnAutoGroupNewTabs_message', [tabs.length]),
+  const listing = configs.warnOnAutoGroupNewTabsWithListing ?
+    Background.tabsToHTMLList(tabs, { maxRows: configs.warnOnAutoGroupNewTabsWithListingMaxRows }) :
+    '';
+  const dialogParams = {
+    content: `
+      <div>${sanitizeForHTMLText(browser.i18n.getMessage('warnOnAutoGroupNewTabs_message', [tabs.length]))}</div>${listing}
+    `.trim(),
     buttons: [
       browser.i18n.getMessage('warnOnAutoGroupNewTabs_close'),
       browser.i18n.getMessage('warnOnAutoGroupNewTabs_cancel')
     ],
     checkMessage: browser.i18n.getMessage('warnOnAutoGroupNewTabs_warnAgain'),
     checked: true
-  });
+  };
+  let result;
+  UserOperationBlocker.blockIn(windowId, { throbber: false });
+  try {
+    if (configs.showDialogInSidebar &&
+        SidebarConnection.isOpen(windowId)/* &&
+        SidebarConnection.hasFocus(windowId)*/) {
+      result = await browser.runtime.sendMessage({
+        type:   Constants.kCOMMAND_SHOW_DIALOG,
+        params: dialogParams,
+        windowId
+      }).catch(ApiTabs.createErrorHandler());
+    }
+    else {
+      result = await RichConfirm.showInPopup(windowId, {
+        ...dialogParams,
+        onShown(container) {
+          setTimeout(() => { // this need to be done on the next tick, to use the height of the box for     calculation of dialog size
+            const style = container.querySelector('ul').style;
+            style.height = '0px'; // this makes the box shrinkable
+            style.maxHeight = 'none';
+            style.minHeight = '0px';
+          }, 0);
+        },
+        modal: true,
+        type:  'common-dialog',
+        url:   '/resources/blank.html', // required on Firefox ESR68
+        title: browser.i18n.getMessage('warnOnAutoGroupNewTabs_title')
+      });
+    }
+  }
+  catch(_error) {
+    result = { buttonIndex: -1 };
+  }
+  finally {
+    UserOperationBlocker.unblockIn(windowId, { throbber: false });
+  }
+
   switch (result.buttonIndex) {
     case 0:
       if (!result.checked)
@@ -445,7 +528,7 @@ async function confirmToAutoGroupNewTabs(tabs) {
     case 1:
       if (!result.checked) {
         configs.warnOnAutoGroupNewTabs = false;
-        configs.autoGroupNewTabs = false;
+        configs.autoGroupNewTabsFromOthers = false;
       }
     default:
       return false;
@@ -463,7 +546,7 @@ async function tryGroupNewTabsFromPinnedOpener(rootTabs) {
     if (!pinnedOpeners.includes(opener))
       pinnedOpeners.push(opener);
   }
-  log('pinnedOpeners ', pinnedOpeners.map(dumpTab));
+  log('pinnedOpeners ', () => pinnedOpeners.map(dumpTab));
 
   // Second, collect tabs opened from pinned openers including existing tabs
   // (which were left ungrouped in previous process).
@@ -498,7 +581,7 @@ async function tryGroupNewTabsFromPinnedOpener(rootTabs) {
   pinnedOpeners = pinnedOpeners.filter(opener => {
     return childrenOfPinnedTabs[opener.id].length > 1 || Tab.getGroupTabForOpener(opener);
   });
-  log(' => ', pinnedOpeners.map(dumpTab));
+  log(' => ', () => pinnedOpeners.map(dumpTab));
 
   // Move newly opened tabs to expected position before grouping!
   switch (configs.insertNewTabFromPinnedTabAt) {
@@ -535,7 +618,7 @@ async function tryGroupNewTabsFromPinnedOpener(rootTabs) {
   const newGroupTabs = new Map();
   for (const opener of pinnedOpeners) {
     const children = childrenOfPinnedTabs[opener.id].sort((a, b) => a.index - b.index);
-    log(`trying to group children of ${dumpTab(opener)}: `, children.map(dumpTab));
+    log(`trying to group children of ${dumpTab(opener)}: `, () => children.map(dumpTab));
     let parent = Tab.getGroupTabForOpener(opener);
     if (!parent) {
       const uri = TabsGroup.makeGroupTabURI({
