@@ -8,6 +8,7 @@
 import {
   log as internalLogger,
   wait,
+  mapAndFilterUniq,
   configs
 } from '/common/common.js';
 
@@ -20,6 +21,9 @@ import * as TSTAPI from '/common/tst-api.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
 import * as Permissions from '/common/permissions.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
+import * as Bookmark from '/common/bookmark.js';
+import * as BrowserTheme from '/common/browser-theme.js';
+import * as ContextualIdentities from '/common/contextual-identities.js';
 
 import Tab from '/common/Tab.js';
 
@@ -27,6 +31,7 @@ import * as Background from './background.js';
 import * as TabsGroup from './tabs-group.js';
 import * as Tree from './tree.js';
 import * as Commands from './commands.js';
+import * as Migration from './migration.js';
 
 function log(...args) {
   internalLogger('background/handle-misc', ...args);
@@ -47,7 +52,7 @@ let mInitialized = false;
 Background.onInit.addListener(() => {
   browser.browserAction.onClicked.addListener(onToolbarButtonClick);
   browser.commands.onCommand.addListener(onShortcutCommand);
-  browser.runtime.onMessageExternal.addListener(onMessageExternal);
+  TSTAPI.onMessageExternal.addListener(onMessageExternal);
 });
 
 Background.onBuilt.addListener(() => {
@@ -57,11 +62,12 @@ Background.onBuilt.addListener(() => {
 
 Background.onReady.addListener(() => {
   mInitialized = true;
+  Bookmark.startTracking();
 });
 
 Background.onDestroy.addListener(() => {
   browser.runtime.onMessage.removeListener(onMessage);
-  browser.runtime.onMessageExternal.removeListener(onMessageExternal);
+  TSTAPI.onMessageExternal.removeListener(onMessageExternal);
   browser.browserAction.onClicked.removeListener(onToolbarButtonClick);
 });
 
@@ -70,7 +76,7 @@ function onToolbarButtonClick(tab) {
   if (Permissions.requestPostProcess())
     return;
 
-  if (SidebarConnection.isOpen(tab.windowId))
+  if (SidebarConnection.isSidebarOpen(tab.windowId))
     browser.sidebarAction.close();
   else
     browser.sidebarAction.open();
@@ -81,7 +87,7 @@ async function onShortcutCommand(command) {
     active:        true,
     currentWindow: true
   }).catch(ApiTabs.createErrorHandler()))[0].id);
-  const selectedTabs = activeTab.$TST.multiselected ? Tab.getSelectedTabs(activeTab.windowId) : [];
+  const selectedTabs = activeTab.$TST.multiselected ? Tab.getSelectedTabs(activeTab.windowId) : [activeTab];
   log('onShortcutCommand ', { command, activeTab, selectedTabs });
 
   switch (command) {
@@ -89,34 +95,40 @@ async function onShortcutCommand(command) {
       return;
 
     case 'reloadTree':
-      Commands.reloadTree(activeTab);
+      Commands.reloadTree(selectedTabs);
       return;
     case 'reloadDescendants':
-      Commands.reloadDescendants(activeTab);
+      Commands.reloadDescendants(selectedTabs);
       return;
     case 'closeTree':
-      Commands.closeTree(activeTab);
+      Commands.closeTree(selectedTabs);
       return;
     case 'closeDescendants':
-      Commands.closeDescendants(activeTab);
+      Commands.closeDescendants(selectedTabs);
       return;
     case 'closeOthers':
-      Commands.closeOthers(activeTab);
+      Commands.closeOthers(selectedTabs);
       return;
     case 'collapseTree':
-      Commands.collapseTree(activeTab);
+      Commands.collapseTree(selectedTabs);
+      return;
+    case 'collapseTreeRecursively':
+      Commands.collapseTree(selectedTabs, { recursively: true });
       return;
     case 'collapseAll':
       Commands.collapseAll(activeTab.windowId);
       return;
     case 'expandTree':
-      Commands.expandTree(activeTab);
+      Commands.expandTree(selectedTabs);
+      return;
+    case 'expandTreeRecursively':
+      Commands.expandTree(selectedTabs, { recursively: true });
       return;
     case 'expandAll':
       Commands.expandAll(activeTab.windowId);
       return;
     case 'bookmarkTree':
-      Commands.bookmarkTree(activeTab);
+      Commands.bookmarkTree(selectedTabs);
       return;
 
     case 'newIndependentTab':
@@ -180,13 +192,17 @@ async function onShortcutCommand(command) {
     case 'focusPreviousSilently': {
       const nextActive = activeTab.$TST.nearestVisiblePrecedingTab ||
         Tab.getLastVisibleTab(activeTab.windowId);
-      TabsInternalOperation.activateTab(nextActive, { silently: /Silently/.test(command) });
+      TabsInternalOperation.activateTab(nextActive, {
+        silently: /Silently/.test(command)
+      });
     }; return;
     case 'focusNext':
     case 'focusNextSilently': {
       const nextActive = activeTab.$TST.nearestVisibleFollowingTab ||
         Tab.getFirstVisibleTab(activeTab.windowId);
-      TabsInternalOperation.activateTab(nextActive, { silently: /Silently/.test(command) });
+      TabsInternalOperation.activateTab(nextActive, {
+        silently: /Silently/.test(command)
+      });
     }; return;
     case 'focusParent':
       TabsInternalOperation.activateTab(activeTab.$TST.parent);
@@ -298,6 +314,12 @@ function onMessage(message, sender) {
         return Tab.get(message.tabId).$TST.promisedUniqueId;
       })();
 
+    case Constants.kCOMMAND_GET_THEME_DECLARATIONS:
+      return browser.theme.getCurrent().then(theme => BrowserTheme.generateThemeDeclarations(theme));
+
+    case Constants.kCOMMAND_GET_CONTEXTUAL_IDENTITIES_COLOR_INFO:
+      return Promise.resolve(ContextualIdentities.getColorInfo());
+
     case Constants.kCOMMAND_GET_CONFIG_VALUE:
       if (Array.isArray(message.keys)) {
         const values = {};
@@ -336,19 +358,28 @@ function onMessage(message, sender) {
 
     case Constants.kCOMMAND_NOTIFY_PERMISSIONS_GRANTED:
       return (async () => {
-        if (JSON.stringify(message.permissions) == JSON.stringify(Permissions.ALL_URLS)) {
+        const grantedPermission = JSON.stringify(message.permissions);
+        if (grantedPermission == JSON.stringify(Permissions.ALL_URLS)) {
           const tabs = await browser.tabs.query({}).catch(ApiTabs.createErrorHandler());
           await Tab.waitUntilTracked(tabs.map(tab => tab.id));
           for (const tab of tabs) {
             Background.tryStartHandleAccelKeyOnTab(Tab.get(tab.id));
           }
         }
+        else if (grantedPermission == JSON.stringify(Permissions.BOOKMARKS)) {
+          Migration.migrateBookmarkUrls();
+          Bookmark.startTracking();
+        }
       })();
 
     case Constants.kCOMMAND_SIMULATE_SIDEBAR_MESSAGE: {
       SidebarConnection.onMessage.dispatch(message.message.windowId, message.message);
     }; break;
 
+    case Constants.kCOMMAND_CONFIRM_TO_CLOSE_TABS:
+      log('kCOMMAND_CONFIRM_TO_CLOSE_TABS: ', { message });
+      return Background.confirmToCloseTabs(message.tabs, message);
+
     default:
       const API_PREFIX_MATCHER = /^treestyletab:api:/;
       if (API_PREFIX_MATCHER.test(message.type)) {
@@ -369,7 +400,10 @@ function onMessageExternal(message, sender) {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
         const cache = {};
         return TSTAPI.formatTabResult(
-          Array.from(tabs, tab => new TSTAPI.TreeItem(tab, { interval: message.interval, cache })),
+          Array.from(tabs, tab => new TSTAPI.TreeItem(tab, {
+            interval: message.interval,
+            cache
+          })),
           message,
           sender.id
         );
@@ -380,7 +414,9 @@ function onMessageExternal(message, sender) {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
         await TSTAPI.doProgressively(
           tabs,
-          tab => Tree.collapseExpandSubtree(tab, { collapsed: true, broadcast: true }),
+          tab => Commands.collapseTree(tab, {
+            recursively: !!message.recursively
+          }),
           message.interval
         );
         return true;
@@ -391,7 +427,23 @@ function onMessageExternal(message, sender) {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
         await TSTAPI.doProgressively(
           tabs,
-          tab => Tree.collapseExpandSubtree(tab, { collapsed: false, broadcast: true }),
+          tab => Commands.expandTree(tab, {
+            recursively: !!message.recursively
+          }),
+          message.interval
+        );
+        return true;
+      })();
+
+    case TSTAPI.kTOGGLE_TREE_COLLAPSED:
+      return (async () => {
+        const tabs = await TSTAPI.getTargetTabs(message, sender);
+        await TSTAPI.doProgressively(
+          tabs,
+          tab => Tree.collapseExpandSubtree(tab, {
+            collapsed: !tab.$TST.subtreeCollapsed,
+            broadcast: true
+          }),
           message.interval
         );
         return true;
@@ -416,6 +468,14 @@ function onMessageExternal(message, sender) {
           insertBefore: Tab.get(message.insertBefore),
           insertAfter:  Tab.get(message.insertAfter)
         });
+        if (child.$TST.collapsed &&
+            !parent.$TST.collapsed &&
+            !parent.$TST.subtreeCollapsed) {
+          await Tree.collapseExpandTabAndSubtree(child, {
+            collapsed: false,
+            bradcast:  true
+          });
+        }
         return true;
       })();
 
@@ -428,6 +488,12 @@ function onMessageExternal(message, sender) {
         await Tree.detachTab(tab, {
           broadcast: true
         });
+        if (tab.$TST.collapsed) {
+          await Tree.collapseExpandTabAndSubtree(tab, {
+            collapsed: false,
+            bradcast:  true
+          });
+        }
         return true;
       })();
 
@@ -519,7 +585,9 @@ function onMessageExternal(message, sender) {
         const tabsArray = await TSTAPI.doProgressively(
           tabs,
           tab => {
-            TabsInternalOperation.activateTab(tab, { silently: message.silently });
+            TabsInternalOperation.activateTab(tab, {
+              silently: message.silently
+            });
             return tab;
           },
           message.interval
@@ -564,7 +632,13 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kGROUP_TABS:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        return TabsGroup.groupTabs(Array.from(tabs), { broadcast: true });
+        const tab = await TabsGroup.groupTabs(Array.from(tabs), {
+          broadcast: true
+        });
+        if (!tab)
+          return null;
+        const treeItem = new TSTAPI.TreeItem(tab);
+        return treeItem.exportFor(sender.id);
       })();
 
     case TSTAPI.kOPEN_IN_NEW_WINDOW:
@@ -579,10 +653,16 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kREOPEN_IN_CONTAINER:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const reopenedTabs = await Commands.reopenInContainer(Array.from(tabs), message.containerId || 'firefox-default');
+        const reopenedTabs = await Commands.reopenInContainer(
+          Array.from(tabs),
+          message.containerId || 'firefox-default'
+        );
         const cache = {};
         return TSTAPI.formatTabResult(
-          reopenedTabs.map(tab => new TSTAPI.TreeItem(tab, { interval: message.interval, cache })),
+          reopenedTabs.map(tab => new TSTAPI.TreeItem(tab, {
+            interval: message.interval,
+            cache
+          })),
           message,
           sender.id
         );
@@ -597,9 +677,11 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kSET_TREE_STRUCTURE:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        await Tree.applyTreeStructureToTabs(Array.from(tabs), message.structure, {
-          broadcast: true
-        });
+        await Tree.applyTreeStructureToTabs(
+          Array.from(tabs),
+          message.structure,
+          { broadcast: true }
+        );
         return Promise.resolve(true);
       })();
 
@@ -650,8 +732,10 @@ function onMessageExternal(message, sender) {
     case TSTAPI.kGRANT_TO_REMOVE_TABS:
       return (async () => {
         const tabs = await TSTAPI.getTargetTabs(message, sender);
-        const grantedRemovingTabIds = configs.grantedRemovingTabIds.concat(Array.from(tabs).filter(TabsStore.ensureLivingTab).map(tab => tab.id));
-        configs.grantedRemovingTabIds = Array.from(new Set(grantedRemovingTabIds));
+        configs.grantedRemovingTabIds = mapAndFilterUniq(configs.grantedRemovingTabIds.concat(tabs), tab => {
+          tab = TabsStore.ensureLivingTab(tab);
+          return tab && tab.id || undefined;
+        });
         return true;
       })();
 
@@ -663,59 +747,11 @@ function onMessageExternal(message, sender) {
           button:   message.button || 0
         });
       })();
+
+    case TSTAPI.kOPEN_ALL_BOOKMARKS_WITH_STRUCTURE:
+      return Commands.openAllBookmarksWithStructure(
+        message.id || message.bookmarkId,
+        { discarded: message.discarded }
+      );
   }
 }
-
-
-SidebarConnection.onMessage.addListener(async (windowId, message) => {
-  switch (message.type) {
-    case Constants.kCOMMAND_SET_SUBTREE_MUTED: {
-      await Tab.waitUntilTracked(message.tabId);
-      log('set muted state: ', message);
-      const root = Tab.get(message.tabId);
-      if (!root)
-        return;
-      const multiselected = root.$TST.multiselected;
-      const tabs = multiselected ?
-        Tab.getSelectedTabs(root.windowId, { iterator: true }) :
-        [root].concat(root.$TST.descendants) ;
-      for (const tab of tabs) {
-        const playing = tab.$TST.soundPlaying;
-        const muted   = tab.$TST.muted;
-        log(`tab ${tab.id}: playing=${playing}, muted=${muted}`);
-        if (!multiselected && playing != message.muted)
-          continue;
-
-        log(` => set muted=${message.muted}`);
-        browser.tabs.update(tab.id, {
-          muted: message.muted
-        }).catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-
-        const add = [];
-        const remove = [];
-        if (message.muted) {
-          add.push(Constants.kTAB_STATE_MUTED);
-          tab.$TST.addState(Constants.kTAB_STATE_MUTED);
-        }
-        else {
-          remove.push(Constants.kTAB_STATE_MUTED);
-          tab.$TST.removeState(Constants.kTAB_STATE_MUTED);
-        }
-
-        if (tab.audible && !message.muted) {
-          add.push(Constants.kTAB_STATE_SOUND_PLAYING);
-          tab.$TST.addState(Constants.kTAB_STATE_SOUND_PLAYING);
-        }
-        else {
-          remove.push(Constants.kTAB_STATE_SOUND_PLAYING);
-          tab.$TST.removeState(Constants.kTAB_STATE_SOUND_PLAYING);
-        }
-
-        // tabs.onUpdated is too slow, so users will be confused
-        // from still-not-updated tabs (in other words, they tabs
-        // are unresponsive for quick-clicks).
-        Tab.broadcastState(tab, { add, remove });
-      }
-    }; break;
-  }
-});
