@@ -17,6 +17,7 @@ import * as ApiTabs from '/common/api-tabs.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
 import * as SidebarConnection from '/common/sidebar-connection.js';
+import * as TSTAPI from '/common/tst-api.js';
 
 import Tab from '/common/Tab.js';
 
@@ -46,14 +47,14 @@ Tree.onAttached.addListener(async (tab, info = {}) => {
     });
   }
 
-  // Because the tab is possibly closing for "reopen" operation,
-  // we need to apply "forceExpand" immediately. Otherwise, when
-  // the tab is closed with "subtree collapsed" state, descendant
-  // tabs are also closed even if "forceExpand" is "true".
   if (info.newlyAttached &&
       mInitialized) {
+    // Because the tab is possibly closing for "reopen" operation,
+    // we need to apply "forceExpand" immediately. Otherwise, when
+    // the tab is closed with "subtree collapsed" state, descendant
+    // tabs are also closed even if "forceExpand" is "true".
     log('newly attached tab');
-    if (info.parent.$TST.subtreeCollapsed &&
+    if (parent.$TST.subtreeCollapsed &&
         !info.forceExpand) {
       log('  the tree is collapsed, but keep collapsed by forceExpand option');
       Tree.collapseExpandTabAndSubtree(tab, {
@@ -64,51 +65,81 @@ Tree.onAttached.addListener(async (tab, info = {}) => {
     }
 
     const isNewTreeCreatedManually = !info.justNow && parent.$TST.childIds.length == 1;
-    if (info.forceExpand) {
+    let parentTreeCollasped = parent.$TST.subtreeCollapsed;
+    let parentCollasped     = parent.$TST.collapsed;
+
+    const cache = {};
+    const allowed = (info.forceExpand || !!info.dontExpand) && await TSTAPI.tryOperationAllowed(
+      TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_ATTACHED_CHILD,
+      { tab: new TSTAPI.TreeItem(parent, { cache }) },
+      { tabProperties: ['tab'] }
+    );
+
+    if (info.forceExpand && allowed) {
       log('  expand by forceExpand option');
-      Tree.collapseExpandSubtree(parent, Object.assign({}, info, {
+      Tree.collapseExpandSubtree(parent, {
+        ...info,
         collapsed: false,
         broadcast: true
-      }));
+      });
+      parentTreeCollasped = false;
     }
     if (!info.dontExpand) {
-      if (configs.autoCollapseExpandSubtreeOnAttach &&
-          (isNewTreeCreatedManually ||
-           parent.$TST.isAutoExpandable)) {
-        log('  collapse others by collapseExpandTreesIntelligentlyFor');
-        Tree.collapseExpandTreesIntelligentlyFor(parent, {
-          broadcast: true
-        });
-      }
-
-      const newAncestors = [parent].concat(parent.$TST.ancestors);
-      if (configs.autoCollapseExpandSubtreeOnSelect ||
-          isNewTreeCreatedManually ||
-          parent.$TST.isAutoExpandable ||
-          info.forceExpand) {
-        log('  expand ancestor tabs');
-        newAncestors.filter(ancestor => ancestor.$TST.subtreeCollapsed).forEach(ancestor => {
-          Tree.collapseExpandSubtree(ancestor, Object.assign({}, info, {
-            collapsed:    false,
-            broadcast:    true
+      if (allowed) {
+        if (configs.autoCollapseExpandSubtreeOnAttach &&
+            (isNewTreeCreatedManually ||
+             parent.$TST.isAutoExpandable)) {
+          log('  collapse others by collapseExpandTreesIntelligentlyFor');
+          Tree.collapseExpandTreesIntelligentlyFor(parent, {
+            broadcast: true
+          });
+        }
+        if (configs.autoCollapseExpandSubtreeOnSelect ||
+            isNewTreeCreatedManually ||
+            parent.$TST.isAutoExpandable ||
+            info.forceExpand) {
+          log('  expand ancestor tabs');
+          parentTreeCollasped = false;
+          parentCollasped     = false;
+          await Promise.all([parent].concat(parent.$TST.ancestors).map(async ancestor => {
+            if (!ancestor.$TST.subtreeCollapsed)
+              return;
+            const allowed = await TSTAPI.tryOperationAllowed(
+              TSTAPI.kNOTIFY_TRY_EXPAND_TREE_FROM_ATTACHED_CHILD,
+              { tab: new TSTAPI.TreeItem(ancestor, { cache }) },
+              { tabProperties: ['tab'] }
+            );
+            if (!allowed) {
+              parentTreeCollasped = true;
+              parentCollasped     = true;
+              return;
+            }
+            Tree.collapseExpandSubtree(ancestor, {
+              ...info,
+              collapsed:    false,
+              broadcast:    true
+            });
+            parentTreeCollasped = false;
           }));
-        });
-      }
-      if (parent.$TST.collapsed) {
-        log('  collapse tab because the parent is collapsed');
-        Tree.collapseExpandTabAndSubtree(tab, Object.assign({}, info, {
-          collapsed:    true,
-          broadcast:    true
-        }));
+        }
       }
     }
     else if (parent.$TST.isAutoExpandable ||
              parent.$TST.collapsed) {
       log('  collapse auto expanded tree');
-      Tree.collapseExpandTabAndSubtree(tab, Object.assign({}, info, {
+      Tree.collapseExpandTabAndSubtree(tab, {
+        ...info,
         collapsed:    true,
         broadcast:    true
-      }));
+      });
+    }
+    if (parentTreeCollasped || parentCollasped) {
+      log('  collapse tab because the parent is collapsed');
+      Tree.collapseExpandTabAndSubtree(tab, {
+        ...info,
+        collapsed: true,
+        broadcast: true
+      });
     }
   }
 
@@ -127,13 +158,15 @@ Tree.onAttached.addListener(async (tab, info = {}) => {
         prev: dumpTab(prevTab)
       });
       if (nextTab)
-        await Tree.moveTabSubtreeBefore(tab, nextTab, Object.assign({}, info, {
+        await Tree.moveTabSubtreeBefore(tab, nextTab, {
+          ...info,
           broadcast:    true
-        }));
+        });
       else
-        await Tree.moveTabSubtreeAfter(tab, prevTab, Object.assign({}, info, {
+        await Tree.moveTabSubtreeAfter(tab, prevTab, {
+          ...info,
           broadcast:    true
-        }));
+        });
     })()
   ]);
 
@@ -236,9 +269,12 @@ function reserveAttachShownTab(tab) {
     reserveAttachShownTab.tabs.clear();
     for (const tab of tabs) {
       if (!TabsStore.ensureLivingTab(tab) ||
-          tab.$TST.hasParent)
+          tab.$TST.hasParent) {
+        tab.$TST.removeState(Constants.kTAB_STATE_SHOWING);
         continue;
+      }
       const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(tab, {
+        context:      Constants.kINSERTION_CONTEXT_SHOWN,
         insertAfter:  tab.$TST.nearestVisiblePrecedingTab,
         // Instead of nearestFollowingForeignerTab, to avoid placing the tab
         // after hidden tabs (too far from the target)
