@@ -3,7 +3,6 @@
 # License, v. 2.0. If a copy of the MPL was not distributed with this
 # file, You can obtain one at http://mozilla.org/MPL/2.0/.
 */
-import '../extlib/TabIdFixer.js';
 
 import {
   log,
@@ -24,9 +23,6 @@ import './handle-moved-tabs.js';
 import './handle-tree-changes.js';
 import './handle-group-tabs.js';
 import './handle-misc.js';
-import './context-menu.js';
-
-import './tab-context-menu.js';
 
 log.context = 'BG';
 
