@@ -12,9 +12,10 @@ const matched = uri && uri.match(Constants.kSHORTHAND_CUSTOM_URI);
 if (matched) {
   const name = matched[1];
   const params = matched[2] || '';
-  switch (name) {
+  const delimiter = params ? '?' : '';
+  switch (name.toLowerCase()) {
     case 'group':
-      location.href = `${Constants.kSHORTHAND_URIS.group}?${params}`;
+      location.href = `${Constants.kSHORTHAND_URIS.group}${delimiter}${params}`;
       break;
 
     case 'startup':
@@ -22,11 +23,16 @@ if (matched) {
       break;
 
     case 'test-runner':
-      location.href = Constants.kSHORTHAND_URIS.testRunner;
+    case 'testrunner':
+      location.href = `${Constants.kSHORTHAND_URIS.testRunner}${delimiter}${params}`;
       break;
 
     case 'options':
       location.href = `${Constants.kSHORTHAND_URIS.options}${params.split('#')[1] || ''}`;
       break;
+
+    case 'tabbar':
+      location.href = `${Constants.kSHORTHAND_URIS.tabbar}${delimiter}${params}`;
+      break;
   }
 }
