@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -32,6 +32,8 @@ import {
   log as internalLogger,
   wait,
   dumpTab,
+  mapAndFilter,
+  countMatched,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -60,8 +62,10 @@ const mTabBar = document.querySelector('#tabbar');
 const mContextualIdentitySelector = document.getElementById(Constants.kCONTEXTUAL_IDENTITY_SELECTOR);
 const mNewTabActionSelector       = document.getElementById(Constants.kNEWTAB_ACTION_SELECTOR);
 
+let mHasMouseOverListeners = false;
+
 Sidebar.onInit.addListener(() => {
-  mTargetWindow = TabsStore.getWindow();
+  mTargetWindow = TabsStore.getCurrentWindowId();
 });
 
 Sidebar.onBuilt.addListener(async () => {
@@ -69,6 +73,7 @@ Sidebar.onBuilt.addListener(async () => {
   document.addEventListener('mouseup', onMouseUp);
   document.addEventListener('click', onClick);
   mTabBar.addEventListener('dblclick', onDblClick);
+  mTabBar.addEventListener('mouseover', onMouseOver);
 
   MetricsData.add('mouse-event-listener: Sidebar.onBuilt: apply configs');
 
@@ -95,7 +100,8 @@ Sidebar.onReady.addListener(() => {
 });
 
 function updateSpecialEventListenersForAPIListeners() {
-  if ((TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEMOVE).length > 0) != onMouseMove.listening) {
+  const shouldListenMouseMove = TSTAPI.hasListenerForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEMOVE);
+  if (shouldListenMouseMove != onMouseMove.listening) {
     if (!onMouseMove.listening) {
       window.addEventListener('mousemove', onMouseMove, { capture: true, passive: true });
       onMouseMove.listening = true;
@@ -106,20 +112,7 @@ function updateSpecialEventListenersForAPIListeners() {
     }
   }
 
-  const shouldListenMouseOut = TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEOUT) > 0;
-  const shouldListenMouseOver = shouldListenMouseOut || TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEOVER) > 0;
-
-  if (shouldListenMouseOver != onMouseOver.listening) {
-    if (!onMouseOver.listening) {
-      window.addEventListener('mouseover', onMouseOver, { capture: true, passive: true });
-      onMouseOver.listening = true;
-    }
-    else {
-      window.removeEventListener('mouseover', onMouseOver, { capture: true, passive: true });
-      onMouseOver.listening = false;
-    }
-  }
-
+  const shouldListenMouseOut = TSTAPI.hasListenerForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEOUT);
   if (shouldListenMouseOut != onMouseOut.listening) {
     if (!onMouseOut.listening) {
       window.addEventListener('mouseout', onMouseOut, { capture: true, passive: true });
@@ -130,6 +123,8 @@ function updateSpecialEventListenersForAPIListeners() {
       onMouseOut.listening = false;
     }
   }
+
+  mHasMouseOverListeners = shouldListenMouseOut || TSTAPI.hasListenerForMessageType(TSTAPI.kNOTIFY_TAB_MOUSEOVER);
 }
 
 
@@ -153,9 +148,21 @@ function onMouseMove(event) {
 }
 onMouseMove = EventUtils.wrapWithErrorHandler(onMouseMove);
 
+let mLastWarmUpTab = -1;
+
 function onMouseOver(event) {
   const tab = EventUtils.getTabFromEvent(event);
 
+  if (tab &&
+      mLastWarmUpTab != tab.id &&
+      typeof browser.tabs.warmup == 'function') {
+    browser.tabs.warmup(tab.id);
+    mLastWarmUpTab = tab.id;
+  }
+
+  if (!mHasMouseOverListeners)
+    return;
+
   // We enter the tab or one of its children, but not from any of the tabs
   // (other) children, so we are now starting to hover this tab (relatedTarget
   // contains the target of the mouseout event or null if there is none). This
@@ -226,25 +233,11 @@ function onMouseDown(event) {
 
   const target = event.target;
   const tab = EventUtils.getTabFromEvent(event) || EventUtils.getTabFromTabbarEvent(event);
-  log('onMouseDown: found target tab: ', tab);
+  log('onMouseDown: found target tab: ', tab, event);
 
-  const mousedownDetail = {
-    targetType:    getMouseEventTargetType(event),
-    tab:           tab && tab.id,
-    tabId:         tab && tab.id,
-    window:        mTargetWindow,
-    windowId:      mTargetWindow,
-    twisty:        EventUtils.isEventFiredOnTwisty(event),
-    soundButton:   EventUtils.isEventFiredOnSoundButton(event),
-    closebox:      EventUtils.isEventFiredOnClosebox(event),
-    button:        event.button,
-    ctrlKey:       event.ctrlKey,
-    shiftKey:      event.shiftKey,
-    altKey:        event.altKey,
-    metaKey:       event.metaKey,
-    isMiddleClick: EventUtils.isMiddleClick(event),
-    isAccelClick:  EventUtils.isAccelAction(event)
-  };
+  const extraContentsInfo = getOriginalExtraContentsTarget(event);
+  const mousedownDetail   = getMouseEventDetail(event, tab);
+  mousedownDetail.$extraContentsInfo = extraContentsInfo;
   log('onMouseDown ', mousedownDetail);
 
   if (mousedownDetail.targetType == 'selector')
@@ -267,15 +260,50 @@ function onMouseDown(event) {
       .catch(ApiTabs.createErrorHandler()),
     (async () => {
       if (mousedownDetail.targetType != 'tab')
-        return;
-      log('Sending message to listeners');
-      return TSTAPI.sendMessage(Object.assign({}, mousedownDetail, {
-        type: TSTAPI.kNOTIFY_TAB_MOUSEDOWN,
-        tab:  mousedown.treeItem
-      }), { tabProperties: ['tab'] });
+        return undefined;
+
+      log('Sending message to mousedown listeners ', { extraContentsInfo });
+      const allowed = await tryMouseOperationAllowedWithExtraContents(
+        TSTAPI.kNOTIFY_TAB_MOUSEDOWN,
+        mousedown,
+        extraContentsInfo
+      );
+      if (!allowed) {
+        log(' => canceled');
+        return true;
+      }
+
+      log(' => allowed');
+      return false;
     })()
   ]).then(results => results[1]);
 
+  // Firefox switches tab focus on mousedown, and keeps
+  // tab multiselection for draging of them together.
+  // We simulate the behavior here.
+  mousedown.promisedMousedownNotified.then(canceled => {
+    if (!EventUtils.getLastMousedown(event.button) ||
+        mousedown.expired ||
+        canceled)
+      return;
+
+    const onRegularArea = (
+      !mousedown.detail.twisty &&
+      !mousedown.detail.soundButton &&
+      !mousedown.detail.closebox
+    );
+    const wasMultiselectionAction = (
+      mousedown.detail.ctrlKey ||
+      mousedown.detail.shiftKey
+    );
+    if (mousedown.detail.button == 0 &&
+        onRegularArea &&
+        !wasMultiselectionAction)
+      TabsInternalOperation.activateTab(tab, {
+        keepMultiselection: true
+      });
+  });
+
   EventUtils.setLastMousedown(event.button, mousedown);
   mousedown.timeout = setTimeout(async () => {
     if (!EventUtils.getLastMousedown(event.button))
@@ -294,8 +322,7 @@ function onMouseDown(event) {
       return;
     }
 
-    if (TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_DRAGREADY).length == 0 ||
-        EventUtils.getElementTarget(event).closest('#tab-drag-handle'))
+    if (TSTAPI.getListenersForMessageType(TSTAPI.kNOTIFY_TAB_DRAGREADY).length == 0)
       return;
 
     if (event.button == 0 &&
@@ -307,6 +334,77 @@ function onMouseDown(event) {
 }
 onMouseDown = EventUtils.wrapWithErrorHandler(onMouseDown);
 
+function getMouseEventDetail(event, tab) {
+  return {
+    targetType:    getMouseEventTargetType(event),
+    tab:           tab && tab.id,
+    tabId:         tab && tab.id,
+    window:        mTargetWindow,
+    windowId:      mTargetWindow,
+    twisty:        EventUtils.isEventFiredOnTwisty(event),
+    soundButton:   EventUtils.isEventFiredOnSoundButton(event),
+    closebox:      EventUtils.isEventFiredOnClosebox(event),
+    button:        event.button,
+    ctrlKey:       event.ctrlKey,
+    shiftKey:      event.shiftKey,
+    altKey:        event.altKey,
+    metaKey:       event.metaKey,
+    isMiddleClick: EventUtils.isMiddleClick(event),
+    isAccelClick:  EventUtils.isAccelAction(event)
+  };
+}
+
+function getOriginalExtraContentsTarget(event) {
+  let target = event.originalTarget;
+  if (target && target.nodeType != Node.ELEMENT_NODE)
+    target = target.parentNode;
+
+  const extraContents = target.closest(`.extra-item`);
+  if (extraContents)
+    return {
+      owners: new Set([extraContents.dataset.owner]),
+      target: target.outerHTML
+    };
+
+  return {
+    owners: new Set(),
+    target: null
+  };
+}
+
+async function tryMouseOperationAllowedWithExtraContents(type, mousedown, extraContentsInfo) {
+  if (extraContentsInfo &&
+      extraContentsInfo.owners &&
+      extraContentsInfo.owners.size > 0) {
+    const allowed = await TSTAPI.tryOperationAllowed(
+      type,
+      {
+        ...mousedown.detail,
+        tab:                mousedown.treeItem,
+        originalTarget:     extraContentsInfo.target,
+        $extraContentsInfo: null
+      },
+      { tabProperties: ['tab'],
+        targets:       extraContentsInfo.owners }
+    );
+    if (!allowed)
+      return false;
+  }
+  const allowed = await TSTAPI.tryOperationAllowed(
+    type,
+    {
+      ...mousedown.detail,
+      tab:                mousedown.treeItem,
+      $extraContentsInfo: null
+    },
+    { tabProperties: ['tab'],
+      except:        extraContentsInfo && extraContentsInfo.owners }
+  );
+  if (!allowed)
+    return false;
+  return true;
+}
+
 function getMouseEventTargetType(event) {
   if (EventUtils.getTabFromEvent(event))
     return 'tab';
@@ -331,6 +429,10 @@ function getMouseEventTargetType(event) {
   return 'blank';
 }
 
+let mLastMouseUpX = -1;
+let mLastMouseUpY = -1;
+let mLastMouseUpOnTab = -1;
+
 async function onMouseUp(event) {
   const unsafeTab = EventUtils.getTabFromEvent(event, { force: true }) || EventUtils.getTabFromTabbarEvent(event, { force: true });
   const tab       = EventUtils.getTabFromEvent(event) || EventUtils.getTabFromTabbarEvent(event);
@@ -344,26 +446,34 @@ async function onMouseUp(event) {
 
   const lastMousedown = EventUtils.getLastMousedown(event.button);
   EventUtils.cancelHandleMousedown(event.button);
+  const extraContentsInfo = lastMousedown && lastMousedown.detail && lastMousedown.detail.$extraContentsInfo;
   if (!lastMousedown)
     return;
 
+  if (tab) {
+    const mouseupInfo = {
+      ...lastMousedown,
+      detail:   getMouseEventDetail(event, tab),
+      treeItem: new TSTAPI.TreeItem(tab)
+    };
+    const mouseupAllowed = await tryMouseOperationAllowedWithExtraContents(
+      TSTAPI.kNOTIFY_TAB_MOUSEUP,
+      mouseupInfo,
+      extraContentsInfo
+    );
+    const clickAllowed = await tryMouseOperationAllowedWithExtraContents(
+      TSTAPI.kNOTIFY_TAB_CLICKED,
+      mouseupInfo,
+      extraContentsInfo
+    );
+    if (!mouseupAllowed ||
+        !clickAllowed)
+      return true;
+  }
+
   let promisedCanceled = null;
   if (lastMousedown.treeItem && lastMousedown.detail.targetType == 'tab')
-    promisedCanceled = Promise.all([
-      TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-        type: TSTAPI.kNOTIFY_TAB_MOUSEUP,
-        tab:  lastMousedown.treeItem
-      }), { tabProperties: ['tab'] }),
-
-      TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-        type: TSTAPI.kNOTIFY_TAB_CLICKED,
-        tab:  lastMousedown.treeItem
-      }), { tabProperties: ['tab'] }),
-
-      lastMousedown.promisedMousedownNotified
-    ])
-      .then(results => results.flat())
-      .then(results => results.some(result => result && result.result));
+    promisedCanceled = lastMousedown.promisedMousedownNotified;
 
   if (lastMousedown.expired ||
       lastMousedown.detail.targetType != getMouseEventTargetType(event) || // when the cursor was moved before mouseup
@@ -388,34 +498,74 @@ async function handleDefaultMouseUp({ lastMousedown, tab, event }) {
       await handleDefaultMouseUpOnTab(lastMousedown, tab))
     return;
 
+  if (tab) {
+    mLastMouseUpX = event.clientX;
+    mLastMouseUpY = event.clientY;
+    mLastMouseUpOnTab = Date.now();
+  }
+
   // following codes are for handlig of click event on the tab bar itself.
-  const actionForNewTabCommand = lastMousedown.detail.isAccelClick ?
+  const actionForNewTabCommand = lastMousedown.detail.isMiddleClick ?
     configs.autoAttachOnNewTabButtonMiddleClick :
-    configs.autoAttachOnNewTabCommand;
+    lastMousedown.detail.isAccelClick ?
+      configs.autoAttachOnNewTabButtonAccelClick :
+      configs.autoAttachOnNewTabCommand;
   if (EventUtils.isEventFiredOnNewTabButton(event)) {
     if (lastMousedown.detail.button != 2) {
       log('onMouseUp: click on the new tab button');
-      handleNewTabAction(event, {
-        action: actionForNewTabCommand
-      });
+      // Simulation of Firefox's built-in behavior.
+      // See also: https://github.com/piroor/treestyletab/issues/2593
+      if (event.shiftKey && !lastMousedown.detail.isAccelClick) {
+        browser.windows.create({});
+      }
+      else {
+        const activeTab = Tab.getActiveTab(mTargetWindow);
+        const cookieStoreId = (actionForNewTabCommand == Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING_WITH_INHERITED_CONTAINER) ? activeTab.cookieStoreId : null
+        handleNewTabAction(event, {
+          action: actionForNewTabCommand,
+          cookieStoreId
+        });
+      }
     }
     return;
   }
 
+  // Multiple middle clicks to close tabs can be detected as a middle click on the tab bar.
+  // We should ignore if the cursor is not moved and the closing tab is still in animation.
+  // See also: https://github.com/piroor/treestyletab/issues/1968
+  if (configs.animation &&
+      Date.now() - mLastMouseUpOnTab <= configs.collapseDuration &&
+      Math.abs(mLastMouseUpX - event.clientX) < configs.acceptableFlickerToIgnoreClickOnTabAndTabbar / 2 &&
+      Math.abs(mLastMouseUpY - event.clientY) < configs.acceptableFlickerToIgnoreClickOnTabAndTabbar / 2)
+    return;
+
   log('onMouseUp: notify as a blank area click to other addons');
-  let results = await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-    type:     TSTAPI.kNOTIFY_TABBAR_MOUSEUP,
-    window:   mTargetWindow,
-    windowId: mTargetWindow,
-    tab:      lastMousedown.treeItem
-  }), { tabProperties: ['tab'] });
-  results = results.concat(await TSTAPI.sendMessage(Object.assign({}, lastMousedown.detail, {
-    type:     TSTAPI.kNOTIFY_TABBAR_CLICKED,
-    window:   mTargetWindow,
-    windowId: mTargetWindow,
-    tab:      lastMousedown.treeItem
-  }), { tabProperties: ['tab'] }));
-  if (results.some(result => result.result))// canceled
+  const mouseUpAllowed = await TSTAPI.tryOperationAllowed(
+    TSTAPI.kNOTIFY_TABBAR_MOUSEUP,
+    {
+      ...lastMousedown.detail,
+      window:             mTargetWindow,
+      windowId:           mTargetWindow,
+      tab:                lastMousedown.treeItem,
+      $extraContentsInfo: null
+    },
+    { tabProperties: ['tab'] }
+  );
+  if (!mouseUpAllowed)
+    return;
+
+  const clickAllowed = await TSTAPI.tryOperationAllowed(
+    TSTAPI.kNOTIFY_TABBAR_CLICKED,
+    {
+      ...lastMousedown.detail,
+      window:             mTargetWindow,
+      windowId:           mTargetWindow,
+      tab:                lastMousedown.treeItem,
+      $extraContentsInfo: null
+    },
+    { tabProperties: ['tab'] }
+  );
+  if (!clickAllowed)
     return;
 
   if (lastMousedown.detail.isMiddleClick) { // Ctrl-click does nothing on Firefox's tab bar!
@@ -435,12 +585,11 @@ async function handleDefaultMouseUpOnTab(lastMousedown, tab) {
     !lastMousedown.detail.soundButton &&
     !lastMousedown.detail.closebox
   );
-  const wasMultiselectionAction = (
-    onRegularArea &&
-    updateMultiselectionByTabClick(tab, lastMousedown.detail)
-  );
+  const wasMultiselectionAction = updateMultiselectionByTabClick(tab, lastMousedown.detail);
   log(' => ', { onRegularArea, wasMultiselectionAction });
 
+  // Firefox clears tab multiselection after the mouseup, so
+  // we simulate the behavior.
   if (lastMousedown.detail.button == 0 &&
       onRegularArea &&
       !wasMultiselectionAction)
@@ -460,10 +609,16 @@ async function handleDefaultMouseUpOnTab(lastMousedown, tab) {
         if (confirmed)
           BackgroundConnection.sendMessage({
             type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
-            tabIds: [tab.id]
+            tabIds: tabs.map(tab => tab.id)
           });
       });
   }
+  else if (wasMultiselectionAction) {
+    // On Firefox's native tabs, Ctrl-Click or Shift-Click always
+    // ignore actions on closeboxes and sound playing icons.
+    // Thus we should simulate the behavior.
+    return true;
+  }
   else if (lastMousedown.detail.twisty) {
     log('clicked on twisty');
     if (tab.$TST.hasChild)
@@ -478,9 +633,8 @@ async function handleDefaultMouseUpOnTab(lastMousedown, tab) {
   else if (lastMousedown.detail.soundButton) {
     log('clicked on sound button');
     BackgroundConnection.sendMessage({
-      type:  Constants.kCOMMAND_SET_SUBTREE_MUTED,
-      tabId: tab.id,
-      muted: tab.$TST.maybeSoundPlaying
+      type:  Constants.kCOMMAND_TOGGLE_MUTED,
+      tabId: tab.id
     });
   }
   else if (lastMousedown.detail.closebox) {
@@ -496,13 +650,15 @@ async function handleDefaultMouseUpOnTab(lastMousedown, tab) {
       TreeBehavior.getClosingTabsFromParent(tab, {
         byInternalOperation: true
       }) ;
-    Sidebar.confirmToCloseTabs(tabsToBeClosed.map(tab => tab.$TST.sanitized))
+    Sidebar.confirmToCloseTabs(tabsToBeClosed.map(tab => tab.$TST.sanitized), {
+      configKey: 'warnOnCloseTabsByClosebox'
+    })
       .then(confirmed => {
         if (!confirmed)
           return;
         BackgroundConnection.sendMessage({
           type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
-          tabIds: multiselected ? tabsToBeClosed.map(tab => tab.id) : [tab.id]
+          tabIds: tabsToBeClosed.map(tab => tab.id)
         });
       });
   }
@@ -557,9 +713,8 @@ function updateMultiselectionByTabClick(tab, event) {
       }
 
       // for better performance, we should not call browser.tabs.update() for each tab.
-      const indices = Array.from(highlightedTabIds)
-        .filter(id => id != activeTab.id)
-        .map(id => Tab.get(id).index);
+      const indices = mapAndFilter(highlightedTabIds,
+                                   id => id != activeTab.id && Tab.get(id).index || undefined);
       if (highlightedTabIds.has(activeTab.id))
         indices.unshift(activeTab.index);
       browser.tabs.highlight({
@@ -592,7 +747,7 @@ function updateMultiselectionByTabClick(tab, event) {
       if (tab == activeTab &&
           tab.$TST.subtreeCollapsed &&
           activeTabDescendants.length > 0) {
-        const highlightedCount  = activeTabDescendants.filter(tab => tab.highlighted).length;
+        const highlightedCount  = countMatched(activeTabDescendants, tab => tab.highlighted);
         const partiallySelected = highlightedCount != 0 && highlightedCount != activeTabDescendants.length;
         toBeHighlighted = partiallySelected || !activeTabDescendants[0].highlighted;
         log(' => ', toBeHighlighted, { partiallySelected });
@@ -629,9 +784,8 @@ function updateMultiselectionByTabClick(tab, event) {
       }
 
       // for better performance, we should not call browser.tabs.update() for each tab.
-      const indices = Array.from(highlightedTabIds)
-        .filter(id => id != activeTab.id)
-        .map(id => Tab.get(id).index);
+      const indices = mapAndFilter(highlightedTabIds,
+                                   id => id != activeTab.id && Tab.get(id).index || undefined);
       if (highlightedTabIds.has(activeTab.id))
         indices.unshift(activeTab.index);
       browser.tabs.highlight({
@@ -669,14 +823,14 @@ function handleNewTabAction(event, options = {}) {
 
   BackgroundConnection.sendMessage({
     type:          Constants.kCOMMAND_NEW_TAB_AS,
-    baseTabId:     TabsStore.activeTabInWindow.get(mTargetWindow).id,
+    baseTabId:     Tab.getActiveTab(mTargetWindow).id,
     as:            options.action,
     cookieStoreId: options.cookieStoreId,
     inBackground:  event.shiftKey
   });
 }
 
-function onDblClick(event) {
+async function onDblClick(event) {
   if (EventUtils.isEventFiredOnNewTabButton(event))
     return;
 
@@ -684,30 +838,61 @@ function onDblClick(event) {
   const livingTab = EventUtils.getTabFromEvent(event);
   log('dblclick tab: ', tab, { living: !!livingTab });
 
-  if (livingTab) {
-    if (configs.simulateCloseTabByDblclick &&
-        event.button == 0 &&
-        !event.altKey &&
-        !event.ctrlKey &&
-        !event.metaKey &&
-        !event.shiftKey) {
-      event.stopPropagation();
-      event.preventDefault();
-      BackgroundConnection.sendMessage({
-        type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
-        tabIds: [livingTab.id],
-      });
+  if (livingTab &&
+      !EventUtils.isEventFiredOnTwisty(event) &&
+      !EventUtils.isEventFiredOnSoundButton(event)) {
+    const detail   = getMouseEventDetail(event, livingTab);
+    const treeItem = new TSTAPI.TreeItem(livingTab);
+    const extraContentsInfo = getOriginalExtraContentsTarget(event);
+    if (extraContentsInfo.owners) {
+      const allowed = await TSTAPI.tryOperationAllowed(
+        TSTAPI.kNOTIFY_TAB_DBLCLICKED,
+        {
+          ...detail,
+          tab:            treeItem,
+          originalTarget: extraContentsInfo.target
+        },
+        { tabProperties: ['tab'],
+          targets:       extraContentsInfo.owners }
+      );
+      if (!allowed)
+        return;
     }
-    else if (configs.collapseExpandSubtreeByDblClick) {
-      event.stopPropagation();
-      event.preventDefault();
-      BackgroundConnection.sendMessage({
-        type:            Constants.kCOMMAND_SET_SUBTREE_COLLAPSED_STATE,
-        tabId:           livingTab.id,
-        collapsed:       !livingTab.$TST.subtreeCollapsed,
-        manualOperation: true,
-        stack:           configs.debug && new Error().stack
-      });
+    const allowed = await TSTAPI.tryOperationAllowed(
+      TSTAPI.kNOTIFY_TAB_DBLCLICKED,
+      {
+        ...detail,
+        tab: treeItem
+      },
+      { tabProperties: ['tab'],
+        except:        extraContentsInfo.owners }
+    );
+    if (!allowed)
+      return;
+
+    if (configs.treeDoubleClickBehavior != Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_NONE) {
+      switch (configs.treeDoubleClickBehavior) {
+        case Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_TOGGLE_COLLAPSED:
+          event.stopPropagation();
+          event.preventDefault();
+          BackgroundConnection.sendMessage({
+            type:            Constants.kCOMMAND_SET_SUBTREE_COLLAPSED_STATE,
+            tabId:           livingTab.id,
+            collapsed:       !livingTab.$TST.subtreeCollapsed,
+            manualOperation: true,
+            stack:           configs.debug && new Error().stack
+          });
+          break;
+
+        case Constants.kTREE_DOUBLE_CLICK_BEHAVIOR_CLOSE:
+          event.stopPropagation();
+          event.preventDefault();
+          BackgroundConnection.sendMessage({
+            type:   Constants.kCOMMAND_REMOVE_TABS_INTERNALLY,
+            tabIds: [livingTab.id],
+          });
+          break;
+      }
     }
     return;
   }
