@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -71,15 +71,22 @@ export function init(scrollPosition) {
   document.addEventListener('wheel', onWheel, { capture: true });
   mTabBar.addEventListener('scroll', onScroll);
   BackgroundConnection.onMessage.addListener(onBackgroundMessage);
-  browser.runtime.onMessageExternal.addListener(onMessageExternal);
+  TSTAPI.onMessageExternal.addListener(onMessageExternal);
 }
 
 /* basics */
 
+//const mAnimationReduced = window.matchMedia('(prefers-reduced-motion: reduce)');
+
+function canApplyAnimation() {
+  return configs.animation /* && !mAnimationReduced.matches // Currently Firefox doesn't deactivate smooth scroll animation with this trigger, so we respect the decision. */;
+}
+
 function scrollTo(params = {}) {
   log('scrollTo ', params);
   if (!params.justNow &&
-      configs.animation && configs.smoothScrollEnabled)
+      canApplyAnimation() &&
+      configs.smoothScrollEnabled)
     return smoothScrollTo(params);
 
   //cancelPerformingAutoScroll();
@@ -213,11 +220,12 @@ export function scrollToNewTab(tab, options = {}) {
     return;
 
   if (configs.scrollToNewTabMode == Constants.kSCROLL_TO_NEW_TAB_IF_POSSIBLE) {
-    const activeTab = Tab.getActiveTab(TabsStore.getWindow());
-    scrollToTab(tab, Object.assign({}, options, {
+    const activeTab = Tab.getActiveTab(TabsStore.getCurrentWindowId());
+    scrollToTab(tab, {
+      ...options,
       anchor:            !activeTab.pinned && isTabInViewport(activeTab) && activeTab,
       notifyOnOutOfView: true
-    }));
+    });
   }
 }
 
@@ -296,14 +304,16 @@ export async function scrollToTab(tab, options = {}) {
         container:      containerRect.height
       });
     }
-    await scrollTo(Object.assign({}, options, {
+    await scrollTo({
+      ...options,
       position: mTabBar.scrollTop + delta
-    }));
+    });
   }
   else {
-    await scrollTo(Object.assign({}, options, {
+    await scrollTo({
+      ...options,
       tab
-    }));
+    });
   }
   // A tab can be moved after the tabbar is scrolled to the tab.
   // To retry "scroll to tab" behavior for such cases, we need to
@@ -437,7 +447,7 @@ function reserveToSaveScrollPosition() {
   reserveToSaveScrollPosition.reserved = setTimeout(() => {
     delete reserveToSaveScrollPosition.reserved;
     browser.sessions.setWindowValue(
-      TabsStore.getWindow(),
+      TabsStore.getCurrentWindowId(),
       Constants.kWINDOW_STATE_SCROLL_POSITION,
       mTabBar.scrollTop
     ).catch(ApiTabs.createErrorSuppressor());
@@ -534,7 +544,7 @@ async function onBackgroundMessage(message) {
       if (message.maybeMoved)
         await SidebarTabs.waitUntilNewTabIsMoved(message.tabId);
       const tab = Tab.get(message.tabId);
-      if (configs.animation) {
+      if (canApplyAnimation()) {
         wait(10).then(() => { // wait until the tab is moved by TST itself
           const parent = tab.$TST.parent;
           if (parent && parent.$TST.subtreeCollapsed) // possibly collapsed by other trigger intentionally
@@ -595,7 +605,7 @@ function onMessageExternal(message, _aSender) {
     case TSTAPI.kSCROLL:
       return (async () => {
         const params = {};
-        const currentWindow = TabsStore.getWindow();
+        const currentWindow = TabsStore.getCurrentWindowId();
         if ('tab' in message) {
           await Tab.waitUntilTracked(message.tab, { element: true });
           params.tab = Tab.get(message.tab);
@@ -603,8 +613,15 @@ function onMessageExternal(message, _aSender) {
             return;
         }
         else {
-          if (message.window != currentWindow)
+          const windowId = message.window || message.windowId;
+          if (windowId == 'active') {
+            const currentWindow = await browser.windows.get(TabsStore.getCurrentWindowId());
+            if (!currentWindow.focused)
+              return;
+          }
+          else if (windowId != currentWindow) {
             return;
+          }
           if ('delta' in message)
             params.delta = message.delta;
           if ('position' in message)
