@@ -9,6 +9,8 @@ import {
   log as internalLogger,
   nextFrame,
   wait,
+  countMatched,
+  toLines,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -28,10 +30,14 @@ import * as CollapseExpand from './collapse-expand.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
+import { TabUpdateTarget } from './components/TabElement.js';
+
 function log(...args) {
   internalLogger('sidebar/sidebar-cache', ...args);
 }
 
+const kCONTENTS_VERSION = 18;
+
 export const onRestored = new EventListenerManager();
 
 let mTracking = false;
@@ -41,7 +47,7 @@ let mTargetWindow;
 let mTabBar;
 
 export function init() {
-  mTargetWindow = TabsStore.getWindow();
+  mTargetWindow = TabsStore.getCurrentWindowId();
   mTabBar       = document.querySelector('#tabbar');
 }
 
@@ -135,7 +141,7 @@ export async function getEffectiveWindowCache(options = {}) {
         version: cache && cache.version
       }));
       log('getEffectiveWindowCache: verify cache (1)', { cache, tabsDirty, collapsedDirty });
-      if (cache && cache.version == Constants.kSIDEBAR_CONTENTS_VERSION) {
+      if (cache && cache.version == kCONTENTS_VERSION) {
         log('getEffectiveWindowCache: restore sidebar from cache');
         cache.tabbar.tabsDirty      = tabsDirty;
         cache.tabbar.collapsedDirty = collapsedDirty;
@@ -174,7 +180,8 @@ export async function getEffectiveWindowCache(options = {}) {
     MetricsData.add('getEffectiveWindowCache: validity check: actual signature failed.');
   }
   else {
-    cache.offset          = actualSignature.replace(cachedSignature, '').trim().split('\n').filter(part => !!part).length;
+    cache.offset= countMatched(actualSignature.replace(cachedSignature, '').trim().split('\n'),
+                               part => part);
     cache.actualSignature = actualSignature;
     log('getEffectiveWindowCache: success ');
     MetricsData.add('getEffectiveWindowCache: validity check: actual signature passed.');
@@ -185,7 +192,8 @@ export async function getEffectiveWindowCache(options = {}) {
 
 function getWindowSignature(tabs) {
   const tabIds = tabs.map(tab => tab.id);
-  return tabs.map(tab => `${tab.openerTabId ? tabIds.indexOf(tab.openerTabId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`).join('\n');
+  return toLines(tabs,
+                 tab => `${tab.openerTabId ? tabIds.indexOf(tab.openerTabId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`);
 }
 
 function trimSignature(signature, ignoreCount) {
@@ -230,7 +238,8 @@ function signatureFromTabsCache(cachedTabs) {
       pinned:        pinnedMatcher.test(classes)
     };
   });
-  return tabs.map(tab => `${tab.parentId ? tabIds.indexOf(tab.parentId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`).join('\n');
+  return toLines(tabs,
+                 tab => `${tab.parentId ? tabIds.indexOf(tab.parentId) : -1 },${tab.cookieStoreId},${tab.incognito},${tab.pinned}`);
 }
 
 
@@ -295,6 +304,8 @@ async function restoreTabsFromCacheInternal(params) {
         tabsMustBeRemoved.map(tab => tab.id));
     log(`restoreTabsFromCacheInternal: => ${container.childNodes.length} tabs`);
     const matched = params.cache.match(/<li/g);
+    if (!matched)
+      throw new Error('restoreTabsFromCacheInternal: invalid cache');
     log(`restoreTabsFromCacheInternal: restore ${matched.length} tabs from cache`);
     if (configs.debug)
       dumpCache(params.cache);
@@ -398,8 +409,7 @@ async function fixupTabsRestoredFromCache(tabElements, tabs, options = {}) {
   for (let i = tabElements.length - 1; i > -1; i--) {
     const tabElement = tabElements[i];
     const tab = tabElement.apiTab;
-    SidebarTabs.applyStatesToElement(tab);
-    SidebarTabs.applyCollapseExpandStateToElement(tab);
+    tab.$TST.updateElement(TabUpdateTarget.CollapseExpandState | TabUpdateTarget.TabProperties);
     if (options.dirty)
       TabsUpdate.updateTab(tab, tab, { forceApply: true });
     if (Date.now() - lastDraw > configs.intervalToUpdateProgressForBlockedUserOperation) {
@@ -502,7 +512,7 @@ async function updateCachedTabbar() {
   log('updateCachedTabbar ', { stack: configs.debug && new Error().stack });
   mLastWindowCacheOwner = getWindowCacheOwner(mTargetWindow);
   updateWindowCache(Constants.kWINDOW_STATE_CACHED_SIDEBAR, {
-    version: Constants.kSIDEBAR_CONTENTS_VERSION,
+    version: kCONTENTS_VERSION,
     tabbar: {
       contents:        SidebarTabs.wholeContainer.innerHTML,
       style:           mTabBar.getAttribute('style'),
