@@ -6,12 +6,13 @@
 'use strict';
 
 import RichConfirm from '/extlib/RichConfirm.js';
-import TabIdFixer from '/extlib/TabIdFixer.js';
 
 import {
   log as internalLogger,
   nextFrame,
-  configs
+  mapAndFilter,
+  configs,
+  loadUserStyleRules
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
@@ -22,26 +23,39 @@ import * as TSTAPI from '/common/tst-api.js';
 import * as ContextualIdentities from '/common/contextual-identities.js';
 import * as Bookmark from '/common/bookmark.js';
 import * as UserOperationBlocker from '/common/user-operation-blocker.js';
+import * as Color from '/common/color.js';
 import * as BrowserTheme from '/common/browser-theme.js';
 import * as MetricsData from '/common/metrics-data.js';
 
 import Tab from '/common/Tab.js';
 import Window from '/common/Window.js';
 
+
+import { TabTwistyElement } from './components/TabTwistyElement.js';
+import { TabCloseBoxElement } from './components/TabCloseBoxElement.js';
+import { TabFaviconElement } from './components/TabFaviconElement.js';
+import { TabLabelElement } from './components/TabLabelElement.js';
+import { TabCounterElement } from './components/TabCounterElement.js';
+import { TabSoundButtonElement } from './components/TabSoundButtonElement.js';
+import {
+  TabElement,
+  TabInvalidationTarget,
+  TabUpdateTarget,
+} from './components/TabElement.js';
+
 import * as BackgroundConnection from './background-connection.js';
 import * as SidebarCache from './sidebar-cache.js';
 import * as SidebarTabs from './sidebar-tabs.js';
 import * as PinnedTabs from './pinned-tabs.js';
 import * as DragAndDrop from './drag-and-drop.js';
-import * as TabDragHandle from './tab-drag-handle.js';
 import * as RestoringTabCount from './restoring-tab-count.js';
 import * as CollapseExpand from './collapse-expand.js';
 import * as Size from './size.js';
-import * as Color from './color.js';
 import * as Indent from './indent.js';
 import * as Scroll from './scroll.js';
 import * as TabContextMenu from './tab-context-menu.js';
 import * as SubPanel from './subpanel.js';
+import './tst-api-frontend.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
@@ -69,6 +83,7 @@ const mStyleLoader                = document.querySelector('#style-loader');
 const mBrowserThemeDefinition     = document.querySelector('#browser-theme-definition');
 const mUserStyleRules             = document.querySelector('#user-style-rules');
 const mContextualIdentitiesStyle  = document.querySelector('#contextual-identity-styling');
+const mLessAnimationMedia         = window.matchMedia('(prefers-reduced-motion: reduce)');
 
 { // apply style ASAP!
   // allow customiation for platform specific styles with selectors like `:root[data-user-agent*="Windows NT 10"]`
@@ -78,7 +93,10 @@ const mContextualIdentitiesStyle  = document.querySelector('#contextual-identity
   if (style)
     applyStyle(style[1]);
   else
-    configs.$loaded.then(() => applyStyle());
+    configs.$loaded.then(async () => {
+      await applyStyle();
+      Size.update();
+    });
 
   configs.$loaded.then(applyUserStyleRules);
 
@@ -91,6 +109,26 @@ export async function init() {
   MetricsData.add('init: start');
   log('initialize sidebar on load');
 
+  // If we call `window.customElements.define(localName, constructor)`;` from a file defining a custom element,
+  // it would be a side-effect and happen accidentally that defining a custom element
+  // when we import a new file which defines a new custom element.
+  // It causes a complex side-effect relations and usually causes a bug. It's tough to fix.
+  //
+  // I have not concluded the best practice about it yet,
+  // but I think that it's safely to call `window.customElements.define(localName, constructor)` separately
+  // in the application initialization phase.
+  //
+  // XXX:
+  //  We define our custom elements at first to avoid a problem which calls a method of custom element
+  //  which has not been defined yet.
+  TabTwistyElement.define();
+  TabCloseBoxElement.define();
+  TabFaviconElement.define();
+  TabLabelElement.define();
+  TabCounterElement.define();
+  TabSoundButtonElement.define();
+  TabElement.define();
+
   // Read caches from existing tabs at first, for better performance.
   // Those promises will be resolved while waiting other operations.
   SidebarCache.tryPreload();
@@ -99,22 +137,29 @@ export async function init() {
   UserOperationBlocker.setProgress(0);
   await Promise.all([
     MetricsData.addAsync('getting native tabs', async () => {
-      const window = await MetricsData.addAsync('browser.windows.getCurrent', browser.windows.getCurrent({ populate: true })).catch(ApiTabs.createErrorHandler());
+      const bundledWindowId = /windowId=([1-9][0-9]*)/i.test(location.search) ? parseInt(RegExp.$1) : 0;
+      const window = await MetricsData.addAsync(
+        'getting window',
+        bundledWindowId >= 1 ?
+          browser.windows.get(bundledWindowId, { populate: true }) :
+          browser.windows.getCurrent({ populate: true })
+      ).catch(ApiTabs.createErrorHandler());
       if (window.focused)
         document.documentElement.classList.add('active');
       const trackedWindow = TabsStore.windows.get(window.id) || new Window(window.id);
       trackedWindow.incognito = window.incognito;
+      if (window.incognito)
+        document.documentElement.classList.add('incognito');
 
       const tabs = window.tabs;
       SidebarCache.tryPreload(tabs.filter(tab => !tab.pinned)[0] || tabs[0]);
       mTargetWindow = tabs[0].windowId;
-      TabsStore.setWindow(mTargetWindow);
+      TabsStore.setCurrentWindowId(mTargetWindow);
       mPromisedTargetWindowResolver(mTargetWindow);
       internalLogger.context   = `Sidebar-${mTargetWindow}`;
 
       // Track only the first tab for now, because it is required to initialize
       // the container and it will be used by the SidebarCache module.
-      TabIdFixer.fixTab(tabs[0]);
       Tab.track(tabs[0]);
 
       promisedAllTabsTracked = MetricsData.addAsync('tracking all native tabs', async () => {
@@ -122,7 +167,6 @@ export async function init() {
         let count = 0;
         const maxCount = tabs.length - 1;
         for (const tab of tabs.slice(1)) {
-          TabIdFixer.fixTab(tab);
           Tab.track(tab);
           if (Date.now() - lastDraw > configs.intervalToUpdateProgressForBlockedUserOperation) {
             UserOperationBlocker.setProgress(Math.round(++count / maxCount * 16) + 16); // 2/6: track all tabs
@@ -176,7 +220,7 @@ export async function init() {
 
       log('Start to process messages including queued ones');
       BackgroundConnection.start();
-      onConfigChange('applyBrowserTheme');
+      browser.theme.getCurrent(mTargetWindow).then(applyBrowserTheme);
 
       SidebarTabs.onSyncFailed.addListener(() => rebuildAll());
 
@@ -192,15 +236,20 @@ export async function init() {
       window.addEventListener('resize', onResize);
       mTabBar.addEventListener('transitionend', onTransisionEnd);
 
-      if (browser.theme && browser.theme.onUpdated) // Firefox 58 and later
-        browser.theme.onUpdated.addListener(onBrowserThemeChanged);
+      browser.theme.onUpdated.addListener(onBrowserThemeChanged);
+
+      // We need to re-calculate mixed colors when the system color scheme is changed.
+      // See also: https://github.com/piroor/treestyletab/issues/2314
+      window.matchMedia('(prefers-color-scheme: dark)').addListener(async _event => {
+        const theme = await browser.theme.getCurrent(mTargetWindow);
+        applyBrowserTheme(theme);
+      });
 
       browser.runtime.onMessage.addListener(onMessage);
 
       onBuilt.dispatch();
 
       DragAndDrop.init();
-      TabDragHandle.init();
     }),
     MetricsData.addAsync('parallel initialization: Size', async () => {
       Size.init();
@@ -233,6 +282,9 @@ export async function init() {
       SidebarTabs.init();
       Indent.reserveToUpdateVisualMaxTreeLevel();
 
+      mLessAnimationMedia.addListener(_event => {
+        onConfigChange('animation');
+      });
       onConfigChange('animation');
       onReady.dispatch();
     }),
@@ -260,9 +312,6 @@ export async function init() {
 function applyStyle(style) {
   mStyle = style || configs.style;
   switch (mStyle) {
-    case 'metal':
-      mStyleLoader.setAttribute('href', 'styles/metal/metal.css');
-      break;
     case 'sidebar':
       mStyleLoader.setAttribute('href', 'styles/sidebar/sidebar.css');
       break;
@@ -272,9 +321,6 @@ function applyStyle(style) {
     case 'vertigo':
       mStyleLoader.setAttribute('href', 'styles/square/vertigo.css');
       break;
-    case 'plain-dark':
-      mStyleLoader.setAttribute('href', 'styles/square/plain-dark.css');
-      break;
     case 'plain':
     case 'flat': // for backward compatibility, fall back to plain.
       mStyleLoader.setAttribute('href', 'styles/square/plain.css');
@@ -296,107 +342,57 @@ function applyStyle(style) {
 }
 
 function applyUserStyleRules() {
-  mUserStyleRules.textContent = configs.userStyleRules || '';
+  mUserStyleRules.textContent = loadUserStyleRules();
 }
 
-function applyBrowserTheme(theme) {
+async function applyBrowserTheme(theme) {
   log('applying theme ', theme);
 
-  const baseColor = Color.parseCSSColor(window.getComputedStyle(document.querySelector('#dummy-favicon-size-box'), null).backgroundColor);
+  let browserThemeStyle = '';
+  if (!theme ||
+      !theme.colors) {
+    // Reset colors
+    mBrowserThemeDefinition.textContent = '';
+  }
+  else {
+    browserThemeStyle = await BrowserTheme.generateThemeDeclarations(theme);
+    // Apply theme color at first, to use given colors as the base of following "face-*" colors.
+    mBrowserThemeDefinition.textContent = browserThemeStyle;
+  }
+
+  const baseColor = Color.parseCSSColor(window.getComputedStyle(document.querySelector('#dummy-tab-color-box'), null).backgroundColor);
   const highlightColor = Color.parseCSSColor(window.getComputedStyle(document.querySelector('#dummy-highlight-color-box'), null).backgroundColor);
   const defaultColors = `:root {
-    --face-highlight-lighter: ${Color.mixCSSColors(baseColor, Object.assign({}, highlightColor, { alpha: 0.35 }),)};
-    --face-highlight-more-lighter: ${Color.mixCSSColors(baseColor, Object.assign({}, highlightColor, { alpha: 0.2 }))};
-    --face-highlight-more-more-lighter: ${Color.mixCSSColors(baseColor, Object.assign({}, highlightColor, { alpha: 0.1 }))};
+    --face-highlight-lighter: ${Color.mixCSSColors(baseColor, { ...highlightColor, alpha: 0.35 })};
+    --face-highlight-more-lighter: ${Color.mixCSSColors(baseColor, { ...highlightColor, alpha: 0.2 })};
+    --face-highlight-more-more-lighter: ${Color.mixCSSColors(baseColor, { ...highlightColor, alpha: 0.1 })};
     --face-gradient-start-active: rgba(${baseColor.red}, ${baseColor.green}, ${baseColor.blue}, 0.4);
     --face-gradient-start-inactive: rgba(${baseColor.red}, ${baseColor.green}, ${baseColor.blue}, 0.2);
     --face-gradient-end: rgba(${baseColor.red}, ${baseColor.green}, ${baseColor.blue}, 0);
   }`;
 
   if (!theme ||
-      !theme.colors ||
-      !configs.applyBrowserTheme) {
+      !theme.colors) {
     mBrowserThemeDefinition.textContent = defaultColors;
-    return;
   }
-  const extraColors = [];
-  const themeFrameColor   = theme.colors.frame || theme.colors.accentcolor /* old name */;
-  const inactiveTextColor = theme.colors.tab_background_text || theme.colors.textcolor /* old name */;
-  const activeTextColor   = theme.colors.bookmark_text || theme.colors.toolbar_text /* old name */ || inactiveTextColor;
-  let bgAlpha = 1;
-  if (theme.images) {
-    const frameImage = theme.images.theme_frame || theme.images.headerURL /* old name */;
-    if (frameImage) {
-      extraColors.push(`--browser-header-url: url(${JSON.stringify(frameImage)})`);
-      extraColors.push('--browser-bg-for-header-image: transparent;');
-      // https://searchfox.org/mozilla-central/rev/532e4b94b9e807d157ba8e55034aef05c1196dc9/browser/themes/shared/tabs.inc.css#537
-      extraColors.push('--browser-bg-hover-for-header-image: rgba(0, 0, 0, 0.1);');
-      // https://searchfox.org/mozilla-central/rev/532e4b94b9e807d157ba8e55034aef05c1196dc9/browser/base/content/browser.css#20
-      extraColors.push('--browser-bg-active-for-header-image: rgba(255, 255, 255, 0.4)');
-      // https://searchfox.org/mozilla-central/rev/532e4b94b9e807d157ba8e55034aef05c1196dc9/toolkit/themes/windows/global/global.css#138
-      if (Color.isBrightColor(themeFrameColor))
-        extraColors.push('--browser-textshadow-for-header-image: 1px 1px 1.5px black'); // for bright text
-      else
-        extraColors.push('--browser-textshadow-for-header-image: 0 -0.5px 1.5px white'); // for dark text
-    }
-    if (Array.isArray(theme.images.additional_backgrounds) &&
-        theme.images.additional_backgrounds.length > 0) {
-      extraColors.push(`--browser-bg-url: url(${JSON.stringify(theme.images.additional_backgrounds[0])})`);
-      bgAlpha = 0.75;
-    }
+  else {
+    mBrowserThemeDefinition.textContent = [
+      defaultColors,
+      browserThemeStyle
+    ].join('\n');
   }
-  const themeBaseColor    = Color.mixCSSColors(themeFrameColor, 'rgba(0, 0, 0, 0)', bgAlpha);
-  let toolbarColor = Color.mixCSSColors(themeBaseColor, 'rgba(255, 255, 255, 0.4)', bgAlpha);
-  if (theme.colors.toolbar)
-    toolbarColor = Color.mixCSSColors(themeBaseColor, theme.colors.toolbar);
-  if (theme.colors.tab_line)
-    extraColors.push(`--browser-tab-highlighter: ${theme.colors.tab_line}`);
-  if (theme.colors.tab_loading)
-    extraColors.push(`--browser-loading-indicator: ${theme.colors.tab_loading}`);
-  extraColors.push(BrowserTheme.generateThemeRules(theme));
-  mBrowserThemeDefinition.textContent = `
-    ${defaultColors}
-    :root {
-      --browser-background:      ${themeFrameColor};
-      --browser-bg-base:         ${themeBaseColor};
-      --browser-bg-less-lighter: ${Color.mixCSSColors(themeBaseColor, 'rgba(255, 255, 255, 0.25)', bgAlpha)};
-      --browser-bg-lighter:      ${toolbarColor};
-      --browser-bg-more-lighter: ${Color.mixCSSColors(toolbarColor, 'rgba(255, 255, 255, 0.6)', bgAlpha)};
-      --browser-bg-lightest:     ${Color.mixCSSColors(toolbarColor, 'rgba(255, 255, 255, 0.85)', bgAlpha)};
-      --browser-bg-less-darker:  ${Color.mixCSSColors(themeBaseColor, 'rgba(0, 0, 0, 0.1)', bgAlpha)};
-      --browser-bg-darker:       ${Color.mixCSSColors(themeBaseColor, 'rgba(0, 0, 0, 0.25)', bgAlpha)};
-      --browser-bg-more-darker:  ${Color.mixCSSColors(themeBaseColor, 'rgba(0, 0, 0, 0.5)', bgAlpha)};
-      --browser-fg:              ${inactiveTextColor};
-      --browser-fg-active:       ${activeTextColor};
-      --browser-border:          ${Color.mixCSSColors(inactiveTextColor, 'rgba(0, 0, 0, 0)', 0.4)};
-      ${extraColors.join(';\n')}
-    }
-  `;
 }
 
 function updateContextualIdentitiesStyle() {
-  const definitions = [];
-  const customColors = [];
-  ContextualIdentities.forEach(identity => {
-    if (!identity.colorCode)
-      return;
-    let colorValue = identity.colorCode;
-    if (identity.color) {
-      const customColor = `--contextual-identity-color-${identity.color}`;
-      customColors.push(`${customColor}: ${identity.colorCode};`);
-      colorValue = `var(${customColor})`;
-    }
-    definitions.push(`
-      .tab.contextual-identity-${identity.cookieStoreId} .contextual-identity-marker {
-        background-color: ${colorValue};
-      }
-    `);
-  });
+  const colorInfo = ContextualIdentities.getColorInfo();
+  const definitions = Object.keys(colorInfo.colors).map(id =>
+    `.tab.contextual-identity-${id} .contextual-identity-marker {
+       background-color: ${colorInfo.colors[id]};
+     }`);
 
   // This is required to map different color for color names.
   // See also: https://github.com/piroor/treestyletab/issues/2296
-  if (customColors.length > 0)
-    definitions.push(`:root { ${customColors.join('\n')} }`);
+  definitions.push(colorInfo.colorDeclarations);
 
   mContextualIdentitiesStyle.textContent = definitions.join('\n');
 }
@@ -498,7 +494,7 @@ export async function rebuildAll(importedTabs, cache) {
   for (const tab of tabs) {
     const trackedTab = Tab.init(tab, { existing: true, inBackground: true });
     TabsUpdate.updateTab(trackedTab, tab, { forceApply: true });
-    SidebarTabs.applyCollapseExpandStateToElement(trackedTab);
+    trackedTab.$TST.updateElement(TabUpdateTarget.CollapseExpandState);
     if (tab.active)
       TabsInternalOperation.setTabActive(trackedTab);
     if (Date.now() - lastDraw > configs.intervalToUpdateProgressForBlockedUserOperation) {
@@ -552,37 +548,42 @@ async function importTabsFromBackground() {
 }
 
 
-export async function confirmToCloseTabs(tabs, _options = {}) {
-  tabs = tabs.filter(tab => !configs.grantedRemovingTabIds.includes(tab.id));
-  const tabIds = tabs.map(tab => tab.id);
+export async function confirmToCloseTabs(tabs, options = {}) {
+  const tabIds = [];
+  const configKey = options.configKey || 'warnOnCloseTabs';
+  tabs = tabs.filter(tab => {
+    if (!configs.grantedRemovingTabIds.includes(tab.id)) {
+      tabIds.push(tab.id);
+      return true;
+    }
+    return false;
+  });
   log('confirmToCloseTabs: ', tabIds);
   const count = tabIds.length;
   if (count <= 1 ||
-      !configs.warnOnCloseTabs)
+      !configs[configKey])
     return true;
 
-  const confirm = new RichConfirm({
-    message: browser.i18n.getMessage('warnOnCloseTabs_message', [count]),
-    buttons: [
-      browser.i18n.getMessage('warnOnCloseTabs_close'),
-      browser.i18n.getMessage('warnOnCloseTabs_cancel')
-    ],
-    checkMessage: browser.i18n.getMessage('warnOnCloseTabs_warnAgain'),
-    checked: true
-  });
-  const result = await confirm.show();
-  switch (result.buttonIndex) {
-    case 0:
-      if (!result.checked)
-        configs.warnOnCloseTabs = false;
+  try {
+    const granted = await browser.runtime.sendMessage({
+      type:     Constants.kCOMMAND_CONFIRM_TO_CLOSE_TABS,
+      windowId: mTargetWindow,
+      configKey,
+      tabs
+    });
+    if (granted) {
       configs.lastConfirmedToCloseTabs = Date.now();
-      configs.grantedRemovingTabIds = Array.from(new Set((configs.grantedRemovingTabIds || []).concat(tabIds)));
+      configs.grantedRemovingTabIds = Array.from(new Set((configs.grantedRemovingTabIds   || []).concat(tabIds)));
       log('confirmToCloseTabs: granted ', configs.grantedRemovingTabIds);
       reserveToClearGrantedRemovingTabs();
       return true;
-    default:
-      return false;
+    }
   }
+  catch(error) {
+    console.error(error);
+  }
+
+  return false;
 }
 TabContextMenu.onTabsClosing.addListener(confirmToCloseTabs);
 
@@ -594,37 +595,6 @@ function reserveToClearGrantedRemovingTabs() {
   }, 1000);
 }
 
-async function confirmToAutoGroupNewTabs(tabIds) {
-  const count = tabIds.length;
-  if (count <= 1 ||
-      !configs.warnOnAutoGroupNewTabs)
-    return true;
-
-  const confirm = new RichConfirm({
-    message: browser.i18n.getMessage('warnOnAutoGroupNewTabs_message', [count]),
-    buttons: [
-      browser.i18n.getMessage('warnOnAutoGroupNewTabs_close'),
-      browser.i18n.getMessage('warnOnAutoGroupNewTabs_cancel')
-    ],
-    checkMessage: browser.i18n.getMessage('warnOnAutoGroupNewTabs_warnAgain'),
-    checked: true
-  });
-  const result = await confirm.show();
-  switch (result.buttonIndex) {
-    case 0:
-      if (!result.checked)
-        configs.warnOnAutoGroupNewTabs = false;
-      return true;
-    case 1:
-      if (!result.checked) {
-        configs.warnOnAutoGroupNewTabs = false;
-        configs.autoGroupNewTabs = false;
-      }
-    default:
-      return false;
-  }
-}
-
 
 export function reserveToUpdateTabbarLayout(options = {}) {
   //log('reserveToUpdateTabbarLayout');
@@ -673,13 +643,13 @@ function updateTabbarLayout(params = {}) {
       // Tab at the end of the tab bar can be hidden completely or
       // partially (newly opened in small tab bar, or scrolled out when
       // the window is shrunken), so we need to scroll to it explicitely.
-      const activeTab = Tab.getActiveTab(TabsStore.getWindow());
+      const activeTab = Tab.getActiveTab(TabsStore.getCurrentWindowId());
       if (activeTab && !Scroll.isTabInViewport(activeTab)) {
         log('scroll to active tab on updateTabbarLayout');
         Scroll.scrollToTab(activeTab);
         return;
       }
-      const lastOpenedTab = Tab.getLastOpenedTab(TabsStore.getWindow());
+      const lastOpenedTab = Tab.getLastOpenedTab(TabsStore.getCurrentWindowId());
       const reasons       = params.reasons || 0;
       if (reasons & Constants.kTABBAR_UPDATE_REASON_TAB_OPEN &&
           !Scroll.isTabInViewport(lastOpenedTab)) {
@@ -762,7 +732,7 @@ function onConfigChange(changedKey) {
         // breaking of initialized tab states.
         for (const tab of Tab.getAllTabs(mTargetWindow, { iterator: true })) {
           TabsUpdate.updateTab(tab, tab, { forceApply: true });
-          tab.$TST.tooltipIsDirty = true;
+          tab.$TST.invalidateElement(TabInvalidationTarget.Tooltip);
         }
       }
       if (configs.debug)
@@ -772,7 +742,7 @@ function onConfigChange(changedKey) {
     }; break;
 
     case 'animation':
-      if (configs.animation)
+      if (configs.animation && !mLessAnimationMedia.matches)
         rootClasses.add('animation');
       else
         rootClasses.remove('animation');
@@ -817,17 +787,6 @@ function onConfigChange(changedKey) {
       document.documentElement.setAttribute('color-scheme', configs.colorScheme);
       break;
 
-    case 'applyBrowserTheme':
-      if (browser.theme && browser.theme.getCurrent) // Firefox 58 and later
-        browser.theme.getCurrent(mTargetWindow).then(applyBrowserTheme);
-      else
-        applyBrowserTheme();
-      break;
-
-    case 'userStyleRules':
-      applyUserStyleRules()
-      break;
-
     case 'inheritContextualIdentityToNewChildTab':
       updateContextualIdentitiesSelector();
       break;
@@ -852,6 +811,11 @@ function onConfigChange(changedKey) {
       else
         rootClasses.remove('simulate-svg-context-fill');
       break;
+
+    default:
+      if (changedKey.startsWith('userStyleRules'))
+        applyUserStyleRules();
+      break;
   }
 }
 
@@ -870,28 +834,19 @@ function onMessage(message, _sender, _respond) {
 
   //log('onMessage: ', message, sender);
   switch (message.type) {
-    case Constants.kCOMMAND_CONFIRM_TO_CLOSE_TABS:
-      log('kCOMMAND_CONFIRM_TO_CLOSE_TABS: ', { message, mTargetWindow });
-      return confirmToCloseTabs(message.tabs);
-
-    case Constants.kCOMMAND_CONFIRM_TO_AUTO_GROUP_NEW_TABS:
-      log('kCOMMAND_CONFIRM_TO_AUTO_GROUP_NEW_TABS: ', { message, mTargetWindow });
-      return confirmToAutoGroupNewTabs(message.tabIds);
-
     case Constants.kCOMMAND_RELOAD:
       location.reload();
       return;
+
+    case Constants.kCOMMAND_SHOW_DIALOG:
+      return RichConfirm.show(message.params);
   }
 }
 
+const BUFFER_KEY_PREFIX = 'sidebar-';
 
 BackgroundConnection.onMessage.addListener(async message => {
   switch (message.type) {
-    case Constants.kCOMMAND_REMOVE_TABS_INTERNALLY:
-      await Tab.waitUntilTracked(message.tabIds, { element: true });
-      TabsInternalOperation.removeTabs(message.tabIds.map(id => Tab.get(id)));
-      break;
-
     case Constants.kCOMMAND_BLOCK_USER_OPERATIONS:
       UserOperationBlocker.blockIn(mTargetWindow, message);
       break;
@@ -922,6 +877,26 @@ BackgroundConnection.onMessage.addListener(async message => {
       });
     }; break;
 
+    case Constants.kCOMMAND_NOTIFY_TAB_SHOWN:
+    case Constants.kCOMMAND_NOTIFY_TAB_HIDDEN: {
+      if (BackgroundConnection.handleBufferedMessage({ type: 'shown/hidden', message }, `${BUFFER_KEY_PREFIX}${message.tabId}`))
+        return;
+      await Tab.waitUntilTracked(message.tabId, { element: true });
+      const lastMessage = BackgroundConnection.fetchBufferedMessage('shown/hidden', `${BUFFER_KEY_PREFIX}${message.tabId}`);
+      if (!lastMessage)
+        return;
+      if (lastMessage.message.type == Constants.kCOMMAND_NOTIFY_TAB_SHOWN) {
+        reserveToUpdateTabbarLayout({
+          reason: Constants.kTABBAR_UPDATE_REASON_TAB_OPEN
+        });
+      }
+      else {
+        reserveToUpdateTabbarLayout({
+          reason: Constants.kTABBAR_UPDATE_REASON_TAB_CLOSE
+        });
+      }
+    }; break;
+
     case Constants.kCOMMAND_NOTIFY_TAB_RESTORING: {
       if (!configs.useCachedTree) // we cannot know when we should unblock on no cache case...
         return;
@@ -945,6 +920,7 @@ BackgroundConnection.onMessage.addListener(async message => {
         ignorePinnedTabs: true
       });
       if (!cache ||
+          !cache.tabbar.cache ||
           (cache.offset &&
            window.element.childNodes.length <= cache.offset)) {
         log('Tabs.onWindowRestoring: no effective cache');
@@ -972,13 +948,19 @@ BackgroundConnection.onMessage.addListener(async message => {
       MetricsData.add('Tabs.onWindowRestoring restore end');
     }; break;
 
-    case Constants.kCOMMAND_BOOKMARK_TAB_WITH_DIALOG:
-      Bookmark.bookmarkTab(Tab.get(message.tabId), { showDialog: true });
-      break;
+    case Constants.kCOMMAND_BOOKMARK_TAB_WITH_DIALOG: {
+      Bookmark.bookmarkTab(Tab.get(message.tabId), {
+        ...(message.options || {}),
+        showDialog: true
+      });
+    }; break;
 
-    case Constants.kCOMMAND_BOOKMARK_TABS_WITH_DIALOG:
-      Bookmark.bookmarkTabs(message.tabIds.map(id => Tab.get(id)), { showDialog: true });
-      break;
+    case Constants.kCOMMAND_BOOKMARK_TABS_WITH_DIALOG: {
+      Bookmark.bookmarkTabs(mapAndFilter(message.tabIds, id => Tab.get(id)), {
+        ...(message.options || {}),
+        showDialog: true
+      });
+    }; break;
   }
 });
 
