@@ -14,6 +14,7 @@ import {
 import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
 import * as TabsStore from '/common/tabs-store.js';
+import * as SidebarConnection from '/common/sidebar-connection.js';
 
 import Tab from '/common/Tab.js';
 
@@ -100,14 +101,35 @@ async function updateInternal(tabId) {
   let successor = null;
   if (renewedTab.active) {
     if (configs.successorTabControlLevel == Constants.kSUCCESSOR_TAB_CONTROL_IN_TREE) {
-      const firstChild = tab.$TST.firstChild;
-      successor = (
-        (firstChild && !firstChild.$TST.collapsed && firstChild) ||
-        (tab.$TST.nextSiblingTab || tab.$TST.nearestVisiblePrecedingTab)
-      );
+      const firstChild = (
+        configs.treatTreeAsExpandedOnClosedWithNoSidebar &&
+        !SidebarConnection.isOpen(tab.windowId)
+      ) ? tab.$TST.firstChild : tab.$TST.firstVisibleChild;
+      successor = firstChild || tab.$TST.nextVisibleSiblingTab || tab.$TST.nearestVisiblePrecedingTab;
+      log(`  possible successor: ${dumpTab(tab)}`);
+      if (successor &&
+          successor.discarded &&
+          configs.avoidDiscardedTabToBeActivatedIfPossible) {
+        log(`  ${dumpTab(successor)} is discarded.`);
+        successor = tab.$TST.nearestLoadedSiblingTab ||
+                      tab.$TST.nearestLoadedTabInTree ||
+                      tab.$TST.nearestLoadedTab ||
+                      successor;
+        log(`  => redirected successor is: ${dumpTab(successor)}`);
+      }
     }
-    else
+    else {
       successor = tab.$TST.nearestVisibleFollowingTab || tab.$TST.nearestVisiblePrecedingTab;
+      log(`  possible successor: ${dumpTab(tab)}`);
+      if (successor &&
+          successor.discarded &&
+          configs.avoidDiscardedTabToBeActivatedIfPossible) {
+        log(`  ${dumpTab(successor)} is discarded.`);
+        successor = tab.$TST.nearestLoadedTab ||
+                      successor;
+        log(`  => redirected successor is: ${dumpTab(successor)}`);
+      }
+    }
   }
   if (successor) {
     log(`  ${dumpTab(tab)} is under control: successor = ${successor.id}`);
@@ -190,10 +212,14 @@ Tab.onCreating.addListener((tab, info = {}) => {
   });
 });
 
-Tab.onCreated.addListener((tab, _info = {}) => {
-  const activeTab = Tab.getActiveTab(tab.windowId);
+function updateActiveTab(windowId) {
+  const activeTab = Tab.getActiveTab(windowId);
   if (activeTab)
     update(activeTab.id);
+}
+
+Tab.onCreated.addListener((tab, _info = {}) => {
+  updateActiveTab(tab.windowId);
 });
 
 Tab.onRemoving.addListener((tab, removeInfo = {}) => {
@@ -212,9 +238,8 @@ Tab.onRemoving.addListener((tab, removeInfo = {}) => {
 });
 
 Tab.onRemoved.addListener((tab, info = {}) => {
-  const activeTab = Tab.getActiveTab(info.windowId);
-  if (activeTab && !info.isWindowClosing)
-    update(activeTab.id);
+  updateActiveTab(info.windowId);
+
   const window = TabsStore.windows.get(info.windowId);
   if (!window)
     return;
@@ -224,9 +249,7 @@ Tab.onRemoved.addListener((tab, info = {}) => {
 });
 
 Tab.onMoved.addListener((tab, info = {}) => {
-  const activeTab = Tab.getActiveTab(tab.windowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(tab.windowId);
 
   if (!info.byInternalOperation) {
     log(`clear lastRelatedTabs for ${tab.windowId} by tabs.onMoved`);
@@ -237,15 +260,11 @@ Tab.onMoved.addListener((tab, info = {}) => {
 });
 
 Tab.onAttached.addListener((_tab, info = {}) => {
-  const activeTab = Tab.getActiveTab(info.newWindowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(info.newWindowId);
 });
 
 Tab.onDetached.addListener((_tab, info = {}) => {
-  const activeTab = Tab.getActiveTab(info.oldWindowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(info.oldWindowId);
 
   const window = TabsStore.windows.get(info.oldWindowId);
   if (window) {
@@ -255,20 +274,36 @@ Tab.onDetached.addListener((_tab, info = {}) => {
   }
 });
 
+Tab.onUpdated.addListener((tab, changeInfo = {}) => {
+  if (!('discarded' in changeInfo))
+    return;
+  updateActiveTab(tab.windowId);
+});
+
+Tab.onShown.addListener(tab => {
+  updateActiveTab(tab.windowId);
+});
+
+Tab.onHidden.addListener(tab => {
+  updateActiveTab(tab.windowId);
+});
+
 Tree.onAttached.addListener((child, _info = {}) => {
-  const activeTab = Tab.getActiveTab(child.windowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(child.windowId);
 });
 
 Tree.onDetached.addListener((child, _info = {}) => {
-  const activeTab = Tab.getActiveTab(child.windowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(child.windowId);
 });
 
 Tree.onSubtreeCollapsedStateChanging.addListener((tab, _info = {}) => {
-  const activeTab = Tab.getActiveTab(tab.windowId);
-  if (activeTab)
-    update(activeTab.id);
+  updateActiveTab(tab.windowId);
+});
+
+SidebarConnection.onConnected.addListener((windowId, _openCount) => {
+  updateActiveTab(windowId);
+});
+
+SidebarConnection.onDisconnected.addListener((windowId, _openCount) => {
+  updateActiveTab(windowId);
 });
