@@ -7,6 +7,7 @@
 
 import {
   log as internalLogger,
+  mapAndFilter,
   configs
 } from '/common/common.js';
 import * as ApiTabs from '/common/api-tabs.js';
@@ -30,6 +31,22 @@ function log(...args) {
 export const onTSTItemClick = new EventListenerManager();
 export const onTSTTabContextMenuShown = new EventListenerManager();
 export const onTSTTabContextMenuHidden = new EventListenerManager();
+export const onTopLevelItemAdded = new EventListenerManager();
+
+const EXTERNAL_TOP_LEVEL_ITEM_MATCHER = /^external-top-level-item:([^:]+):(.+)$/;
+function getExternalTopLevelItemId(ownerId, itemId) {
+  return `external-top-level-item:${ownerId}:${itemId}`;
+}
+
+const SAFE_MENU_PROPERTIES = [
+  'checked',
+  'enabled',
+  'icons',
+  'parentId',
+  'title',
+  'type',
+  'visible'
+];
 
 const mItemsById = {
   'context_reloadTab': {
@@ -37,10 +54,12 @@ const mItemsById = {
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_reload_label_multiselected')
   },
   'context_topLevel_reloadTree': {
-    title:              browser.i18n.getMessage('context_reloadTree_label')
+    title:              browser.i18n.getMessage('context_reloadTree_label'),
+    titleMultiselected: browser.i18n.getMessage('context_reloadTree_label_multiselected')
   },
   'context_topLevel_reloadDescendants': {
-    title:              browser.i18n.getMessage('context_reloadDescendants_label')
+    title:              browser.i18n.getMessage('context_reloadDescendants_label'),
+    titleMultiselected: browser.i18n.getMessage('context_reloadDescendants_label_multiselected')
   },
   'context_toggleMuteTab-mute': {
     title:              browser.i18n.getMessage('tabContextMenu_mute_label'),
@@ -73,7 +92,8 @@ const mItemsById = {
     titleMultiselected: browser.i18n.getMessage('tabContextMenu_bookmark_label_multiselected')
   },
   'context_topLevel_bookmarkTree': {
-    title: browser.i18n.getMessage('context_bookmarkTree_label')
+    title:              browser.i18n.getMessage('context_bookmarkTree_label'),
+    titleMultiselected: browser.i18n.getMessage('context_bookmarkTree_label_multiselected')
   },
   'context_reopenInContainer': {
     title: browser.i18n.getMessage('tabContextMenu_reopenInContainer_label')
@@ -97,23 +117,24 @@ const mItemsById = {
   'context_separator:afterSendTab': {
     type: 'separator'
   },
-  'context_bookmarkAllTabs': {
-    title: browser.i18n.getMessage('tabContextMenu_bookmarkAll_label')
-  },
-  'context_reloadAllTabs': {
-    title: browser.i18n.getMessage('tabContextMenu_reloadAll_label')
-  },
-  'context_separator:afterReloadAll': {
-    type: 'separator'
-  },
   'context_topLevel_collapseTree': {
-    title: browser.i18n.getMessage('context_collapseTree_label')
+    title:              browser.i18n.getMessage('context_collapseTree_label'),
+    titleMultiselected: browser.i18n.getMessage('context_collapseTree_label_multiselected')
+  },
+  'context_topLevel_collapseTreeRecursively': {
+    title:              browser.i18n.getMessage('context_collapseTreeRecursively_label'),
+    titleMultiselected: browser.i18n.getMessage('context_collapseTreeRecursively_label_multiselected')
   },
   'context_topLevel_collapseAll': {
     title: browser.i18n.getMessage('context_collapseAll_label')
   },
   'context_topLevel_expandTree': {
-    title: browser.i18n.getMessage('context_expandTree_label')
+    title:              browser.i18n.getMessage('context_expandTree_label'),
+    titleMultiselected: browser.i18n.getMessage('context_expandTree_label_multiselected')
+  },
+  'context_topLevel_expandTreeRecursively': {
+    title:              browser.i18n.getMessage('context_expandTreeRecursively_label'),
+    titleMultiselected: browser.i18n.getMessage('context_expandTreeRecursively_label_multiselected')
   },
   'context_topLevel_expandAll': {
     title: browser.i18n.getMessage('context_expandAll_label')
@@ -121,23 +142,33 @@ const mItemsById = {
   'context_separator:afterCollapseExpand': {
     type: 'separator'
   },
+  'context_closeMultipleTabs': {
+    title: browser.i18n.getMessage('tabContextMenu_closeMultipleTabs_label')
+  },
   'context_closeTabsToTheEnd': {
+    parentId: 'context_closeMultipleTabs',
     title: browser.i18n.getMessage('tabContextMenu_closeTabsToBottom_label')
   },
   'context_closeOtherTabs': {
+    parentId: 'context_closeMultipleTabs',
     title: browser.i18n.getMessage('tabContextMenu_closeOther_label')
   },
   'context_topLevel_closeTree': {
-    title: browser.i18n.getMessage('context_closeTree_label')
+    title:              browser.i18n.getMessage('context_closeTree_label'),
+    titleMultiselected: browser.i18n.getMessage('context_closeTree_label_multiselected')
   },
   'context_topLevel_closeDescendants': {
-    title: browser.i18n.getMessage('context_closeDescendants_label')
+    title:              browser.i18n.getMessage('context_closeDescendants_label'),
+    titleMultiselected: browser.i18n.getMessage('context_closeDescendants_label_multiselected')
   },
   'context_topLevel_closeOthers': {
-    title: browser.i18n.getMessage('context_closeOthers_label')
+    title:              browser.i18n.getMessage('context_closeOthers_label'),
+    titleMultiselected: browser.i18n.getMessage('context_closeOthers_label_multiselected')
   },
   'context_undoCloseTab': {
-    title: browser.i18n.getMessage('tabContextMenu_undoClose_label')
+    title: browser.i18n.getMessage('tabContextMenu_undoClose_label'),
+    titleRegular: browser.i18n.getMessage('tabContextMenu_undoClose_label'),
+    titleMultipleTabsRestorable: browser.i18n.getMessage('tabContextMenu_undoClose_label_multiple')
   },
   'context_closeTab': {
     title:              browser.i18n.getMessage('tabContextMenu_close_label'),
@@ -171,11 +202,7 @@ const mItemsById = {
 
 const mExtraItems = new Map();
 
-// Imitation native context menu items depend on https://bugzilla.mozilla.org/show_bug.cgi?id=1280347
-const mNativeContextMenuAvailable = typeof browser.menus.overrideContext == 'function';
-let mNativeMultiselectionAvailable = true;
-
-const SIDEBAR_URL_PATTERN = mNativeContextMenuAvailable ? [`moz-extension://${location.host}/*`] : null;
+const SIDEBAR_URL_PATTERN = [`${Constants.kSHORTHAND_URIS.tabbar}*`];
 
 function getItemPlacementSignature(item) {
   if (item.placementSignature)
@@ -186,17 +213,13 @@ function getItemPlacementSignature(item) {
 }
 export async function init() {
   browser.runtime.onMessage.addListener(onMessage);
-  browser.runtime.onMessageExternal.addListener(onExternalMessage);
+  TSTAPI.onMessageExternal.addListener(onMessageExternal);
 
   window.addEventListener('unload', () => {
     browser.runtime.onMessage.removeListener(onMessage);
-    browser.runtime.onMessageExternal.removeListener(onExternalMessage);
+    TSTAPI.onMessageExternal.removeListener(onMessageExternal);
   }, { once: true });
 
-  browser.runtime.getBrowserInfo().then(browserInfo => {
-    mNativeMultiselectionAvailable = parseInt(browserInfo.version.split('.')[0]) >= 63;
-  }).catch(ApiTabs.createErrorSuppressor());
-
   const itemIds = Object.keys(mItemsById);
   for (const id of itemIds) {
     const item = mItemsById[id];
@@ -242,23 +265,22 @@ export async function init() {
       title:    item.title,
       type:     item.type || 'normal',
       contexts: ['tab'],
-      viewTypes: ['sidebar'],
+      viewTypes: ['sidebar', 'tab', 'popup'],
       visible:  false, // hide all by default
       documentUrlPatterns: SIDEBAR_URL_PATTERN
     };
     if (item.parentId)
       info.parentId = item.parentId;
-    if (mNativeContextMenuAvailable && !item.fakeMenu)
+    if (!item.fakeMenu)
       browser.menus.create(info);
-    onExternalMessage({
+    onMessageExternal({
       type: TSTAPI.kCONTEXT_MENU_CREATE,
       params: info
     }, browser.runtime);
   }
-  if (mNativeContextMenuAvailable) {
-    browser.menus.onShown.addListener(onShown);
-    browser.menus.onClicked.addListener(onClick);
-  }
+  browser.menus.onShown.addListener(onShown);
+  browser.menus.onHidden.addListener(onHidden);
+  browser.menus.onClicked.addListener(onClick);
   onTSTItemClick.addListener(onClick);
 
   await ContextualIdentities.init();
@@ -268,15 +290,19 @@ export async function init() {
   });
 }
 
+let mMultipleTabsRestorable = false;
+Tab.onChangeMultipleTabsRestorability.addListener(multipleTabsRestorable => {
+  mMultipleTabsRestorable = multipleTabsRestorable;
+});
+
 const mContextualIdentityItems = new Set();
 function updateContextualIdentities() {
   for (const item of mContextualIdentityItems.values()) {
     const id = item.id;
     if (id in mItemsById)
       delete mItemsById[id];
-    if (mNativeContextMenuAvailable)
-      browser.menus.remove(id).catch(ApiTabs.createErrorSuppressor());
-    onExternalMessage({
+    browser.menus.remove(id).catch(ApiTabs.createErrorSuppressor());
+    onMessageExternal({
       type: TSTAPI.kCONTEXT_MENU_REMOVE,
       params: id
     }, browser.runtime);
@@ -288,12 +314,11 @@ function updateContextualIdentities() {
     id:        'context_reopenInContainer:firefox-default',
     title:     browser.i18n.getMessage('tabContextMenu_reopenInContainer_noContainer_label'),
     contexts:  ['tab'],
-    viewTypes: ['sidebar'],
+    viewTypes: ['sidebar', 'tab', 'popup'],
     documentUrlPatterns: SIDEBAR_URL_PATTERN
   };
-  if (mNativeContextMenuAvailable)
-    browser.menus.create(defaultItem);
-  onExternalMessage({
+  browser.menus.create(defaultItem);
+  onMessageExternal({
     type: TSTAPI.kCONTEXT_MENU_CREATE,
     params: defaultItem
   }, browser.runtime);
@@ -304,12 +329,11 @@ function updateContextualIdentities() {
     id:        'context_reopenInContainer_separator',
     type:      'separator',
     contexts:  ['tab'],
-    viewTypes: ['sidebar'],
+    viewTypes: ['sidebar', 'tab', 'popup'],
     documentUrlPatterns: SIDEBAR_URL_PATTERN
   };
-  if (mNativeContextMenuAvailable)
-    browser.menus.create(defaultSeparator);
-  onExternalMessage({
+  browser.menus.create(defaultSeparator);
+  onMessageExternal({
     type: TSTAPI.kCONTEXT_MENU_CREATE,
     params: defaultSeparator
   }, browser.runtime);
@@ -322,14 +346,13 @@ function updateContextualIdentities() {
       id:       id,
       title:    identity.name.replace(/^([a-z0-9])/i, '&$1'),
       contexts: ['tab'],
-      viewTypes: ['sidebar'],
+      viewTypes: ['sidebar', 'tab', 'popup'],
       documentUrlPatterns: SIDEBAR_URL_PATTERN
     };
     if (identity.iconUrl)
       item.icons = { 16: identity.iconUrl };
-    if (mNativeContextMenuAvailable)
-      browser.menus.create(item);
-    onExternalMessage({
+    browser.menus.create(item);
+    onMessageExternal({
       type: TSTAPI.kCONTEXT_MENU_CREATE,
       params: item
     }, browser.runtime);
@@ -350,7 +373,9 @@ function updateItem(id, state = {}) {
     visible: 'visible' in state ? !!state.visible : true,
     enabled: 'enabled' in state ? !!state.enabled : true
   };
-  const title = state.multiselected ? item.titleMultiselected || item.title : item.title;
+  if ('checked' in state)
+    updateInfo.checked = state.checked;
+  const title = state.title || (state.multiselected && item.titleMultiselected) || item.title;
   if (title) {
     updateInfo.title = title;
     modified = title != item.lastTitle;
@@ -361,9 +386,8 @@ function updateItem(id, state = {}) {
                  updateInfo.enabled != item.lastEnabled;
   item.lastVisible = updateInfo.visible;
   item.lastEnabled = updateInfo.enabled;
-  if (mNativeContextMenuAvailable)
-    browser.menus.update(id, updateInfo).catch(ApiTabs.createErrorSuppressor());
-  onExternalMessage({
+  browser.menus.update(id, updateInfo).catch(ApiTabs.createErrorSuppressor());
+  onMessageExternal({
     type: TSTAPI.kCONTEXT_MENU_UPDATE,
     params: [id, updateInfo]
   }, browser.runtime);
@@ -389,6 +413,8 @@ function hasVisiblePrecedingItem(separator) {
   );
 }
 
+let mOverriddenContext = null;
+
 async function onShown(info, contextTab) {
   contextTab = contextTab && Tab.get(contextTab.id);
   const windowId              = contextTab ? contextTab.windowId : (await browser.windows.getLastFocused({}).catch(ApiTabs.createErrorHandler())).id;
@@ -399,8 +425,17 @@ async function onShown(info, contextTab) {
   const hasMultipleTabs       = Tab.hasMultipleTabs(windowId);
   const hasMultipleNormalTabs = Tab.hasMultipleTabs(windowId, { normal: true });
   const multiselected         = contextTab && contextTab.$TST.multiselected;
+  const contextTabs           = multiselected ?
+    Tab.getSelectedTabs(windowId) :
+    contextTab ?
+      [contextTab] :
+      [];
+  const hasChild              = contextTab && contextTabs.some(tab => tab.$TST.hasChild);
 
-  let modifiedItemsCount = 0;
+  if (mOverriddenContext)
+    return onOverriddenMenuShown(info, contextTab, windowId);
+
+  let modifiedItemsCount = cleanupOverriddenMenu();
 
   // ESLint reports "short circuit" error for following codes.
   //   https://eslint.org/docs/rules/no-unused-expressions#allowshortcircuit
@@ -414,11 +449,13 @@ async function onShown(info, contextTab) {
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_reloadTree', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_reloadTree
+    visible: emulate && contextTab && configs.context_topLevel_reloadTree,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_reloadDescendants', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_reloadDescendants,
-    enabled: contextTab && contextTab.$TST.hasChild
+    visible: emulate && contextTab && configs.context_topLevel_reloadDescendants,
+    enabled: hasChild,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_toggleMuteTab-mute', {
     visible: emulate && contextTab && (!contextTab.mutedInfo || !contextTab.mutedInfo.muted),
@@ -442,7 +479,7 @@ async function onShown(info, contextTab) {
   }) && modifiedItemsCount++;
 
   updateItem('context_selectAllTabs', {
-    visible: mNativeContextMenuAvailable && emulate && contextTab,
+    visible: emulate && contextTab,
     enabled: contextTab && Tab.getSelectedTabs(windowId).length != Tab.getVisibleTabs(windowId).length,
     multiselected
   }) && modifiedItemsCount++;
@@ -451,7 +488,8 @@ async function onShown(info, contextTab) {
     multiselected: multiselected || !contextTab
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_bookmarkTree', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_bookmarkTree
+    visible: emulate && contextTab && configs.context_topLevel_bookmarkTree,
+    multiselected
   }) && modifiedItemsCount++;
 
   let showContextualIdentities = false;
@@ -469,7 +507,7 @@ async function onShown(info, contextTab) {
       showContextualIdentities = true;
   }
   updateItem('context_reopenInContainer', {
-    visible: emulate && contextTab && showContextualIdentities,
+    visible: emulate && contextTab && showContextualIdentities && !contextTab.incognito,
     multiselected
   }) && modifiedItemsCount++;
 
@@ -491,32 +529,41 @@ async function onShown(info, contextTab) {
     multiselected
   }) && modifiedItemsCount++;
 
-  // workaround for https://github.com/piroor/treestyletab/issues/2056
-  updateItem('context_bookmarkAllTabs', {
-    visible: emulate && !mNativeMultiselectionAvailable
-  }) && modifiedItemsCount++;
-  updateItem('context_reloadAllTabs', {
-    visible: emulate && !mNativeMultiselectionAvailable
-  }) && modifiedItemsCount++;
-
   updateItem('context_topLevel_collapseTree', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_collapseTree,
-    enabled: contextTab && contextTab.$TST.hasChild
+    visible: emulate && contextTab && configs.context_topLevel_collapseTree,
+    enabled: hasChild,
+    multiselected
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_collapseTreeRecursively', {
+    visible: emulate && contextTab && configs.context_topLevel_collapseTreeRecursively,
+    enabled: hasChild,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_collapseAll', {
     visible: emulate && !multiselected && contextTab && configs.context_topLevel_collapseAll
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_expandTree', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_expandTree,
-    enabled: contextTab && contextTab.$TST.hasChild
+    visible: emulate && contextTab && configs.context_topLevel_expandTree,
+    enabled: hasChild,
+    multiselected
+  }) && modifiedItemsCount++;
+  updateItem('context_topLevel_expandTreeRecursively', {
+    visible: emulate && contextTab && configs.context_topLevel_expandTreeRecursively,
+    enabled: hasChild,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_expandAll', {
     visible: emulate && !multiselected && contextTab && configs.context_topLevel_expandAll
   }) && modifiedItemsCount++;
 
+  updateItem('context_closeMultipleTabs', {
+    visible: emulate && contextTab,
+    enabled: hasMultipleNormalTabs,
+    multiselected
+  }) && modifiedItemsCount++;
   updateItem('context_closeTabsToTheEnd', {
     visible: emulate && contextTab,
-    enabled: hasMultipleNormalTabs && nextTab,
+    enabled: nextTab,
     multiselected
   }) && modifiedItemsCount++;
   updateItem('context_closeOtherTabs', {
@@ -526,17 +573,22 @@ async function onShown(info, contextTab) {
   }) && modifiedItemsCount++;
 
   updateItem('context_topLevel_closeTree', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeTree
+    visible: emulate && contextTab && configs.context_topLevel_closeTree,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_closeDescendants', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeDescendants,
-    enabled: contextTab && contextTab.$TST.hasChild
+    visible: emulate && contextTab && configs.context_topLevel_closeDescendants,
+    enabled: hasChild,
+    multiselected
   }) && modifiedItemsCount++;
   updateItem('context_topLevel_closeOthers', {
-    visible: emulate && !multiselected && contextTab && configs.context_topLevel_closeOthers
+    visible: emulate && contextTab && configs.context_topLevel_closeOthers,
+    multiselected
   }) && modifiedItemsCount++;
 
+  const undoCloseTabLabel = mItemsById.context_undoCloseTab[configs.undoMultipleTabsClose && mMultipleTabsRestorable ? 'titleMultipleTabsRestorable' : 'titleRegular'];
   updateItem('context_undoCloseTab', {
+    title:   undoCloseTabLabel,
     visible: emulate && contextTab,
     multiselected
   }) && modifiedItemsCount++;
@@ -549,13 +601,14 @@ async function onShown(info, contextTab) {
     visible: emulate && !contextTab
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_bookmarkSelected', {
-    visible: emulate && !contextTab && mNativeMultiselectionAvailable
+    visible: emulate && !contextTab
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_selectAllTabs', {
     visible: emulate && !contextTab,
     enabled: !contextTab && Tab.getSelectedTabs(windowId).length != Tab.getVisibleTabs(windowId).length
   }) && modifiedItemsCount++;
   updateItem('noContextTab:context_undoCloseTab', {
+    title:   undoCloseTabLabel,
     visible: emulate && !contextTab
   }) && modifiedItemsCount++;
 
@@ -565,7 +618,6 @@ async function onShown(info, contextTab) {
 
   updateSeparator('context_separator:afterDuplicate') && modifiedItemsCount++;
   updateSeparator('context_separator:afterSendTab') && modifiedItemsCount++;
-  updateSeparator('context_separator:afterReloadAll') && modifiedItemsCount++;
   updateSeparator('context_separator:afterCollapseExpand') && modifiedItemsCount++;
   updateSeparator('context_separator:afterClose') && modifiedItemsCount++;
 
@@ -577,10 +629,127 @@ async function onShown(info, contextTab) {
 
   /* eslint-enable no-unused-expressions */
 
-  if (mNativeContextMenuAvailable && modifiedItemsCount)
+  if (modifiedItemsCount > 0)
     browser.menus.refresh().catch(ApiTabs.createErrorSuppressor());
 }
 
+let mLastOverriddenContextOwner = null;
+
+function onOverriddenMenuShown(info, contextTab, windowId) {
+  if (!mLastOverriddenContextOwner) {
+    for (const itemId of Object.keys(mItemsById)) {
+      if (!mItemsById[itemId].lastVisible)
+        continue;
+      mItemsById[itemId].lastVisible = false;
+      browser.menus.update(itemId, { visible: false });
+    }
+    mLastOverriddenContextOwner = mOverriddenContext.owner;
+  }
+
+  for (const item of (mExtraItems.get(mLastOverriddenContextOwner) || [])) {
+    if (item.$topLevel &&
+        item.lastVisible) {
+      browser.menus.update(
+        getExternalTopLevelItemId(mOverriddenContext.owner, item.id),
+        { visible: true }
+      );
+    }
+  }
+
+  const message = {
+    type: TSTAPI.kCONTEXT_MENU_SHOWN,
+    info: {
+      bookmarkId:    info.bookmarkId || null,
+      button:        info.button,
+      checked:       info.checked,
+      contexts:      contextTab ? ['tab'] : info.bookmarkId ? ['bookmark'] : [],
+      editable:      false,
+      frameId:       null,
+      frameUrl:      null,
+      linkText:      null,
+      linkUrl:       null,
+      mediaType:     null,
+      menuIds:       [],
+      menuItemId:    null,
+      modifiers:     [],
+      pageUrl:       null,
+      parentMenuItemId: null,
+      selectionText: null,
+      srcUrl:        null,
+      targetElementId: null,
+      viewType:      'sidebar',
+      wasChecked:    false
+    },
+    tab: contextTab && new TSTAPI.TreeItem(contextTab, { isContextTab: true }) || null,
+    windowId
+  }
+  TSTAPI.sendMessage(message, {
+    targets: [mOverriddenContext.owner],
+    tabProperties: ['tab']
+  });
+  TSTAPI.sendMessage({
+    ...message,
+    type: TSTAPI.kFAKE_CONTEXT_MENU_SHOWN
+  }, {
+    targets: [mOverriddenContext.owner],
+    tabProperties: ['tab']
+  });
+
+  reserveRefresh();
+}
+
+function cleanupOverriddenMenu() {
+  if (!mLastOverriddenContextOwner)
+    return 0;
+
+  let modifiedItemsCount = 0;
+
+  const owner = mLastOverriddenContextOwner;
+  mLastOverriddenContextOwner = null;
+
+  for (const itemId of Object.keys(mItemsById)) {
+    if (!mItemsById[itemId].lastVisible)
+      continue;
+    mItemsById[itemId].lastVisible = true;
+    browser.menus.update(itemId, { visible: true });
+    modifiedItemsCount++;
+  }
+
+  for (const item of (mExtraItems.get(owner) || [])) {
+    if (item.$topLevel &&
+        item.lastVisible) {
+      browser.menus.update(
+        getExternalTopLevelItemId(owner, item.id),
+        { visible: false }
+      );
+      modifiedItemsCount++;
+    }
+  }
+
+  return modifiedItemsCount;
+}
+
+function onHidden() {
+  const owner = mOverriddenContext && mOverriddenContext.owner;
+  const windowId = mOverriddenContext && mOverriddenContext.windowId;
+  if (mLastOverriddenContextOwner &&
+      owner == mLastOverriddenContextOwner) {
+    mOverriddenContext = null;
+    TSTAPI.sendMessage({
+      type: TSTAPI.kCONTEXT_MENU_HIDDEN,
+      windowId
+    }, {
+      targets: [owner]
+    });
+    TSTAPI.sendMessage({
+      type: TSTAPI.kFAKE_CONTEXT_MENU_HIDDEN,
+      windowId
+    }, {
+      targets: [owner]
+    });
+  }
+}
+
 async function onClick(info, contextTab) {
   contextTab = contextTab && Tab.get(contextTab.id);
   const window    = await browser.windows.getLastFocused({ populate: true }).catch(ApiTabs.createErrorHandler());
@@ -673,7 +842,8 @@ async function onClick(info, contextTab) {
       browser.tabs.highlight({
         windowId,
         populate: false,
-        tabs:     [activeTab.index].concat(tabs.filter(tab => !tab.active).map(tab => tab.index))
+        tabs:     [activeTab.index].concat(mapAndFilter(tabs,
+                                                        tab => !tab.active ? tab.index : undefined))
       }).catch(ApiTabs.createErrorSuppressor());
     }; break;
     case 'context_bookmarkTab':
@@ -682,16 +852,6 @@ async function onClick(info, contextTab) {
     case 'context_bookmarkSelected':
       Commands.bookmarkTab(contextTab || activeTab);
       break;
-    case 'context_bookmarkAllTabs':
-      Commands.bookmarkTabs(Tab.getTabs(contextTab.windowId));
-      break;
-    case 'context_reloadAllTabs': {
-      const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
-      for (const tab of tabs) {
-        browser.tabs.reload(tab.id)
-          .catch(ApiTabs.createErrorHandler(ApiTabs.handleMissingTabError));
-      }
-    }; break;
     case 'context_closeTabsToTheEnd': {
       const tabs = await browser.tabs.query({ windowId }).catch(ApiTabs.createErrorHandler());
       let after = false;
@@ -725,7 +885,8 @@ async function onClick(info, contextTab) {
           multiselectedTabs.map(tab => tab.id) :
           [contextTab.id]
       );
-      const closeTabs = tabs.filter(tab => !tab.pinned && !keptTabIds.has(tab.id)).map(tab => Tab.get(tab.id));
+      const closeTabs = mapAndFilter(tabs,
+                                     tab => !tab.pinned && !keptTabIds.has(tab.id) && Tab.get(tab.id) || undefined);
       const canceled = (await browser.runtime.sendMessage({
         type: Constants.kCOMMAND_NOTIFY_TABS_CLOSING,
         tabs: closeTabs.map(tab => tab.$TST.sanitized),
@@ -759,6 +920,46 @@ async function onClick(info, contextTab) {
       if (contextTab &&
           contextualIdentityMatch)
         Commands.reopenInContainer(contextTab, contextualIdentityMatch[1]);
+
+      if (EXTERNAL_TOP_LEVEL_ITEM_MATCHER.test(info.menuItemId)) {
+        const owner      = RegExp.$1;
+        const menuItemId = RegExp.$2;
+        const tab = contextTab && (await (new TSTAPI.TreeItem(contextTab, { isContextTab: true })).exportFor      (owner)) || null;
+        const message = {
+          type: TSTAPI.kCONTEXT_MENU_CLICK,
+          info: {
+            bookmarkId:    info.bookmarkId || null,
+            button:        info.button,
+            checked:       info.checked,
+            editable:      false,
+            frameId:       null,
+            frameUrl:      null,
+            linkText:      null,
+            linkUrl:       null,
+            mediaType:     null,
+            menuItemId,
+            modifiers:     [],
+            pageUrl:       null,
+            parentMenuItemId: null,
+            selectionText: null,
+            srcUrl:        null,
+            targetElementId: null,
+            viewType:      'sidebar',
+            wasChecked:    info.wasChecked
+          },
+          tab
+        };
+        if (owner == browser.runtime.id) {
+          browser.runtime.sendMessage(message).catch(ApiTabs.createErrorSuppressor());
+        }
+        else if (TSTAPI.isSafeAtIncognito(owner, { tab: contextTab, windowId: TabsStore.getCurrentWindowId() })) {
+          browser.runtime.sendMessage(owner, message).catch(ApiTabs.createErrorSuppressor());
+          browser.runtime.sendMessage(owner, {
+            ...message,
+            type: TSTAPI.kFAKE_CONTEXT_MENU_CLICK
+          }).catch(ApiTabs.createErrorSuppressor());
+        }
+      }
     }; break;
   }
 }
@@ -783,7 +984,7 @@ function exportExtraItems() {
 
 async function notifyUpdated() {
   await browser.runtime.sendMessage({
-    type:  TSTAPI.kCONTEXT_MENU_UPDATED,
+    type:  Constants.kCOMMAND_NOTIFY_CONTEXT_MENU_UPDATED,
     items: exportExtraItems()
   }).catch(ApiTabs.createErrorSuppressor());
 }
@@ -808,10 +1009,19 @@ function reserveNotifyUpdated() {
   });
 }
 
+function reserveRefresh() {
+  if (reserveRefresh.reserved)
+    clearTimeout(reserveRefresh.reserved);
+  reserveRefresh.reserved = setTimeout(() => {
+    reserveRefresh.reserved = null;;
+    browser.menus.refresh();
+  }, 0);
+}
+
 function onMessage(message, _sender) {
   log('tab-context-menu: internally called:', message);
   switch (message.type) {
-    case TSTAPI.kCONTEXT_MENU_GET_ITEMS:
+    case Constants.kCOMMAND_GET_CONTEXT_MENU_ITEMS:
       return Promise.resolve(exportExtraItems());
 
     case TSTAPI.kCONTEXT_MENU_CLICK:
@@ -827,7 +1037,7 @@ function onMessage(message, _sender) {
       onTSTTabContextMenuHidden.dispatch();
       return;
 
-    case TSTAPI.kCONTEXT_ITEM_CHECKED_STATUS_CHANGED:
+    case Constants.kCOMMAND_NOTIFY_CONTEXT_ITEM_CHECKED_STATUS_CHANGED:
       for (const itemData of mExtraItems.get(message.ownerId)) {
         if (!itemData.id != message.id)
           continue;
@@ -835,15 +1045,22 @@ function onMessage(message, _sender) {
         break;
       }
       return;
+
+    case Constants.kCOMMAND_NOTIFY_CONTEXT_OVERRIDDEN:
+      mOverriddenContext = message.context || null;
+      if (mOverriddenContext) {
+        mOverriddenContext.owner = message.owner;
+        mOverriddenContext.windowId = message.windowId;
+      }
+      break;
   }
 }
 
-export function onExternalMessage(message, sender) {
-  if (!message)
-    return;
-  log('API called:', message, { id: sender.id, url: sender.url });
+export function onMessageExternal(message, sender) {
   switch (message.type) {
-    case TSTAPI.kCONTEXT_MENU_CREATE: {
+    case TSTAPI.kCONTEXT_MENU_CREATE:
+    case TSTAPI.kFAKE_CONTEXT_MENU_CREATE: {
+      log('TSTAPI.kCONTEXT_MENU_CREATE:', message, { id: sender.id, url: sender.url });
       const items  = getItemsFor(sender.id);
       let params = message.params;
       if (Array.isArray(params))
@@ -870,23 +1087,83 @@ export function onExternalMessage(message, sender) {
         }
       }
       mExtraItems.set(sender.id, items);
+      params.$topLevel = (
+        Array.isArray(params.viewTypes) &&
+        params.viewTypes.includes('sidebar')
+      );
+      if (sender.id != browser.runtime.id &&
+          params.$topLevel) {
+        params.lastVisible = params.visible !== false;
+        const visible = !!(
+          params.lastVisible &&
+          mOverriddenContext &&
+          mLastOverriddenContextOwner == sender.id
+        );
+        const createParams = {
+          id:        getExternalTopLevelItemId(sender.id, params.id),
+          type:      params.type || 'normal',
+          visible,
+          viewTypes: ['sidebar', 'tab', 'popup'],
+          contexts:  (params.contexts || []).filter(context => context == 'tab' || context == 'bookmark'),
+          documentUrlPatterns: SIDEBAR_URL_PATTERN
+        };
+        if (params.parentId)
+          createParams.parentId = getExternalTopLevelItemId(sender.id, params.parentId);
+        for (const property of SAFE_MENU_PROPERTIES) {
+          if (property in params)
+            createParams[property] = params[property];
+        }
+        browser.menus.create(createParams);
+        reserveRefresh();
+        onTopLevelItemAdded.dispatch();
+      }
       return reserveNotifyUpdated();
     }; break;
 
-    case TSTAPI.kCONTEXT_MENU_UPDATE: {
+    case TSTAPI.kCONTEXT_MENU_UPDATE:
+    case TSTAPI.kFAKE_CONTEXT_MENU_UPDATE: {
+      log('TSTAPI.kCONTEXT_MENU_UPDATE:', message, { id: sender.id, url: sender.url });
       const items = getItemsFor(sender.id);
       for (let i = 0, maxi = items.length; i < maxi; i++) {
         const item = items[i];
         if (item.id != message.params[0])
           continue;
-        items.splice(i, 1, Object.assign({}, item, message.params[1]));
+        const params = message.params[1];
+        const updateProperties = {};
+        for (const property of SAFE_MENU_PROPERTIES) {
+          if (property in params)
+            updateProperties[property] = params[property];
+        }
+        if (sender.id != browser.runtime.id &&
+            item.$topLevel) {
+          if ('visible' in updateProperties)
+            item.lastVisible = updateProperties.visible;
+          if (!mOverriddenContext ||
+              mLastOverriddenContextOwner != sender.id)
+            delete updateProperties.visible;
+        }
+        items.splice(i, 1, {
+          ...item,
+          ...updateProperties
+        });
+        if (sender.id != browser.runtime.id &&
+            item.$topLevel &&
+            Object.keys(updateProperties).length > 0) {
+          browser.menus.update(
+            getExternalTopLevelItemId(sender.id, item.id),
+            updateProperties
+          );
+          reserveRefresh()
+        }
         break;
       }
       mExtraItems.set(sender.id, items);
       return reserveNotifyUpdated();
     }; break;
 
-    case TSTAPI.kCONTEXT_MENU_REMOVE: {
+    case TSTAPI.kCONTEXT_MENU_REMOVE:
+    case TSTAPI.kFAKE_CONTEXT_MENU_REMOVE: {
+      log('TSTAPI.kCONTEXT_MENU_REMOVE:', message, { id: sender.id, url: sender.url });
       let items = getItemsFor(sender.id);
       let id    = message.params;
       if (Array.isArray(id))
@@ -901,13 +1178,19 @@ export function onExternalMessage(message, sender) {
         parent.children = parent.children.filter(childId => childId != id);
       if (item.children) {
         for (const childId of item.children) {
-          onExternalMessage({ type: message.type, params: childId }, sender);
+          onMessageExternal({ type: message.type, params: childId }, sender);
         }
       }
+      if (sender.id != browser.runtime.id &&
+          item.$topLevel) {
+        browser.menus.remove(getExternalTopLevelItemId(sender.id, item.id));
+        reserveRefresh();
+      }
       return reserveNotifyUpdated();
     }; break;
 
     case TSTAPI.kCONTEXT_MENU_REMOVE_ALL:
+    case TSTAPI.kFAKE_CONTEXT_MENU_REMOVE_ALL:
     case TSTAPI.kUNREGISTER_SELF: {
       delete mExtraItems.delete(sender.id);
       return reserveNotifyUpdated();
