@@ -10,12 +10,14 @@ import MenuUI from '/extlib/MenuUI.js';
 import {
   log as internalLogger,
   wait,
+  notify,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
 import * as ApiTabs from '/common/api-tabs.js';
 import * as TabsStore from '/common/tabs-store.js';
 import * as TSTAPI from '/common/tst-api.js';
+import * as Permissions from '/common/permissions.js';
 import * as EventUtils from './event-utils.js';
 import * as BackgroundConnection from './background-connection.js';
 
@@ -54,10 +56,10 @@ export function init() {
   });
 
   browser.runtime.onMessage.addListener(onMessage);
-  browser.runtime.onMessageExternal.addListener(onExternalMessage);
+  TSTAPI.onMessageExternal.addListener(onMessageExternal);
 
   browser.runtime.sendMessage({
-    type: TSTAPI.kCONTEXT_MENU_GET_ITEMS
+    type: Constants.kCOMMAND_GET_CONTEXT_MENU_ITEMS
   }).then(items => {
     importExtraItems(items);
     mIsDirty = true;
@@ -83,9 +85,11 @@ async function rebuild() {
     return;
 
   const extraItemNodes = document.createDocumentFragment();
-  const incognitoParams = { windowId: TabsStore.getWindow() };
+  const incognitoParams = { windowId: TabsStore.getCurrentWindowId() };
   for (const [id, extraItems] of mExtraItems.entries()) {
-    if (!TSTAPI.isSafeAtIncognito(id, incognitoParams))
+    const addon = TSTAPI.getAddon(id);
+    if (!TSTAPI.isSafeAtIncognito(id, incognitoParams) ||
+        !addon)
       continue;
     let addonItem = document.createElement('li');
     const name = getAddonName(id);
@@ -116,7 +120,7 @@ async function rebuild() {
     const topLevelItems = toBeBuiltItems.filter(item => !item.parentId);
     if (topLevelItems.length == 1 &&
         !topLevelItems[0].icons)
-      topLevelItems[0].icons = TSTAPI.getAddon(id).icons || {};
+      topLevelItems[0].icons = addon.icons || {};
 
     const addonSubMenu = addonItem.lastChild;
     const knownItems   = {};
@@ -342,6 +346,7 @@ async function onCommand(item, event) {
       linkUrl:          null,
       mediaType:        null,
       menuItemId:       id,
+      button:           event.button,
       modifiers:        modifiers,
       pageUrl:          null,
       parentMenuItemId: null,
@@ -351,10 +356,18 @@ async function onCommand(item, event) {
     },
     tab
   };
-  if (owner == browser.runtime.id)
+  if (owner == browser.runtime.id) {
     await browser.runtime.sendMessage(message).catch(ApiTabs.createErrorSuppressor());
-  else if (TSTAPI.isSafeAtIncognito(owner, { tab: contextTab, windowId: TabsStore.getWindow() }))
-    await browser.runtime.sendMessage(owner, message).catch(ApiTabs.createErrorSuppressor());
+  }
+  else if (TSTAPI.isSafeAtIncognito(owner, { tab: contextTab, windowId: TabsStore.getCurrentWindowId() })) {
+    await Promise.all([
+      browser.runtime.sendMessage(owner, message).catch(ApiTabs.createErrorSuppressor()),
+      browser.runtime.sendMessage(owner, {
+        ...message,
+        type: TSTAPI.kFAKE_CONTEXT_MENU_CLICK
+      }).catch(ApiTabs.createErrorSuppressor())
+    ]);
+  }
 
   if (item.matches('.checkbox')) {
     item.classList.toggle('checked');
@@ -363,7 +376,7 @@ async function onCommand(item, event) {
         continue;
       itemData.checked = item.matches('.checked');
       browser.runtime.sendMessage({
-        type:    TSTAPI.kCONTEXT_ITEM_CHECKED_STATUS_CHANGED,
+        type:    Constants.kCOMMAND_NOTIFY_CONTEXT_ITEM_CHECKED_STATUS_CHANGED,
         id:      item.dataset.itemId,
         ownerId: item.dataset.itemOwnerId,
         checked: itemData.checked
@@ -395,7 +408,7 @@ async function onCommand(item, event) {
         if (radioItem)
           radioItem.classList.toggle('checked', itemData.checked);
         browser.runtime.sendMessage({
-          type:    TSTAPI.kCONTEXT_ITEM_CHECKED_STATUS_CHANGED,
+          type:    Constants.kCOMMAND_NOTIFY_CONTEXT_ITEM_CHECKED_STATUS_CHANGED,
           id:      item.dataset.itemId,
           ownerId: item.dataset.itemOwnerId,
           checked: itemData.checked
@@ -420,27 +433,37 @@ async function onShown(contextTab) {
       srcUrl:           null,
       contexts:         ['tab'],
       menuIds:          [],
-      viewType:         'sidebar'
+      viewType:         'sidebar',
+      bookmarkId:       null
     },
     tab: contextTab && new TSTAPI.TreeItem(contextTab, { isContextTab: true }) || null,
-    windowId: TabsStore.getWindow()
+    windowId: TabsStore.getCurrentWindowId()
   };
   return Promise.all([
-    browser.runtime.sendMessage(Object.assign({}, message, {
+    browser.runtime.sendMessage({
+      ...message,
       tab: message.tab && await message.tab.exportFor(browser.runtime.id)
-    })).catch(ApiTabs.createErrorSuppressor()),
-    TSTAPI.sendMessage(message, { tabProperties: ['tab'] })
+    }).catch(ApiTabs.createErrorSuppressor()),
+    TSTAPI.sendMessage(message, { tabProperties: ['tab'] }),
+    TSTAPI.sendMessage({
+      ...message,
+      type: TSTAPI.kFAKE_CONTEXT_MENU_SHOWN
+    }, { tabProperties: ['tab'] })
   ]);
 }
 
 async function onHidden() {
   const message = {
     type: TSTAPI.kCONTEXT_MENU_HIDDEN,
-    windowId: TabsStore.getWindow()
+    windowId: TabsStore.getCurrentWindowId()
   };
   return Promise.all([
     browser.runtime.sendMessage(message).catch(ApiTabs.createErrorSuppressor()),
-    TSTAPI.sendMessage(message)
+    TSTAPI.sendMessage(message),
+    TSTAPI.sendMessage({
+      ...message,
+      type: TSTAPI.kFAKE_CONTEXT_MENU_HIDDEN
+    })
   ]);
 }
 
@@ -450,11 +473,11 @@ function onMessage(message, _sender) {
     case Constants.kCOMMAND_NOTIFY_TABS_CLOSING:
       // Don't respond to message for other windows, because
       // the sender receives only the firstmost response.
-      if (message.windowId != TabsStore.getWindow())
+      if (message.windowId != TabsStore.getCurrentWindowId())
         return;
       return Promise.resolve(onTabsClosing.dispatch(message.tabs));
 
-    case TSTAPI.kCONTEXT_MENU_UPDATED: {
+    case Constants.kCOMMAND_NOTIFY_CONTEXT_MENU_UPDATED: {
       importExtraItems(message.items);
       mIsDirty = true;
       if (mUI.opened)
@@ -470,14 +493,17 @@ function importExtraItems(importedItems) {
   }
 }
 
-function onExternalMessage(message, sender) {
-  log('API called:', message, { id: sender.id, url: sender.url });
+let mReservedOverrideContext = null;
+
+function onMessageExternal(message, sender) {
   switch (message.type) {
     case TSTAPI.kCONTEXT_MENU_OPEN:
+    case TSTAPI.kFAKE_CONTEXT_MENU_OPEN:
+      log('TSTAPI.kCONTEXT_MENU_OPEN:', message, { id: sender.id, url: sender.url });
       return (async () => {
         const tab      = message.tab ? Tab.get(message.tab) : null ;
         const windowId = message.window || tab && tab.windowId;
-        if (windowId != TabsStore.getWindow())
+        if (windowId != TabsStore.getCurrentWindowId())
           return;
         await onShown(tab);
         await wait(25);
@@ -487,16 +513,110 @@ function onExternalMessage(message, sender) {
           top:      message.top
         });
       })();
+
+    case TSTAPI.kOVERRIDE_CONTEXT:
+      if (message.windowId != TabsStore.getCurrentWindowId())
+        return;
+      mReservedOverrideContext = (
+        message.context == 'bookmark' ?
+          { context:    'bookmark',
+            bookmarkId: message.bookmarkId } :
+          message.context == 'tab' ?
+            { context:    'tab',
+              tabId:      message.tabId } :
+            null
+      );
+      if (mReservedOverrideContext) {
+        if (reserveToActivateSubpanel.reserved) {
+          clearTimeout(reserveToActivateSubpanel.reserved);
+          reserveToActivateSubpanel.reserved = null;
+        }
+        browser.runtime.sendMessage({
+          type:     Constants.kCOMMAND_NOTIFY_CONTEXT_OVERRIDDEN,
+          context:  mReservedOverrideContext,
+          windowId: message.windowId,
+          owner:    sender.id
+        });
+        // We need to ignore mouse events on the iframe, to handle
+        // the contextmenu event on this parent frame side.
+        document.getElementById('subpanel').style.pointerEvents = 'none';
+      }
+      break;
   }
 }
 
+function reserveToActivateSubpanel() {
+  if (reserveToActivateSubpanel.reserved)
+    clearTimeout(reserveToActivateSubpanel.reserved);
+  reserveToActivateSubpanel.reserved = setTimeout(() => {
+    reserveToActivateSubpanel.reserved = null;
+    document.getElementById('subpanel').style.pointerEvents = '';
+  }, 100);
+}
+reserveToActivateSubpanel.reserved = null;
+
+// safe guard
+window.addEventListener('mouseup', _event => {
+  reserveToActivateSubpanel();
+});
 
 async function onContextMenu(event) {
-  if (EventUtils.getElementTarget(event).closest('input, textarea'))
+  reserveToActivateSubpanel();
+
+  const context = mReservedOverrideContext;
+  mReservedOverrideContext = null;
+
+  const target         = EventUtils.getElementTarget(event);
+  const originalTarget = EventUtils.getElementOriginalTarget(event);
+  const onInputField   = (
+    target.closest('input, textarea') ||
+    originalTarget.closest('input, textarea')
+  );
+
+  if (!onInputField && context && context.context) {
+    try {
+      browser.menus.overrideContext(context);
+    }
+    catch(error) {
+      if (context.context == 'bookmark' &&
+          !(await Permissions.isGranted(Permissions.BOOKMARKS)))
+        notify({
+          title:   browser.i18n.getMessage('bookmarkContext_notification_notPermitted_title'),
+          message: browser.i18n.getMessage('bookmarkContext_notification_notPermitted_message'),
+          url:     `moz-extension://${location.host}/options/options.html#bookmarksPermissionGranted_context`
+        });
+      else
+        console.error(error);
+    }
+    return;
+  }
+
+  browser.runtime.sendMessage({
+    type:    Constants.kCOMMAND_NOTIFY_CONTEXT_OVERRIDDEN,
+    context: null
+  });
+
+  if (onInputField)
     return;
 
-  const tab = EventUtils.getTabFromEvent(event);
   const modifierKeyPressed = /^Mac/i.test(navigator.platform) ? event.metaKey : event.ctrlKey;
+
+  const originalTargetBookmarkElement = originalTarget && originalTarget.closest('[data-bookmark-id]');
+  const bookmarkId = originalTargetBookmarkElement && originalTargetBookmarkElement.dataset.bookmarkId;
+  if (bookmarkId &&
+      !modifierKeyPressed &&
+      typeof browser.menus.overrideContext == 'function') {
+    browser.menus.overrideContext({
+      context:    'bookmark',
+      bookmarkId: bookmarkId
+    });
+    return;
+  }
+
+  const originalTargetTabElement = originalTarget && originalTarget.closest('[data-tab-id]');
+  const tab = originalTargetTabElement ?
+    TabsStore.ensureLivingTab(Tab.get(parseInt(originalTargetTabElement.dataset.tabId))) :
+    EventUtils.getTabFromEvent(event);
   if (tab &&
       !modifierKeyPressed &&
       typeof browser.menus.overrideContext == 'function') {
@@ -506,8 +626,10 @@ async function onContextMenu(event) {
     });
     return;
   }
+
   if (!configs.emulateDefaultContextMenu)
     return;
+
   event.stopPropagation();
   event.preventDefault();
   await onShown(tab);
