@@ -19,11 +19,11 @@ function log(...args) {
 
 let mTargetWindow;
 
-export function setWindow(targetWindow) {
+export function setCurrentWindowId(targetWindow) {
   return mTargetWindow = targetWindow;
 }
 
-export function getWindow() {
+export function getCurrentWindowId() {
   return mTargetWindow;
 }
 
@@ -175,27 +175,32 @@ function matchedWithQuery(tab, query) {
     if (attribute in query &&
         !matched(tab[attribute], query[attribute]))
       return false;
-    if (`!${attribute}` in query &&
-        matched(tab[attribute], query[`!${attribute}`]))
+    const invertedAttribute = `!${attribute}`;
+    if (invertedAttribute in query &&
+        matched(tab[attribute], query[invertedAttribute]))
       return false;
   }
 
   if (!tab.$TST)
     return false;
 
-  if ('states' in query && tab.$TST.states) {
-    for (let i = 0, maxi = query.states.length; i < maxi; i += 2) {
-      const state   = query.states[i];
-      const pattern = query.states[i+1];
-      if (!matched(tab.$TST.states.has(state), pattern))
+  const tabStates = tab.$TST.states;
+  if ('states' in query && tabStates) {
+    const queryStates = query.states;
+    for (let i = 0, maxi = queryStates.length; i < maxi; i += 2) {
+      const state   = queryStates[i];
+      const pattern = queryStates[i+1];
+      if (!matched(tabStates.has(state), pattern))
         return false;
     }
   }
-  if ('attributes' in query && tab.$TST.attributes) {
-    for (let i = 0, maxi = query.attributes.length; i < maxi; i += 2) {
-      const attribute = query.attributes[i];
-      const pattern   = query.attributes[i+1];
-      if (!matched(tab.$TST.attributes[attribute], pattern))
+  const tabAttributes = tab.$TST.attributes;
+  if ('attributes' in query && tabAttributes) {
+    const queryAttributes = query.attributes;
+    for (let i = 0, maxi = queryAttributes.length; i < maxi; i += 2) {
+      const attribute = queryAttributes[i];
+      const pattern   = queryAttributes[i+1];
+      if (!matched(tabAttributes[attribute], pattern))
         return false;
     }
   }
@@ -205,18 +210,18 @@ function matchedWithQuery(tab, query) {
     return false;
   if (query.normal &&
       (tab.hidden ||
-       tab.$TST.states.has(Constants.kTAB_STATE_SHOWING) ||
+       tabStates.has(Constants.kTAB_STATE_SHOWING) ||
        tab.pinned))
     return false;
   if (query.visible &&
-      ((tab.$TST.states.has(Constants.kTAB_STATE_COLLAPSED) &&
-        !tab.$TST.states.has(Constants.kTAB_STATE_EXPANDING)) ||
+      ((tabStates.has(Constants.kTAB_STATE_COLLAPSED) &&
+        !tabStates.has(Constants.kTAB_STATE_EXPANDING)) ||
        tab.hidden ||
-       tab.$TST.states.has(Constants.kTAB_STATE_SHOWING)))
+       tabStates.has(Constants.kTAB_STATE_SHOWING)))
     return false;
   if (query.controllable &&
       (tab.hidden ||
-       tab.$TST.states.has(Constants.kTAB_STATE_SHOWING)))
+       tabStates.has(Constants.kTAB_STATE_SHOWING)))
     return false;
   if ('hasChild' in query &&
       query.hasChild != tab.$TST.hasChild)
@@ -224,6 +229,12 @@ function matchedWithQuery(tab, query) {
   if ('hasParent' in query &&
       query.hasParent != tab.$TST.hasParent)
     return false;
+  if ('childOf' in query &&
+      !tab.$TST.parentId != query.childOf)
+    return false;
+  if ('descendantOf' in query &&
+      !tab.$TST.ancestorIds.includes(query.descendantOf))
+    return false;
 
   return true;
 }
@@ -488,12 +499,16 @@ export function removeTabFromIndexes(tab) {
 }
 
 function addTabToIndex(tab, indexes) {
+  if (!tab)
+    throw new Error(`TabsStore.addTabToIndex gets non-tab parameter!: ${JSON.stringify(tab)} : ${new Error().stack}`);
   const tabs = indexes.get(tab.windowId);
   if (tabs)
     tabs.set(tab.id, tab);
 }
 
 function removeTabFromIndex(tab, indexes) {
+  if (!tab)
+    throw new Error(`TabsStore.removeTabFromIndex gets non-tab parameter!: ${JSON.stringify(tab)} : ${new Error().stack}`);
   const tabs = indexes.get(tab.windowId);
   if (tabs)
     tabs.delete(tab.id);
