@@ -1,7 +1,7 @@
 /*
-# This Source Code Form is subject to the terms of the Mozilla Public
-# License, v. 2.0. If a copy of the MPL was not distributed with this
-# file, You can obtain one at http://mozilla.org/MPL/2.0/.
+ license: The MIT License, Copyright (c) 2020 YUKI "Piro" Hiroshi
+ original:
+   https://github.com/piroor/treestyletab/blob/master/webextensions/tests/assert.js
 */
 'use strict';
 
