@@ -8,7 +8,9 @@
 import {
   log as internalLogger,
   dumpTab,
-  configs
+  toLines,
+  configs,
+  wait
 } from '/common/common.js';
 
 import * as Constants from '/common/constants.js';
@@ -18,10 +20,14 @@ import * as SidebarConnection from '/common/sidebar-connection.js';
 import * as MetricsData from '/common/metrics-data.js';
 import * as UserOperationBlocker from '/common/user-operation-blocker.js';
 import * as TreeBehavior from '/common/tree-behavior.js';
+import * as TabsInternalOperation from '/common/tabs-internal-operation.js';
 
 import Tab from '/common/Tab.js';
 
 import * as Tree from './tree.js';
+import * as TabsOpen from './tabs-open.js';
+import * as TabsMove from './tabs-move.js';
+import * as Commands from './commands.js';
 
 import EventListenerManager from '/extlib/EventListenerManager.js';
 
@@ -31,6 +37,9 @@ function log(...args) {
 
 export const onTabAttachedFromRestoredInfo = new EventListenerManager();
 
+let mRecentlyClosedTabs = [];
+let mRecentlyClosedTabsTreeStructure = [];
+
 export function startTracking() {
   Tab.onCreated.addListener((tab, _info) => { reserveToSaveTreeStructure(tab.windowId); });
   Tab.onRemoved.addListener((tab, info) => {
@@ -92,7 +101,8 @@ export async function loadTreeStructure(windows, restoredFromCacheResults) {
         uniqueIds = uniqueIds.map(id => id.id);
         let tabsOffset;
         if (structure[0].id) {
-          tabsOffset = uniqueIds.join('\n').indexOf(structure.map(item => item.id).join('\n'));
+          const structureSignature = toLines(structure, item => item.id);
+          tabsOffset = uniqueIds.join('\n').indexOf(structureSignature);
           windowStateCompletelyApplied = tabsOffset > -1;
         }
         else {
@@ -148,10 +158,11 @@ async function reserveToAttachTabFromRestoredInfo(tab, options = {}) {
     const bulk = tasks.length > 1;
     await Promise.all(uniqueIds.map((uniqueId, index) => {
       const task = tasks[index];
-      return attachTabFromRestoredInfo(task.tab, Object.assign({}, task.options, {
+      return attachTabFromRestoredInfo(task.tab, {
+        ...task.options,
         uniqueId,
         bulk
-      })).catch(error => {
+      }).catch(error => {
         console.log(`TreeStructure.reserveToAttachTabFromRestoredInfo: Fatal error on processing task ${index}, ${error}`, error.stack);
       });
     }));
@@ -169,7 +180,7 @@ reserveToAttachTabFromRestoredInfo.promisedDone = null;
 
 
 async function attachTabFromRestoredInfo(tab, options = {}) {
-  log('attachTabFromRestoredInfo ', tab);
+  log('attachTabFromRestoredInfo ', tab, options);
   let uniqueId, insertBefore, insertAfter, ancestors, children, states, collapsed /* for backward compatibility */;
   // eslint-disable-next-line prefer-const
   [uniqueId, insertBefore, insertAfter, ancestors, children, states, collapsed] = await Promise.all([
@@ -199,12 +210,20 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   insertAfter  = Tab.getByUniqueId(insertAfter);
   ancestors    = ancestors.map(Tab.getByUniqueId);
   children     = children.map(Tab.getByUniqueId);
-  log(' => references: ', {
+  log(' => references: ', () => ({
     insertBefore: dumpTab(insertBefore),
     insertAfter:  dumpTab(insertAfter),
     ancestors:    ancestors.map(dumpTab).join(', '),
     children:     children.map(dumpTab).join(', ')
-  });
+  }));
+  if (configs.fixupTreeOnTabVisibilityChanged) {
+    ancestors = ancestors.filter(ancestor => ancestor && (ancestor.hidden == tab.hidden));
+    children = children.filter(child => child && (child.hidden == tab.hidden));
+    log(' ==> references: ', () => ({
+      ancestors: ancestors.map(dumpTab).join(', '),
+      children:  children.map(dumpTab).join(', ')
+    }));
+  }
 
   // clear wrong positioning information
   if (tab.pinned ||
@@ -221,6 +240,7 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
   for (const ancestor of ancestors) {
     if (!ancestor)
       continue;
+    log(' attach to old ancestor: ', { child: tab, parent: ancestor });
     const promisedDone = Tree.attachTabTo(tab, ancestor, {
       insertBefore,
       insertAfter,
@@ -273,6 +293,8 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
     });
   }
   if (options.children && !options.bulk) {
+    let firstInTree = tab.$TST.firstChild || tab;
+    let lastInTree  = tab.$TST.lastDescendant || tab;
     for (const child of children) {
       if (!child)
         continue;
@@ -280,8 +302,13 @@ async function attachTabFromRestoredInfo(tab, options = {}) {
         dontExpand:  !child.active,
         forceExpand: active,
         insertAt:    Constants.kINSERT_NEAREST,
+        dontMove:    child.index >= firstInTree.index && child.index <= lastInTree.index + 1,
         broadcast:   true
       });
+      if (child.index < firstInTree.index)
+        firstInTree = child;
+      else if (child.index > lastInTree.index)
+        lastInTree = child;
     }
   }
 
@@ -363,6 +390,8 @@ Tab.onRestored.addListener(tab => {
           setTimeout(() => { // unblock in the next event loop, after other asynchronous operations are finished
             UserOperationBlocker.unblockIn(tab.windowId, { throbber: true });
           }, 0);
+
+          tryRestoreClosedSetFor(tab);
         }
         else {
           mRestoringTabs.set(tab.windowId, count);
@@ -380,3 +409,168 @@ Tab.onRestored.addListener(tab => {
   if (mProcessingTabRestorations.length == 1)
     mProcessingTabRestorations[0]();
 });
+
+
+// Implementation for the "Undo Close Tab*s*" feature
+// https://github.com/piroor/treestyletab/issues/2627
+
+const mPendingRecentlyClosedTabsInfo = {
+  tabs:      [],
+  structure: []
+};
+
+Tab.onRemoved.addListener((_tab, _info) => {
+  const currentlyRestorable = mRecentlyClosedTabs.length > 1;
+
+  mRecentlyClosedTabs = [];
+  mRecentlyClosedTabsTreeStructure = [];
+
+  const newlyRestorable = mRecentlyClosedTabs.length > 1;
+  if (currentlyRestorable != newlyRestorable)
+    Tab.onChangeMultipleTabsRestorability.dispatch(newlyRestorable);
+});
+
+Tab.onMultipleTabsRemoving.addListener(tabs => {
+  mPendingRecentlyClosedTabsInfo.tabs = tabs.map(tab => ({
+    originalId:    tab.id,
+    uniqueId:      tab.$TST.uniqueId.id,
+    windowId:      tab.windowId,
+    title:         tab.title,
+    url:           tab.url,
+    cookieStoreId: tab.cookieStoreId
+  }));
+  mPendingRecentlyClosedTabsInfo.structure = TreeBehavior.getTreeStructureFromTabs(tabs, {
+    full:                 true,
+    keepParentOfRootTabs: true
+  });
+});
+
+Tab.onMultipleTabsRemoved.addListener(tabs => {
+  log('multiple tabs are removed');
+  const currentlyRestorable = mRecentlyClosedTabs.length > 1;
+
+  const tabIds = new Set(tabs.map(tab => tab.id));
+  mRecentlyClosedTabs = mPendingRecentlyClosedTabsInfo.tabs.filter(info => tabIds.has(info.originalId));
+  mRecentlyClosedTabsTreeStructure = mPendingRecentlyClosedTabsInfo.structure.filter(structure => tabIds.has(structure.originalId));
+  log('  structure: ', mRecentlyClosedTabsTreeStructure);
+
+  const newlyRestorable = mRecentlyClosedTabs.length > 1;
+  if (currentlyRestorable != newlyRestorable)
+    Tab.onChangeMultipleTabsRestorability.dispatch(newlyRestorable);
+
+  mPendingRecentlyClosedTabsInfo.tabs = [];
+  mPendingRecentlyClosedTabsInfo.structure = [];
+});
+
+async function tryRestoreClosedSetFor(tab) {
+  const lastRecentlyClosedTabs = mRecentlyClosedTabs;
+  const lastRecentlyClosedTabsTreeStructure = mRecentlyClosedTabsTreeStructure;
+  mRecentlyClosedTabs = [];
+  mRecentlyClosedTabsTreeStructure = [];
+  if (lastRecentlyClosedTabs.length > 1)
+    Tab.onChangeMultipleTabsRestorability.dispatch(false);
+
+  if (!configs.undoMultipleTabsClose)
+    return;
+
+  const alreadRestoredIndex = lastRecentlyClosedTabs.findIndex(info => info.uniqueId == tab.$TST.uniqueId.id && info.windowId == tab.windowId);
+  log('tryRestoreClosedSetFor ', tab, lastRecentlyClosedTabs, lastRecentlyClosedTabsTreeStructure);
+  if (alreadRestoredIndex < 0) {
+    log(' => not a member of restorable tab set.');
+    return;
+  }
+
+  const toBeRestoredTabsCount = lastRecentlyClosedTabs.length - 1;
+  if (toBeRestoredTabsCount == 0) {
+    log(' => no more tab to be restored.');
+    return;
+  }
+
+  const sessions = (await browser.sessions.getRecentlyClosed({
+    maxResults: browser.sessions.MAX_SESSION_RESULTS
+  }).catch(ApiTabs.createErrorHandler())).filter(session => session.tab);
+
+  let restoredTabs;
+  if (toBeRestoredTabsCount <= sessions.length) {
+    log('tryRestoreClosedSetFor: restore tabs with the sessions API');
+    restoredTabs = await Commands.restoreTabs(toBeRestoredTabsCount);
+    restoredTabs.push(tab);
+    Tab.sort(restoredTabs)
+  }
+  else {
+    log('tryRestoreClosedSetFor: recreate tabs');
+    const tabOpenOptions = {
+      windowId:     lastRecentlyClosedTabs[0].windowId,
+      isOrphan:     true,
+      inBackground: true,
+      discarded:    true,
+      fixPositions: true
+    };
+    const beforeTabs = await TabsOpen.openURIsInTabs(
+      lastRecentlyClosedTabs.slice(0, alreadRestoredIndex).map(info => ({
+        title:         info.title,
+        url:           info.url,
+        cookieStoreId: info.cookieStoreId
+      })),
+      tabOpenOptions
+    );
+    const afterTabs = await TabsOpen.openURIsInTabs(
+      lastRecentlyClosedTabs.slice(alreadRestoredIndex + 1).map(info => ({
+        title:         info.title,
+        url:           info.url,
+        cookieStoreId: info.cookieStoreId
+      })),
+      tabOpenOptions
+    );
+    // We need to move tabs after they are opened instead of
+    // specifying the "index" option for TabsOpen.openURIsInTabs(),
+    // because the given restored tab can be moved while this
+    // operation.
+    if (beforeTabs.length > 0)
+      await TabsMove.moveTabsBefore(
+        beforeTabs,
+        tab,
+        { broadcast: true }
+      );
+    if (afterTabs.length > 0)
+      await TabsMove.moveTabsAfter(
+        afterTabs,
+        tab,
+        { broadcast: true }
+      );
+    restoredTabs = [...beforeTabs, tab, ...afterTabs];
+    await TabsInternalOperation.activateTab(restoredTabs[0]);
+  }
+
+  const rootTabs = restoredTabs.filter((tab, index) => lastRecentlyClosedTabsTreeStructure[index].parent == TreeBehavior.STRUCTURE_KEEP_PARENT || lastRecentlyClosedTabsTreeStructure[index].parent == TreeBehavior.STRUCTURE_NO_PARENT);
+  for (const rootTab of rootTabs) {
+    const referenceTabs = TreeBehavior.calculateReferenceTabsFromInsertionPosition(rootTab, {
+      context:      Constants.kINSERTION_CONTEXT_MOVED,
+      insertAfter:  rootTab.$TST.previousTab,
+      insertBefore: restoredTabs[restoredTabs.length - 1].$TST.nextTab
+    });
+    log(`tryRestoreClosedSetFor: referenceTabs for ${rootTab.id} => `, referenceTabs);
+    if (referenceTabs.parent)
+      await Tree.attachTabTo(rootTab, referenceTabs.parent, {
+        dontExpand:  true,
+        insertAfter: referenceTabs.insertAfter,
+        dontMove:    true,
+        broadcast:   true
+      });
+  }
+
+  await Tree.applyTreeStructureToTabs(
+    restoredTabs,
+    lastRecentlyClosedTabsTreeStructure
+  );
+
+  // Firefox itself activates the initially restored tab with delay,
+  // so we need to activate the first tab of the restored tabs again.
+  const onActivated = activeInfo => {
+    browser.tabs.onActivated.removeListener(onActivated);
+    if (activeInfo.id != restoredTabs[0].id)
+      TabsInternalOperation.activateTab(restoredTabs[0]);
+  };
+  browser.tabs.onActivated.addListener(onActivated);
+  wait(100).then(() => onActivated({ id: -1 })); // failsafe
+}
