@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -26,12 +26,11 @@
  * ***** END LICENSE BLOCK ******/
 'use strict';
 
-import TabIdFixer from '/extlib/TabIdFixer.js';
-
 import {
   log as internalLogger,
   wait,
   dumpTab,
+  mapAndFilter,
   configs
 } from '/common/common.js';
 import * as Constants from '/common/constants.js';
@@ -99,7 +98,7 @@ export async function attachTabTo(child, parent, options = {}) {
     log('=> could not attach tab to a parent in different window');
     return false;
   }
-  const ancestors = [parent].concat(child.$TST.ancestors);
+  const ancestors = [parent].concat(parent.$TST.ancestors);
   if (ancestors.includes(child)) {
     log('=> canceled for recursive request');
     return false;
@@ -124,7 +123,8 @@ export async function attachTabTo(child, parent, options = {}) {
     insertAfter:  options.insertAfter
   });
 
-  await Tab.waitUntilTrackedAll(child.windowId);
+  if (!options.synchronously)
+    await Tab.waitUntilTrackedAll(child.windowId);
 
   parent = TabsStore.ensureLivingTab(parent);
   child = TabsStore.ensureLivingTab(child);
@@ -149,12 +149,13 @@ export async function attachTabTo(child, parent, options = {}) {
     log('=> already attached');
 
   if (newlyAttached) {
-    detachTab(child, Object.assign({}, options, {
+    detachTab(child, {
+      ...options,
       // Don't broadcast this detach operation, because this "attachTabTo" can be
       // broadcasted. If we broadcast this detach operation, the tab is detached
       // twice in the sidebar!
       broadcast: false
-    }));
+    });
 
     log('attachTabTo: setting child information to ', parent.id);
     // we need to set its children via the "children" setter, to invalidate cached information.
@@ -164,9 +165,8 @@ export async function attachTabTo(child, parent, options = {}) {
     // "children" setter updates the child itself automatically.
 
     const parentLevel = parseInt(parent.$TST.getAttribute(Constants.kLEVEL) || 0);
-    if (!options.dontUpdateIndent) {
-      updateTabsIndent(child, parentLevel + 1);
-    }
+    if (!options.dontUpdateIndent)
+      updateTabsIndent(child, parentLevel + 1, { justNow: options.synchronously });
 
     SidebarConnection.sendMessage({
       type:     Constants.kCOMMAND_NOTIFY_CHILDREN_CHANGED,
@@ -187,18 +187,19 @@ export async function attachTabTo(child, parent, options = {}) {
     }
   }
 
-  onAttached.dispatch(child, Object.assign({}, options, {
+  onAttached.dispatch(child, {
+    ...options,
     parent,
     insertBefore: options.insertBefore,
     insertAfter:  options.insertAfter,
     newIndex, newlyAttached
-  }));
+  });
 
   return !options.dontMove && moved;
 }
 
 export function getReferenceTabsForNewChild(child, parent, options = {}) {
-  log('getReferenceTabsForNewChild ', child, parent, options);
+  log('getReferenceTabsForNewChild ', { child, parent, options });
   let insertAt = options.insertAt;
   if (typeof insertAt !== 'number')
     insertAt = configs.insertNewChildAt;
@@ -210,7 +211,7 @@ export function getReferenceTabsForNewChild(child, parent, options = {}) {
   let insertBefore, insertAfter;
   if (descendants.length > 0) {
     const firstChild     = descendants[0];
-    const lastDescendant = descendants[descendants.length-1];
+    const lastDescendant = descendants[descendants.length - 1];
     switch (insertAt) {
       case Constants.kINSERT_END:
       default:
@@ -224,6 +225,7 @@ export function getReferenceTabsForNewChild(child, parent, options = {}) {
       case Constants.kINSERT_NEAREST: {
         const allTabs = Tab.getOtherTabs(parent.windowId, options.ignoreTabs);
         const index = allTabs.indexOf(child);
+        log('  insertAt=kINSERT_NEAREST ', { allTabs, index });
         if (index < allTabs.indexOf(firstChild)) {
           insertBefore = firstChild;
           insertAfter  = parent;
@@ -338,10 +340,13 @@ export function detachTab(child, options = {}) {
   // "children" setter removes the parent ifself from the detached child
   // automatically.
 
-  updateTabsIndent(child);
+  if (!options.toBeRemoved && !options.toBeDetached)
+    updateTabsIndent(child);
 
   onDetached.dispatch(child, {
-    oldParentTab: parent
+    oldParentTab: parent,
+    toBeRemoved:  !!options.toBeRemoved,
+    toBeDetached: !!options.toBeDetached
   });
 }
 
@@ -355,12 +360,20 @@ export async function detachTabsFromTree(tabs, options = {}) {
     const parent   = tab.$TST.parent;
     for (const child of children) {
       if (!tabs.includes(child)) {
-        if (parent)
-          promisedAttach.push(attachTabTo(child, parent, Object.assign({}, options, {
+        if (parent) {
+          promisedAttach.push(attachTabTo(child, parent, {
+            ...options,
             dontMove: true
-          })));
-        else
+          }));
+        }
+        else {
           detachTab(child, options);
+          if (child.$TST.collapsed)
+            await collapseExpandTabAndSubtree(child, {
+              ...options,
+              collapsed: false
+            });
+        }
       }
     }
   }
@@ -368,81 +381,104 @@ export async function detachTabsFromTree(tabs, options = {}) {
     await Promise.all(promisedAttach);
 }
 
-export function detachAllChildren(tab, options = {}) {
-  log('detachAllChildren: ', tab.id);
+export function detachAllChildren(
+  tab = null,
+  { children, parent, nearestFollowingRootTab, newParent, behavior, dontExpand,
+    ...options } = {}
+) {
+  log('detachAllChildren: ',
+      tab && tab.id,
+      { children, parent, nearestFollowingRootTab, newParent, behavior, dontExpand },
+      options);
   // the "children" option is used for removing tab.
-  const children = options.children ? options.children.map(TabsStore.ensureLivingTab) : tab.$TST.children;
+  children = children ? children.map(TabsStore.ensureLivingTab) : tab.$TST.children;
+
+  if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD &&
+      newParent &&
+      !children.includes(newParent))
+    children.unshift(newParent);
+
   if (!children.length)
     return;
-  log(' => children to be detached: ', children.map(dumpTab));
+  log(' => children to be detached: ', () => children.map(dumpTab));
 
-  if (!('behavior' in options))
-    options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_SIMPLY_DETACH_ALL_CHILDREN;
-  if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
-    options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
+  if (behavior === undefined)
+    behavior = Constants.kCLOSE_PARENT_BEHAVIOR_SIMPLY_DETACH_ALL_CHILDREN;
+  if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_CLOSE_ALL_CHILDREN)
+    behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD;
 
   options.dontUpdateInsertionPositionInfo = true;
 
   // the "parent" option is used for removing tab.
-  const parent = TabsStore.ensureLivingTab(options.parent) || tab.$TST.parent;
-  if (tab.$TST.isGroupTab &&
+  parent = TabsStore.ensureLivingTab(parent) || (tab && tab.$TST.parent);
+  if (tab &&
+      tab.$TST.isGroupTab &&
       Tab.getRemovingTabs(tab.windowId).length == children.length) {
-    options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN;
+    behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN;
     options.dontUpdateIndent = false;
   }
 
   let nextTab = null;
-  if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN &&
+  if (tab &&
+      behavior == Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN &&
       !configs.moveTabsToBottomWhenDetachedFromClosedParent) {
     nextTab = tab.$TST.nearestFollowingRootTab;
   }
 
-  if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB) {
+  if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_REPLACE_WITH_GROUP_TAB) {
     // open new group tab and replace the detaching tab with it.
-    options.behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN;
+    behavior = Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN;
   }
 
-  if (options.behavior != Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN)
-    collapseExpandSubtree(tab, Object.assign({}, options, {
+  if (!dontExpand &&
+      tab &&
+      behavior != Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN)
+    collapseExpandSubtree(tab, {
+      ...options,
       collapsed: false
-    }));
+    });
 
   let count = 0;
   for (const child of children) {
     if (!child)
       continue;
-    if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN) {
+    if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_DETACH_ALL_CHILDREN) {
       detachTab(child, options);
       moveTabSubtreeBefore(child, nextTab, options);
     }
-    else if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD) {
+    else if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_FIRST_CHILD) {
       detachTab(child, options);
       if (count == 0) {
         if (parent) {
-          attachTabTo(child, parent, Object.assign({}, options, {
-            dontExpan: true,
-            dontMove:  true
-          }));
+          attachTabTo(child, parent, {
+            ...options,
+            dontExpand: true,
+            dontMove:   true
+          });
         }
-        collapseExpandSubtree(child, Object.assign({}, options, {
+        collapseExpandSubtree(child, {
+          ...options,
           collapsed: false
-        }));
+        });
         //deleteTabValue(child, Constants.kTAB_STATE_SUBTREE_COLLAPSED);
       }
       else {
-        attachTabTo(child, children[0], Object.assign({}, options, {
+        attachTabTo(child, children[0], {
+          ...options,
           dontExpand: true,
           dontMove:   true
-        }));
+        });
       }
     }
-    else if (options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN && parent) {
-      attachTabTo(child, parent, Object.assign({}, options, {
+    else if (behavior == Constants.kCLOSE_PARENT_BEHAVIOR_PROMOTE_ALL_CHILDREN &&
+             parent) {
+      attachTabTo(child, parent, {
+        ...options,
         dontExpand: true,
         dontMove:   true
-      }));
+      });
     }
-    else { // options.behavior == Constants.kCLOSE_PARENT_BEHAVIOR_SIMPLY_DETACH_ALL_CHILDREN
+    else { // behavior == Constants.kCLOSE_PARENT_BEHAVIOR_SIMPLY_DETACH_ALL_CHILDREN
       detachTab(child, options);
     }
     count++;
@@ -450,30 +486,36 @@ export function detachAllChildren(tab, options = {}) {
 }
 
 // returns moved (or not)
-export async function behaveAutoAttachedTab(tab, options = {}) {
-  const baseTab = options.baseTab || Tab.getActiveTab(TabsStore.getWindow() || tab.windowId);
-  log('behaveAutoAttachedTab ', tab.id, baseTab.id, options);
+export async function behaveAutoAttachedTab(
+  tab,
+  { baseTab, behavior, broadcast, dontMove } = {}
+) {
+  if (!configs.autoAttach)
+    return false;
+
+  baseTab = baseTab || Tab.getActiveTab(TabsStore.getCurrentWindowId() || tab.windowId);
+  log('behaveAutoAttachedTab ', tab.id, baseTab.id, { baseTab, behavior });
 
   if (baseTab &&
       baseTab.$TST.ancestors.includes(tab)) {
     log(' => ignore possibly restored ancestor tab to avoid cyclic references');
-    return;
+    return false;
   }
 
   if (baseTab.pinned) {
     if (!tab.pinned)
       return false;
-    options.behavior = Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING;
+    behavior = Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING;
     log(' => override behavior for pinned tabs');
   }
-  switch (options.behavior) {
+  switch (behavior) {
     default:
       return false;
 
     case Constants.kNEWTAB_OPEN_AS_ORPHAN:
       log(' => kNEWTAB_OPEN_AS_ORPHAN');
       detachTab(tab, {
-        broadcast: options.broadcast
+        broadcast
       });
       if (tab.$TST.nextTab)
         return TabsMove.moveTabAfter(tab, Tab.getLastTab(tab.windowId), {
@@ -484,10 +526,10 @@ export async function behaveAutoAttachedTab(tab, options = {}) {
     case Constants.kNEWTAB_OPEN_AS_CHILD:
       log(' => kNEWTAB_OPEN_AS_CHILD');
       return attachTabTo(tab, baseTab, {
-        dontMove:    options.dontMove || configs.insertNewChildAt == Constants.kINSERT_NO_CONTROL,
+        dontMove:    dontMove || configs.insertNewChildAt == Constants.kINSERT_NO_CONTROL,
         forceExpand: true,
         delayedMove: true,
-        broadcast:   options.broadcast
+        broadcast
       });
 
     case Constants.kNEWTAB_OPEN_AS_SIBLING: {
@@ -496,13 +538,13 @@ export async function behaveAutoAttachedTab(tab, options = {}) {
       if (parent) {
         await attachTabTo(tab, parent, {
           delayedMove: true,
-          broadcast: options.broadcast
+          broadcast
         });
         return true;
       }
       else {
         detachTab(tab, {
-          broadcast: options.broadcast
+          broadcast
         });
         return TabsMove.moveTabAfter(tab, Tab.getLastTab(tab.windowId), {
           delayedMove: true
@@ -510,8 +552,9 @@ export async function behaveAutoAttachedTab(tab, options = {}) {
       }
     };
 
-    case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING: {
-      log(' => kNEWTAB_OPEN_AS_NEXT_SIBLING');
+    case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING:
+    case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING_WITH_INHERITED_CONTAINER: {
+      log(' => kNEWTAB_OPEN_AS_NEXT_SIBLING(_WITH_INHERITED_CONTAINER)');
       let nextSibling = baseTab.$TST.nextSiblingTab;
       if (nextSibling == tab)
         nextSibling = null;
@@ -521,22 +564,22 @@ export async function behaveAutoAttachedTab(tab, options = {}) {
           insertBefore: nextSibling,
           insertAfter:  baseTab.$TST.lastDescendant || baseTab,
           delayedMove:  true,
-          broadcast:    options.broadcast
+          broadcast
         });
       }
       else {
         detachTab(tab, {
-          broadcast: options.broadcast
+          broadcast
         });
         if (nextSibling)
           return TabsMove.moveTabBefore(tab, nextSibling, {
             delayedMove: true,
-            broadcast: options.broadcast
+            broadcast
           });
         else
           return TabsMove.moveTabAfter(tab, baseTab.$TST.lastDescendant, {
             delayedMove: true,
-            broadcast: options.broadcast
+            broadcast
           });
       }
     };
@@ -550,7 +593,7 @@ export async function behaveAutoAttachedTabs(tabs, options = {}) {
 
     case Constants.kNEWTAB_OPEN_AS_ORPHAN:
       if (options.baseTabs && !options.baseTab)
-        options.baseTab = options.baseTabs[options.baseTabs.length-1];
+        options.baseTab = options.baseTabs[options.baseTabs.length - 1];
       for (const tab of tabs) {
         await behaveAutoAttachedTab(tab, options);
       }
@@ -569,7 +612,7 @@ export async function behaveAutoAttachedTabs(tabs, options = {}) {
     case Constants.kNEWTAB_OPEN_AS_SIBLING:
     case Constants.kNEWTAB_OPEN_AS_NEXT_SIBLING: {
       if (options.baseTabs && !options.baseTab)
-        options.baseTab = options.baseTabs[options.baseTabs.length-1];
+        options.baseTab = options.baseTabs[options.baseTabs.length - 1];
       let moved = false;
       for (const tab of tabs.reverse()) {
         moved = (await behaveAutoAttachedTab(tab, options)) || moved;
@@ -579,7 +622,7 @@ export async function behaveAutoAttachedTabs(tabs, options = {}) {
   }
 }
 
-function updateTabsIndent(tabs, level = undefined) {
+function updateTabsIndent(tabs, level = undefined, options = {}) {
   if (!tabs)
     return;
 
@@ -597,32 +640,39 @@ function updateTabsIndent(tabs, level = undefined) {
     if (!item || item.pinned)
       continue;
 
-    updateTabIndent(item, level);
+    updateTabIndent(item, level, options);
   }
 }
 
 // this is called multiple times on a session restoration, so this should be throttled for better performance
-function updateTabIndent(tab, level = undefined) {
+function updateTabIndent(tab, level = undefined, options = {}) {
   let timer = updateTabIndent.delayed.get(tab.id);
   if (timer)
     clearTimeout(timer);
+  if (options.justNow) {
+    return updateTabIndentNow(tab, level, options);
+  }
   timer = setTimeout(() => {
     updateTabIndent.delayed.delete(tab.id);
-    if (!TabsStore.ensureLivingTab(tab))
-      return;
-    tab.$TST.setAttribute(Constants.kLEVEL, level);
-    updateTabsIndent(tab.$TST.children, level + 1);
-    SidebarConnection.sendMessage({
-      type:     Constants.kCOMMAND_NOTIFY_TAB_LEVEL_CHANGED,
-      windowId: tab.windowId,
-      tabId:    tab.id,
-      level
-    });
+    updateTabIndentNow(tab, level);
   }, 100);
   updateTabIndent.delayed.set(tab.id, timer);
 }
 updateTabIndent.delayed = new Map();
 
+function updateTabIndentNow(tab, level = undefined, options = {}) {
+  if (!TabsStore.ensureLivingTab(tab))
+    return;
+  tab.$TST.setAttribute(Constants.kLEVEL, level);
+  updateTabsIndent(tab.$TST.children, level + 1, options);
+  SidebarConnection.sendMessage({
+    type:     Constants.kCOMMAND_NOTIFY_TAB_LEVEL_CHANGED,
+    windowId: tab.windowId,
+    tabId:    tab.id,
+    level
+  });
+}
+
 
 // collapse/expand tabs
 
@@ -636,8 +686,15 @@ export async function collapseExpandSubtree(tab, params = {}) {
   logCollapseExpand('collapseExpandSubtree: ', dumpTab(tab), tab.$TST.subtreeCollapsed, params);
   await collapseExpandSubtreeInternal(tab, params);
   onSubtreeCollapsedStateChanged.dispatch(tab, { collapsed: !!params.collapsed });
+  if (TSTAPI.hasListenerForMessageType(TSTAPI.kNOTIFY_TREE_COLLAPSED_STATE_CHANGED)) {
+    TSTAPI.sendMessage({
+      type:      TSTAPI.kNOTIFY_TREE_COLLAPSED_STATE_CHANGED,
+      tab:       new TSTAPI.TreeItem(tab),
+      collapsed: !!params.collapsed
+    }, { tabProperties: ['tab'] }).catch(_error => {});
+  }
 }
-function collapseExpandSubtreeInternal(tab, params = {}) {
+async function collapseExpandSubtreeInternal(tab, params = {}) {
   if (!params.force &&
       tab.$TST.subtreeCollapsed == params.collapsed)
     return;
@@ -658,7 +715,7 @@ function collapseExpandSubtreeInternal(tab, params = {}) {
     if (!params.collapsed &&
         !params.justNow &&
         i == lastExpandedTabIndex) {
-      collapseExpandTabAndSubtree(childTab, {
+      await collapseExpandTabAndSubtree(childTab, {
         collapsed: params.collapsed,
         justNow:   params.justNow,
         anchor:    tab,
@@ -667,7 +724,7 @@ function collapseExpandSubtreeInternal(tab, params = {}) {
       });
     }
     else {
-      collapseExpandTabAndSubtree(childTab, {
+      await collapseExpandTabAndSubtree(childTab, {
         collapsed: params.collapsed,
         justNow:   params.justNow,
         broadcast: false
@@ -696,35 +753,51 @@ export function manualCollapseExpandSubtree(tab, params = {}) {
   }
 }
 
-export function collapseExpandTabAndSubtree(tab, params = {}) {
+export async function collapseExpandTabAndSubtree(tab, params = {}) {
   if (!tab)
     return;
 
-  const parent = tab.$TST.parent;
-  if (!parent)
+  // allow to expand root collapsed tab
+  if (!tab.$TST.collapsed &&
+      !tab.$TST.parent)
     return;
 
   collapseExpandTab(tab, params);
 
-  if (params.collapsed && tab.active) {
-    const newSelection = tab.$TST.nearestVisibleAncestorOrSelf;
-    logCollapseExpand('current tab is going to be collapsed, switch to ', newSelection.id);
-    TabsInternalOperation.activateTab(newSelection, { silently: true });
+  if (params.collapsed &&
+      tab.active &&
+      configs.unfocusableCollapsedTab) {
+    logCollapseExpand('current tree is going to be collapsed');
+    const allowed = await TSTAPI.tryOperationAllowed(
+      TSTAPI.kNOTIFY_TRY_MOVE_FOCUS_FROM_COLLAPSING_TREE,
+      { tab: new TSTAPI.TreeItem(tab) },
+      { tabProperties: ['tab'] }
+    );
+    if (allowed) {
+      let newSelection = tab.$TST.nearestVisibleAncestorOrSelf;
+      if (configs.avoidDiscardedTabToBeActivatedIfPossible && newSelection.discarded)
+        newSelection = newSelection.$TST.nearestLoadedTabInTree ||
+                         newSelection.$TST.nearestLoadedTab ||
+                         newSelection;
+      logCollapseExpand('=> switch to ', newSelection.id);
+      TabsInternalOperation.activateTab(newSelection, { silently: true });
+    }
   }
 
   if (!tab.$TST.subtreeCollapsed) {
     const children = tab.$TST.children;
-    children.forEach((child, index) => {
+    await Promise.all(children.map((child, index) => {
       const last = params.last &&
                      (index == children.length - 1);
-      collapseExpandTabAndSubtree(child, Object.assign({}, params, {
+      return collapseExpandTabAndSubtree(child, {
+        ...params,
         collapsed: params.collapsed,
         justNow:   params.justNow,
         anchor:    last && params.anchor,
         last:      last,
         broadcast: params.broadcast
-      }));
-    });
+      });
+    }));
   }
 }
 
@@ -739,18 +812,22 @@ export async function collapseExpandTab(tab, params = {}) {
   // synchronous "collapse" operation, it can produce an expanded
   // child tab under "subtree-collapsed" parent. So this is a failsafe.
   if (!params.collapsed &&
-      tab.$TST.ancestors.some(ancestor => ancestor.$TST.subtreeCollapsed))
+      tab.$TST.ancestors.some(ancestor => ancestor.$TST.subtreeCollapsed)) {
+    log('collapseExpandTab: canceled to avoid expansion under collapsed tree ',
+        tab.$TST.ancestors.find(ancestor => ancestor.$TST.subtreeCollapsed));
     return;
+  }
 
   const stack = `${configs.debug && new Error().stack}\n${params.stack || ''}`;
   logCollapseExpand(`collapseExpandTab ${tab.id} `, params, { stack })
   const last = params.last &&
                  (!tab.$TST.hasChild || tab.$TST.subtreeCollapsed);
   const byAncestor = tab.$TST.ancestors.some(ancestor => ancestor.$TST.subtreeCollapsed) == params.collapsed;
-  const collapseExpandInfo = Object.assign({}, params, {
+  const collapseExpandInfo = {
+    ...params,
     anchor: last && params.anchor,
     last
-  });
+  };
 
   if (params.collapsed) {
     tab.$TST.addState(Constants.kTAB_STATE_COLLAPSED);
@@ -826,14 +903,16 @@ export function collapseExpandTreesIntelligentlyFor(tab, options = {}) {
 
     const manuallyExpanded = collapseTab.$TST.states.has(Constants.kTAB_STATE_SUBTREE_EXPANDED_MANUALLY);
     if (!dontCollapse && !manuallyExpanded)
-      collapseExpandSubtree(collapseTab, Object.assign({}, options, {
+      collapseExpandSubtree(collapseTab, {
+        ...options,
         collapsed: true
-      }));
+      });
   }
 
-  collapseExpandSubtree(tab, Object.assign({}, options, {
+  collapseExpandSubtree(tab, {
+    ...options,
     collapsed: false
-  }));
+  });
   window.doingIntelligentlyCollapseExpandCount--;
 }
 
@@ -853,9 +932,10 @@ export async function fixupSubtreeCollapsedState(tab, options = {}) {
   });
   if (collapsedStateMismatched) {
     log(' => set collapsed state');
-    await collapseExpandSubtree(tab, Object.assign({}, options, {
+    await collapseExpandSubtree(tab, {
+      ...options,
       collapsed: childrenCollapsed
-    }));
+    });
     fixed = true;
   }
   if (!nextIsFirstChild) {
@@ -943,15 +1023,15 @@ export async function moveTabs(tabs, options = {}) {
   if (tabs.length == 0)
     return [];
 
-  log('moveTabs: ', tabs.map(dumpTab), options);
+  log('moveTabs: ', () => ({ tabs: tabs.map(dumpTab), options }));
 
-  const windowId = parseInt(tabs[0].windowId || TabsStore.getWindow());
+  const windowId = parseInt(tabs[0].windowId || TabsStore.getCurrentWindowId());
 
   let newWindow = options.destinationPromisedNewWindow;
 
   let destinationWindowId = options.destinationWindowId;
   if (!destinationWindowId && !newWindow)
-    destinationWindowId = TabsStore.getWindow();
+    destinationWindowId = TabsStore.getCurrentWindowId();
 
   const isAcrossWindows = windowId != destinationWindowId || !!newWindow;
 
@@ -961,6 +1041,9 @@ export async function moveTabs(tabs, options = {}) {
   const structure = TreeBehavior.getTreeStructureFromTabs(tabs);
   log('original tree structure: ', structure);
 
+  if (!options.duplicate)
+    await detachTabsFromTree(tabs, options);
+
   if (isAcrossWindows || options.duplicate) {
     if (mSlowDuplication)
       UserOperationBlocker.blockIn(windowId, { throbber: true });
@@ -1018,7 +1101,7 @@ export async function moveTabs(tabs, options = {}) {
                 return null;
               }
             })).then(tabs => {
-              log(`ids from API responses are resolved in ${Date.now() - startTime}msec: `, tabs.map(dumpTab));
+              log(`ids from API responses are resolved in ${Date.now() - startTime}msec: `, () => tabs.map(dumpTab));
               return tabs;
             });
             movedTabs = await promisedDuplicatedTabs;
@@ -1027,6 +1110,16 @@ export async function moveTabs(tabs, options = {}) {
             movedTabs = movedTabs.map(tab => Tab.get(tab.id));
             movedTabIds = movedTabs.map(tab => tab.id);
           }
+          else {
+            for (const tab of movedTabs) {
+              if (tab.$TST.parent &&
+                  !movedTabs.includes(tab.$TST.parent))
+                detachTab(tab, {
+                  broadcast:    true,
+                  toBeDetached: true
+                });
+            }
+          }
         })()
       ]);
       log('moveTabs: all windows and tabs are ready, ', movedTabIds, destinationWindowId);
@@ -1075,8 +1168,8 @@ export async function moveTabs(tabs, options = {}) {
       const startTime = Date.now();
       const maxDelay = configs.maximumAcceptableDelayForTabDuplication;
       while (Date.now() - startTime < maxDelay) {
-        newTabs = movedTabs.map(tab => Tab.get(TabIdFixer.fixTab(tab).id));
-        newTabs = newTabs.filter(tab => !!tab);
+        newTabs = mapAndFilter(movedTabs,
+                               tab => Tab.get(tab.id) || undefined);
         if (mSlowDuplication)
           UserOperationBlocker.setProgress(Math.round(newTabs.length / tabs.length * 50) + 50, windowId);
         if (newTabs.length < tabs.length) {
@@ -1115,8 +1208,7 @@ export async function moveTabs(tabs, options = {}) {
   }
 
 
-  movedTabs = movedTabs.map(tab => Tab.get(tab.id));
-  movedTabs = movedTabs.filter(tab => !!tab);
+  movedTabs = mapAndFilter(movedTabs, tab => Tab.get(tab.id) || undefined);
   if (options.insertBefore) {
     await TabsMove.moveTabsBefore(
       movedTabs,
@@ -1136,8 +1228,16 @@ export async function moveTabs(tabs, options = {}) {
   }
   // Tabs can be removed while waiting, so we need to
   // refresh the array of tabs.
-  movedTabs = movedTabs.map(tab => Tab.get(tab.id));
-  movedTabs = movedTabs.filter(tab => !!tab);
+  movedTabs = mapAndFilter(movedTabs, tab => Tab.get(tab.id) || undefined);
+
+  if (isAcrossWindows) {
+    for (const tab of movedTabs) {
+      if (tab.$TST.parent ||
+          parseInt(tab.$TST.getAttribute(Constants.kLEVEL) || 0) == 0)
+        continue;
+      updateTabIndent(tab, 0);
+    }
+  }
 
   return movedTabs;
 }
@@ -1166,20 +1266,20 @@ export async function openNewWindowFromTabs(tabs, options = {}) {
     })
     .catch(ApiTabs.createErrorHandler());
   tabs = tabs.filter(TabsStore.ensureLivingTab);
-  const movedTabs = await moveTabs(tabs, Object.assign({}, options, {
+  const movedTabs = await moveTabs(tabs, {
+    ...options,
     destinationPromisedNewWindow: promsiedNewWindow
-  }));
+  });
 
   log('closing needless tabs');
   browser.windows.get(newWindow.id, { populate: true })
     .then(window => {
-      log('moved tabs: ', movedTabs.map(dumpTab));
-      const movedTabIds     = movedTabs.map(tab => tab.id);
-      const allTabsInWindow = window.tabs.map(tab => TabIdFixer.fixTab(tab));
-      const removeTabs      = allTabsInWindow
-        .filter(tab => !movedTabIds.includes(tab.id))
-        .map(tab => Tab.get(tab.id));
-      log('removing tabs: ', removeTabs.map(dumpTab));
+      const movedTabIds = new Set(movedTabs.map(tab => tab.id));
+      log('moved tabs: ', movedTabIds);
+      const removeTabs = mapAndFilter(window.tabs, tab =>
+        !movedTabIds.has(tab.id) && Tab.get(tab.id) || undefined
+      );
+      log('removing tabs: ', removeTabs);
       TabsInternalOperation.removeTabs(removeTabs);
       UserOperationBlocker.unblockIn(newWindow.id);
     })
@@ -1189,39 +1289,47 @@ export async function openNewWindowFromTabs(tabs, options = {}) {
 }
 
 
+/* "treeStructure" is an array of integers, meaning:
+  [A]     => TreeBehavior.STRUCTURE_NO_PARENT (parent is not in this tree)
+    [B]   => 0 (parent is 1st item in this tree)
+    [C]   => 0 (parent is 1st item in this tree)
+      [D] => 2 (parent is 2nd in this tree)
+  [E]     => TreeBehavior.STRUCTURE_NO_PARENT (parent is not in this tree, and this creates another tree)
+    [F]   => 0 (parent is 1st item in this another tree)
+  See also getTreeStructureFromTabs() in tree-behavior.js
+*/
 export async function applyTreeStructureToTabs(tabs, treeStructure, options = {}) {
   if (!tabs || !treeStructure)
     return;
 
   MetricsData.add('applyTreeStructureToTabs: start');
 
-  log('applyTreeStructureToTabs: ', tabs.map(dumpTab), treeStructure, options);
+  log('applyTreeStructureToTabs: ', () => ({ tabs: tabs.map(dumpTab), treeStructure, options }));
   tabs = tabs.slice(0, treeStructure.length);
   treeStructure = treeStructure.slice(0, tabs.length);
 
   let expandStates = tabs.map(tab => !!tab);
   expandStates = expandStates.slice(0, tabs.length);
   while (expandStates.length < tabs.length)
-    expandStates.push(-1);
+    expandStates.push(TreeBehavior.STRUCTURE_NO_PARENT);
 
   MetricsData.add('applyTreeStructureToTabs: preparation');
 
-  let parentTab = null;
+  let parent = null;
   let tabsInTree = [];
   const promises   = [];
   for (let i = 0, maxi = tabs.length; i < maxi; i++) {
     const tab = tabs[i];
     /*
     if (tab.$TST.collapsed)
-      collapseExpandTabAndSubtree(tab, Object.assign({}, options, {
+      collapseExpandTabAndSubtree(tab, {
+        ...options,
         collapsed: false,
         justNow: true
-      }));
+      });
     */
-    detachTab(tab, { justNow: true });
-
     const structureInfo = treeStructure[i];
-    let parentIndexInTree = -1;
+    let parentIndexInTree = TreeBehavior.STRUCTURE_NO_PARENT;
     if (typeof structureInfo == 'number') { // legacy format
       parentIndexInTree = structureInfo;
     }
@@ -1229,30 +1337,28 @@ export async function applyTreeStructureToTabs(tabs, treeStructure, options = {}
       parentIndexInTree = structureInfo.parent;
       expandStates[i]   = !structureInfo.collapsed;
     }
-    if (parentIndexInTree < 0) { // there is no parent, so this is a new parent!
-      parentTab  = tab.id;
+    log(`  applyTreeStructureToTabs: parent for ${tab.id} => ${parentIndexInTree}`);
+    if (parentIndexInTree == TreeBehavior.STRUCTURE_NO_PARENT ||
+        parentIndexInTree == TreeBehavior.STRUCTURE_KEEP_PARENT) {
+      // there is no parent, so this is a new parent!
+      parent = null;
       tabsInTree = [tab];
     }
-
-    let parent = null;
-    if (parentIndexInTree > -1) {
-      parent = Tab.get(parentTab);
-      if (parent) {
-        //log('existing tabs in tree: ', {
-        //  size:   tabsInTree.length,
-        //  parent: parentIndexInTree
-        //});
-        parent = parentIndexInTree < tabsInTree.length ? tabsInTree[parentIndexInTree] : parent ;
-        tabsInTree.push(tab);
-      }
+    else {
+      tabsInTree.push(tab);
+      parent = parentIndexInTree < tabsInTree.length ? tabsInTree[parentIndexInTree] : null;
     }
-    if (parent) {
+    log('   => parent = ', parent);
+    if (parentIndexInTree != TreeBehavior.STRUCTURE_KEEP_PARENT)
+      detachTab(tab, { justNow: true });
+    if (parent && tab != parent) {
       parent.$TST.removeState(Constants.kTAB_STATE_SUBTREE_COLLAPSED); // prevent focus changing by "current tab attached to collapsed tree"
-      promises.push(attachTabTo(tab, parent, Object.assign({}, options, {
+      promises.push(attachTabTo(tab, parent, {
+        ...options,
         dontExpand: true,
         dontMove:   true,
         justNow:    true
-      })));
+      }));
     }
   }
   if (promises.length > 0)
@@ -1260,14 +1366,15 @@ export async function applyTreeStructureToTabs(tabs, treeStructure, options = {}
   MetricsData.add('applyTreeStructureToTabs: attach/detach');
 
   log('expandStates: ', expandStates);
-  for (let i = tabs.length-1; i > -1; i--) {
+  for (let i = tabs.length - 1; i > -1; i--) {
     const tab = tabs[i];
     const expanded = expandStates[i];
-    collapseExpandSubtree(tab, Object.assign({}, options, {
+    collapseExpandSubtree(tab, {
+      ...options,
       collapsed: expanded === undefined ? !tab.$TST.hasChild : !expanded ,
       justNow:   true,
       force:     true
-    }));
+    });
   }
   MetricsData.add('applyTreeStructureToTabs: collapse/expand');
 }
