@@ -14,7 +14,7 @@
  * The Original Code is the Tree Style Tab.
  *
  * The Initial Developer of the Original Code is YUKI "Piro" Hiroshi.
- * Portions created by the Initial Developer are Copyright (C) 2011-2019
+ * Portions created by the Initial Developer are Copyright (C) 2011-2020
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): YUKI "Piro" Hiroshi <piro.outsider.reflex@gmail.com>
@@ -46,6 +46,20 @@ function log(...args) {
 }
 
 export const onInitialized = new EventListenerManager();
+export const onRegistered   = new EventListenerManager();
+export const onUnregistered = new EventListenerManager();
+export const onMessageExternal = {
+  $listeners: new Set(),
+  addListener(listener) {
+    this.$listeners.add(listener);
+  },
+  removeListener(listener) {
+    this.$listeners.delete(listener);
+  },
+  dispatch(...args) {
+    return Array.from(this.$listeners, listener => listener(...args));
+  }
+};
 
 export const kREGISTER_SELF         = 'register-self';
 export const kUNREGISTER_SELF       = 'unregister-self';
@@ -56,6 +70,7 @@ export const kNOTIFY_SHUTDOWN       = 'shutdown'; // defined but not notified fo
 export const kNOTIFY_SIDEBAR_SHOW   = 'sidebar-show';
 export const kNOTIFY_SIDEBAR_HIDE   = 'sidebar-hide';
 export const kNOTIFY_TAB_CLICKED    = 'tab-clicked'; // for backward compatibility
+export const kNOTIFY_TAB_DBLCLICKED = 'tab-dblclicked';
 export const kNOTIFY_TAB_MOUSEDOWN  = 'tab-mousedown';
 export const kNOTIFY_TAB_MOUSEUP    = 'tab-mouseup';
 export const kNOTIFY_TABBAR_CLICKED = 'tabbar-clicked'; // for backward compatibility
@@ -72,10 +87,20 @@ export const kNOTIFY_TAB_DRAGEXIT   = 'tab-dragexit';
 export const kNOTIFY_TAB_DRAGEND    = 'tab-dragend';
 export const kNOTIFY_TREE_ATTACHED  = 'tree-attached';
 export const kNOTIFY_TREE_DETACHED  = 'tree-detached';
+export const kNOTIFY_TREE_COLLAPSED_STATE_CHANGED = 'tree-collapsed-state-changed';
 export const kNOTIFY_NATIVE_TAB_DRAGSTART = 'native-tab-dragstart';
+export const kNOTIFY_NATIVE_TAB_DRAGEND   = 'native-tab-dragend';
 export const kNOTIFY_PERMISSIONS_CHANGED = 'permissions-changed';
 export const kSTART_CUSTOM_DRAG     = 'start-custom-drag';
-export const kNOTIFY_TRY_MOVE_FOCUS_FROM_CLOSING_CURRENT_TAB = 'try-move-focus-from-closing-current-tab';
+export const kNOTIFY_TRY_MOVE_FOCUS_FROM_COLLAPSING_TREE = 'try-move-focus-from-collapsing-tree';
+export const kNOTIFY_TRY_REDIRECT_FOCUS_FROM_COLLAPSED_TAB = 'try-redirect-focus-from-collaped-tab';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_PARENT = 'try-expand-tree-from-focused-parent';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_BUNDLED_PARENT = 'try-expand-tree-from-focused-bundled-parent';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_ATTACHED_CHILD = 'try-expand-tree-from-attached-child';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_FOCUSED_COLLAPSED_TAB = 'try-expand-tree-from-focused-collapsed-tab';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_LONG_PRESS_CTRL_KEY = 'try-expand-tree-from-long-press-ctrl-key';
+export const kNOTIFY_TRY_EXPAND_TREE_FROM_END_TAB_SWITCH = 'try-expand-tree-from-end-tab-switch';
+export const kNOTIFY_TRY_FIXUP_TREE_ON_TAB_MOVED = 'try-fixup-tree-on-tab-moved';
 export const kGET_TREE              = 'get-tree';
 export const kATTACH                = 'attach';
 export const kDETACH                = 'detach';
@@ -98,6 +123,7 @@ export const kGET_TREE_STRUCTURE    = 'get-tree-structure';
 export const kSET_TREE_STRUCTURE    = 'set-tree-structure';
 export const kCOLLAPSE_TREE         = 'collapse-tree';
 export const kEXPAND_TREE           = 'expand-tree';
+export const kTOGGLE_TREE_COLLAPSED = 'toggle-tree-collapsed';
 export const kADD_TAB_STATE         = 'add-tab-state';
 export const kREMOVE_TAB_STATE      = 'remove-tab-state';
 export const kSCROLL                = 'scroll';
@@ -107,19 +133,29 @@ export const kNOTIFY_SCROLLED       = 'scrolled';
 export const kBLOCK_GROUPING        = 'block-grouping';
 export const kUNBLOCK_GROUPING      = 'unblock-grouping';
 export const kGRANT_TO_REMOVE_TABS  = 'grant-to-remove-tabs';
+export const kOPEN_ALL_BOOKMARKS_WITH_STRUCTURE = 'open-all-bookmarks-with-structure';
+export const kSET_EXTRA_TAB_CONTENTS   = 'set-extra-tab-contents';
+export const kCLEAR_EXTRA_TAB_CONTENTS = 'clear-extra-tab-contents';
+export const kCLEAR_ALL_EXTRA_TAB_CONTENTS = 'clear-all-extra-tab-contents';
+export const kGET_DRAG_DATA         = 'get-drag-data';
 
-export const kCONTEXT_MENU_UPDATED    = 'fake-contextMenu-updated';
-export const kCONTEXT_MENU_GET_ITEMS  = 'fake-contextMenu-get-items';
-export const kCONTEXT_MENU_OPEN       = 'fake-contextMenu-open';
-export const kCONTEXT_MENU_CREATE     = 'fake-contextMenu-create';
-export const kCONTEXT_MENU_UPDATE     = 'fake-contextMenu-update';
-export const kCONTEXT_MENU_REMOVE     = 'fake-contextMenu-remove';
-export const kCONTEXT_MENU_REMOVE_ALL = 'fake-contextMenu-remove-all';
-export const kCONTEXT_MENU_CLICK      = 'fake-contextMenu-click';
-export const kCONTEXT_MENU_SHOWN      = 'fake-contextMenu-shown';
-export const kCONTEXT_MENU_HIDDEN     = 'fake-contextMenu-hidden';
-
-export const kCONTEXT_ITEM_CHECKED_STATUS_CHANGED = 'fake-contextMenu-item-checked-status-changed';
+export const kCONTEXT_MENU_OPEN       = 'contextMenu-open';
+export const kCONTEXT_MENU_CREATE     = 'contextMenu-create';
+export const kCONTEXT_MENU_UPDATE     = 'contextMenu-update';
+export const kCONTEXT_MENU_REMOVE     = 'contextMenu-remove';
+export const kCONTEXT_MENU_REMOVE_ALL = 'contextMenu-remove-all';
+export const kCONTEXT_MENU_CLICK      = 'contextMenu-click';
+export const kCONTEXT_MENU_SHOWN      = 'contextMenu-shown';
+export const kCONTEXT_MENU_HIDDEN     = 'contextMenu-hidden';
+export const kFAKE_CONTEXT_MENU_OPEN       = 'fake-contextMenu-open';
+export const kFAKE_CONTEXT_MENU_CREATE     = 'fake-contextMenu-create';
+export const kFAKE_CONTEXT_MENU_UPDATE     = 'fake-contextMenu-update';
+export const kFAKE_CONTEXT_MENU_REMOVE     = 'fake-contextMenu-remove';
+export const kFAKE_CONTEXT_MENU_REMOVE_ALL = 'fake-contextMenu-remove-all';
+export const kFAKE_CONTEXT_MENU_CLICK      = 'fake-contextMenu-click';
+export const kFAKE_CONTEXT_MENU_SHOWN      = 'fake-contextMenu-shown';
+export const kFAKE_CONTEXT_MENU_HIDDEN     = 'fake-contextMenu-hidden';
+export const kOVERRIDE_CONTEXT        = 'override-context';
 
 export const kCOMMAND_BROADCAST_API_REGISTERED   = 'treestyletab:broadcast-registered';
 export const kCOMMAND_BROADCAST_API_UNREGISTERED = 'treestyletab:broadcast-unregistered';
@@ -129,6 +165,7 @@ export const kCOMMAND_REQUEST_CONTROL_STATE      = 'treestyletab:request-control
 export const kCOMMAND_GET_ADDONS                 = 'treestyletab:get-addons';
 export const kCOMMAND_SET_API_PERMISSION         = 'treestyletab:set-api-permisssion';
 export const kCOMMAND_NOTIFY_PERMISSION_CHANGED  = 'treestyletab:notify-api-permisssion-changed';
+export const kCOMMAND_UNREGISTER_ADDON           = 'treestyletab:unregister-addon';
 
 // https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/manifest.json/permissions
 // https://developer.mozilla.org/en-US/docs/Mozilla/Add-ons/WebExtensions/API/tabs/Tab
@@ -179,13 +216,20 @@ export class TreeItem {
         (sourceTab.incognito &&
          !permissions.has(kPERMISSION_INCOGNITO)))
       return null;
-
     const cacheKey = `${sourceTab.id}:${commonCacheKey}`;
-    if (cacheKey in this.cache.tabs)
-      return this.cache.tabs[cacheKey];
+    // The promise is cached here instead of the result,
+    // to avoid cache miss caused by concurrent call.
+    if (!(cacheKey in this.cache.tabs)) {
+      this.cache.tabs[cacheKey] = this._exportTab(sourceTab, permissions, commonCacheKey);
+    }
+    return await this.cache.tabs[cacheKey];
+  }
 
+  async _exportTab(sourceTab, permissions, commonCacheKey = '') {
     const [effectiveFavIconUrl, children] = await Promise.all([
-      (sourceTab.id in this.cache.effectiveFavIconUrls) ? this.cache.effectiveFavIconUrls[sourceTab.id] : TabFavIconHelper.getLastEffectiveFavIconURL(sourceTab),
+      (sourceTab.id in this.cache.effectiveFavIconUrls) ?
+        this.cache.effectiveFavIconUrls[sourceTab.id] :
+        TabFavIconHelper.getLastEffectiveFavIconURL(sourceTab).catch(ApiTabs.handleMissingTabError),
       doProgressively(
         sourceTab.$TST.children,
         child => this.exportTab(child, permissions, commonCacheKey),
@@ -196,11 +240,13 @@ export class TreeItem {
     if (!(sourceTab.id in this.cache.effectiveFavIconUrls))
       this.cache.effectiveFavIconUrls[sourceTab.id] = effectiveFavIconUrl;
 
+    const tabStates = sourceTab.$TST.states;
     const exportedTab = {
-      states:         Array.from(Constants.kTAB_SAFE_STATES).filter(state => sourceTab.$TST.states.has(state)),
+      states:         Constants.kTAB_SAFE_STATES_ARRAY.filter(state => tabStates.has(state)),
       indent:         parseInt(sourceTab.$TST.getAttribute(Constants.kLEVEL) || 0),
-      ancestorTabIds: sourceTab.$TST.ancestors.map(tab => tab.id),
-      children
+      ancestorTabIds: sourceTab.$TST.ancestorIds,
+      children,
+      bundledTabId:   sourceTab.$TST.bundledTabId
     };
 
     let allowedProperties = [
@@ -242,8 +288,10 @@ export class TreeItem {
       ]);
       exportedTab.effectiveFavIconUrl = effectiveFavIconUrl;
     }
-    if (permissions.has(kPERMISSION_COOKIES))
+    if (permissions.has(kPERMISSION_COOKIES)) {
       allowedProperties.push('cookieStoreId');
+      exportedTab.cookieStoreName = sourceTab.$TST.cookieStoreName;
+    }
 
     allowedProperties = new Set(allowedProperties);
     for (const key of allowedProperties) {
@@ -251,8 +299,6 @@ export class TreeItem {
         exportedTab[key] = sourceTab[key];
     }
 
-    this.cache.tabs[cacheKey] = exportedTab;
-
     return exportedTab;
   }
 }
@@ -305,8 +351,11 @@ function registerAddon(id, addon) {
       addon.grantedPermissions.size != addon.requestedPermissions.size)
     notifyPermissionRequest(addon, addon.requestedPermissions);
 
+  addon.id = id;
   addon.lastRegistered = Date.now();
   mAddons.set(id, addon);
+
+  onRegistered.dispatch(addon);
 }
 
 const mPermissionNotificationForAddon = new Map();
@@ -363,6 +412,131 @@ function notifyPermissionChanged(addon) {
   }).catch(ApiTabs.createErrorHandler());
 }
 
+function unregisterAddon(id) {
+  const addon = getAddon(id);
+  log('addon is unregistered: ', id, addon);
+  onUnregistered.dispatch(addon);
+  mAddons.delete(id);
+  delete mScrollLockedBy[id];
+  delete mGroupingBlockedBy[id];
+}
+
+export function getAddons() {
+  return mAddons.entries();
+}
+
+const mConnections = new Map();
+
+function onCommonCommand(message, sender) {
+  if (!message ||
+      typeof message.type != 'string')
+    return;
+
+  const addon = getAddon(sender.id);
+
+  switch (message.type) {
+    case kSCROLL_LOCK:
+      mScrollLockedBy[sender.id] = true;
+      if (!addon)
+        registerAddon(sender.id, sender);
+      return Promise.resolve(true);
+
+    case kSCROLL_UNLOCK:
+      delete mScrollLockedBy[sender.id];
+      return Promise.resolve(true);
+
+    case kBLOCK_GROUPING:
+      mGroupingBlockedBy[sender.id] = true;
+      if (!addon)
+        registerAddon(sender.id, sender);
+      return Promise.resolve(true);
+
+    case kUNBLOCK_GROUPING:
+      delete mGroupingBlockedBy[sender.id];
+      return Promise.resolve(true);
+
+    case kSET_EXTRA_TAB_CONTENTS:
+      if (!addon)
+        registerAddon(sender.id, sender);
+      break;
+  }
+}
+
+
+// =======================================================================
+// for backend
+// =======================================================================
+
+export async function initAsBackend() {
+  // We must listen API messages from other addons here beacause:
+  //  * Before notification messages are sent to other addons.
+  //  * After configs are loaded and TST's background page is almost completely initialized.
+  //    (to prevent troubles like breakage of `configs.cachedExternalAddons`, see also:
+  //     https://github.com/piroor/treestyletab/issues/2300#issuecomment-498947370 )
+  browser.runtime.onMessageExternal.addListener(onBackendCommand);
+
+  const manifest = browser.runtime.getManifest();
+  registerAddon(browser.runtime.id, {
+    id:         browser.runtime.id,
+    internalId: browser.runtime.getURL('').replace(/^moz-extension:\/\/([^\/]+)\/.*$/, '$1'),
+    icons:      manifest.icons,
+    listeningTypes: [],
+    bypassPermissionCheck: true
+  });
+  browser.runtime.onConnectExternal.addListener(port => {
+    const sender = port.sender;
+    mConnections.set(sender.id, port);
+    port.onMessage.addListener(message => {
+      onMessageExternal.dispatch(message, sender);
+      SidebarConnection.sendMessage({
+        type: 'external',
+        message,
+        sender
+      });
+    });
+    port.onDisconnect.addListener(_message => {
+      mConnections.delete(sender.id);
+      onBackendCommand({
+        type: kUNREGISTER_SELF,
+        sender
+      }).catch(ApiTabs.createErrorSuppressor());
+    });
+  });
+  const respondedAddons = [];
+  const notifiedAddons = {};
+  const notifyAddons = configs.knownExternalAddons.concat(configs.cachedExternalAddons);
+  log('initAsBackend: notifyAddons = ', respondedAddons);
+  await Promise.all(notifyAddons.map(async id => {
+    if (id in notifiedAddons)
+      return;
+    notifiedAddons[id] = true;
+    try {
+      id = await new Promise((resolve, reject) => {
+        let responded = false;
+        browser.runtime.sendMessage(id, {
+          type: kNOTIFY_READY
+        }).then(() => {
+          responded = true;
+          resolve(id);
+        }).catch(ApiTabs.createErrorHandler(reject));
+        setTimeout(() => {
+          if (!responded)
+            reject(new Error(`TSTAPI.initAsBackend: addon ${id} does not respond.`));
+        }, 3000);
+      });
+      if (id)
+        respondedAddons.push(id);
+    }
+    catch(e) {
+      console.log(`TSTAPI.initAsBackend: failed to send "ready" message to "${id}":`, e);
+    }
+  }));
+  log('initAsBackend: respondedAddons = ', respondedAddons);
+  configs.cachedExternalAddons = respondedAddons;
+
+  onInitialized.dispatch();
+}
+
 if (mIsBackend) {
   browser.notifications.onClicked.addListener(notificationId => {
     for (const [addonId, id] of mPermissionNotificationForAddon.entries()) {
@@ -385,19 +559,21 @@ if (mIsBackend) {
     }
   });
 
-  SidebarConnection.onConnected.addListener(windowId => {
+  SidebarConnection.onConnected.addListener((windowId, openCount) => {
     sendMessage({
       type:   kNOTIFY_SIDEBAR_SHOW,
       window: windowId,
-      windowId
+      windowId,
+      openCount
     });
   });
 
-  SidebarConnection.onDisconnected.addListener(windowId => {
+  SidebarConnection.onDisconnected.addListener((windowId, openCount) => {
     sendMessage({
       type:   kNOTIFY_SIDEBAR_HIDE,
       window: windowId,
-      windowId
+      windowId,
+      openCount
     });
   });
 
@@ -461,89 +637,12 @@ if (mIsBackend) {
     port.onDisconnect.addListener(onDisconnected);
   });
   */
-}
 
-function unregisterAddon(id) {
-  log('addon is unregistered: ', id, getAddon(id));
-  mAddons.delete(id);
-  delete mScrollLockedBy[id];
-  delete mGroupingBlockedBy[id];
-}
-
-export function getAddons() {
-  return mAddons.entries();
-}
-
-const mConnections = new Map();
-
-export async function initAsBackend() {
-  // We must listen API messages from other addons here beacause:
-  //  * Before notification messages are sent to other addons.
-  //  * After configs are loaded and TST's background page is almost completely initialized.
-  //    (to prevent troubles like breakage of `configs.cachedExternalAddons`, see also:
-  //     https://github.com/piroor/treestyletab/issues/2300#issuecomment-498947370 )
-  browser.runtime.onMessageExternal.addListener(onBackendCommand);
-
-  const manifest = browser.runtime.getManifest();
-  registerAddon(browser.runtime.id, {
-    id:         browser.runtime.id,
-    internalId: browser.runtime.getURL('').replace(/^moz-extension:\/\/([^\/]+)\/.*$/, '$1'),
-    icons:      manifest.icons,
-    listeningTypes: [],
-    bypassPermissionCheck: true
-  });
-  browser.runtime.onConnectExternal.addListener(port => {
-    const sender = port.sender;
-    mConnections.set(sender.id, port);
-    port.onDisconnect.addListener(_message => {
-      mConnections.delete(sender.id);
-      onBackendCommand({
-        type: kUNREGISTER_SELF,
-        sender
-      }).catch(ApiTabs.createErrorSuppressor());
-    });
-  });
-  const respondedAddons = [];
-  const notifiedAddons = {};
-  const notifyAddons = configs.knownExternalAddons.concat(configs.cachedExternalAddons);
-  log('initAsBackend: notifyAddons = ', respondedAddons);
-  await Promise.all(notifyAddons.map(async id => {
-    if (id in notifiedAddons)
+  browser.runtime.onMessage.addListener((message, _sender) => {
+    if (!message ||
+        typeof message.type != 'string')
       return;
-    notifiedAddons[id] = true;
-    try {
-      id = await new Promise((resolve, reject) => {
-        let responded = false;
-        browser.runtime.sendMessage(id, {
-          type: kNOTIFY_READY
-        }).then(() => {
-          responded = true;
-          resolve(id);
-        }).catch(ApiTabs.createErrorHandler(reject));
-        setTimeout(() => {
-          if (!responded)
-            reject(new Error(`TSTAPI.initAsBackend: addon ${id} does not respond.`));
-        }, 3000);
-      });
-      if (id)
-        respondedAddons.push(id);
-    }
-    catch(e) {
-      console.log(`TSTAPI.initAsBackend: failed to send "ready" message to "${id}":`, e);
-    }
-  }));
-  log('initAsBackend: respondedAddons = ', respondedAddons);
-  configs.cachedExternalAddons = respondedAddons;
 
-  onInitialized.dispatch();
-}
-
-browser.runtime.onMessage.addListener((message, _sender) => {
-  if (!message ||
-      typeof message.type != 'string')
-    return;
-
-  if (mIsBackend) {
     switch (message.type) {
       case kCOMMAND_REQUEST_INITIALIZE:
         return Promise.resolve({
@@ -578,28 +677,13 @@ browser.runtime.onMessage.addListener((message, _sender) => {
       case kCOMMAND_NOTIFY_PERMISSION_CHANGED:
         notifyPermissionChanged(getAddon(message.id));
         break;
-    }
-  }
-  else if (mIsFrontend) {
-    switch (message.type) {
-      case kCOMMAND_BROADCAST_API_REGISTERED:
-        registerAddon(message.sender.id, message.message);
-        if (message.message.style)
-          installStyleForAddon(message.sender.id, message.message.style);
-        break;
 
-      case kCOMMAND_BROADCAST_API_UNREGISTERED:
-        uninstallStyleForAddon(message.sender.id)
-        unregisterAddon(message.sender.id);
+      case kCOMMAND_UNREGISTER_ADDON:
+        unregisterAddon(message.id);
         break;
-
-      case kCOMMAND_BROADCAST_API_PERMISSION_CHANGED: {
-        const addon = getAddon(message.id);
-        addon.grantedPermissions = new Set(message.permissions);
-      }; break;
     }
-  }
-});
+  });
+}
 
 const mPromisedOnBeforeUnload = new Promise((resolve, _reject) => {
   // If this promise doesn't do anything then there seems to be a timeout so it only works if TST is disabled within about 10 seconds after this promise is used as a response to a message. After that it will not throw an error for the waiting extension.
@@ -611,9 +695,15 @@ const mWaitingShutdownMessages = new Map();
 
 function onBackendCommand(message, sender) {
   if (!message ||
+      typeof message != 'object' ||
       typeof message.type != 'string')
     return;
 
+  const results = onMessageExternal.dispatch(message, sender);
+  const firstPromise = results.find(result => result instanceof Promise);
+  if (firstPromise)
+    return firstPromise;
+
   switch (message.type) {
     case kPING:
       return Promise.resolve(true);
@@ -702,30 +792,6 @@ function onBackendCommand(message, sender) {
   }
 }
 
-function onCommonCommand(message, sender) {
-  if (!message ||
-      typeof message.type != 'string')
-    return;
-
-  switch (message.type) {
-    case kSCROLL_LOCK:
-      mScrollLockedBy[sender.id] = true;
-      return Promise.resolve(true);
-
-    case kSCROLL_UNLOCK:
-      delete mScrollLockedBy[sender.id];
-      return Promise.resolve(true);
-
-    case kBLOCK_GROUPING:
-      mGroupingBlockedBy[sender.id] = true;
-      return Promise.resolve(true);
-
-    case kUNBLOCK_GROUPING:
-      delete mGroupingBlockedBy[sender.id];
-      return Promise.resolve(true);
-  }
-}
-
 function exportAddons() {
   const exported = {};
   for (const [id, addon] of getAddons()) {
@@ -734,7 +800,17 @@ function exportAddons() {
   return exported;
 }
 
+export function isGroupingBlocked() {
+  return Object.keys(mGroupingBlockedBy).length > 0;
+}
+
+
+// =======================================================================
+// for frontend
+// =======================================================================
+
 export async function initAsFrontend() {
+  log('initAsFrontend: start');
   let response;
   while (true) {
     response = await browser.runtime.sendMessage({ type: kCOMMAND_REQUEST_INITIALIZE });
@@ -742,20 +818,53 @@ export async function initAsFrontend() {
       break;
     await wait(10);
   }
-  browser.runtime.onMessageExternal.addListener(onCommonCommand);
+  browser.runtime.onMessageExternal.addListener((message, sender) => {
+    if (message &&
+        typeof message == 'object' &&
+        typeof message.type == 'string') {
+      const results = onMessageExternal.dispatch(message, sender);
+      const firstPromise = results.find(result => result instanceof Promise);
+      if (firstPromise)
+        return firstPromise;
+    }
+    if (configs.incognitoAllowedExternalAddons.includes(sender.id) ||
+        !document.documentElement.classList.contains('incognito'))
+      return onCommonCommand(message, sender);
+  });
+  log('initAsFrontend: response = ', response);
   importAddons(response.addons);
-  for (const [id, addon] of getAddons()) {
-    // Install stylesheet always, even if the addon is not allowed to access
-    // private windows, because the client addon can be alloed on private
-    // windows by Firefox itself and extra context menu commands may be called
-    // via Firefox's native context menu (or shortcuts).
-    if (addon.style)
-      installStyleForAddon(id, addon.style);
+  for (const [, addon] of getAddons()) {
+    onRegistered.dispatch(addon);
   }
   mScrollLockedBy    = response.scrollLocked;
   mGroupingBlockedBy = response.groupingLocked;
 
   onInitialized.dispatch();
+  log('initAsFrontend: finish');
+}
+
+if (mIsFrontend) {
+  browser.runtime.onMessage.addListener((message, _sender) => {
+    if (!message ||
+        typeof message != 'object' ||
+        typeof message.type != 'string')
+      return;
+
+    switch (message.type) {
+      case kCOMMAND_BROADCAST_API_REGISTERED:
+        registerAddon(message.sender.id, message.message);
+        break;
+
+      case kCOMMAND_BROADCAST_API_UNREGISTERED:
+        unregisterAddon(message.sender.id);
+        break;
+
+      case kCOMMAND_BROADCAST_API_PERMISSION_CHANGED: {
+        const addon = getAddon(message.id);
+        addon.grantedPermissions = new Set(message.permissions);
+      }; break;
+    }
+  });
 }
 
 function importAddons(addons) {
@@ -769,28 +878,6 @@ function importAddons(addons) {
   }
 }
 
-const mAddonStyles = new Map();
-
-function installStyleForAddon(id, style) {
-  let styleElement = mAddonStyles.get(id);
-  if (!styleElement) {
-    styleElement = document.createElement('style');
-    styleElement.setAttribute('type', 'text/css');
-    document.head.insertBefore(styleElement, document.querySelector('#addons-style-rules'));
-    mAddonStyles.set(id, styleElement);
-  }
-  styleElement.textContent = style;
-}
-
-function uninstallStyleForAddon(id) {
-  const styleElement = mAddonStyles.get(id);
-  if (!styleElement)
-    return;
-  document.head.removeChild(styleElement);
-  mAddonStyles.delete(id);
-}
-
-
 export function isScrollLocked() {
   return Object.keys(mScrollLockedBy).length > 0;
 }
@@ -798,7 +885,7 @@ export function isScrollLocked() {
 export async function notifyScrolled(params = {}) {
   const lockers = Object.keys(mScrollLockedBy);
   const tab     = params.tab;
-  const window  = TabsStore.getWindow();
+  const window  = TabsStore.getCurrentWindowId();
   const cache   = {};
   const allTreeItems = Tab.getTabs(window).map(tab => new TreeItem(tab, { cache }));
   const results = await sendMessage({
@@ -808,7 +895,9 @@ export async function notifyScrolled(params = {}) {
     window,
     windowId: window,
 
+    deltaX:       params.event.deltaX,
     deltaY:       params.event.deltaY,
+    deltaZ:       params.event.deltaZ,
     deltaMode:    params.event.deltaMode,
     scrollTop:    params.scrollContainer.scrollTop,
     scrollTopMax: params.scrollContainer.scrollTopMax,
@@ -831,55 +920,62 @@ export async function notifyScrolled(params = {}) {
 }
 
 
-export function isGroupingBlocked() {
-  return Object.keys(mGroupingBlockedBy).length > 0;
-}
+// =======================================================================
+// Common utilities to send notification messages to other addons
+// =======================================================================
 
-
-/* Utilities to send notification messages to other addons */
-
-export function hasListenerForMessageType(type) {
-  return getListenersForMessageType(type).length > 0;
-}
-
-export function getListenersForMessageType(type) {
-  const uniqueTargets = {};
-  for (const [id, addon] of getAddons()) {
-    if (addon.listeningTypes.includes(type))
-      uniqueTargets[id] = true;
+export async function tryOperationAllowed(type, message = {}, options = {}) {
+  if (!hasListenerForMessageType(type, options)) {
+    //log(`=> ${type}: no listener, always allowed`);
+    return true;
   }
-  return Object.keys(uniqueTargets).map(id => getAddon(id));
+  const results = await sendMessage({ ...message, type }, options).catch(_error => {});
+  if (results.flat().some(result => result && result.result)) {
+    log(`=> ${type}: canceled by some helper addon`);
+    return false;
+  }
+  log(`=> ${type}: allowed by all helper addons`);
+  return true;
+}
+
+export function hasListenerForMessageType(type, options = {}) {
+  return getListenersForMessageType(type, options).length > 0;
+}
+
+export function getListenersForMessageType(type, { targets, except } = {}) {
+  targets = targets instanceof Set ? targets : new Set(Array.isArray(targets) ? targets : targets ? [targets] : []);
+  except  = except instanceof Set ? except : new Set(Array.isArray(except) ? except : except ? [except] : []);
+  const finalTargets = new Set();
+  for (const [id, addon] of getAddons()) {
+    if (addon.listeningTypes.includes(type) &&
+        (targets.size == 0 || targets.has(id)) &&
+        !except.has(id))
+      finalTargets.add(id);
+  }
+  //log('getListenersForMessageType ', { type, targets, except, finalTargets, all: mAddons });
+  return Array.from(finalTargets, getAddon);
 }
 
 export async function sendMessage(message, options = {}) {
-  const uniqueTargets = new Set();
-  const listenerAddons = getListenersForMessageType(message.type);
-  for (const addon of listenerAddons) {
-    uniqueTargets.add(addon.id);
-  }
+  const listenerAddons = getListenersForMessageType(message.type, options);
   const tabProperties = options.tabProperties || [];
   log(`sendMessage: sending message for ${message.type}: `, {
     message,
     listenerAddons,
-    targets: options.targets,
     tabProperties
   });
-  if (options.targets) {
-    if (!Array.isArray(options.targets))
-      options.targets = [options.targets];
-    for (const id of options.targets) {
-      uniqueTargets.add(id);
-    }
-  }
 
-  const promisedResults = spawnMessages(uniqueTargets, { message, tabProperties });
+  const promisedResults = spawnMessages(new Set(listenerAddons.map(addon => addon.id)), {
+    message,
+    tabProperties
+  });
   return Promise.all(promisedResults).then(results => {
     log(`sendMessage: got responses for ${message.type}: `, results);
     return results;
   }).catch(ApiTabs.createErrorHandler());
 }
 
-function* spawnMessages(targetSet, params) {
+function* spawnMessages(targets, params) {
   const message = params.message;
   const tabProperties = params.tabProperties || [];
 
@@ -912,6 +1008,15 @@ function* spawnMessages(targetSet, params) {
     }
     catch(e) {
       console.log(`Error on sending message to ${id}`, allowedMessage, e);
+      if (e &&
+          e.message == 'Could not establish connection. Receiving end does not exist.') {
+        browser.runtime.sendMessage(id, { type: kNOTIFY_READY }).catch(_error => {
+          console.log(`Unregistering missing helper addon ${id}...`);
+          unregisterAddon(id);
+          if (mIsFrontend)
+            browser.runtime.sendMessage({ type: kCOMMAND_UNREGISTER_ADDON, id });
+        });
+      }
       return {
         id,
         error: e
@@ -919,7 +1024,7 @@ function* spawnMessages(targetSet, params) {
     }
   };
 
-  for (const id of targetSet) {
+  for (const id of targets) {
     yield send(id);
   }
 }
@@ -960,11 +1065,13 @@ async function sanitizeMessage(message, params) {
     }
   }
   await Promise.all(tasks);
-  return Object.assign({}, message, sanitizedProperties);
+  return { ...message, ...sanitizedProperties };
 }
 
 
-/* Utilities for request-response type API call */
+// =======================================================================
+// Common utilities for request-response type API call
+// =======================================================================
 
 export async function getTargetTabs(message, sender) {
   await Tab.waitUntilTrackedAll(message.window || message.windowId);
@@ -979,7 +1086,7 @@ export async function getTargetTabs(message, sender) {
         message.tabIds == '*')
       return Tab.getAllTabs(message.window || message.windowId, { iterator: true });
     else
-      return Tab.getRoots(message.window || message.windowId, { iterator: true });
+      return Tab.getRootTabs(message.window || message.windowId, { iterator: true });
   }
   if (message.tab == '*' ||
       message.tabId == '*' ||
@@ -1011,16 +1118,59 @@ async function getTabsFromWrongIds(ids, sender) {
       case 'active':
       case 'current':
         return Tab.getActiveTab(activeWindow.id);
+
       case 'next':
         return Tab.getActiveTab(activeWindow.id).$TST.nextTab;
+      case 'nextcyclic':
+        return Tab.getActiveTab(activeWindow.id).$TST.nextTab || Tab.getFirstTab(activeWindow.id);
+
       case 'previous':
       case 'prev':
         return Tab.getActiveTab(activeWindow.id).$TST.previousTab;
+      case 'previouscyclic':
+      case 'prevcyclic':
+        return Tab.getActiveTab(activeWindow.id).$TST.previousTab || Tab.getLastTab(activeWindow.id);
+
       case 'nextsibling':
         return Tab.getActiveTab(activeWindow.id).$TST.nextSiblingTab;
+      case 'nextsiblingcyclic': {
+        const active = Tab.getActiveTab(activeWindow.id);
+        const nextSibling = active.$TST.nextSiblingTab;
+        if (nextSibling)
+          return nextSibling;
+        const parent = active.$TST.parent;
+        if (parent)
+          return parent.$TST.firstChild;
+        return Tab.getFirstTab(activeWindow.id);
+      }
+
       case 'previoussibling':
       case 'prevsibling':
         return Tab.getActiveTab(activeWindow.id).$TST.previousSiblingTab;
+      case 'previoussiblingcyclic':
+      case 'prevsiblingcyclic': {
+        const active = Tab.getActiveTab(activeWindow.id);
+        const previousSiblingTab = active.$TST.previousSiblingTab;
+        if (previousSiblingTab)
+          return previousSiblingTab;
+        const parent = active.$TST.parent;
+        if (parent)
+          return parent.$TST.lastChild;
+        return Tab.getLastRootTab(activeWindow.id);
+      }
+
+      case 'nextvisible':
+        return Tab.getActiveTab(activeWindow.id).$TST.nearestVisibleFollowingTab;
+      case 'nextvisiblecyclic':
+        return Tab.getActiveTab(activeWindow.id).$TST.nearestVisibleFollowingTab || Tab.getFirstVisibleTab(activeWindow.id);
+
+      case 'previousvisible':
+      case 'prevvisible':
+        return Tab.getActiveTab(activeWindow.id).$TST.nearestVisiblePrecedingTab;
+      case 'previousvisiblecyclic':
+      case 'prevvisiblecyclic':
+        return Tab.getActiveTab(activeWindow.id).$TST.nearestVisiblePrecedingTab || Tab.getLastVisibleTab(activeWindow.id);
+
       case 'sendertab':
         return sender.tab && Tab.get(sender.tab.id) || null;
       case 'highlighted':
@@ -1063,6 +1213,7 @@ export async function formatTabResult(treeItems, originalMessage, senderId) {
   if (Array.isArray(originalMessage.tabs) ||
       originalMessage.tab == '*' ||
       originalMessage.tabs == '*')
-    return Promise.all(treeItems.map(treeItem => treeItem.exportFor(senderId))).then(tabs => tabs.filter(tab => !!tab));
+    return Promise.all(treeItems.map(treeItem => treeItem.exportFor(senderId)))
+      .then(tabs => tabs.filter(tab => !!tab));
   return treeItems.length > 0 ? treeItems[0].exportFor(senderId) : null ;
 }
